# encoding: utf-8
"""
data_filters.py -- Select proper data for analysis
"""

# Library imports
import numpy as np

# Package imports
from scanr.meta import get_day_list
from scanr.data import get_node
from scanr.session import SessionData
from scanr.spike import parse_cell_name
from tools.misc import memoize

# --------------------------------------------------------------------------------

@memoize
def latest_qualified_day(rat,expt_type):
    """
    Return the latest day to be included in scan-related analysis

    Most rats ran double-rotation task for 4 consecutive days, but some longer.
    To minimize individual differences, we restrict ourselves to use only the
    first 4 experiment days in record.
    """
    MAX_DAY = { 'DR'    : 4 }

    if int(rat) != rat or rat < 0:
        raise ValueError, "Bad rat number: " + str(rat)
    if expt_type not in MAX_DAY.keys():
        raise ValueError, "Bad experiment type: " + str(expt_type)

    session_table = get_node('/metadata','sessions')
    days = np.sort( list(set( [rec['day'] for rec in \
        session_table.where('(expt_type=="%s")&(rat==%d)'%(expt_type,rat))] )) )

    return days[ min(len(days),MAX_DAY[expt_type])-1 ]
# end of FUNCTION latest_qualified_day

@memoize
def first_exp_day(rat,expt_type):
    """
    Return the first recording day of the specified rat and experiment type
    """
    if int(rat) != rat or rat < 0:
        raise ValueError, "Bad rat number: " + str(rat)

    session_table = get_node('/metadata','sessions')
    days = np.sort( list(set( [rec['day'] for rec in \
        session_table.where('(expt_type=="%s")&(rat==%d)'%(expt_type,rat))] )) )

    return days[0]
# end of FUNCTION first_exp_day

@memoize
def last_exp_day(rat,expt_type):
    """
    Return the first recording day of the specified rat and experiment type
    """
    if int(rat) != rat or rat < 0:
        raise ValueError, "Bad rat number: " + str(rat)

    session_table = get_node('/metadata','sessions')
    days = np.sort( list(set( [rec['day'] for rec in \
        session_table.where('(expt_type=="%s")&(rat==%d)'%(expt_type,rat))] )) )

    return days[-1]
# end of FUNCTION first_exp_day

def collect_unique_index(ses, table, query, repeat_filter, debug=False):
    table_idx, tt_counts = [], {}
    for field in table.where(query):
        rat, day  = field['rat'], field['day']
        if day > latest_qualified_day(rat,'DR'):
            continue
        table_idx.append( field['id'] )
        if repeat_filter is not None:
            # count for each (rat,day,mis_angle,tt), how many place fields 
            # were recorded
            tt, cl = parse_cell_name( field['tc'] )
            key = (rat, day, field['mis_angle'], tt)
            if key not in tt_counts.keys():
                tt_counts[key] = 1
            else:
                tt_counts[key] += 1
    if repeat_filter is None:
        return table_idx

    if repeat_filter is not None:
        # growing the list based on field counts (high to low)
        tt_counts = [ key+(tt_counts[key],) for key in tt_counts.keys() ]
        rats, days, mis_angles, tts, counts = zip(*tt_counts)
        insert_order = np.argsort(counts)[::-1] # from max counts to min
        tt_list = {} # tt_list[ (rat,day) ] = [tt1, tt2, ... ]
        rtm_list = [] # record whether (rat, tt, mis_angle) have encountered
        for i in insert_order:
            rat, day, mis_angle, tt = rats[i], days[i], mis_angles[i], tts[i]
            # if True then current (rat,tt,mis_angle) has most fields
            if (rat,tt,mis_angle) not in rtm_list:
                key = (rat,day,mis_angle)
                if key not in tt_list.keys():
                    tt_list[key] = [tt, ]
                else:
                    tt_list[key].append( tt )
                rtm_list.append( (rat,tt,mis_angle) )
        if debug:
            print tt_list

    # collect qualified field id
    index = []
    keep_idx, maze_rate = {}, {}
    for i in table_idx:
        field = table[i]
        rat, day, maze, tc  = \
            field['rat'], field['day'], field['session'], field['tc']
        tt, cl = parse_cell_name( field['tc'] )
        key = (rat, day, field['mis_angle'])
        if (key not in tt_list.keys()) or (tt not in tt_list[key]):
            continue
        if (ses=='STD') and repeat_filter=='highest_mean':
            clust = (rat, day, tc)
            # Calculate within-field mean firing rate
            session = SessionData.get( (rat,day,maze) )
            # running filter
            R = session.get_cluster_ratemap( tc,bins=360,\
                blur_width=4.3,**session.running_filter() )
            # off-track filter
            #R = session.get_cluster_ratemap( tc,bins=360,\
            #    blur_width=4.3,exclude_off_track=True, \
            #        exclude=session.extended_scan_and_pause_list )
            end = int( np.floor(field['field_edges'][0]) )
            start = int( np.floor(field['field_edges'][1]) )
            if start > end:
                rate = np.mean( R[end:start+1] )
            else:
                rate = np.mean( np.r_[R[-(360-end):],R[:start+1]] )

            if clust not in keep_idx.keys():
                keep_idx[clust] = [i,]
                maze_rate[clust] = (maze, rate)
            else:
                if rate > maze_rate[clust][1]:
                    maze_rate[clust] = (maze, rate)
                    if maze == maze_rate[clust][0]:
                        keep_idx[clust].append(i)
                    else:
                        keep_idx[clust] = [i,]
                elif maze == maze_rate[clust][0]:
                    keep_idx[clust].append(i)
        else:
            index.append( i )
    if (ses=='STD') and repeat_filter=='highest_mean':
        for value in keep_idx.values():
            index += value

    if debug:
        print index 

    return index
# end of FUNCTION collect_unique_index
