# encoding: utf-8

"""
misc.py
"""

# Library imports
import matplotlib.pyplot as plt
import numpy as np

# Local imports

# ---------------------------------------------------------------------------------

def star_to_significance_level( sig, one_side=False, multi_n=None ):
    """
    Get the p value threshold for given star sign. Default a two-tail test and no
    multiple-comparison correction.

    sig: [* / ** / ***] significance level sign
    one_side: [True/False] True if run a one-tail test
    multi_n: [integer n] correct for n comparisons
    """
    if sig == '*':
        a = 0.05
    elif sig == '**':
        a = 0.01
    elif sig == '***':
        a = 0.001
    else:
        raise ValueError('the significance level should be */**/***!')

    if not one_side:
        a /= 2.0
    if multi_n is not None:
        a /= float(multi_n)

    return a
# end of FUNCTION star_to_significance_level

# ---------------------------------------------------------------------------------

def plot_resample_distribution(ax,obs,data,exp=0,resample='bootstrap',nbin=20,xlim=None,one_side=False,larger=True,multi_n=None,bs_cv='centered',sig_level='*',tail=None,show_obs_diff=True):
    data = np.array(data) # transform to numpy array if not already is
    if show_obs_diff and (resample=='bootstrap'):
        hist_data = obs + (obs-data)
    else:
        hist_data = data.copy()

    # Plot resample distribution
    if xlim is None:
        ax.hist(hist_data,color='white')
    else:
        ax.hist(hist_data,color='white',bins=np.linspace(xlim[0],xlim[1],nbin+1))
        ax.set_xlim( xlim )
    ax.axvline(x=obs,lw=2,color='k')
    ax.axvline(x=exp,lw=2,color='gray',ls='--')

    if resample != 'bootstrap':
        if one_side:
            if larger:
                outside, n = np.sum(data>=obs), len(data)
                rel = 'larger than or equal to'
            else:
                outside, n = np.sum(data<=obs), len(data)
                rel = 'smaller than or equal to'
        else:
            if obs >= exp:
                outside, n = np.sum(data>=obs), len(data)
                rel = 'larger than or equal to'
            else:
                outside, n = np.sum(data<=obs), len(data)
                rel = 'smaller than or equal to'
        p = float(outside) / n
        print( '%d out of %d shuffled examples %s observation' % (outside,n,rel) )

    # Decide and plot confidence intervals
    sig_level = ['***', '**', '*'] if sig_level is None else sig_level
    for i, sig in enumerate(sig_level):
        a = star_to_significance_level( sig,one_side,multi_n )

        if resample == 'permutation':
            left_tail = np.percentile( data, 100*a )
            right_tail = np.percentile( data, 100-100*a )
            if one_side:
                tail = right_tail if larger else left_tail
                if p < a: # pass significance
                    ax.axvline(x=tail,alpha=0.5,ls=':',lw=1,color='gray')
                    ax.text(0.05,0.8,'%s'%sig,fontsize=10,transform=ax.transAxes)
                    return
                elif i == (len(sig_level)-1):
                    ax.axvline(x=tail,alpha=0.5,ls=':',lw=1,color='gray')
                    ax.text(0.05,0.8,'n.s.',fontsize=10,transform=ax.transAxes)
            else:
                if p < a: # pass significance
                    for CI in [ left_tail, right_tail ]:
                        ax.axvline(x=CI,alpha=0.5,ls=':',lw=1,color='gray')
                    ax.text(0.05,0.8,'%s'%sig,fontsize=10,transform=ax.transAxes)
                    return
                elif i == (len(sig_level)-1):
                    for CI in [ left_tail, right_tail ]:
                        ax.axvline(x=CI,alpha=0.5,ls=':',lw=1,color='gray')
                    ax.text(0.05,0.8,'n.s.',fontsize=10,transform=ax.transAxes)
        elif resample == 'bootstrap':
            if tail is not None:
                left_tail = tail
            elif bs_cv == 'centered':
                left_tail = 2*obs - np.percentile( data, 100-100*a )
            elif bs_cv == 'percentile':
                left_tail = np.percentile( data, 100*a )
            if one_side:
                if exp < left_tail: # pass significance
                    ax.axvline(x=left_tail,alpha=0.5,ls=':',lw=1,color='gray')
                    ax.text(0.05,0.8,sig,fontsize=10,transform=ax.transAxes)
                    print('CI: %.4f'%left_tail)
                    return
                elif i == (len(sig_level)-1):
                    ax.axvline(x=left_tail,alpha=0.5,ls=':',lw=1,color='gray')
                    ax.text(0.05,0.8,'n.s.',fontsize=10,transform=ax.transAxes)
                    print('CI: %.4f'%left_tail)
            else:
                if bs_cv == 'centered':
                    right_tail = 2*obs - np.percentile( data, 100*a )
                elif bs_cv == 'percentile':
                    right_tail = np.percentile( data, 100-100*a )
                if (exp<left_tail) or (exp>right_tail): # pass significance
                    for CI in [ left_tail, right_tail ]:
                        ax.axvline(x=CI,alpha=0.5,ls=':',lw=1,color='gray')
                    ax.text(0.05,0.8,sig,fontsize=10,transform=ax.transAxes)
                    print('CI: [%.4f, %.4f]' % (left_tail,right_tail))
                    return
                elif i == (len(sig_level)-1):
                    for CI in [ left_tail, right_tail ]:
                        ax.axvline(x=CI,alpha=0.5,ls=':',lw=1,color='gray')
                    ax.text(0.05,0.8,'n.s.',fontsize=10,transform=ax.transAxes)
                    print('CI: [%.4f, %.4f]' % (left_tail,right_tail))
    return
# end of FUNCTION plot_resample_distribution

# ---------------------------------------------------------------------------------

def reset_plot():
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)
# end of FUNCTION reset_plot
