# encoding: utf-8
"""
F1B.py

Plot out the session ratemaps of individual clusters in stacks to see the distributions of fields
"""

import tables as tb
import numpy as np
import matplotlib.pylab as plt

from scanr.lib import *
from scanr.data import get_node, get_unique_row

from analysis_shan.data_filters import collect_unique_index
from boundary_paper.misc import ModifiedSessionData

BINS = 360 # number of spatial bins
BLUR_WIDTH = 4.3 # smoothing kernel width
BD_ANGLES = { 'local':[45,135,225,315], 'global':[0,45,135,180,270] }

def plot_ratemap_stack(ax,area,sess,repeat_filter='highest_mean'):
    """
    For the specified subarea and session type, plot out the rate map stack
    """

    # Load the table containing the session and place field information
    session_table = get_node('/metadata','sessions')
    field_file = tb.openFile('/home/paramecium/output/fields.h5','r')
    field_table = field_file.root.place_fields
    query = '(area=="%s")&(type=="%s")&(expt_type=="DR")&(exclude==False)'\
        % (area,sess)
    title_str = area + '-' + sess

    # collect the cluster and place field information
    print 'Collecting rate maps for %s...' % title_str 
    data = []
    clust_count = 0
    index = collect_unique_index(sess,field_table,query,repeat_filter)
    index.sort()
    for field in field_table[index]:
        rat, day, maze, tc, com = \
            field['rat'], field['day'], field['session'], field['tc'], field['COM']
        if field['field_num'] == 1:
            clust_count += 1

        mis = 0
        if sess == 'MIS':
            session = get_unique_row( session_table, \
                '(rat==%d)&(day==%d)&(session==%d)' %(rat,day,maze) )
            mis = session['mis_angle'] / 2.0
            com = (com-mis+360) % 360
        data.append( ((rat,day,maze,tc), com, mis) )
    N = len(data)
    print '%d fields collected from %d cells for %s-%s' % (N,clust_count,area,sess)

    # Load the firing rate maps based on the sorted order 
    angles = np.array( zip(*data)[1] )
    sorted_index = np.argsort(angles)

    ratemat = None 
    for i in sorted_index:
        rat, day, maze, tc = data[i][0]
        session = ModifiedSessionData.get( (rat,day,maze) )
        R = session.get_cluster_ratemap(tc,bins=BINS,mis=data[i][-1], \
            rot='local',blur_width=BLUR_WIDTH,**session.running_filter() )
        R /= float( np.max(R) )
        if ratemat is None:
            ratemat = R.copy()
        else:
            ratemat = np.vstack( [ratemat, R] )

    # Plot rate map stacks
    ax.imshow(ratemat,interpolation='none',aspect='auto',cmap='gray_r', \
        origin='lower',extent=[0,360,0,N])
    for cue in BD_ANGLES['local']:
        ax.axvline(x=cue,lw=1,color='g',alpha=0.1)
    ax.set_title(title_str,fontsize=12,fontname='Arial')
    ax.set_xticks(np.linspace(0,360,3).astype(int))
    ax.set_xticklabels(np.linspace(0,360,3).astype(int),fontname='Arial')
    ymin, ymax = ax.get_ylim()
    ax.set_yticks(np.round(np.linspace(ymin,ymax,6),0).astype(int))
    ax.set_yticklabels( \
        np.round(np.linspace(ymin,ymax,6),0).astype(int),fontname='Arial')
    ax.tick_params(axis='both',labelsize=8)
    ax.set_xlim([0, 360])

    field_file.close()
    print 'Done!'
# end of function plot_ratemap_stack

def panels():
    """
    Go through each subarea-session to print out the rate map stack
    """
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    fn = 'F1B.png'
    fg_cnt = 1
    for area in ['CA1', 'CA3']:
        for sess in ['STD', 'MIS']:
            ax = plt.subplot(1,4,fg_cnt)
            plot_ratemap_stack(ax,area,sess,repeat_filter='highest_mean')
            fg_cnt += 1

    plt.subplots_adjust(left=0.07,right=0.96,bottom=0.08,top=0.93,wspace=0.35)
    f = plt.gcf()
    f.set_size_inches(4.5,3,forward=True)
    f.savefig(fn,dpi=1000)
# end of function panels
