# encoding: utf-8
"""
F1C.py

Plot out the trajectory-and-spike map and rate map for the selected examples
"""

import os
import numpy as np
import tables as tb
from itertools import product
import matplotlib.pylab as plt
from matplotlib.gridspec import GridSpec

from tools.misc import contiguous_groups
from core.report import BaseReport
from core.analysis import AbstractAnalysis

from scanr.lib import * 
from scanr.session import SessionData
from scanr.time import select_from
from scanr.data import get_unique_row
from scanr.tracking import plot_track_underlay
from analysis_shan.data_filters import latest_qualified_day

# Configure
INNER_DIAMETER = Config['track']['inner_diameter']
OUTER_DIAMETER = Config['track']['outer_diameter']
track_fmt = dict(lw=0.5,ls='dotted',zorder=-100,ec='grey')

# ---------------------------------------------------------------------------------

def plot_quadrant_boundaries(D_in=INNER_DIAMETER, D_out=OUTER_DIAMETER, mis=0):
    """
    Mark quadrant boundaries with green lines

    D_in: inner diameter of the track
    D_out: outer diameter of the track
    mis: the mismatch angle
    """
    bd_color = 'g'

    R_in, R_out = D_in/2.0, D_out/2.0
    angles = [ np.pi/4.0, np.pi*3/4, np.pi*5/4, np.pi*7/4 ]
    angles += np.radians(mis)
    for i,angle in enumerate(angles):
        plt.plot( [R_in*np.cos(angle),R_out*np.cos(angle)],
            [R_in*np.sin(angle),R_out*np.sin(angle)],bd_color,lw=2, \
                zorder=2,alpha=1 )
# end of FUNCTION plot_quadrant_boundaries

# ---------------------------------------------------------------------------------

def plot_field_edges(D_in=INNER_DIAMETER,D_out=OUTER_DIAMETER,table=None,query=None):

    """
    Mark place field edges with blue lines

    D_in: inner diameter of the track
    D_out: outer diameter of the track
    table: table which contains place field edge angle information
    query: filter conditions
    """
    R_in, R_out = D_in/2.0, D_out/2.0
    for field in table.where(query):
        for angle in field['field_edges']:
            rad = np.deg2rad(angle)
            plt.plot([0,R_out*np.cos(rad)],[0,R_out*np.sin(rad)],'b', \
                lw=0.5,zorder=5)
# end of FUNCTION plot_field_edges

# ---------------------------------------------------------------------------------

class ExampleField1D(AbstractAnalysis):
    """
    Plot the trajectory-and-spike plots and the firing-rate maps of the example clusters 
    """
    datadir = './Example1D'
    label = 'example field 1d'

    def collect_data(self):
        """
        Generate the cluster list
        """
        self.results = {}
        self.results['item_list'] = [ (113,3,1,'t4c4'), (113,4,3,'t1c11'), \
                                      (119,3,3,'t6c1'), (434,1,6,'t10c3'), \
                                      (435,2,1,'t10c2'), (72,1,1,'t17c4'), \
                                      (195,4,4,'t16c2'), (435,2,4,'t12c2') ]

        self.out('Done')
    # end of METHOD collect_data

    def process_data(self):
        """
        For the specified clusters, print out the trajectory-and-spike maps and rate maps
        """
        self.figure = {}

        # Import the place field information
        table_file = tb.openFile('/home/paramecium/output/fields.h5','r')
        field_table = table_file.root.place_fields

        RUN, NRUN = False, True

        # Set up the subplot layout
        f = plt.figure()
        gs = GridSpec(4,6)
        gs.update(left=0.01,right=0.98,top=0.94,bottom=0.08,wspace=0.1,hspace=0.35)

        cell_count = 1
        item_list = self.results['item_list']
        plt_types = ['Spikes','Ratemap']
        data_list = [ data for data in product(item_list,plt_types) ]

        for item, plt_type in data_list:
            rat, day, maze, tc = item

            # mismatch angle
            metadata = get_session_metadata(rat, day, maze)
            mis=metadata['mis_angle']/2.0

            # identify the subplot to be drawn
            ix = cell_count - 1
            if plt_type == 'Spikes':
                ax = plt.subplot( gs[ix%4, (ix/4)*3] )
            elif plt_type == 'Ratemap':
                ax = plt.subplot( gs[ix%4, (ix/4)*3+1:(ix/4)*3+3] )

            # Plot out the field edges and local-cue boundaries
            query = '(rat==%d)&(day==%d)&(session==%d)&(tc=="%s")' % item 
            for field in field_table.where(query):
                if plt_type == 'Spikes':
                    plot_field_edges(table=field_table,query='(id==%d)'%field['id'])
                    plot_quadrant_boundaries( mis=mis )
                elif plt_type == 'Ratemap':
                    edges = field['field_edges'] # (end, start)
                    offset, onset = edges
                    cross = True if onset < offset else False
                    # plot field ranges 
                    plt.axvline(x=onset,lw=0.5,color='b')
                    plt.axvline(x=offset,lw=0.5,color='b')
                    plt.axvline(x=onset+360,lw=0.5,color='b')
                    plt.axvline(x=offset+360,lw=0.5,color='b')
                    if cross:
                        plt.axvspan(offset,360+onset,alpha=0.5,color='c')
                        plt.axvspan(0,onset,alpha=0.5,color='c')
                        plt.axvspan(offset+360,720,alpha=0.5,color='c')
                    else:
                        plt.axvspan(offset,onset,alpha=0.5,color='c')
                        plt.axvspan(offset+360,onset+360,alpha=0.5,color='c')
                    # plot local-cue boundaries
                    angles = np.array([45, 135, 225, 315]) + mis
                    for angle in angles:
                        plt.axvline(x=angle,alpha=0.5,color='g')
                        plt.axvline(x=angle+360,alpha=0.5,color='g')

            session =  SessionData.get( (rat,day,maze) )
            clust = session.clusts[tc]

            if plt_type == 'Ratemap': # Plot out the firing-rate map
                # Load the firing-rate map
                R = session.get_cluster_ratemap(tc,bins=360,blur_width=4.3, \
                    **session.running_filter())
                plt.plot(np.r_[R,R],'k',lw=1)
                threshold = R.max() * 0.1
                plt.axhline(y=threshold,lw=1,color='grey') # firing-rate threshold
                # Formatting the plot
                plt.xlim(0,720)
                if cell_count%4 != 0:
                    ax.set_xticklabels([])
                else:
                    ax.set_xticks( np.linspace(0,720,5) )
                if cell_count == 4:
                    plt.ylabel( 'Rate (Hz)',\
                        fontsize=10,fontname='Arial',labelpad=0 )
                    plt.xlabel( 'Track Angle ($^\circ$)',\
                        fontsize=10,fontname='Arial',labelpad=0)
                ymin, ymax = ax.get_ylim()
                ax.set_yticks( np.round(np.linspace(ymin,ymax,3),0).astype(int) )
                ax.set_yticklabels( \
                    np.round(np.linspace(ymin,ymax,3),0).astype(int), \
                        fontname='Arial' )
                plt.tick_params(axis='both',labelsize=8)
                cell_count += 1
            elif plt_type == 'Spikes': # Plot out the spike-and-trajectory plot
                traj = session.trajectory
                ts = traj.ts

                # Extract the running/not running segments
                run_idx = session.filter_tracking_data( ts,boolean_index=True, \
                    **session.running_filter() )
                groups = np.array( \
                    [(g[0],g[1]-1) for g in contiguous_groups(run_idx)] )

                # For each segment print out the spikes within the segment
                spk, spk_x, spk_y = clust.spikes, clust.x, clust.y
                colors = ['r','grey']
                orders = [3,2]
                ix = select_from(spk, ts[groups])
                for group_type in [RUN, NRUN]:
                    ix = group_type - ix
                    x, y = spk_x[ix], spk_y[ix]
                    plt.plot( x,y,'o',mfc='None',mec=colors[group_type], \
                        markersize=2,zorder=orders[group_type] )

                # Plot out the track edges
                fmt = dict(lw=0.5,ls='dotted',zorder=-100,ec='grey')
                plot_track_underlay(ax,inner=fmt,outer=fmt)

                # Plot out the trajectory
                plt.plot(traj.x,traj.y,'-',zorder=1,c='k',lw=0.1)

                # Formatting the picture
                ax.axis( [-55,55,-55,55] )
                ax.set_aspect('equal')
                ax.set_title( 'cell %d: rat %d, day %d, session %d, %s' \
                    % (cell_count,rat,day,maze,tc), size=10, fontname='Arial',\
                        loc='left', position=[0,1.005] )
                ax.set_axis_off()
        # end of get_plot

        table_file.close()

        f.set_size_inches(7,4,forward=True)
        f.savefig( os.path.join(self.datadir,'F1C.png'),dpi=1000 )

        self.out('Done')
    # end of METHOD process_data

# end of FUNCTION ExampleField1D

# ---------------------------------------------------------------------------------
