# encoding: utf-8
"""
F1D.py

Plot out the correlograms of the firing-rate map matrix
"""

import tables as tb
import numpy as np
import matplotlib.pylab as plt

from matplotlib.gridspec import GridSpec
from scipy.stats import sem

from scanr.lib import *
from analysis_shan.data_filters import collect_unique_index
from boundary_paper.misc import ModifiedSessionData

BD_ANGLES = { 'local':[45, 135, 225, 315], 'global':[0, 45, 135, 180, 270] }

def population_ratemap_correlogram():
    """
    Plot out the spatial ratemap correlograms
    """

    # Clean up plot residues
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    bins = 360 # number of spatial bins
    blur_width = 4.3 # smooth kernel width

    # Import the place field information
    table_file = tb.openFile('/home/paramecium/output/fields.h5','r')
    field_table = table_file.root.place_fields
    table_query = '(expt_type=="DR")&(exclude==False)'

    fig_cnt = 1
    for i, area in enumerate( ["CA1", "CA3"] ):
        for j, sess_type in enumerate( ["STD", "MIS"] ):
            # collect filtered field list
            query = table_query + '&(area=="%s")&(type=="%s")' % (area,sess_type)
            index = collect_unique_index(sess_type,field_table,query,'highest_mean')
            index.sort()

            # construct ratemap matrix
            ratemap = []
            for ix in index:
                field = field_table[ix]
                rat, day, maze, tc = \
                    field['rat'], field['day'], field['session'], field['tc']
                # avoid including the same rate map repetitively
                if field['field_num'] > 1:
                    continue

                session = ModifiedSessionData.get( (rat,day,maze) )
                R = session.get_cluster_ratemap( tc, bins=bins, \
                    mis=field['mis_angle']/2.0, blur_width=blur_width, \
                        **session.running_filter() )
                R /= float( np.max(R) )
                ratemap.append( R ) 

            n = len(ratemap)
            if not n:
                fig_cnt += 1
                continue
            print '%d fields collected from %d cells for %s-%s sessions' \
                % (len(index),len(ratemap),area,sess_type)

            # Calculate and plot the correlograms
            plt.subplot(2,2,fig_cnt)
            ax = plt.gca()
            C = np.corrcoef(ratemap,rowvar=0)
            plt.imshow( C,interpolation='none',aspect='equal',origin='lower')
            ax.set_aspect('equal')
            for angle in BD_ANGLES['local']:
                plt.axvline(angle,color='white',ls='--',lw=0.5)
                plt.axhline(angle,color='white',ls='--',lw=0.5)
            if i == 1: # bottom
                ax.set_xticks([0,180,360])
                ax.set_xticklabels([0,180,360],fontname='Arial',fontsize=8)
            else:
                ax.set_xticks([0,180,360])
                ax.set_xticklabels([])
                ax.set_title('%s'%sess_type,fontname='Arial',fontsize=12)
            if j == 1: # right
                ax.yaxis.tick_right()
                ax.set_yticks([0,180,360])
                ax.set_yticklabels([0,180,360],fontname='Arial',fontsize=8)
            else:
                ax.set_yticks([0,180,360])
                ax.set_yticklabels([])
                ax.set_ylabel('%s'%area,fontname='Arial',fontsize=12)

            fig_cnt += 1
    plt.subplots_adjust( \
        left=0.13,right=0.9,bottom=0.05,top=0.93,wspace=0.14,hspace=0.006 )

    f = plt.gcf()
    f.set_size_inches(2.5,2.4,forward=True)
    f.savefig( './F1D.png', dpi=1000 )

    table_file.close()
# end of FUNCTION population_ratemap_correlogram
