# encoding: utf-8
"""
F2.py

Check how the locations of place field edges were affected by the local-cue
boundries on double rotation track
"""

import tables as tb
import numpy as np
import matplotlib.pylab as plt
import random
import sys

from core.analysis import AbstractAnalysis

from scanr.lib import *

from analysis_shan.data_filters import collect_unique_index
from analysis_shan.misc import plot_resample_distribution

BD_ANGLES = { 'local':[45, 135, 225, 315], 'global':[0, 45, 135, 180, 270] } # the locations of the cue boundaries
MARGIN = 15 # the half-width of the local cue boundary window

# --------------------------------------------------------------------------------

class SpatialDistrAna(AbstractAnalysis):
    """
    A general class for generating 1D spatial distribution histograms, and
    within-window proportion shuffling and bootstrap tests
    """

    def collect_data(self):
        """
        Collect the specified data (ex: field edge)
        """
        # to be implemented in child class
        pass
    # end of METHOD collect_data

    def process_data(self):
        """
        Interface to generate the histograms
        """
        # to be implemented in child class
        pass
    # end of METHOD process_data

    def plot_paper(self, **kwds):
        """
        Call plot_panel_paper to generate the histograms for each subarea-session

        kwds: see plot_panel_paper
        """

        # Set the panel and subplot list
        self._reset_plot()
        panel_list = self.results['panel_list']
        self.figure = {}
        f, axes = plt.subplots( nrows=len(panel_list), ncols=3 )

        # Plot and format the figures
        for i, panel in enumerate( panel_list ):
            area, stype = panel.split('_')
            for j, plt_type in enumerate( ['edge', 'shuffle', 'bootstrap'] ):
                ax = axes[i][j]
                # The main function which generates the histogram
                self.plot_panel_paper( ax, panel, plt_type, **kwds )
                # Formatting the plots
                ax.tick_params(axis='both',labelsize=8,pad=3)
                if j == 0:
                    ax.set_ylabel( \
                        area+'-'+stype,fontsize=12,fontname='Arial',ha='center')
                    ax.yaxis.set_label_coords(-0.15,0.5)
                if i < 3:
                    ax.set_xticks([])
                else:
                    if j == 0:
                        ax.set_xlabel( 'Track Angle ($^\circ$)', \
                            fontname='Arial',fontsize=10,va='top' )
                    else:
                        ax.set_xlabel( 'Proportion',fontname='Arial',fontsize=10 )
                ymin, ymax = ax.get_ylim()
                ax.set_yticks( np.round(np.linspace(ymin,ymax,4),0).astype(int) )
                ax.set_yticklabels( \
                    np.round(np.linspace(ymin,ymax,4),0).astype(int), \
                        fontname='Arial' )
        plt.subplots_adjust( \
            left=0.08,right=0.97,bottom=0.1,top=0.98,wspace=0.2,hspace=0.15)
    # end of METHOD plot_paper 

    def plot_panel_paper(self,ax,panel,plt_type,one_side=True,multi_n=4,xlim=None,nbin=20):
        """
        For the specified panel plot out the distribution histogram, the bootstrap
        distribution, and the random shuffled distribution

        panel [string]: subarea-session
        plt_type ['shuffle'/'bootstrap'/'edge']: type of histogram
        (kwds in plot_paper:)
        one_side [True/False]: one-sided statistic test or not
        multi_n [integer]: number of multiple comparison
        xlim [array]: [xmin, xmax] for the simulated distributions
        nbin [integer]: number of spatial (angle) bins
        """

        # Load collected data
        if panel in self.results['data'].keys(): 
            data = self.results['data'][panel]
        else:
            print 'Data not found!'
            return

        # Calculate the expected edge proportion with homogeneous distribution hypothesis
        r_exp = (len(BD_ANGLES['local'])*2*MARGIN) / 360.0 
        obs = data['r_obs']

        if plt_type == 'shuffle': # plot the random shuffle distribution
            ratios = np.array( data['r_shuffle'] )
            plot_resample_distribution( \
                ax,obs,ratios,exp=r_exp,resample='permutation',nbin=nbin, \
                    xlim=xlim,one_side=one_side,multi_n=multi_n,show_obs_diff=True )
            if xlim is not None:
                ax.set_xticks( np.linspace(xlim[0],xlim[-1],3) )
                ax.set_xticklabels( np.linspace(xlim[0],xlim[-1],3), \
                    fontname='Arial',fontsize=8 )
            self.out( '%s:' % panel )
            self.out( '%d out of %d shuffle larger than observation' \
                % (np.sum(ratios>obs),len(ratios)) )
        elif plt_type == 'bootstrap': # plot the bootstrapped distribution
            ratios = np.array( data['r_bootstrap'] )
            plot_resample_distribution( \
                ax,obs,ratios,exp=r_exp,resample='bootstrap',nbin=nbin, \
                    xlim=xlim,one_side=one_side,multi_n=multi_n,show_obs_diff=True )
            if xlim is not None:
                ax.set_xticks( np.linspace(xlim[0],xlim[-1],3) )
                ax.set_xticklabels( np.linspace(xlim[0],xlim[-1],3), \
                    fontname='Arial',fontsize=8 )
        else: # plot the distribution of the field edges
            objs = data['edge']
            ax.hist(objs,color='w',bins=np.linspace(0,360,36+1))
            ax.set_xlim( [0, 360] )
            ax.set_xticks( np.linspace(0,360,4+1).astype(int) )
            ax.set_xticklabels( np.linspace(0,360,4+1).astype(int), \
                fontname='Arial',fontsize=8 )
            for angle in BD_ANGLES['local']: 
                ax.axvline(x=angle,lw=2,color='k')
            return
    # end of METHOD plot_panel_paper

    def _calc_ratio(self, obj_data, margin=MARGIN):
        """
        Calculate the within-window field edge proportion

        margin: the half width of the local-cue boundary window
        """
        n = len(obj_data)
        cues = np.array( BD_ANGLES['local'] )
        cues = np.r_[cues, cues+360, cues+720]

        in_cnt, out_cnt = 0, 0
        for data in obj_data:
            bd = False
            for angle in cues:
                if abs(data+360-angle) < margin:
                    in_cnt += 1
                    bd = True
                    break
            if not bd:
                out_cnt += 1
        if(in_cnt+out_cnt) != n:
            print ('Error!!')

        return float(in_cnt)/n
    # end of METHOD _calc_ratio

    def _reset_plot(self):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)
    # end of METHOD _reset_plot

# end of CLASS SpatialDistrAna

# --------------------------------------------------------------------------------

class FieldSpatialDistr(SpatialDistrAna):
    """
    Examine whether more field edges were observed near local cue boundaries
    Generate the place field edge (track angle, degree) distribution plots and 
    statistic tests (random shuffle, bootstrap)
    """

    datadir = 'FieldSpatialDistr'
    label = 'field spatial distr'

    def collect_data(self, N=1000):
        """
        Collect (rotated) field edges (in track angle, degree), shuffled data, 
        bootstrapped data, and the proportion of field edges observed within 
        the global/local cue window (the half window width defined by MARGIN)

        N: the numbers of simulations
        """
        self.results = {}

        table_file = tb.openFile('/home/paramecium/output/fields.h5','r')
        field_table = table_file.root.place_fields
        table_query = '(expt_type=="DR")&(exclude==False)'

        data = {}
        panel_list = []
        for area in ['CA1', 'CA3']:
            for ses in ['STD', 'MIS']:
                panel = area + '_' + ses
                self.out( 'Collecting field %s locations for %s:' % (obj,panel) )

                # collect qualified fields
                query = table_query + '&(area=="%s")&(type=="%s")' % (area,ses)
                index = collect_unique_index(ses, field_table, query, 'highest_mean')
                index.sort()

                clust_count = 0
                txt_file = open('./%s_%s.txt'%(area,ses),'w')
                for field in field_table[index]:
                    if field['field_num'] == 1:
                        txt_file.write( '%d %d %d %d %s\n' % \
                            (field['id'],field['rat'],field['day'],field['session'],field['tc']) )
                        clust_count += 1
                txt_file.close()

                n = len(index)
                if n == 0:
                    continue
                self.out('%d fields collected from %d cells' % (n,clust_count))

                # collect field edge locations (track angles, in degrees)
                objs = self._collect_bd( field_table, index, shuffle=False )
                r_obs = self._calc_ratio( objs )

                # collect shuffled data
                self.out( 'Shuffling field edges...' )
                shuffle, r_shuffle = [], []
                for i in xrange(N):
                    if np.mod(i,10) == 0:
                        sys.stdout.write('.')
                        sys.stdout.flush()
                    shuffled_edges = self._collect_bd(field_table,index,shuffle=True)
                    shuffle.append( shuffled_edges[:] )
                    r_shuffle.append( self._calc_ratio(shuffled_edges) )
                print('\n')

                # collect bootstrapped data
                self.out( 'Bootstrapping field edges...' )
                bootstrap, r_bootstrap = [], []
                new_index = np.random.randint(n, size=(N,n))
                for i, idx in enumerate(new_index):
                    if np.mod(i,10) == 0:
                        sys.stdout.write('.')
                        sys.stdout.flush()
                    ix = np.array(index)[idx]
                    bootstrapped_edges = self._collect_bd(field_table,ix,shuffle=False)
                    bootstrap.append( bootstrapped_edges[:] )
                    r_bootstrap.append( self._calc_ratio(bootstrapped_edges) )
                print('\n')

                data[panel] = {}
                data[panel][obj] = objs
                data[panel]['r_obs'] = r_obs
                data[panel]['shuffle'] = shuffle
                data[panel]['r_shuffle'] = r_shuffle 
                data[panel]['bootstrap'] = bootstrap 
                data[panel]['r_bootstrap'] = r_bootstrap

                panel_list.append( panel )

        table_file.close()

        self.results['N'] = N
        self.results['panel_list'] = panel_list
        self.results['data'] = data

        self.out('Done')
    # end of METHOD collect_data

    def _collect_bd(self, table, index, shuffle):
        """
        For each specified field collect its field edges in angles

        table: the table containing the field information
        index: the indices of selected fields
        shuffle: whether random shuffle the edges
        """
        random.seed()

        data = []
        for ix in index:
            field = table[ix]
            angle = np.array( field['field_edges'] )
            rotate = field['mis_angle']/2.0 if field['type']=='MIS' else 0
            sign = -1
            
            rnd = 0 if not shuffle else random.uniform(0,360)
            angle = (angle + (rotate*sign) + 360) % 360
            angle = np.mod( angle+rnd, 360 )

            for theta in angle:
                data.append( theta )

        return data
    # end of METHOD _collect_bd 

# end of CLASS FieldSpatialDistr

# --------------------------------------------------------------------------------
