# encoding: utf-8
"""
F4D.py

Generate the correlograms of the firing-rate map matrices of the 2D data
"""

import os
import itertools

import tables as tb
import numpy as np
import matplotlib.pyplot as plt

from core.analysis import AbstractAnalysis

from tdb.data import get_group
from tdb.tdb_misc import get_panel_clust_list
from tdb.ratemap import get_ratemap, smooth_ratemap

#==================================================================================

class SLCorrAna(AbstractAnalysis):
    """
    Generate the correlograms of the firing-rate map matrices collected from 
    different types of 2D platforms
    """
    datadir = './SLCorr'
    label = 'single line corr'

    def collect_data(self):
        """
        Collect the firing-rate map matrices for each 2D platform
        """
        self.results = {}

        # Set panel list
        platforms = [ 'leather', 'tape' ]
        stypes = [ 'STD', 'SHF' ]
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        # For each panel collect the firing-rate maps
        for panel in panels:
            # Collect filtered cluster information for the specified panel 
            self.out('Collecting data for %s...' % panel)
            clust_list = get_panel_clust_list( panel,min_quality='Fair',\
                repeat_filter='highest_mean',debug=False )
            # element in clust_list: (rat, day, maze, tc, threshold)
            clust_list = list( set(zip(*zip(*clust_list)[:-1])) )
            # element in clust_list: (rat, day, maze, tc)

            # Get the binning grids specific for this analysis
            grids = self.create_ratemap_grids(panel)

            # Calculate and collect the firing-rate map for each cluster
            H = None
            for clust in clust_list:
                # Load session data
                rat, day, maze, tc = clust
                rds = (rat, day, maze)
                session = get_group(rds=rds)
                spikes = session._v_children[tc]

                # Calculate and smooth rate map
                H_rate, mask = get_ratemap(session, spikes, rds, grids=grids)
                H_rate = smooth_ratemap(H_rate, mask, smooth='gaussian')
                H_rate[mask] = 0
                # Normalize rate map
                if H_rate.max() > 0: # field might be out of sample window
                    H_rate = H_rate / float(H_rate.max())

                # Since the cue boundary orientations are differnt on the leather
                # and the tape boards, we stack the rate maps along different
                # dimensions to construct the rate map matrix such that the following
                # calculation process could be consistent
                if 'leather' in panel:
                    H = H_rate.copy() if H is None else np.vstack( (H,H_rate) )
                else:
                    H = H_rate.copy() if H is None else np.hstack( (H,H_rate) )

            self.results[panel] = H if 'leather' in panel else H.transpose()

            # Check cluster and field list used in analysis
            rats = list( set([data[0] for data in clust_list]) )
            days = list( set([data[0:2] for data in clust_list]) )
            mazes = list( set([data[0:3] for data in clust_list]) )
            clusts = list( set([data[0:4] for data in clust_list]) )
            self.out( '%d rats, %d days, %d mazes, %d cells collected' \
                %(len(rats), len(days), len(mazes), len(clusts)) )

        self.out('Done')
    # end of METHOD collect_data

    def plot_paper(self):
        """
        Plot out the correlograms
        """
        # Clean up plot residues
        plt.close('all')
        plt_cnt = 1

        # For each panel calculate the correlogram and visualize it
        for i, platform in enumerate(['leather', 'tape']):
            for j, stype in enumerate(['STD', 'SHF']):
                # Load the firing-rate matrices and calculate the correlogram
                full_stype = 'Standard' if stype=='STD' else 'Shift'
                panel = '%s_%s' % (platform,stype)
                H = self.results[panel]
                C = np.corrcoef(H, rowvar=0)

                # Plot the correlogram
                ax = plt.subplot(2,2,plt_cnt)
                plt.imshow( C,interpolation='none',aspect=1.0,origin='lower' )

                # labeling the cue boundary locations
                bd = 15 if stype=='STD' else 25
                plt.axvline(bd,color='white',ls='--',lw=0.5)
                plt.axhline(bd,color='white',ls='--',lw=0.5)

                # Format the plots
                ax.set_xticks([])
                ax.set_yticks([])
                if i < 1:
                    ax.set_title('%s'%full_stype, fontname='Arial', fontsize=12)
                if j < 1:
                    ax.set_ylabel('%s'%platform.capitalize(), \
                        fontname='Arial', fontsize=12)

                plt_cnt += 1
        plt.subplots_adjust( \
            left=0.13,right=0.9,bottom=0.05,top=0.93,wspace=0.14,hspace=0.004)

        f = plt.gcf()
        f.set_size_inches(2.5,2.4,forward=True)
        f.savefig( os.path.join(self.datadir,'F4D.png'), dpi=1000 )

        self.out('Done')
    # end of METHOD plot_paper

    #------------------------------------------------------------------------------

    def create_ratemap_grids(self, panel):
        """
        Return the coefficients defining the cue boundary
        """

        consts = tuple( np.linspace(-170,166,43) )
        x_consts = consts[:]
        y_consts = consts[:]

        if 'leather' in panel:
            x_coef = (-4, 1) # -4x + y = k, the equation describes the cue boundary
            y_coef = (1, 4)  # x + 4y = k
        elif 'tape' in panel:
            x_coef = (4, 1)  # 4x + y = k
            y_coef = (-1, 4) # -x + 4y = k

        return (x_coef, x_consts, y_coef, y_consts) 
    # end of METHOD create_ratemap_grids
# end of CLASS SLCorrAna
