# encoding: utf-8
"""
F5.py

Plot out the distribution of place field edges and check whether the edges tend to
concentrate near the cue boundary
"""

import os
import sys
import itertools

import tables as tb
import numpy as np
import matplotlib.pyplot as plt

from scipy.ndimage.filters import gaussian_filter
from scipy.stats import norm

from core.analysis import AbstractAnalysis

from tdb.data import get_group
from tdb.td_maps import bin_edges, get_coordinate_range, plot_boundaries, zone_filter, plot_boundary_window
from tdb.tdb_misc import get_panel_clust_list
from tdb.field import label_place_fields

from analysis_shan.misc import plot_resample_distribution

#=================================================================================

class PopFieldAna(AbstractAnalysis):
    datadir = './PopField'
    label = 'population field'

    def collect_data(self, **traits):
        self.results = {}

        H = {} # population map (summation of field binary maps)
        B = {} # stack of field binary maps 
        N = {} # number of fields
        clust_list = {} # list of clusters

        # Set the board list
        platforms = ['leather', 'tape']
        stypes = ['STD', 'SHF']
        sl = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]
        panels = sl + ['plain-all',] # the name of the board for human read
        patterns = sl + ['TEX',] # key words defined in the database

        # For each board collect the clust information, field edge map, and the 
        # number of fields collected (variables as defined above)
        for panel, pattern in zip(panels, patterns):
            clust_list[panel], H[panel], B[panel], N[panel] = self.collect_panel(\
                panel, pattern, **traits )
            
        # Export data
        self.results['clust_list'] = clust_list
        self.results['H'] = H
        self.results['B'] = B
        self.results['N'] = N
        self.results['panels'] = panels
        self.results['patterns'] = patterns

        self.out('Done!')
    # end of METHOD collect_data

    def plot_paper_heatmap(self,control='plain-all',**traits):
        """
        Create the place field edge heat map

        control: if None plot the data collected from the textured board, 
                 otherwise plot the data collected from the specified control
                 board
        """

        # Clean up plot residues
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        self.figure = {}

        fg_cnt = 1
        for platform in ['leather', 'tape']:
            for stype in ['STD', 'SHF']:
                panel = platform + '_' + stype
                [y_edges, x_edges] = bin_edges(pattern=panel,**traits)

                # Plot the smoothed summation of the field edge map
                ax = plt.subplot(1,4,fg_cnt)
                panel_h = panel if control is None else control
                H = self.results['H'][panel_h].copy().astype(float)
                H = gaussian_filter(H,2.5)
                ax.imshow( H,origin='lower',interpolation='none', \
                    aspect='equal',cmap=cmap,\
                        extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )

                # Format figures and plot the outer edges and cue boundaries
                bd_c = 'k'
                combine_kw = traits.copy()
                combine_kw.update( **{'c':bd_c,'alpha':0.5,'lw':2} )
                plot_boundaries( pattern=panel, **combine_kw )

                bd_c = 'k'
                combine_kw = traits.copy()
                combine_kw.update( **{'c':bd_c,'alpha':0.5,'lw':1} )
                plot_boundary_window( pattern=panel, edge_margin=10, **combine_kw )

                stype_str = 'Standard' if stype=='STD' else 'Shift'
                title_str = platform.capitalize() + '-' + stype_str
                ax.set_title(title_str,fontsize=12,fontname='Arial')
                plt.axis( [x_edges[0], x_edges[-1], y_edges[0], y_edges[-1]] )
                ax.set_axis_off()
                ax.set_aspect('equal')

                fg_cnt += 1

        plt.subplots_adjust(left=0.06,right=0.97,bottom=0.15,top=0.93,wspace=0.3)

        f = plt.gcf()
        f.set_size_inches(4.4,1,forward=True)
        f.savefig( os.path.join(self.datadir,'F5A.png'),dpi=1000 )

        self.figure['F5A'] = plt.gcf()
    # end of METHOD plot_paper_heatmap

    def plot_paper_BPI(self,test_type='boot_strap',xlim=[-1,1]):
        """
        Plot out the simulation histograms

        test_type ['boot_strap'/'permutation']: type of simulation
        xlim [array]: (xmin, xmax) of the histogram
        """
        # Clean up plot residues
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        # Set panel list
        platforms = ['leather', 'tape']
        stypes = ['STD','SHF']
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        ttype = 'bootstrap' if test_type=='boot_strap' else 'permutation'

        for i, panel in enumerate(panels):
            ax = plt.subplot(1,4,i+1)
            
            # Load data
            if ttype == 'bootstrap':
                bs_ptr = self.results['boot_strap'][panel]
                obs_ctrl = bs_ptr[ctrl][ 'obs_plain-all' ]
                bs_ctrl = bs_ptr[ctrl][ 'bs_plain-all' ]
                obs_exp = bs_ptr[panel]['obs']
                bs_exp = bs_ptr[panel]['bs']
                obs = obs_exp - obs_ctrl
                rs = bs_exp - bs_ctrl
            elif ttype == 'permutation':
                ptr = self.results['shuffle'][panel]
                obs = ptr[ 'obs' ]
                rs = ptr[ 'shuffle' ]
            plot_resample_distribution( ax,obs,rs,exp=0,nbin=15,show_obs_diff=True,\
                resample=ttype,xlim=xlim,one_side=True,multi_n=4 )

            bins = np.linspace(xlim[0],xlim[1],3)
            ax.set_xticks(bins)
            ax.set_xticklabels(bins,fontname='Arial',fontsize=8)

            ymin, ymax = ax.get_ylim()
            bins = np.round(np.linspace(ymin,ymax,3),0).astype(int)
            ax.set_yticks(bins)
            ax.set_yticklabels(bins,fontname='Arial',fontsize=8)

        plt.subplots_adjust(left=0.06,right=0.97,bottom=0.15,top=0.93,wspace=0.3)

        f = plt.gcf()
        f.set_size_inches(4.4,1,forward=True)
        fn = 'F5B_' + ttype + '.png'
        f.savefig( os.path.join(self.datadir,fn),dpi=1000 )
    # end of METHOD plot_paper_BPI

    def run_bootstrap_test(self):
        """
        An interface to self.bootstrap() METHOD
        Run the bootstrap simulations and store the results
        """
        # Reset analysis log
        self.log = False
        self.log = True
        self.out.timestamp = False

        # Simple boards
        panels = [ panel for panel in self.results['panels'] \
            if not 'plain' in panel ]

        # calculate textured board BPI --> calculate plain board BPI --> BPI diff
        # ctrl_pair = ( ctrl_switch, ctrl_diff )
        # ctrl_switch [True/False]: calculate the BPI based on the ctrl board?
        # ctrl_diff [True/False]: calculate the BPI difference?
        ctrl_pairs = zip( [False,True,True], [False,False,True] )
        for i, bs_panel in enumerate(panels):
            for j, ctrl_pair in enumerate(ctrl_pairs):
                ctrl_switch, ctrl_diff = ctrl_pair
                self.bootstrap( bs_panel, ctrl=ctrl_switch,\
                    ctrl_diff=ctrl_diff, ax=ax, scale_plain=True )
    # end of METHOD run_bootstrap_test

    def run_permutation_test(self, ctrl, N=1000, nbin=40, **traits):
        """
        Run the permutation simulations and store the results
        """

        # Reset analysis log
        self.log = False
        self.log = True
        self.out.timestamp = False

        self.results['shuffle'] = {} 

        for platform in ['leather', 'tape']:
            for exp_type in ['STD', 'SHF']:
                # Set data storage structure
                panel = platform + '_' + exp_type
                self.out('%s sessions:' % panel)
                self.results['shuffle'][panel] = {} 
                ptr = self.results['shuffle'][panel]
                obs_str, shuffle_str, index_str = 'obs', 'shuffle', 'index'

                # Load the field edge maps
                H_obs = self.results['B'][panel].copy().astype(float)
                H_ctrl = self.results['B'][ctrl].copy().astype(float)
                H = np.dstack( (H_obs,H_ctrl) )
                N_obs = H_obs.shape[2]
                N_ctrl = H_ctrl.shape[2]
                index = np.array( range(N_obs+N_ctrl) )

                # Calculate the observed BPI difference 
                H_in = np.sum(H_obs,2)
                H_in = gaussian_filter(H_in,2.5)
                pi_obs = get_auc( H_in,panel,**traits )
                H_in = np.sum(H_ctrl,2)
                H_in = gaussian_filter(H_in,2.5)
                pi_ctrl = get_auc( H_in,panel,**traits )
                obs = pi_obs - pi_ctrl
                ptr[obs_str] = obs

                # Simulate the observed BPI difference with permutation test
                shuffle, index_rec = [], []
                for i in range(N):
                    shuffle_index = np.random.permutation(index)
                    H_in = np.sum(H[:,:,shuffle_index[:N_obs]],2)
                    H_in = gaussian_filter(H_in,2.5)
                    pi_obs = get_auc( H_in,panel,**traits )
                    H_in = np.sum(H[:,:,shuffle_index[N_obs:]],2)
                    H_in = gaussian_filter(H_in,2.5)
                    pi_ctrl = get_auc( H_in,panel,**traits )
                    shuffle.append( pi_obs-pi_ctrl )
                ptr[index_str] = index_rec[:]
                ptr[shuffle_str] = shuffle[:] 

                self.out('Observed preference index difference = %.4f' % obs)
                self.out('%d out of %d (p=%.4f) shuffle samples >= observed PI' \
                    % (np.sum(shuffle>=obs),N,np.sum(shuffle>=obs)/float(N)))

        self.save_data()

    # end of METHOD run_permutation_test

    #-----------------------------------------------------------------------------

    def collect_panel( self, panel, pattern, **traits):
        """
        For the specified board and surface pattern calculate the field edge map
        """

        # Collect cluster list for the specified board
        self.out('Collecting data for %s...' % panel)
        clust_list = get_panel_clust_list( \
            panel,min_quality='Fair',repeat_filter='highest_mean')
        if not len(clust_list):
            self.out('No data collected!')
            return None, 0
        
        # For each cluster calculate the place field edge
        self.out('Detecting place field contours...')
        N_field, H, H_bool = 0, None, None
        for i, clust in enumerate(clust_list):
            if np.mod(i,10) == 0:
                sys.stdout.write('.')
                sys.stdout.flush()
            rat, day, maze, tc, threshold = clust
            rds = (rat, day, maze)
            session = get_group(rds=rds)
            spikes = session._v_children[tc]
            # Detect place fields and edges
            field_num, labels, contours, n_edges, tmp = label_place_fields( \
                rds,session,spikes,threshold=threshold,turn_off_rim=False, \
                    smooth=True,smooth_kernel='gaussian',**traits )
            unify_map = (contours>1).astype(int) # overlapped field edge map

            # Stack the field edge map together
            if H is None:
                H = unify_map.copy()
                H_bool = np.zeros( H.shape, dtype=bool ) # H[:,:,0] = False matrix
            else:
                H += unify_map
            for num in range(field_num):
                H_bool = np.dstack( (H_bool,contours==num+1) )
            N_field += field_num
        H_bool = H_bool[:,:,1:] # remove the False matrix
        sys.stdout.write('\n')

        # Report counts of the collected data
        rats, days, mazes, tcs, tmp = zip(*clust_list)
        N_rat = len( list(set(rats)) )
        N_dataset = len( list(set(zip(rats,days))) )
        N_session = len( list(set(zip(rats,days,mazes))) )
        N_clust = len(clust_list)
        self.out('Data collected from %d rats, %d datasets, %d sessions, %d clusters and %d fields' % (N_rat,N_dataset,N_session,N_clust,N_field))

        return clust_list, H, H_bool, N_field
    # end of METHOD collect_panel

    def bootstrap(self, bs_panel, ctrl=False, ctrl_diff=False, nbin=30, **traits):
        """
        Collect the BPI and the BPI difference
        """

        if not ctrl:
            self.out('Bootstrap data for %s sessions...' % bs_panel)
        elif ctrl_diff:
            self.out('Bootstrap difference between %s and plain-all...' \
                % bs_panel )
        else:
            self.out('Bootstrap control data (plain-all) for %s sessions...' \
                % bs_panel )
        self.out('\n')

        obs_bs, bs_bs = self.get_bootstrap_data( bs_panel,bs_panel,**traits )
        if ctrl_panel is not None:
           obs_ctrl, bs_ctrl = self.get_bootstrap_data( \
                'plain-all',bs_panel,**traits )
        if ctrl_diff:
            obs = obs_bs - obs_ctrl
            bs = bs_bs - bs_ctrl
            x_min, x_max = min(min(obs,0),min(bs)), max(max(obs,0),max(bs))
        else:
            x_min = np.min( [0,obs_bs,obs_ctrl,bs_bs.min(),bs_ctrl.min()] )
            x_max = np.max( [0,obs_bs,obs_ctrl,bs_bs.max(),bs_ctrl.max()] )
            if ctrl:
                obs, bs = obs_ctrl, bs_ctrl
            else:
                obs, bs = obs_bs, bs_bs
    # end of METHOD bootstrap

    def get_bootstrap_data(self, panel, bs_panel, N=1000, **traits):
        """
        panel: the data being bootstrapped
        bs_panel: the surface pattern used to calculate BPI
        N: number of simulation sample
        """
        # Set data storage structure
        self.results['boot_strap'] = {} 
        self.results['boot_strap'][bs_panel] = {} 
        self.results['boot_strap'][bs_panel][panel] = {}
        bs_ptr = self.results['boot_strap'][bs_panel][panel]

        # Load field edge maps
        H_bool = self.results['B'][panel]
        length = H_bool.shape[2]

        if 'plain' in panel: # calculate BPI for control data 
            obs_str, bs_str, xmin, xmax, pattern, ctrl_pattern = \
                'obs_'+panel, 'bs_'+panel, 'min_'+panel, 'max_'+panel, \
                    bs_panel, 'square_PLA'
        else:
            obs_str, bs_str, xmin, xmax, pattern, ctrl_pattern = \
                'obs', 'bs', 'min', 'max', panel, None

        H = np.sum(H_bool,2)
        H = gaussian_filter(H,2.5)
        obs = get_auc( H,pattern,**traits )
        bs_ptr[obs_str] = obs

        # Calculate BPI
        index = np.random.choice(range(length),size=(N,length),replace=True)
        bs = []
        for idx in index:
            H = np.sum(H_bool[:,:,idx],2)
            H = gaussian_filter(H,2.5)
            bs.append( get_auc(H,pattern,**traits) )
        bs = np.array(bs)
        bs_ptr[bs_str] = bs
        bs_ptr['index'] = index

        self.save_data()

        return obs, bs 
    # end of METHOD get_bootstrap_data

# end of CLASS PopFieldEdgeAna

#=================================================================================

def get_rim_filter( map_shape, pattern, debug=False, edge_margin=0, **traits ):
    """
    return a boolean rim matrix which has out-of-arena bins filled with True
    """
    [y_edges, x_edges] = bin_edges(pattern=pattern,**traits)
    del_y, del_x = y_edges[1]-y_edges[0], x_edges[1]-x_edges[0]
    xmin, xmax, ymin, ymax = get_coordinate_range( \
        pattern=pattern, margin=0, **traits )
    rim_xmin = round( (xmin+edge_margin-x_edges[0])/float(del_x) )
    rim_xmax = round( (xmax-edge_margin-x_edges[0])/float(del_x) )
    rim_ymin = round( (ymin+edge_margin-y_edges[0])/float(del_y) )
    rim_ymax = round( (ymax-edge_margin-y_edges[0])/float(del_y) )

    H_rim = np.ones( map_shape ).astype(bool)
    H_rim[rim_ymin:rim_ymax+1,rim_xmin:rim_xmax+1] = False

    if debug:
        plt.imshow( H_rim,origin='lower',interpolation='none', \
            aspect='auto',cmap='gray_r',\
                extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )

    return H_rim
# end of FUNCTION get_rim_filter

def get_auc( H, pattern, debug=False, shuffle=False, scale_plain=False, absolute_diff=False, return_cdf=False, **traits ):
    H_rim = get_rim_filter( H.shape, pattern, scale_plain=scale_plain, **traits )

    cnt = [[], []]
    for bd in [True, False]:
        Z = zone_filter( pattern=pattern, bd=bd, scale_plain=scale_plain )
        F = (Z*~H_rim).astype(int)
        cnt[bd] = H[F.astype(bool)]

    Cmax = max( max(cnt[0]), max(cnt[1]) )
    bins = np.linspace( 0, Cmax+1, Cmax+2)
    diff = np.zeros( Cmax+1 )
    cdf = [ None, None ]
    for bd in [True, False]:
        n,nbin,patches = plt.hist(cnt[bd],bins=bins,cumulative=True,visible=False)
        n = np.array(n).astype(float) / float(n[-1])
        cdf[bd] = n
        diff = n - diff
    auc = np.sum(diff) / float(Cmax+1)

    if return_cdf:
        return cdf, auc
    else:
        return auc
# end of FUNCTION get_auc
