# encoding: utf-8
"""
F6.py
"""

import os
import itertools

import tables as tb
import numpy as np
import matplotlib.pyplot as plt

from scipy.stats import chi2_contingency, mannwhitneyu

from core.analysis import AbstractAnalysis

from analysis_shan.misc import star_to_significance_level, reset_plot, plot_resample_distribution

from tdb.data import get_group
from tdb.td_maps import bin_edges, get_boundary_line_params, zone_filter
from tdb.tdb_misc import get_panel_field_list
from tdb.field import label_place_fields
from boundary_paper.F5 import get_rim_filter

#=================================================================================

class CountQualifiedFieldAna(AbstractAnalysis):
    datadir = './CountQualField'
    label = 'count qualified field analysis'

    def collect_data(self,N=None,field_type=None,**traits):
        """
        Count the proportion of field with characteristic of interest and compare
        the proportion to the control board based on chi2 and permutation test

        N: Number of simulation sample
        field_type: Characteristics of interest
            'abutt': Whether the field is abutting the cue boundary
            'overlap': Whether the field overlap with the cue boundary window and
                       increase the within-window field edge density
            'not_cross': Whether the field did not cross the cue boundary
        """
        self.results = {}

        platforms = ['leather', 'tape']
        stypes = ['STD', 'SHF']
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        # Set field detection parameters based on type of intereset
        if field_type == 'abutt': # field abutting cue boundary
            min_r, min_d = 0.8, 10
        elif field_type == 'overlap': # field overlap with cue boundary window
            min_r, min_d = 0, 10 
        elif field_type == 'not_cross': # field did not cross cue boundary
            min_r, min_d = 0.8, np.inf 
        else:
            print 'Warning!'

        # collect (filtered) cell list for the plain board control
        self.out('Collecting cell list for control (plain-all) board...')
        ctrl_list = get_panel_field_list( 'plain-all', \
            min_quality='Fair', repeat_filter='highest_mean' )

        for panel in panels:
            self.results[panel] = {}

            # collect (filtered) cell list for the specified panel 
            self.out('Collecting data for %s sessions...' % panel)
            obs_list = get_panel_field_list( \
                panel, min_quality='Fair', repeat_filter='highest_mean')

            for control_data in [True, False]:
                qual = []
                field_list = ctrl_list if control_data else obs_list
                for rat, day, maze, tc, threshold, field_no in field_list:
                    qual.append( self._detect_qualified_field( \
                        panel,(rat,day,maze),tc,threshold,field_no, \
                            min_r=min_r,min_d=min_d,**traits ) )

                n_all = len(qual)
                if control_data:
                    ctrl = np.array(qual)
                    n_ctrl = n_all 
                    data_type = 'control'
                else:
                    obs = np.array(qual)
                    n_obs = n_all 
                    data_type = 'observation'

                n_qual = np.sum(qual)
                self.out( '%d out of %d fields qualified - %s (%s)' \
                    % (n_qual,n_all,field_type,data_type) )
                self.results[panel][data_type] = (n_qual, n_all, np.array(qual))

            cont_table = [ [np.sum(ctrl),np.sum(~ctrl)], \
                           [np.sum(obs),np.sum(~obs)] ]
            chi2, p, dof, ex = chi2_contingency(cont_table,correction=False)
            self.out('Observed vs. control: chi = %.4f, p = %.4f' % (chi2,p))

            if N is not None: # do permutation test
                self.out('start permutaion...')
                cat = np.r_[ctrl, obs]
                perm = []
                for i in range(N):
                    index = np.random.permutation(n_ctrl+n_obs)
                    # ((p-ctrl_qualified,p-ctrl_all), (p-obs_qualified,p-obs_all))
                    perm.append( ((np.sum(cat[index][:n_ctrl]),n_ctrl), \
                        (np.sum(cat[index][n_ctrl:]),n_obs)) )
                self.results[panel]['perm'] = perm

        self.results['panels'] = panels
        self.results['control'] = 'plain-all'
        self.results['N'] = N
        self.out('Done')
    # end of METHOD collect_data

    def plot_paper(self, nbin=10):
        reset_plot()

        panel_list = self.results['panels']
        bins = np.linspace(-0.4,0.4,3)
        f, axes = plt.subplots( nrows=1, ncols=len(panel_list) )
        for i, panel in enumerate(panel_list):
            platform, stype = panel.split('_')
            print stype
            stype = 'Shift' if stype=='SHF' else 'Standard'
            title_str = platform.capitalize() + '-' + stype
            ax = axes[i]

            data = self.results[panel]
            perm_diff = []
            for ctrl_data, obs_data in data['perm']:
                obs = float(obs_data[0]) / obs_data[1] 
                ctrl = float(ctrl_data[0]) / ctrl_data[1]
                diff = obs - ctrl
                perm_diff.append( diff )

            obs = float(data['observation'][0]) / data['observation'][1] 
            ctrl = float(data['control'][0]) / data['control'][1]
            obs_diff = obs - ctrl

            plot_resample_distribution( ax,obs_diff,perm_diff,exp=0,nbin=15,\
                resample='permutation',xlim=[-0.4,0.4],one_side=False,multi_n=4 )
            ax.tick_params(axis='both',labelsize=8,pad=3)
            ax.set_title(title_str,fontsize=12,fontname='Arial',va='bottom')
            ax.set_xticks(bins)
            ax.set_xticklabels(bins,fontname='Arial')
            ax.set_yticks( np.linspace(0,300,4) )
            ax.set_yticklabels(np.linspace(0,300,4).astype(int),fontname='Arial')
            ax.set_ylim( [0,330] )
        plt.subplots_adjust(left=0.05,right=0.97,bottom=0.1,top=0.8,wspace=0.25)

        #f = plt.gcf()
        f.set_size_inches(5.6,1.6,forward=True)
        f.savefig( os.path.join(self.datadir,'F6B.png'), dpi=1000)
    # end of METHOD plot_paper

    def _detect_qualified_field(self,panel,rds,tc,threshold,field_no,min_r=0.8,min_d=5,params_out=False,debug=False,**traits):
        a, b, c = get_boundary_line_params(pattern=panel)

        session = get_group(rds=rds)
        spikes = session._v_children[tc]

        [y_edges, x_edges] = bin_edges(rds,**traits)
        x_centers = (x_edges[:-1]+x_edges[1:]) / 2.0
        y_centers = (y_edges[:-1]+y_edges[1:]) / 2.0
        Hx, Hy = np.meshgrid(x_centers, y_centers)

        field_num, labels, contours, n_edges, threshold = label_place_fields( \
            rds,session,spikes,threshold=threshold,turn_off_rim=False,\
                smooth=True,smooth_kernel='gaussian',**traits )

        # calculate the larger area ratio of a field on one side of the cue boundary
        select = labels==field_no
        xy = zip( Hx[select], Hy[select] )
        values = np.array( [a*x+b*y+c for x, y in xy] )
        pos = np.sum(values>0)+np.sum(values==0)/2.0 # no of pixels on positive side
        ratio = float(pos) / len(xy) 
        ratio = max( ratio, 1-ratio) # reverse ratio if pos is the smaller side

        # calculate the shortest distance from the field to the cue boundary
        select = contours==field_no
        min_dist = min( [abs(a*x+b*y+c)/float(np.sqrt(a*a+b*b)) \
            for x, y in zip(Hx[select],Hy[select])] )

        if debug:
            return ratio, min_dist, labels, contours

        if (min_dist<=min_d) and (ratio>=min_r):
            if params_out:
                return True, ratio, min_dist
            else:
                return True
        else:
            if params_out:
                return False, ratio, min_dist
            else:
                return False
    # end of METHOD _detect_qualified_field

# end of CLASS CountQualifiedFieldAna

#=================================================================================

class OverlapContourAna(AbstractAnalysis):
    datadir = './OverlapContour'
    label = 'overlap contour analysis'

    def collect_data(self,N=None,**traits):
        """
        Calcualte the contour length (in pixel) within the cue-boundary-window
        and compare the average length to the control board data based on 
        permutation test

        N: Number of simulation sample
        """
        self.results = {}

        platforms = ['leather', 'tape']
        stypes = ['STD', 'SHF']
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        # collect (filtered) cell list for the plain board control
        self.out('Collecting cell list for control (plain-all) board...')
        ctrl_list = get_panel_field_list( \
            'plain-all', min_quality='Fair', repeat_filter='highest_mean' )

        for panel in panels:
            self.results[panel] = {}

            # collect (filtered) cell list for the specified panel
            self.out('Collecting data for %s sessions...' % panel)
            obs_list = get_panel_field_list( \
                panel, min_quality='Fair', repeat_filter='highest_mean')

            for control_data in [True, False]:
                oc = [] # overlapped contour length
                field_list = ctrl_list if control_data else obs_list
                for rat, day, maze, tc, threshold, field_no in field_list:
                    oc.append( self._overlap_contour( panel,(rat,day,maze), \
                        tc,threshold,field_no,**traits ) )
                oc = np.array(oc)
                oc = oc[oc>0]

                if control_data:
                    ctrl = oc
                    data_type = 'control'
                else:
                    obs = oc
                    data_type = 'observation'
                self.results[panel][data_type] = oc
                self.out('%d fields collected (%s)' % (len(oc),data_type))

            u, p = mannwhitneyu(obs, ctrl)
            self.out('Observed vs. control: U = %.4f, p = %.4f' % (u,p))

            if N is not None: # do permutation test
                self.out('start permutaion...')
                cat = np.r_[ctrl, obs]
                perm = []
                for i in range(N):
                    index = np.random.permutation( len(cat) )
                    perm.append( (cat[index[:len(ctrl)]],cat[index[len(ctrl):]]) )
                    pdiff = np.nanmean(perm[-1][1]) - np.nanmean(perm[-1][0])
                    txt_file.write('%.3f '%pdiff)
                self.results[panel]['perm'] = perm
            txt_file.write('\r\n')

            txt_file.close()

        self.results['panels'] = panels
        self.results['control'] = 'plain-all'
        self.results['N'] = N
        self.out('Done')
    # end of METHOD collect_data

    def plot_paper(self, nbin=10):
        """
        Plot out the histograms
        """
        reset_plot()
        panel_list = self.results['panels']

        # Plot the distributions of overlapped bin number
        bins = np.linspace(0,90,nbin+1).astype(int)
        f, axes = plt.subplots( nrows=1, ncols=len(panel_list) )
        for i, panel in enumerate(panel_list):
            platform, stype = panel.split('_')
            stype = 'Shift' if stype=='shf' else 'Standard'
            title_str = platform.capitalize() + '-' + stype
            ax = axes[i]
            data = [ self.results[panel][data_type] \
                for data_type in ['observation', 'control'] ]
            data[0] = data[0][~np.isnan(data[0])]
            data[1] = data[1][~np.isnan(data[1])]

            ax.hist(data,color=['k','w'],bins=bins,normed=norm)
            ax.axvline(x=np.median(data[0]),lw=1,c='k',alpha=0.5,zorder=99)
            ax.axvline(x=np.median(data[1]),lw=1,c='k',alpha=0.5,ls='--',zorder=99)
            ax.tick_params(axis='both',labelsize=8,pad=3)
            ax.set_title(title_str,fontsize=12,fontname='Arial',va='bottom')
            ax.set_xlim( [bins[0], bins[-1]] )
            ax.set_xticks( np.linspace(bins[0],bins[-1],3) )
            ax.set_xticklabels( np.linspace(bins[0],bins[-1],3).astype(int), \
                fontname='Arial' )
            ax.set_yticks( np.linspace(0,24,5) )
            ax.set_yticklabels( np.linspace(0,24,5).astype(int), fontname='Arial' )
        plt.subplots_adjust(left=0.05,right=0.97,bottom=0.1,top=0.8,wspace=0.3)

        f.set_size_inches(5.6,1.6,forward=True)
        f.savefig( os.path.join(self.datadir,'F6C_oc.png'),dpi=1000)
        
        # Plot the permutation distribution
        reset_plot()
        bins = np.linspace(-15,15,3)
        f, axes = plt.subplots( nrows=1, ncols=len(panel_list) )
        for i, panel in enumerate(panel_list):
            platform, stype = panel.split('_')
            stype = 'Shift' if stype=='shf' else 'Standard'
            title_str = platform.capitalize() + '-' + stype
            ax = axes[i]

            data = self.results[panel]
            perm_diff = [ np.nanmean(obs)-np.nanmean(ctrl) \
                for ctrl, obs in data['perm'] ]
            obs_diff = np.nanmean(data['observation']) - np.nanmean(data['control'])

            plot_resample_distribution( ax,obs_diff,perm_diff,exp=0,nbin=15,\
                resample='permutation',xlim=[-15,15],one_side=False,multi_n=4 )
            ax.tick_params(axis='both',labelsize=8,pad=3)
            ax.set_title(title_str,fontsize=12,fontname='Arial',va='bottom')
            ax.set_xticks(bins)
            ax.set_xticklabels(bins.astype(int),fontname='Arial')
            ax.set_yticks( np.linspace(0,450,5) )
            ax.set_yticklabels( np.linspace(0,450,5).astype(int),fontname='Arial' )
        plt.subplots_adjust(left=0.05,right=0.97,bottom=0.1,top=0.8,wspace=0.3)

        f.set_size_inches(5.6,1.6,forward=True)
        f.savefig( os.path.join(self.datadir,'F6C_permutation.png'),dpi=1000)
    # end of METHOD plot_paper

    def _overlap_contour(self,panel,rds,tc,threshold,field_no,hw=10,**traits):
        session = get_group(rds=rds)
        spikes = session._v_children[tc]
        field_num, labels, contours, n_edges, threshold = label_place_fields( \
            rds,session,spikes,threshold=threshold,turn_off_rim=False,\
                smooth=True,smooth_kernel='gaussian',**traits )
        select = contours==field_no

        H_rim = get_rim_filter(select.shape,panel,edge_margin=10,**traits)

        F_b = zone_filter(pattern=panel,bd=True,hw=10,**traits)
        return np.sum( (select*F_b)*(~H_rim) )
    # end of METHOD _overlap_contour

# end of CLASS OverlapContourdAna

#=================================================================================
