# encoding: utf-8
"""
F7.py

Calculate the directions at which the neural representation of location decorrelate the most,
and visualize the directions with color map.
The distributions of the directions within the cue-boundary window and outside of the
cue-boundary window are being compared.
"""

import os
import sys
import itertools

import tables as tb
import numpy as np
import matplotlib.pyplot as plt

from matplotlib.colors import rgb_to_hsv, hsv_to_rgb
from matplotlib.ticker import FormatStrFormatter

from core.report import BaseReport
from tools.images import masked_array_to_rgba

from tdb.data import get_tdb_file, get_group
from tdb.ratemap import get_ratemap, smooth_ratemap
from tdb.td_maps import bin_edges, get_boundary_line, plot_boundaries, zone_filter, plot_boundary_window
from tdb.tdb_misc import get_panel_clust_list
from tdb.population_field import get_rim_filter
from analysis_shan.misc import plot_resample_distribution

#--------------------------------------------------------------------------------

class NeuralSpaceGradReport(BaseReport):
    datadir = './NSpaceGrad'
    label = 'neural space grad'

    ncols = 4
    nrows = 6

    def collect_data(self,N=1000):
        """
        Construct the direction of change map and calculate the related statistics
        for boundary vs. non-boundary comparisons

        N: the number of bootstrap sample
        """
        self.results = {}

        # Set panel list
        platforms = [ 'leather', 'tape' ]
        stypes = [ 'STD', 'SHF' ]
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]
        panels += ['plain-all',]

        # For each panel (board) list collect the firing-rate maps and compute
        # the directions of change and the statistics
        for panel in panels:
            self.out('Collecting data for %s...' % panel)
            self.results[panel] = {}

            # Collect firing-rate maps from the selected clusters
            H, M, n = None, None, 0 # rate map, not enough occ mask, number of map
            clust_list = get_panel_clust_list(panel,min_quality='Fair',\
                repeat_filter='highest_mean')
            for clust in clust_list:
                if np.mod(n,10) == 0:
                    sys.stdout.write('.')
                    sys.stdout.flush()

                rat, day, maze, tc, threshold = clust # threhold not used
                rds = (rat,day,maze)
                session = get_group(rds=rds)
                spikes = session._v_children[tc]
                H_rate, mask = get_ratemap(session, spikes, rds)
                H_rate = smooth_ratemap(H_rate, mask, smooth='gaussian')
                H_rate /= H_rate.max()

                H = H_rate.copy() if H is None else np.dstack( (H,H_rate) )
                M = mask.copy().astype(int) if M is None else M+mask.astype(int)
                n += 1
            sys.stdout.write('\n')

            # Calculate the related parameters and run the bootstrap test
            if (N is not None) and ('plain' not in panel): # doing bootstrap
                self.out('Bootstrap data')
                obs_pair = self._get_vector_stats( panel,H,exclude_rim=True )

                cnt = 0
                # r: mean vector length, p: perpendicular proportion
                r_pairs, x_pairs, t_pairs, p_pairs = [], [], [], []
                index = np.random.choice(range(n),size=(N,n),replace=True)
                for idx in index:
                    if np.mod(cnt,10) == 0:
                        sys.stdout.write('.')
                        sys.stdout.flush()
                    H_bs = H[:,:,idx]
                    r_pair, x_pair, t_pair, p_pair = self._get_vector_stats( \
                        panel,H_bs,exclude_rim=exclude_rim )
                    r_pairs.append( r_pair )
                    x_pairs.append( x_pair )
                    t_pairs.append( t_pair )
                    p_pairs.append( p_pair )
                    cnt += 1
                sys.stdout.write('\n')

                # Store data
                self.results[panel]['obs'] = {}
                self.results[panel]['obs']['r'] = obs_pair[0]
                self.results[panel]['obs']['x'] = obs_pair[1]
                self.results[panel]['obs']['t'] = obs_pair[2]
                self.results[panel]['obs']['p'] = obs_pair[3]
                self.results[panel]['bootstrap'] = {}
                self.results[panel]['bootstrap']['index'] = index
                self.results[panel]['bootstrap']['r'] = r_pairs
                self.results[panel]['bootstrap']['x'] = x_pairs
                self.results[panel]['bootstrap']['t'] = t_pairs
                self.results[panel]['bootstrap']['p'] = p_pairs

            self.results[panel]['H'] = H.copy()
            self.results[panel]['M'] = M.copy()
            self.results[panel]['clust_list'] = clust_list

            # Count the numbers of data used
            rats = list( set([data[0] for data in clust_list]) )
            days = list( set([data[:2] for data in clust_list]) )
            mazes = list( set([data[:3] for data in clust_list]) )
            clusts = list( set([data[:4] for data in clust_list]) )
            self.out( '%d rats, %d days, %d mazes, %d cells collected' \
                %(len(rats), len(days), len(mazes), len(clusts)) )

        self.out('Done')
    # end of METHOD collect_data

    def plot_paper(self,nbin=20,xlim=None,r_max=1,**traits):
        """
        Plot out the figures

        nbin: histogram number of bins
        xlim: histogram (xmin, xmax)
        r_max: maximum r value for the polar plot
        """

        # Clean up figure residues
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        self.figure = {}

        platforms = [ 'leather', 'tape' ]
        stypes = [ 'STD', 'SHF' ]

        # For each board output the plot
        for i, item in enumerate( itertools.product(platforms,stypes) ):
            platform, stype = item
            stype_str = 'Standard' if stype=='STD' else 'Shift'
            title_str = platform.capitalize() + '\n' + stype_str
            panel = '%s_%s' % (platform,stype)

            data = self.results[panel]

            # Get the parameters describing the location and orientation of the
            # cue boundary and the rim of the board
            bd_line, theta = get_boundary_line( pattern=panel )
            [y_edges, x_edges] = bin_edges(pattern=panel,**traits)
            extent = [ x_edges[0]+20,x_edges[-1]-20,y_edges[0]+20,y_edges[-1]-20 ]

            # Calculate the direction of change map
            gamma, amp = self.get_nspace_grad( data['H'], theta )
            mask = gamma < 0
            gim = masked_array_to_rgba( gamma,mask=mask,mask_color='k',cmap='hsv', \
                norm=False,cmin=theta,cmax=theta+180 )

            # Remove the outer rim too closed to the rim of the board
            H_rim = get_rim_filter(gamma.shape,panel,edge_margin=10,**traits)
            gamma[H_rim] = -1 

            # Plot the heat map
            ax = plt.subplot(4,4,4*i+1)
            plt.imshow( gim,origin='lower',interpolation='nearest', \
                aspect='equal',extent=extent )
            bd_traits = traits.copy() 
            bd_traits.update( {'c':'k','alpha':0.5,'lw':2} )
            plot_boundaries( pattern=panel, **bd_traits )
            bd_traits = traits.copy()
            bd_traits.update( {'c':'k','alpha':0.9,'lw':1} )
            plot_boundary_window( pattern=panel,edge_margin=10,**bd_traits )
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_ylabel( \
                title_str,fontname='Arial',fontsize=12,visible=True,ha='center' )

            # Plot the polar plot
            ax = plt.subplot(4,4,4*i+2,polar=True)
            colors = ['r', 'b']
            for j, in_window in enumerate([True,False]):
                F = zone_filter(pattern=panel,bd=in_window,hw=10,**traits) >= 0.5
                angles = (gamma[F*~H_rim]-theta) / 180.0 # normed directions [0,1]
                rads, n = angles*2*np.pi, float(len(angles))
                h, edges = np.histogram( \
                    angles,bins=np.linspace(0,1,20+1),density=False)
                h = h.astype(float) / n
                x = ((edges[:-1]+edges[1:])/2.0) * 2*np.pi 
                ax.plot(np.r_[x,x[0]],np.r_[h,h[0]],colors[j])
                ax.set_rmax(r_max)
                ax.set_yticks( np.linspace(0,r_max,3)[1:] )
                ax.set_yticklabels( np.linspace(0,r_max,3)[1:],fontname='Arial' )
                ax.yaxis.set_major_formatter(FormatStrFormatter('%.2f'))
                ax.set_thetagrids( angles=np.linspace(0,1,9)[:-1:2]*360, \
                    labels=['0$^\circ$','45$^\circ$','90$^\circ$','135$^\circ$'], \
                        frac=1.3)
                plt.tick_params(axis='both',labelsize=8)

            # Plot the mvl and proportion distribution 
            for j, stat in enumerate( ['r', 'p'] ):
                ax = plt.subplot(4,4,4*i+3+j)

                a, b = data['obs'][stat]
                obs_diff = a - b
                bs_diff = [a-b for a, b in data['bootstrap'][stat]]

                plot_resample_distribution( ax,obs_diff,bs_diff,exp=0,\
                    resample='bootstrap',nbin=nbin,show_obs_diff=True,\
                        xlim=xlim,one_side=False,multi_n=4,bs_cv='centered' )
                plt.tick_params(axis='both',labelsize=8)
                xvals, yvals = ax.get_xlim(), ax.get_ylim()
                ax.set_aspect(0.9*((xvals[1]-xvals[0])/float(yvals[1]-yvals[0])))
                if i == 3:
                    ax.set_xticks(np.linspace(xvals[0],xvals[1],3))
                    ax.set_xticklabels( \
                        np.linspace(xvals[0],xvals[1],3),fontname='Arial' )
                else:
                    ax.set_xticks([])
                ax.set_yticks( \
                    np.round(np.linspace(yvals[0],yvals[1],3),0).astype(int) )
                ax.set_yticklabels( \
                    np.round(np.linspace(yvals[0],yvals[1],3),0).astype(int),\
                        fontname='Arial' )
        plt.subplots_adjust( \
            left=0.08,right=0.95,bottom=0.05,top=0.98,wspace=0.07,hspace=0.1 )
        f = plt.gcf()
        allaxes = f.get_axes()
        for i in range(4):
            ax = allaxes[i*4+1]
            bb = ax.get_position()
            ax.set_position( \
                pos=[ bb.xmin+0.15*bb.width,bb.ymin+0.15*bb.height,\
                      0.7*bb.width,0.7*bb.height ] )

        f.set_size_inches(7,5,forward=True)
        f.savefig( os.path.join(self.datadir,'F7.png'), dpi=1000 )

        self.out('Done')
    # end of METHOD plot_paper

    #------------------------------------------------------------------------------

    def _get_vector_stats(self,panel,H,ctrl=None,exclude_rim=False,**traits):
        center, half_width = 90.0/180.0, 15.0/180.0
        
        pattern = 'square_PLA' if 'plain' in panel else panel
        if ctrl:
            bd_line, theta = get_boundary_line( pattern=ctrl )
        elif 'plain' in panel:
            bd_line, theta = None, 0
        else:
            bd_line, theta = get_boundary_line( pattern=pattern )
        [y_edges, x_edges] = bin_edges(pattern=pattern,**traits)
        extent = [ x_edges[0], x_edges[-1], y_edges[0], y_edges[-1] ]

        gamma, amp = self.get_nspace_grad( H, theta )
        H_rim = np.zeros(gamma.shape).astype(bool) # no rim bins
        if exclude_rim:
            H_rim = get_rim_filter(gamma.shape,pattern,edge_margin=10,**traits)

        mvl, x, t, bin_prop = [], [], [], []
        for in_window in [True, False]:
            if ctrl: 
                F = zone_filter(pattern=ctrl,bd=in_window,hw=10,**traits) >= 0.5
            else:
                F = zone_filter(pattern=pattern,bd=in_window,hw=10,**traits) >= 0.5

            angles = (gamma[F*~H_rim]-theta) / 180.0 # normed directions [0,1]

            cnt = np.sum( abs(angles-center)<half_width ) # perpendicular or nearly 
            proportion = float(cnt) / len(angles)

            # calculate mean vector length difference and 90-component difference
            rads, n = angles*2*np.pi, float(len(angles))
            X = np.sum( np.cos(rads) ) / n
            Y = np.sum( np.sin(rads) ) / n
            r = np.sqrt( X**2 + Y**2 )
            
            if X == 0:
                if Y == 0:
                    T = np.nan
                elif Y > 0:
                    T = np.pi / 2.0
                else:
                    T = np.pi * 3 / 2.0
            if X < 0:
                T = np.arctan( float(Y)/X ) + np.pi
            else:
                if Y >= 0:
                    T = np.arctan( float(Y)/X )
                else:
                    T = np.arctan( float(Y)/X ) + 2*np.pi
            T = T / (2.0*np.pi) * 180 # [0, 180]

            mvl.append(r)
            x.append( r*np.cos((90-T)*np.pi/180.0) )
            t.append(T)
            bin_prop.append(proportion)

        return mvl, x, t, bin_prop
    # end of METHOD _get_vector_stats

    def get_nspace_grad(self, H, theta):
        """
        Calculate the direction of change map
        """
        m, n, d = H.shape

        Dx, Dy = None, None
        for i in range(d):
            dy, dx = np.gradient( H[:,:,i] ) # derivatives of the i_th cluster
            Dx = dx if Dx is None else np.dstack( (Dx,dx) )
            Dy = dy if Dy is None else np.dstack( (Dy,dy) )
        # Jacobian[x_0, y_0] = [ Dx[x_0,y_0,:] Dy[x_0,y_0,:] ]

        gamma, amp = np.zeros((m,n)), np.zeros((m,n))
        for x in range(n):
            for y in range(m):
                j_x, j_y = Dx[y,x,:], Dy[y,x,:]
                j_x_square, j_y_square = np.dot(j_x,j_x), np.dot(j_y,j_y)
                A = ( j_x_square - j_y_square ) * 0.5
                B = np.dot( j_x, j_y )
                if A==0 and B==0:
                    # amp == sqrt(jx^2) == sqrt(jy^2)
                    gamma[y,x], amp[y,x] = -1, np.sqrt(j_x_square)
                else:
                    sin_alpha = float(A) / np.sqrt(A**2+B**2)
                    beta = np.arcsin(sin_alpha) # alpha = beta or pi-beta
                    angles = [ np.pi/4.0-beta/2.0, np.pi*5/4.0-beta/2.0, \
                               np.pi*3/4.0+beta/2.0, np.pi*7/4.0+beta/2.0 ]
                    max_dist, omega = 0, -1
                    for angle in angles:
                        dist = np.linalg.norm( np.cos(angle)*j_x+np.sin(angle)*j_y )
                        if dist > max_dist:
                            omega = (angle*180.0/np.pi+360) % 360 # 0 <= omega < 360
                            max_dist = dist
                    if (omega<theta) or (omega>theta+180):
                        omega = (omega+180) % 360
                    gamma[y,x], amp[y,x] = omega, max_dist

        return gamma, amp
    # end of METHOD get_nspace_grad
# end of CLASS NeuralSpaceGradReport

#--------------------------------------------------------------------------------
