# encoding: utf-8
"""
derivative_map.py
"""

import tables as tb
import numpy as np
import matplotlib.pylab as plt

from random import randint
from scipy.interpolate import interp1d

from core.analysis import AbstractAnalysis
from core.report import BaseReport
from tools.misc import memoize

from analysis_shan.texture_boundary import ModifiedSessionData
from analysis_shan.data_filters import latest_qualified_day

from ldb.stack_ratemaps import RmapStackAna

# should be consistant with PlaceFieldReport
BINS = 360 # need to fix pop_grad_shuffle_test if not 360
BLUR_WIDTH = 4.3
BD_ANGLES = { 'local':[45, 135, 225, 315], 'global':[0, 45, 135, 180, 270] }
MARGIN = 15 # in degree

PF_FILE = '/home/paramecium/output/fields.h5'

# ================================================================================

def run_field_edge_vs_derivatives(**traits):
    for area in ['CA1', 'CA3']:
        for stype in ['STD', 'MIS']:
            field_edge_vs_derivatives(area=area,stype=stype,**traits)
# end of FUNCTION run_field_edge_vs_derivatives

def field_edge_vs_derivatives(area=None,stype=None,map_type='rate',ext=10,n=400,norm=True):
    """
    Check the average onset/offset map normalized by place field span 

    map_type: ['rate', 'binary']
        use the rectified second order derivative map or the binary map
    ext: integer between 0 and 100
        extend %ext percent of the field on both sides
          ex: ext=10, ext_start = start + 10% of field width
                      ext_end = end - 10% of field width
              the extended field would be 20% larger wider
    n: positive integer
        map bin number
    norm: [True, False]
        whether normalize map by the peak value
    """
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    table_file = tb.openFile(PF_FILE,'r')
    field_table = table_file.root.place_fields

    table_query = '(expt_type=="DR")&(exclude==False)'
    fname_str = 'check_derivatives'
    if area is not None:
        table_query += '&(area=="%s")' % area
        fname_str += '_%s' % area
    if stype is not None:
        table_query += '&(type=="%s")' % stype
        fname_str += '_%s' % stype 

    rate, d1, onset, offset = np.zeros(n), np.zeros(n), np.zeros(n), np.zeros(n)

    cnt = 0
    for field in field_table.where( table_query ):
        rat, day, maze = field['rat'], field['day'], field['session']
        if day > latest_qualified_day(rat,"DR"):
            continue

        end, start = field['field_edges']
        ext_start = (start + ext/100.0 * ((start-end+360)%360)) % 360
        ext_end = (end - ext/100.0 * ((start-end+360)%360) + 360) % 360
        if ext_start < ext_end: # wrap around if field cross 0/360
            ext_start += 360

        buff_start = np.floor(ext_start) + 1
        buff_end = np.floor(ext_end-1+360) % 360
        if ext_start < buff_end:
            ext_start += 360
            buff_start += 360
        
        R = get_1d_ratemap( (rat,day,maze), field['tc'], 0) # ratemap not rotated
        R = np.r_[R, R] # for field crossing 0/360

        r = R[buff_end:buff_start]
        dr = r[:-1] - r[1:]
        ddr = dr[:-1] - dr[1:]

        r = r[1:-1]
        dr = dr[:-1]

        on_index = (dr>0) * (ddr>0)
        off_index = (dr<0) * (ddr>0)
        if map_type == 'rate':
            field_on = np.abs( ddr )
            field_on[~on_index] = 0
            field_off = np.abs( ddr )
            field_off[~off_index] = 0
        elif map_type == 'binary':
            field_on = on_index.astype(int)
            field_off = off_index.astype(int)

        rate += norm_feature( r, n=n, kind='linear', norm=norm)
        d1 += norm_feature( dr, n=n, kind='linear', norm=norm )
        onset += norm_feature( field_on, n=n, kind='nearest', norm=norm )
        offset += norm_feature( field_off, n=n, kind='nearest', norm=norm )
        cnt += 1

    features = [rate, d1, onset, offset]
    titles = [ 'Firing Rate (r)', 'r\'', \
        'Increasing firing rate increment', 'Decreasing firing rate decrement' ]
    fig, axes = plt.subplots(4,1,figsize=(4,8))
    for ax, feature, title in zip( axes, features, titles):
        ax.plot(feature/float(cnt))
        if title == 'Firing Rate (r)':
            ax.axvline(x=n*ext/100.0,c='r',lw=1)
            ax.axvline(x=n-n*ext/100.0,c='r',lw=1)
        if title == 'r\'':
            ax.axhline(y=0,c='k',lw=1)
        ax.xaxis.set_ticklabels([])
        ax.tick_params(axis='y',labelsize=7)
        ax.set_xlim( [0, n] )
        ax.set_title(title)
    fig.suptitle('Population Map Features',size='large')

    plt.savefig('./%s.png' % fname_str)

    table_file.close()

# end of FUNCTION field_edge_vs_derivatives

def norm_feature( v, n=360, kind='linear', norm=True):
    v_interp = interp1d( np.arange(v.size), v, kind=kind )
    norm_v = v_interp( np.linspace(0,v.size-1,n) )
    if norm:
        if norm_v.max() > 0:
            norm_v /= norm_v.max()

    return norm_v
# end of FUNCTION norm_feature

# ================================================================================

def run_population_demap_panels(**traits):
    for area in ['CA1', 'CA3']:
        for rot in ['local', 'global']:
            for ft in ['onset', 'offset']:
                population_demap( area=area,rot=rot,ft=ft,**traits)
# end of FUNCTION run_population_demap_panels

def population_demap(area='CA1',rot='local',ft='onset',plot='curve',set_vmin=False,**ft_params):
    """
    Plot the derivative heat maps for STD and MIS sessions separately, and for
    each different MIS angle

    ft: ['onset', 'offset']
        derivative features, onset or offset
    rot: ['local', 'global']
        rotate the MIS data following the specified cues
    plot: ['ring', 'curve']
        display the population derivative map in ring plots or curves 
    set_vmin : [True, False]
        set the minimum value to be zero for display purpose if TRUE 

    ft_params:
    bins: integer
        map size (number of bins)
    scale: ['one', 'cell', 'bin_count']
        'one' - rescale the population map to [0,1]
        'cell' - average by cell number
        'bin_count' - normed by occupancy map
    win_type: ['all', 'edge', 'nonfield', 'field']:
        'all' - include all data
        'edge' - extract [-win_hw, win_hw] data centered at speciried field edge
        'nonfield' - ~'field'
        'field' - extract data within field plus win_hw margins on both side
    win_hw: integer
        half window width in number of bins
    map_type: ['rate', 'binary']
        returns the rectified second order derivative map or the binary map
    """
    
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    stypes = ['STD','MIS','MIS','MIS','MIS','MIS']
    mis_angles = [None, None, 45, 90, 135, 180]

    for i, panel in enumerate( zip(stypes,mis_angles) ):
        stype, mis_angle = panel
        title_str = stype if mis_angle is None else stype+' - '+str(mis_angle)

        pop_ft = get_population_derivative( area,stype,mis_angle=mis_angle, \
            rot=rot,ft=ft,**ft_params )

        if plot == 'ring':
            r_min, r_max = 1.5, 2.0
            theta = np.linspace(0,360,BINS+1) * np.pi / 180.0 # in rad
            rad = np.array( [r_min, r_max] )
            X, Y = np.meshgrid(theta,rad)
            plt.subplot(3,2,i+1,projection='polar')
            if set_vmin:
                plt.pcolormesh( theta, rad, \
                    np.tile(pop_ft,(2,1)),cmap='gray_r',vmin=0 )
            else:
                plt.pcolormesh( theta, rad, np.tile(pop_ft,(2,1)),cmap='gray_r' )
            adj = 0 if mis_angle is None else mis_angle/2.0
            if rot=='local' or stype=='STD' or mis_angle!=None:
                for alpha in BD_ANGLES['local']:
                    a = (alpha+adj) * np.pi / 180.0
                    plt.plot( [a, a], [0.5,r_min], 'r' )
            if rot=='global' or stype=='STD' or mis_angle!=None:
                for alpha in BD_ANGLES['global']:
                    a = (alpha-adj) * np.pi / 180.0
                    plt.plot( [a, a], [1.0,r_min], 'g', lw=5, zorder=-1 )
            plt.colorbar()
            plt.yticks([])
            plt.ylim([0,r_max])
        elif plot == 'curve':
            plt.subplot(3,2,i+1)
            plt.plot(np.linspace(0.5,359.5,360),pop_ft)
            adj = 0 if mis_angle is None else mis_angle/2.0
            adj_sign = 1 if rot=='local' else -1
            for alpha in BD_ANGLES[rot]:
                a = (alpha+adj*adj_sign+360) % 360
                plt.axvline(x=a,lw=2,color='r',alpha=0.1)
            if set_vmin:
                plt.gca().set_ylim(bottom=0)
            plt.xlim( [0,360] )
            plt.gcf().set_size_inches(8,12,forward=True)

        plt.title(title_str)

    plt.suptitle('%s - %s (%s)' % (area,ft,rot))
    plt.savefig('./%s_%s_%s.png' % (area,ft,rot))
# end of FUNCTION population_demap 

# ================================================================================

def run_PSTH_panels(**traits):
    for area in ['CA1', 'CA3']:
        for stype in ['STD', 'MIS']:
            for ft in ['onset', 'offset']:
                for norm in [True, False]:
                    print 'Processing %s_%s_%s (%d)' % (area,stype,ft,2-norm)
                    pop_ft_PSTH_field( \
                        area=area,stype=stype,ft=ft,norm=norm,**traits)
# end of FUNCTION run_PSTH_panels

def pop_ft_PSTH_field(area='CA1',stype='STD',mis_angle=None,ft='onset',ref_cue='both',norm=True,clean=False,bins=BINS,half_win=45,win_hw=15,map_type='rate'):
    """
    Plot the average derivative maps (onset/offset) near the local or global cues

    ft: ['onset', 'offset']
        derivative features, onset or offset
    ref_cue: ['both', 'local', 'global']
        use only local/global cues as references (center) or both 
        (two sets of plots on the same figure)
    norm: [True, False]
        normalize the figure by bin counts
    clean: [True, False]
        use only local/global cues not overlapping with global/local cues
    half_win: integer 
        half width of the PSTH window
    win_hw: integer
        half window width of field-related window in number of bins
    map_type: ['rate', 'binary']
        use the rectified second order derivative map or the binary map
    """

    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    if ref_cue == 'both':
        panels = ['local', 'global']
        colors = ['r', 'g'] # local: red, global: green
    elif ref_cue == 'local':
        panels = ['local', ]
        colors = ['r', ]
    elif ref_cue == 'global':
        panels = ['global', ]
        colors = ['g', ]

    title_str = area + '-' + stype
    if mis_angle is not None:
        title_str += ' (%d)' % mis_angle
    cell_list = collect_cell_list(area,stype,mis_angle) 
    adj = 0 if mis_angle is None else mis_angle/2.0

    peaks = [0, 0, 0, 0]
    for i, panel in enumerate( panels ):
        if panel == 'local':
            angles = (np.array(BD_ANGLES[panel])+adj)%360 + 360
        elif panel == 'global':
            angles = (np.array(BD_ANGLES[panel])-adj+360)%360 + 360

        for j, win_type in enumerate( ['all', 'notfield', 'edge', 'field'] ):
            acc_ft = np.zeros(2*half_win)
            bin_cnt = np.zeros(2*half_win)
            count = 0
            for rds, tc, mis in cell_list:
                data = get_derivative_map( \
                    rds,tc,ft=ft,map_type=map_type,mis=mis,rot=panel,bins=bins)
                if win_type != 'all':
                    cnt = get_field_related_window( \
                        rds,tc,win_type,bins,ft=ft,mis=mis,rot=panel,win_hw=win_hw)
                    data[ ~cnt.astype(bool) ] = 0
                    ext_cnt = np.tile(cnt,3)
                ext_D = np.tile(data,3)

                if clean:
                    sel_angles = get_nonoverlap_cues( panel,stype, \
                        cues=tuple(angles),mis_angle=mis_angle,half_mis=mis)
                else:
                    sel_angles = angles.copy()

                for alpha in sel_angles:
                    extract = ext_D[alpha-half_win:alpha+half_win]
                    acc_ft += extract
                    if win_type != 'all':
                        bin_cnt += ext_cnt[alpha-half_win:alpha+half_win]
                    count += 1
            print '%d %s cues detected for %d cells...' \
                % (count,panel,len(cell_list))
            if (win_type!='all') and norm:
                acc_ft /= bin_cnt.astype(float)
            else:
                acc_ft /= float(count)

            if np.max(acc_ft) > peaks[j]:
                peaks[j] = np.max(acc_ft)

            x = np.linspace(-half_win,half_win-1,2*half_win) + 0.5
            plt.subplot(2,2,j+1)
            plt.plot(x,acc_ft,colors[i],lw=2,zorder=-100)
            plt.axvline(x=0,lw=1,color='k')
            plt.xlim( [x[0],x[-1]] )
            plt.ylim( [0,peaks[j]*1.2] )
            plt.title(win_type)

    plt.suptitle(title_str)
    fig = plt.gcf()
    fig.set_size_inches(8,12,forward=True)

    fname_str = '%s_%s_%s_cmp' % (area,stype,ft)
    if norm:
        fname_str += '_norm'
    fname_str += '_0'
    plt.savefig('./%s.png' % fname_str)

# end of FUNCTION pop_ft_PSTH_field

# ================================================================================

class DerivativeMapStackAna(RmapStackAna):

    datadir = './DeMapStack'
    label = 'derivative map stack'

    def process_data(self,**traits):
        super(DerivativeMapStackAna, self).process_data(**traits)

    def build_map(self,rat,day,maze,tc,mis,rotate,ft_type='onset',win_type='all',fd_hw=20,map_type='rate'):
        table_file = tb.openFile(PF_FILE,'r')
        field_table = table_file.root.place_fields

        angles = np.array( BD_ANGLES[rotate] )
        rds = (rat,day,maze)

        R = get_derivative_map( \
            rds,tc,ft=ft_type,mis=mis,rot=rotate,map_type=map_type )
        if win_type == 'all':
            table_file.close()
            return R

        query = '(rat==%d)&(day==%d)&(session==%d)&(tc=="%s")' \
            % (rat,day,maze,tc)
        cnt = np.zeros(len(R))
        for field in field_table.where(query):
            # the same as get_field_related_window
            end, start = field['field_edges']
            if rotate == 'local':
                start = (start-mis+360) % 360
                end = (end-mis+360) % 360
            elif rotate == 'global':
                start = (start+mis) % 360
                end = (end+mis) % 360

            if win_type == 'edge':
                edge = start if ft_type=='onset' else end
                s = (edge+win_hw) % 360
                e = (edge-win_hw+360) % 360
            elif (win_type=='field') or (win_type=='notfield'):
                s = (start+win_hw) % 360
                e = (end-win_hw+360) % 360
            elif win_type == 'overlap':
                s = start
                e = end
            s, e = np.floor(s), np.floor(e)

            if s < e:
                cnt[:s] = 1 
                cnt[e:] = 1 
            else:
                cnt[e:s] = 1 
        if win_type == 'notfield':
            cnt = 1 - cnt # reverse field 
        R[ ~cnt.astype(bool) ] = 0

        table_file.close()

        return R
    # end of METHOD build_map

# end of CLASS DerivativeMapStackAna

# ================================================================================

def check_overlap_peak_scatter(area='CA1',stype='STD',ft='offset', nbin=30 ):
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    table_file = tb.openFile(PF_FILE,'r')
    field_table = table_file.root.place_fields

    table_query = '(expt_type=="DR")&(exclude==False)'
    if area is not None:
        table_query += '&(area=="%s")' % area
    if stype is not None:
        table_query += '&(type=="%s")' % stype

    title_str = '%s_%s (%s)' % (area,stype,ft)
    angles = np.array( BD_ANGLES['local'] )

    points = []
    for field in field_table.where( table_query ):
        rat, day, maze = field['rat'], field['day'], field['session']
        if day > latest_qualified_day(rat,"DR"):
            continue

        rds = (rat, day, maze)
        tc = field['tc']
        end, start = np.floor( field['field_edges'] )
        hm = field['mis_angle'] / 2.0

        # get derivative map
        if ft == 'onset':
            data = get_onset_map(rds,tc,0,rot='local')
        elif ft == 'offset':
            data = get_offset_map(rds,tc,0,rot='local')
        data = np.r_[data, data] # for field crossing 0/360

        start = start+360 if start<end else start # if field crossing 0/360
        field_size = start - end
        if field_size < 0:
            print start, end

        # extract within field derivative map and get the track angles where
        # the binary derivative map is true
        theta = np.array( range(720) ) + 0.5
        extract = data[end:start]
        ext_theta = theta[end:start]
        alpha = ext_theta[ extract.astype(bool) ]

        cues = (angles+hm) % 360
        for ta in alpha:
            extend_cues = np.r_[ (cues, cues+360, cues+720) ]
            dist = ta - extend_cues[ np.argmin(abs(ta-extend_cues)) ]
            pct = float(ta-end) / field_size
            if pct < 0:
                print ta, start, end, field_size
            points.append( (pct,dist) )

    x, y = zip(*points)
    plt.hist2d(x,y,bins=nbin)
    plt.title( title_str )
    plt.show()
    
    table_file.close()

    print 'Done'
# end of FUNCTION check_overlap_peak_scatter

# --------------------------------------------------------------------------------

# move to other module??
def collect_cell_list(area, stype, mis_angle=None):
    """
    Return qualified cell list (rds, tc, half_mis_angle) with specified area, 
    session type and mismatch angle
    """
    table_file = tb.openFile(PF_FILE,'r')
    field_table = table_file.root.place_fields
    table_query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area

    cell_list = []
    for field in field_table.where(table_query+'&(type=="%s")'%stype):
        rat, day = field['rat'], field['day']
        if day > latest_qualified_day(rat,"DR"):
            continue
        if mis_angle is not None and field['mis_angle']!=mis_angle:
            continue
        mis = 0 if stype=='STD' or mis_angle!=None else field['mis_angle']/2.0
        clust = ((rat,day,field['session']),field['tc'],mis) # (rds,tc,half_mis)
        if clust not in cell_list:
            cell_list.append(clust)

    table_file.close()

    return cell_list
# end of FUNCTION collect_cell_list

@memoize
def get_nonoverlap_cues(rot, stype, cues=None, mis_angle=None, half_mis=None):
    """
    Return the track angles (in degree) of the local (global) cues which do not 
    overlap with any of the global (local) cues

    rot: ['local', 'global']
        Returning the filtered local or global cue locations
    stype: ['STD', 'MIS']
        Session type
    mis_angel: [None, 45, 90, 135, 180]
        The mismatch angle of the session. Does not matter if stype == 'STD'
    half_mis: [None, 22.5, 45, 67.5, 90]
        The rotation angle of the cues. (mismatch angle / 2.0)
    """
    # Illegal to have conflicting (mis_angel, half_mis) pair, while it is legal 
    # to have either mis_angle or half_mis equal to 0
    if mis_angle and half_mis and (mis_angle/2.0!=half_mis): 
        raise ValueError('half_mis should be half of mis_angle')

    if cues is None:
        cues = np.array( BD_ANGLES[rot] )
    else:
        cues = np.array(cues)

    if rot == 'local':
        if stype == 'STD':
            nonoverlap_cues = cues[ np.array([2,3]) ]
        elif stype == 'MIS':
            if mis_angle==45 or half_mis==22.5:
                nonoverlap_cues = np.array( [cues[0],] )
            elif mis_angle==90 or half_mis==45:
                nonoverlap_cues = cues[ np.array([1,2]) ]
            elif mis_angle==135 or half_mis==67.5:
                nonoverlap_cues = np.array( [cues[3],] )
            elif mis_angle==180 or half_mis==90:
                nonoverlap_cues = cues[ np.array([0,1]) ]
    elif rot == 'global':
        if stype == 'STD':
            nonoverlap_cues = cues[ np.array([0,3,4]) ]
        elif stype == 'MIS':
            if mis_angle==45 or half_mis==22.5:
                nonoverlap_cues = cues[ np.array([1,2]) ]
            elif mis_angle==90 or half_mis==45:
                nonoverlap_cues = cues[ np.array([0,3,4]) ]
            elif mis_angle==135 or half_mis==67.5:
                nonoverlap_cues = cues[ np.array([1,2]) ]
            elif mis_angle==180 or half_mis==90:
                nonoverlap_cues = cues[ np.array([0,3,4]) ]

    return nonoverlap_cues
# end of FUNCTION get_nonoverlap_cues

@memoize
def get_overlap_cues(rot, stype, cues=None, mis_angle=None, half_mis=None):
    if mis_angle and half_mis and (mis_angle/2.0!=half_mis): 
        raise ValueError('half_mis should be half of mis_angle')

    if cues is None:
        cues = np.array( BD_ANGLES[rot] )
    else:
        cues = np.array(cues)

    if rot == 'local':
        if stype == 'STD':
            #nonoverlap_cues = cues[ np.array([2,3]) ]
            overlap_cues = cues[ np.array([0,1]) ]
        elif stype == 'MIS':
            if mis_angle==45 or half_mis==22.5:
                #nonoverlap_cues = np.array( [cues[0],] )
                overlap_cues = cues[ np.array([1,2,3]) ]
            elif mis_angle==90 or half_mis==45:
                #nonoverlap_cues = cues[ np.array([1,2]) ]
                overlap_cues = cues[ np.array([0,3]) ]
            elif mis_angle==135 or half_mis==67.5:
                #nonoverlap_cues = np.array( [cues[3],] )
                overlap_cues = cues[ np.array([0,1,2]) ]
            elif mis_angle==180 or half_mis==90:
                #nonoverlap_cues = cues[ np.array([0,1]) ]
                overlap_cues = cues[ np.array([2,3]) ]
    elif rot == 'global':
        if stype == 'STD':
            #nonoverlap_cues = cues[ np.array([0,3,4]) ]
            overlap_cues = cues[ np.array([1,2]) ]
        elif stype == 'MIS':
            if mis_angle==45 or half_mis==22.5:
                #nonoverlap_cues = cues[ np.array([1,2]) ]
                overlap_cues = cues[ np.array([0,3,4]) ]
            elif mis_angle==90 or half_mis==45:
                #nonoverlap_cues = cues[ np.array([0,3,4]) ]
                overlap_cues = cues[ np.array([1,2]) ]
            elif mis_angle==135 or half_mis==67.5:
                #nonoverlap_cues = cues[ np.array([1,2]) ]
                overlap_cues = cues[ np.array([0,3,4]) ]
            elif mis_angle==180 or half_mis==90:
                #nonoverlap_cues = cues[ np.array([0,3,4]) ]
                overlap_cues = cues[ np.array([1,2]) ]

    return overlap_cues
# end of FUNCTION get_overlap_cues

# --------------------------------------------------------------------------------

def get_population_derivative(area,stype,mis_angle=None,bins=BINS,scale='one',map_type='rate',win_type='all',win_hw=15,**traits):
    """
    Return the population derivative map (onset/offset)
    Check get_derivative_map() for derivative map details

    scale: ['one', 'cell', 'bin_count']
        'one' - rescale the population map to [0,1]
        'cell' - average by cell number
        'bin_count' - normed by occupancy map
    win_type: ['all', 'edge', 'nonfield', 'field']:
        'all' - include all data
        'edge' - extract [-win_hw, win_hw] data centered at speciried field edge
        'nonfield' - ~'field'
        'field' - extract data within field plus win_hw margins on both side
    win_hw: integer
        half window width in number of bins
    map_type: ['rate', 'binary']
        returns the rectified second order derivative map or the binary map

    traits:

    ft: ['onset', 'offset']
        derivative features, onset or offset
    rot: ['local', 'global']
        correction for local/global cues
    """
    pop_ft = np.zeros(bins)
    bin_cnt = np.zeros(bins)

    cell_list = collect_cell_list(area,stype,mis_angle=mis_angle) 
    for rds, tc, mis in cell_list:
        data = get_derivative_map( \
            rds,tc,mis=mis,bins=bins,map_type=map_type,**traits)
        if win_type != 'all':
            cnt = get_field_related_window( \
                rds,tc,win_type,bins,win_hw=win_hw,mis=mis,**traits)
            data[ ~cnt.astype(bool) ] = 0
            bin_cnt += cnt
        pop_ft += data

    if scale == 'one': # re-scale to [0,1]
        pd_max, pd_min = np.max(pop_ft), np.min(pop_ft)
        pop_ft = (pop_ft-pd_min) / float(pd_max-pd_min)
    elif scale == 'cell': # average by cell number
        pop_ft /= len(cell_list)
    elif scale == 'bin_count':
        pop_ft /= bin_cnt

    return pop_ft
# end of FUNCTION get_population_derivative

def get_field_related_window(rds,tc,win_type,N,win_hw=15,rot=None,mis=None,ft=None):
    """
    Extract data within specified window type

    N: integer
        length of data (map bin number)
    mis: float
        cue rotation angle, equals to half mismatch angle
    win_type: ['all', 'edge', 'nonfield', 'field']:
        'all' - include all data
        'edge' - extract [-win_hw, win_hw] data centered at speciried field edge
        'nonfield' - ~'field'
        'field' - extract data within field plus win_hw margins on both side
    win_hw: integer
        half window width in number of bins
    ft: ['onset', 'offset']
        derivative features, onset or offset
    rot: ['local', 'global']
        correction for local/global cues
    """
    table_file = tb.openFile(PF_FILE,'r')
    field_table = table_file.root.place_fields

    query = '(rat==%d)&(day==%d)&(session==%d)&(tc=="%s")' % (rds+(tc,))

    cnt = np.zeros(N)
    for field in field_table.where(query):
        end, start = field['field_edges']
        if rot == 'local':
            start = (start-mis+360) % 360
            end = (end-mis+360) % 360
        elif rot == 'global':
            start = (start+mis) % 360
            end = (end+mis) % 360

        if win_type == 'edge': # window centered at specified edge
            edge = start if ft=='onset' else end
            s = (edge+win_hw) % 360
            e = (edge-win_hw+360) % 360
        elif (win_type=='field') or (win_type=='notfield'): # notfield == exclude
            s = (start+win_hw) % 360
            e = (end-win_hw+360) % 360
        elif win_type == 'overlap':
            s = start
            e = end
        s, e = np.floor(s), np.floor(e)

        if s < e:
            cnt[:s] = 1 
            cnt[e:] = 1 
        else:
            cnt[e:s] = 1 
    if win_type == 'notfield':
        cnt = 1 - cnt # reverse field 

    table_file.close()

    return cnt
# end of FUNCTION get_field_related_window 

def get_derivative_map(rds, tc, ft='onset', map_type='rate', **traits):
    """
    Return the derivative map (onset/offset) of the specified cell-session

    The ONSET is defined as the angle bins where the firing rate increasing speeds
    up; and the OFFSET is defined as the angle bins where the firing rate
    decreasing slows down.

    ft: ['onset', 'offset']
        derivative features, onset or offset
    map_type: ['rate', 'binary']
        returns the rectified second order derivative map or the binary map

    traits:
    mis, rot, bins, ratemap_filter
    """
    R = get_1d_ratemap(rds, tc, **traits) # ratemap
    D = np.roll(R,1) - R # first order derivative
    DD = D - np.roll(D,-1) # second order derivative

    if ft == 'onset':
        index = (D>0) * (DD>0)
    elif ft == 'offset':
        index = (D<0) * (DD>0)

    if map_type == 'binary':
        derivative_map = index.astype(int)
    elif map_type == 'rate':
        derivative_map = np.abs(DD) / float( np.max(R) )
        derivative_map[~index] = 0

    return derivative_map
# end of get_derivative_map

@memoize
def get_1d_ratemap(rds,tc,mis=0,rot='local',bins=BINS,ratemap_filter='running'):
    session = ModifiedSessionData.get(rds)
    if ratemap_filter == 'running':
        R = session.get_cluster_ratemap( tc,bins=bins,mis=mis,rot=rot, \
            blur_width=BLUR_WIDTH,**session.running_filter() )
    elif ratemap_filter == 'off_track':
        R = session.get_cluster_ratemap( tc,bins=bins,mis=mis,rot=rot,\
            blur_width=BLUR_WIDTH,velocity_filter=False, exclude_off_track=True )
    elif ratemap_filter == 'none':
        R = session.get_cluster_ratemap( tc,bins=bins,mis=mis,rot=rot, \
            blur_width=BLUR_WIDTH,velocity_filter=False )
    return R
# end of FUNCTION get_ratemap

# --------------------------------------------------------------------------------

#def get_onset_map(rds, tc, mis, map_type='rate', **traits):
#    R = get_1d_ratemap(rds, tc, mis, **traits)
#    D = np.roll(R,1) - R
#    DD = D - np.roll(D,-1)
#
#    index = (D>0) * (DD>0)
#    if map_type == 'binary':
#        field_on = index.astype(int)
#    elif map_type == 'rate':
#        field_on = np.abs(DD) / float( np.max(R) )
#        field_on[~index] = 0
#
#    return field_on 
## end of FUNCTION get_onset_map
#
#def get_offset_map(rds, tc, mis, map_type='rate', **traits):
#    R = get_1d_ratemap(rds, tc, mis, **traits)
#    D = np.roll(R,1) - R
#    DD = D - np.roll(D,-1)
#
#    index = (D<0) * (DD>0)
#    if map_type == 'binary':
#        field_off = index.astype(int)
#    elif map_type == 'rate':
#        field_off = np.abs(DD) / float( np.max(R) )
#        field_off[~index] = 0
#
#    return field_off 
## end of FUNCTION get_offset_map

#def calculate_avg_diff( pd, select ):
#    # normalize population rate difference
#    pd_max, pd_min = np.max(pd), np.min(pd)
#    norm_pd = (pd-pd_min) / float(pd_max-pd_min)
#
#    avg_in = np.sum(norm_pd[select]) / float(np.sum(select))
#    avg_out = np.sum(norm_pd[~select]) / float(np.sum(~select))
#
#    return avg_in - avg_out
# end of FUNCTION calculate_avg_diff

#@memoize
#def get_1d_rate_diff_map(rds, tc, mis, norm=True, rec=None, **traits):
#    R = get_1d_ratemap(rds, tc, mis, **traits)
#    D = np.roll(R,-1) - R
#
#    if rec == 'abs': # collect the absolute value of diff
#        D = abs(D)
#    elif rec == 'inc':
#        D[D>0] = 0 # collect when the firing rate is increasing  
#        D *= -1
#    elif rec == 'dec':
#        D[D<0] = 0 # collect when the firing rate is decreasing  
#
#    if norm:
#        D = D / float(np.max(D))
#    return D
## end of FUNCTION get_1d_rate_diff_map

#def get_population_derivative(area,stype,mis_angle=None,scale='one',**traits):
#    pop_diff = np.zeros(BINS)
#
#    cell_list = collect_cell_list(area,stype,mis_angle) 
#    for rds, tc, mis in cell_list:
#        D = get_1d_rate_diff_map(rds, tc, mis, **traits)
#        pop_diff += D
#
#    if scale == 'one': # re=scale to [0,1]
#        pd_max, pd_min = np.max(pop_diff), np.min(pop_diff)
#        pop_diff = (pop_diff-pd_min) / float(pd_max-pd_min)
#    if scale == 'cell': # average by cell number
#        pop_diff /= len(cell_list)
#
#    return pop_diff
## end of FUNCTION get_population_derivative

# --------------------------------------------------------------------------------

class ChkStartDecrePntReport(BaseReport):
    datadir = './StartDecre'
    label = 'rate start decrease'

    ncols = 2 
    nrows = 8
    ynorm = False
    
    def collect_data(self,min_change_ratio=0.33,tol_ratio=0.33):
        self.results = {}

        table_file = tb.openFile(PF_FILE,'r')
        field_table = table_file.root.place_fields
        query = '((area=="CA1")|(area=="CA3"))&(expt_type=="DR")&(exclude==False)'

        pre = None
        for field in field_table.where(query):
            rat, day = field['rat'], field['day']
            if day > latest_qualified_day(rat,"DR"):
                continue

            maze, tc = field['session'], field['tc']
            cur = (rat, day, maze, tc)
            if cur != pre:
                if pre is None:
                    data = []
                else:
                    data.append( (pre,R_ext,fields,points) )
                pre = cur
                fields = []
                points = []

                R = get_1d_ratemap( (rat,day,maze), tc, 0) # not rotated
                R_ext = np.r_[R, R]
                D = np.roll(R_ext,1) - R_ext
                DD = D - np.roll(D,-1)

            end, start = np.floor( field['field_edges'] )
            if start < end:
                start += 360
            fields.append( (start,end) )

            min_change = max(R_ext[end:start]) * min_change_ratio
            tol = min_change * tol_ratio

            head, tail = end, end
            # when at field edge, min_R = 0
            max_R, min_R, peak, in_seg = 0, 0, 0, False
            while tail <= start:
                if tail == start: # end segment anyway
                    if max_R-min_R >= min_change:
                        points.append( index )
                    tail += 1
                elif D[tail] < 0: # extending segment
                    if (DD[tail]<0) and (R_ext[tail]>peak):
                        peak = R_ext[tail]
                        index = tail
                    if R_ext[tail] > max_R:
                        max_R = R_ext[tail]
                    if R_ext[tail] < min_R:
                        min_R = R_ext[tail]
                    local_max = R_ext[tail]
                    tail += 1
                    in_seg = True
                else:
                    #if in_seg and (counts<tol_size): # extending segment
                    if in_seg and (local_max-R_ext[tail]<tol): # extending segment
                        tail += 1
                        peak = local_max + tol
                        continue
                    if max_R-min_R >= min_change:
                        points.append( index )
                    tail += 1
                    head = tail
                    max_R, min_R, peak, in_seg = 0, np.inf, 0, False
        data.append( (cur,R_ext,fields,points) )

        table_file.close()

        self.results['data'] = data
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self):
        data = self.results['data']

        def plot_data(item, ax):
            clust, R_ext, fields, points = item
            for start, end in fields:
                ax.axvline(x=start+0.5,color='r',lw=1)
                ax.axvline(x=end+0.5,color='r',lw=1)
                plt.axvspan(end+0.5,start+0.5,alpha=0.5,color='r')
            if type(points) is not list:
                points = [points, ]
            for point in points:
                ax.axvline(x=point+0.5,color='b',alpha=0.5,lw=1)
            plt.plot( R_ext, 'k', lw=1)
            plt.tick_params(axis='both', labelsize=7)
            plt.xlim(0, 720)

        for item, ax in self.get_plot(data):
            plot_data(item, ax)

        self.out('Done')
    # end of METHOD process_data

def start_decrease_distr(rot='local',nbin=36,min_change_ratio=0.33,tol_ratio=0.33):
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    table_file = tb.openFile(PF_FILE,'r')
    field_table = table_file.root.place_fields
    table_query = '(expt_type=="DR")&(exclude==False)'

    bins = np.linspace( 0, 360, nbin+1 )

    plt_cnt = 1
    for area in ['CA1', 'CA3']:
        for stype in ['STD', 'MIS']:
            query = table_query + '&(area=="%s")&(type=="%s")' % (area, stype)

            pre = None
            data = []
            for field in field_table.where(query):
                rat, day = field['rat'], field['day']
                if day > latest_qualified_day(rat,"DR"):
                    continue

                maze, tc, hm = \
                    field['session'], field['tc'], field['mis_angle']/2.0
                cur = (rat, day, maze, tc)
                if cur != pre:
                    R = get_1d_ratemap( (rat,day,maze), tc, 0) # not rotated
                    R_ext = np.r_[R, R]
                    D = np.roll(R_ext,1) - R_ext
                    DD = D - np.roll(D,-1)
                    pre = cur

                end, start = np.floor( field['field_edges'] )
                if start < end:
                    start += 360

                min_change = max(R_ext[end:start]) * min_change_ratio
                tol = min_change * tol_ratio

                head, tail = end, end
                # when at field edge, min_R = 0
                max_R, min_R, peak, in_seg = 0, 0, 0, False
                while tail <= start:
                    if tail == start: # end segment anyway
                        if max_R-min_R >= min_change:
                            data.append( index )
                        tail += 1
                    elif D[tail] < 0: # extending segment
                        if (DD[tail]<0) and (R_ext[tail]>peak):
                            peak = R_ext[tail]
                            if rot == 'local':
                                index = (tail+0.5-hm+360) % 360
                            elif rot == 'global':
                                index = (tail+0.5+hm) % 360
                        if R_ext[tail] > max_R:
                            max_R = R_ext[tail]
                        if R_ext[tail] < min_R:
                            min_R = R_ext[tail]
                        local_max = R_ext[tail]
                        tail += 1
                        in_seg = True
                    else:
                        # extending segment
                        if in_seg and (local_max-R_ext[tail]<tol):
                            tail += 1
                            peak = local_max + tol
                            continue
                        if max_R-min_R >= min_change:
                            data.append( index )
                        tail += 1
                        head = tail
                        max_R, min_R, peak, in_seg = 0, np.inf, 0, False
            plt.subplot(2,2,plt_cnt)
            plt.hist(data,bins=bins)
            for cue in BD_ANGLES[rot]:
                plt.axvline(x=cue,color='r',lw=1)
            plt.xlim( [0, 360] )
            plt.tick_params(axis='both',labelsize=7)
            plt.title('%s - %s' % (area,stype))

            plt_cnt += 1

    table_file.close()

    plt.show()

    print 'Done'
# end of FUNCTION start_decrease_distr 

# --------------------------------------------------------------------------------

def check_ft_rmapstack(area=None,stype=None,mis_angle=None,ft='onset',window=[-10,20],clean=False,rot='local',norm=True,**traits):
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)
    
    table_file = tb.openFile(PF_FILE,'r')
    field_table = table_file.root.place_fields

    adj = 0 if mis_angle is None else mis_angle/2.0
    if rot == 'local':
        angles = (np.array(BD_ANGLES[rot])+adj)%360 + 360
    elif rot == 'global':
        angles = (np.array(BD_ANGLES[rot])-adj+360)%360 + 360

    clusts = []
    cell_list = collect_cell_list(area,stype,mis_angle) 
    for rds, tc, mis in cell_list:
        if ft == 'onset':
            data = get_onset_map(rds,tc,mis,rot=rot,**traits)
            threshold = 0
        elif ft == 'offset':
            data = get_offset_map(rds,tc,mis,rot=rot,**traits)
            threshold = 0
        ext_D = np.tile(data,3)

        if clean:
            sel_angles = get_nonoverlap_cues( \
                rot,stype,cues=tuple(angles),mis_angle=mis_angle,half_mis=mis)
        else:
            sel_angles = angles.copy()

        for alpha in sel_angles:
            extract = ext_D[alpha+window[0]:alpha+window[1]]
            if np.sum(extract) > threshold: # contributing to specified window 
                clusts.append( (rds,tc,mis,alpha) )

    not_sort, to_be_sort = [], []
    for rds, tc, mis, alpha in clusts:
        rat, day, maze = rds
        query = '(rat==%d)&(day==%d)&(session==%d)&(tc=="%s")' % \
            (rat, day, maze, tc)
        min_dist, nearest_edge = np.inf, None
        for field in field_table.where(query):
            start, end = field['field_edges'][1], field['field_edges'][0]
            if rot == 'local':
                start = (start-mis+360) % 360
                end = (end-mis+360) % 360
                hw = 45
            elif rot == 'global':
                start = (start+mis) % 360
                end = (end+mis) % 360
                hw = 22.5

            if ft == 'onset':
                edges = [ start, start+360, start+720 ]
            elif ft == 'offset':
                edges = [ end, end+360, end+720 ]

            dist = edges[ np.argmin(np.abs(np.array(edges)-alpha)) ] - alpha
            if abs(dist) > 360: # detect error
                print dist

            if abs(dist) < abs(min_dist): # if current field closer to cue 
                if (dist>=-hw) and (dist<hw): # field edge within window
                    min_dist = dist
                    nearest_edge = edges[1]
                else: # check whether field overlap with window 
                    e = end
                    s = start+360 if start<end else start
                    win_e = (alpha+window[0]+360) % 360
                    win_s = (alpha+window[1]+360) % 360
                    if win_s < win_e:
                        win_s += 360
                    if (s>win_e) and (e<win_s):
                        min_dist = dist
                        nearest_edge = edges[1]
            
        if np.isinf(min_dist): # no field in window or overlap w window
            not_sort.append( ((rat,day,maze,tc,mis),alpha%360,-1,-1) )
        else:
            to_be_sort.append( \
                ((rat,day,maze,tc,mis),alpha%360,min_dist,nearest_edge%360) )

    distances = np.array( zip(*to_be_sort)[2] )
    index = np.argsort(distances)
    combined = []
    for i in index:
        combined.append( to_be_sort[i] )
    combined += not_sort

    ratemat = None
    for clust, angle, dist, edge in combined:
        rat, day, maze, tc, mis = clust
        session = ModifiedSessionData.get( (rat,day,maze) )
        R = session.get_cluster_ratemap( tc,bins=BINS,mis=mis,rot=rot,\
            blur_width=BLUR_WIDTH,**session.running_filter() )
        R = np.roll( R, -( int(angle+window[0]-1+360+180)%360 ) )
        if norm:
            R /= np.max(R)
        if ratemat is None:
            ratemat = R.copy()
        else:
            ratemat = np.vstack( [ratemat, R] )
    plt.imshow(ratemat,interpolation='none',aspect='auto',cmap='gray_r', \
        origin='lower',extent=[0,360,0,len(combined)-1])
    plt.axvline(x=180,lw=4,color='r',alpha=0.1) # 0
    plt.axvline(x=180+window[0],lw=2,color='m',alpha=0.1) # window[0] 
    plt.axvline(x=180+window[1],lw=2,color='m',alpha=0.1) # window[1]
    plt.axvline(x=180+hw,lw=1,color='b',alpha=0.1)
    plt.axvline(x=180-hw,lw=1,color='b',alpha=0.1)

    plt.tick_params(axis='both',labelsize=7)
    plt.axis('tight')
    plt.show()

    table_file.close()
# end of FUNCTION check_ft_rmapstack
