# encoding: utf-8
"""
field.py

"""

import tables as tb
import numpy as np
import os

from scipy import ndimage

from core.report import BaseReport

from scanr.lib import *
from scanr.data import get_node, get_unique_row
from scanr.session import SessionData
from scanr.cluster import get_min_quality_criterion, AND, PlaceCellCriteria
from scanr.spike import TetrodeSelect
from scanr.field import field_extent, center_of_mass

Cfg = { 'bins'             : 360,
        'blur_width'       : 4.3,
        'min_rate'         : 1.5,
        'min_size'         : 15,
        'max_size'         : 330,
        'min_space'        : 1,
        'floor_fraction'   : 0.1 }

# --------------------------------------------------------------------------------

class PlaceFieldReport(BaseReport):
    """
    Modified place field detection algorithm for texture boundary analyses.  
    (The field edges detected by Joe's criterion seem to be too loose)
    """

    datadir = 'PlaceField'
    label = 'place field report'

    def collect_data(self,ratemap_filter='running',bins=Cfg['bins'],blur_width=Cfg['blur_width'],min_rate=Cfg['min_rate'],floor_fraction=Cfg['floor_fraction'],**kwds):

        ColDescr = { 'id'           : tb.UInt16Col(pos=1),
                     'rat'          : tb.UInt16Col(pos=2),
                     'day'          : tb.UInt16Col(pos=3),
                     'session'      : tb.UInt16Col(pos=4),
                     'expt_type'    : tb.StringCol(itemsize=4, pos=5),
                     'type'         : tb.StringCol(itemsize=5, pos=6),
                     'mis_angle'    : tb.UInt16Col(pos=7),
                     'exptr'        : tb.StringCol(itemsize=10, pos=8),
                     'tc'           : tb.StringCol(itemsize=8, pos=9),
                     'area'         : tb.StringCol(itemsize=4, pos=10),
                     'subdiv'       : tb.StringCol(itemsize=15, pos=11),
                     'exclude'      : tb.BoolCol(pos=12),
                     'field_num'    : tb.UInt16Col(pos=13),
                     'COM'          : tb.FloatCol(pos=14),
                     'field_edges'  : tb.FloatCol(shape=(2,),pos=15),
                     'outlier'      : tb.BoolCol(pos=16) }

        self.out('Creating place field table...')
        file_path = os.path.join(self.datadir, 'fields.h5')
        data_file = tb.openFile(file_path,'w')
        field_table = data_file.createTable('/', 'place_fields', 
            ColDescr, title='Place Field Data')
        row = field_table.row

        session_table = get_node('/metadata','sessions')
        tetrode_table = get_node('/metadata','tetrodes')

        # Quality criterion
        Quality = get_min_quality_criterion('fair')
        area_query = '((area=="CA1")|(area=="CA3")|(area=="CA2")|(area=="DG"))'
        
        count, err_cnt = 0, 0
        for dataset in TetrodeSelect.datasets(area_query):
            rat, day = dataset

            Tetrodes = TetrodeSelect.criterion(dataset, area_query)
            Criteria = AND(Quality, Tetrodes, PlaceCellCriteria)

            for session in session_table.where('(rat==%d)&(day==%d)'%dataset):
                rds = rat, day, session['session']
                session_data = SessionData.get(rds)
                session_data.cluster_criteria = Criteria
                meta = get_session_metadata(*rds)
            
                for tc in session_data.get_clusters(Criteria):
                    tt, cl = parse_cell_name(tc)
                    tetrode = get_unique_row( tetrode_table, 
                        '(rat==%d)&(day==%d)&(tt==%d)' % (rat, day, tt) )
                    outlier = tetrode['outlier']

                    if ratemap_filter == 'running':
                        R = session_data.get_cluster_ratemap( tc, bins=bins, \
                            blur_width=blur_width, **session_data.running_filter() )
                    elif ratemap_filter == 'off_track':
                        R = session_data.get_cluster_ratemap( tc, bins=bins, \
                            blur_width=blur_width, velocity_filter=False, \
                                exclude_off_track=True )
                    elif ratemap_filter == 'none':
                        R = session_data.get_cluster_ratemap( tc, bins=bins, \
                            blur_width=blur_width,velocity_filter=False ) 
                    if R.max() < min_rate:
                        continue
                    threshold = R.max() * floor_fraction

                    field_num, labels = label_place_field( R, threshold, **kwds )

                    for i in range(field_num):
                        selected = labels == (i+1)
                        fd_size = np.sum(selected)

                        # Add row to place-field table
                        row['id'] = count
                        row['rat'] = rat
                        row['day'] = day
                        row['session'] = session['session']
                        row['expt_type'] = session['expt_type']
                        row['type'] = session['type']
                        #row['mis_angle'] = meta['parameter'] #meta['mis_angle']
                        if meta['type']=="STD":
                            row['mis_angle'] = 0
                        elif meta['type']=="MIS":
                            row['mis_angle'] = meta['mis_angle']
                            if meta['mis_angle'] not in [45,90,135,180]:
                                err_cnt += 1
                                print 'Error!!!!!'
                        row['exptr'] = session['exptr']
                        row['tc'] = tc
                        row['area'] = tetrode['area']
                        row['subdiv'] = tetrode['subdiv']
                        row['exclude'] = tetrode['ambiguous'] or \
                            tetrode['unsure'] or (outlier=='too anterior') \
                                or (outlier=='too posterior')
                        row['field_num'] = i + 1 
                        row['COM'] = center_of_mass(R,F=selected,degrees=True)
                        row['field_edges'] = field_extent( selected )
                        row['outlier'] = (outlier=='anterior') or \
                            (outlier=='posterior') or \
                                (outlier=='too anterior')\
                                    or (outlier=='too posterior')
                        row.append()
                        count += 1

                    field_table.flush()
                self.out.printf('\n')

        field_table.close()
        self.out('%d fields detected' % count)
        self.out('Error found: %d' % err_cnt)
        self.out('All done!')
    # end of METHOD collect_data

# end of CLASS PlaceFieldTable()

# --------------------------------------------------------------------------------

def label_place_field(R,threshold,min_space=Cfg['min_space'],min_size=Cfg['min_size'],max_size=Cfg['max_size'],min_rate=Cfg['min_rate']):
    H = R >= threshold
    labels, num = ndimage.label(H)
    if (num==0) or np.sum(labels>0)>max_size:
        return 0, labels

    # merge fields if the interval between them <= min_sapce
    wrapped_labels = np.r_[labels, labels[:min_space+1]]
    w_len = len(wrapped_labels)
    intvl, n_intvl = ndimage.label( wrapped_labels==0 )
    if intvl[0] > 0:
        intvl[ intvl==intvl[0] ] = 0
    if intvl[-1] > 0:
        intvl[ intvl==intvl[-1] ] = 0
    for i in range(n_intvl):
        n = i + 1
        selected = intvl == n
        intvl_size = np.sum(selected)
        if (intvl_size>0) and (intvl_size<=min_space):
            numbering = np.array(range(w_len))[selected]
            m, M = min(numbering), max(numbering)
            old, new = wrapped_labels[M+1], wrapped_labels[m-1]
            if old == 1: # wrapped field
                old, new = (new, old) # swap(new, old)
            wrapped_labels[ wrapped_labels==old ] = new 
            wrapped_labels[ selected ] = new
    labels = wrapped_labels[:len(labels)]

    if (labels[0]!=labels[-1]) and (labels[0]*labels[-1]>0):
        labels[ labels==labels[-1] ] = labels[0] # wrap

    field_num = 1
    for i in range(num):
        selected = labels == (i+1)
        fd_size = np.sum(selected)
        if (fd_size<min_size) or (fd_size>max_size) or \
           (np.max(R[selected])<min_rate):
            labels[selected] = 0
        else:
            labels[selected] = field_num
            field_num += 1

    return field_num-1, labels
# end of FUNCTION mark_place_field

# --------------------------------------------------------------------------------
