# encoding: utf-8
"""
field_distribution.py

Check how the place field edges are spatially distributed and whether the
distribution correlate with the quadrant boundaries

Should replace analysis_shan/texture_boundaries.py in the future
"""

import tables as tb
import numpy as np
import matplotlib.pylab as plt

from itertools import product

from analysis_shan.data_filters import first_exp_day, last_exp_day, latest_qualified_day
from ldb.ld_gradient_map import collect_cell_list

BINS = 360
BD_ANGLES = { 'local':[45, 135, 225, 315], 'global':[0, 45, 135, 180, 270] }

# --------------------------------------------------------------------------------

def plot_field_count_map(area='CA1',scale='one'):
    """
    Plot the field count heatmap for the STD and MIS sessions.

    scale   - ['one', 'cell']
              'one': rescale the population map to [0,1]
              'cell': average by cell number
    """

    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    plt.figure(figsize=(12,4))

    table_file = tb.openFile('/home/paramecium/output/fields.h5', 'r')
    field_table = table_file.root.place_fields
    table_query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area

    stypes = ['STD', 'MIS', 'MIS']
    rotations = [None, 'local', 'global']

    for i, panel in enumerate( zip(stypes,rotations) ):
        stype, rot = panel
        title_str = stype if rot is None else stype + ' - ' + rot

        C = np.zeros(BINS)
        cell_list = []
        for field in field_table.where(table_query+'&(type=="%s")'%stype):
            rat, day = field['rat'], field['day']
            if day > latest_qualified_day(rat,"DR"):
                continue

            hm = 0 if stype=="STD" else field['mis_angle']/2.0
            start, end = field['field_edges'][1], field['field_edges'][0]
            if rot == 'local':
                start = (start-hm+360) % 360
                end = (end-hm+360) % 360
            elif rot == 'global':
                start = (start+hm) % 360
                end = (end+hm) % 360

            if start > end:
                C[end:start] += 1
            else:
                C[:start] += 1
                C[end:] += 1

            cell_list.append( (rat,day,field['session'],field['tc']) )

        N = len( list(set(cell_list)) )
        if scale == 'one': # re-scale to [0,1]
            cnt_max, cnt_min = np.max(C), np.min(C)
            C = (C-cnt_min) / float(cnt_max-cnt_min)
        elif scale == 'cell': # average by cell number
            C /= N

        r_min, r_max = 1.5, 2.0
        theta = np.linspace(0,360,BINS+1) * np.pi / 180.0 # in rad
        rad = np.array( [r_min, r_max] )
        X, Y = np.meshgrid(theta,rad)
        plt.subplot(1,3,i+1,projection='polar')
        plt.pcolormesh( theta, rad, np.tile(C,(2,1)),cmap='gray_r' )

        # Label cue locations
        if rot=='local' or stype=='STD':
            for alpha in BD_ANGLES['local']:
                a = alpha * np.pi / 180.0
                plt.plot( [a, a], [0.5,1.5], 'r' )
        if rot=='global' or stype=='STD':
            for alpha in BD_ANGLES['global']:
                a = alpha * np.pi / 180.0
                plt.plot( [a, a], [1.0,1.5], 'g', lw=5, zorder=-1 )

        plt.colorbar()
        plt.yticks([])
        plt.ylim([0,r_max])
        plt.title(title_str)

    plt.show()
    print 'Done'

    # end of FUNCTION plot_field_count_map

# --------------------------------------------------------------------------------

def field_edge_PSTH(area='CA1',stype='STD',mis_angle=None,cues='local',side='both',half_win=45,clean=False,nbins=20):
    """
    Plot the field count heatmap for the STD and MIS sessions.

    scale   - ['one', 'cell']
              'one': rescale the population map to [0,1]
              'cell': average by cell number
    """

    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)
    plt.figure(figsize=(4,8))

    table_file = tb.openFile('/home/paramecium/output/fields.h5', 'r')
    field_table = table_file.root.place_fields
    table_query = \
        '(area=="%s")&(expt_type=="DR")&(type=="%s")&(exclude==False)' \
            % (area, stype)

    title_str = area + '-' + stype
    if mis_angle is not None:
        title_str += ' (%d)' % mis_angle
    bins = None if nbins is None else np.linspace(-half_win,half_win,nbins+1)
    color = 'r' if cues=='local' else 'g'

    data = []
    count = 0
    for field in field_table.where(table_query):
        rat, day = field['rat'], field['day']
        #if day == first_exp_day(rat,"DR"):
        #    continue
        #if day != last_exp_day(rat,"DR"):
        #    continue
        if day > latest_qualified_day(rat,"DR"):
            continue
        if mis_angle is not None and (field['mis_angle']!=mis_angle):
            continue

        hm = 0 if stype=="STD" else field['mis_angle']/2.0
        start = calc_angle_dist( field['field_edges'][1],hm,cues,half_win,clean )
        end = calc_angle_dist( field['field_edges'][0],hm,cues,half_win,clean )
        if side == 'start':
            if start is not None:
                data.append( start )
        elif side == 'end':
            if end is not None:
                data.append( end )
        elif side == 'both':
            if start is not None:
                data.append( start )
            if end is not None:
                data.append( end )
        count += 1
    print '%d edges collected from %d fields' % (len(data), count)

    if bins is None:
        plt.hist(data)
    else:
        plt.hist(data,bins=bins,color='c')
    plt.xlim( [bins[0], bins[-1]] )
    plt.axvline(x=0,lw=2,color=color)
    plt.tick_params(axis='both',labelsize=7)
    plt.title(title_str)

    #plt.show()
    plt.savefig('output.png')
    table_file.close()
    print 'Done'

    # end of FUNCTION plot_field_count_map

def calc_angle_dist( x, half_mis, cues, half_win, clean):
    if clean:
        angles = BD_ANGLES[cues]
        if cues == 'local':
            if half_mis == 0: # STD
                sel_cues = [ angles[2], angles[3] ]
            elif half_mis == 22.5:
                sel_cues = [ angles[0], ]
            elif half_mis == 45:
                sel_cues = [ angles[1], angles[2] ]
            elif half_mis == 67.5:
                sel_cues = [ angles[3], ]
            elif half_mis == 90:
                sel_cues = [ angles[0], angles[1] ]
        elif cues == 'global':
            if half_mis == 0: # STD
                sel_cues = [ angles[0], angles[3], angles[4] ]
            elif half_mis == 22.5:
                sel_cues = [ angles[1], angles[2] ]
            elif half_mis == 45:
                sel_cues = [ angles[0], angles[3], angles[4] ]
            elif half_mis == 67.5:
                sel_cues = [ angles[1], angles[2] ]
            elif half_mis == 90:
                sel_cues = [ angles[0], angles[3], angles[4] ]
    else:
        sel_cues = BD_ANGLES[cues]
    sel_cues = np.array(sel_cues) + 360

    if cues == 'local':
        x = (x-half_mis+360) % 360
    elif cues == 'global':
        x = (x+half_mis) % 360

    xs = np.array( [x, x+360, x+720] )
    dist_pair = [ (a,b,abs(a-b)) for a, b in product(xs,sel_cues) ]
    a, b = dist_pair[ np.argmin(zip(*dist_pair)[2]) ][:2]
    shift = a - b

    if (shift<-half_win) or (shift>=half_win):
        return None
    else:
        return a - b

# --------------------------------------------------------------------------------
