# encoding: utf-8
"""
field_rotation.py

Check whether there is a correlation between how close the field is to a qudrant 
boundary and whether the field follows local cues in MIS sessions
"""

import tables as tb
import numpy as np
import matplotlib.pylab as plt

from tools.misc import memoize
from scipy.stats import chi2_contingency, mannwhitneyu

from scanr.data import get_node, get_unique_row
from analysis_shan.data_filters import first_exp_day, last_exp_day, latest_qualified_day

BD_ANGLES = { 'local':[45, 135, 225, 315], 'global':[0, 45, 135, 180, 270] }

# --------------------------------------------------------------------------------

def distance_to_cue_scatter(area='CA1',cues='local',side='start',match_cue=False,abs_dist=False,mis_angle=None,show_opp=True):

    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    colors = { '45': 'b', '90': 'c', '135': 'grey', '180': 'm'}

    title_str = area
    if mis_angle is not None:
        title_str += ' (%d)' % mis_angle
    title_str += ' - Distance to nearest %s cue (degree)' % cues

    query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area
    # (rat, day, tc, [ (m1,hm1,s1,e1), (m2,hm2,s2,e2)] )
    field_pairs = get_field_pair_list(query) 

    data = []
    for field_pair in field_pairs:
        rat, day, tc = field_pair[:3]
        if day > latest_qualified_day(rat,"DR"):
            continue
        point = []
        cue_id = None
        for maze, hm, s, e in field_pair[3]:
            param = s if side=='start' else e
            cue_id, dist = nearest_cue(param,hm,cues,cue_id=cue_id)
            if not match_cue:
                cue_id = None
            if abs_dist:
                dist = abs(dist)
            point.append( dist )
            if hm != 0:
                point.append( colors[str(int(hm*2))] )
        data.append( point )
    N = len(data)
    print '%d field pairs collected' % N

    x, y, c = zip(*data)
    jitter = np.random.rand(2,N) / 10.0
    x = np.array(x) + jitter[0]
    y = np.array(y) + jitter[1]
    c = np.array(c)

    if mis_angle is None:
        idx = np.ones(N).astype(bool)
    else:
        idx = c == colors[str(mis_angle)]
    plt.scatter(x[idx],y[idx],color=list(c[idx]),s=2)
    if abs_dist:
        plt.plot([-1,46],[-1,46],'r',lw=3,alpha=0.2)
        plt.axvline(x=15,color='k',lw=1,alpha=0.2)
        plt.axhline(y=15,color='k',lw=1,alpha=0.2)
        plt.xlim( -1, 46 )
        if match_cue:
            plt.ylim( -1, 181 )
        else:
            plt.ylim( -1, 46 )
    else:
        plt.plot([-46,46],[-46,46],'r',lw=5,alpha=0.2)

        if mis_angle is not None:
            if cues == 'local':
                for drift in [-90, -180, 90, 180]:
                    plt.plot([-46,46],[-46+drift,46+drift],'r',lw=2,alpha=0.2)
                if show_opp:
                    plt.plot( [-46,46],[-46-mis_angle,46-mis_angle], \
                        colors[str(mis_angle)],lw=1,alpha=0.2 )
                else:
                    plt.plot( [-46,46],[-46-mis_angle/2.0,46-mis_angle/2.0], \
                        colors[str(mis_angle)],lw=1,alpha=0.2 )
            elif cues == 'global':
                if show_opp:
                    plt.plot( [-46,46],[-46+mis_angle,46+mis_angle], \
                        colors[str(mis_angle)],lw=1,alpha=0.2 )
                else:
                    plt.plot( [-46,46],[-46+mis_angle/2.0,46+mis_angle/2.0], \
                        colors[str(mis_angle)],lw=1,alpha=0.2 )

        plt.axvline(x=0,color='k',lw=1,alpha=0.2)
        plt.axhline(y=0,color='k',lw=1,alpha=0.2)
        plt.xlim( -46, 46 )
        if match_cue:
            plt.ylim( -181, 181 )
        else:
            plt.ylim( -46, 46 )
    plt.xlabel('STD',size='small')
    plt.ylabel('MIS',size='small')
    plt.tick_params(axis='both',labelsize=7)
    plt.title(title_str)

    #plt.show()
    file_name = '%s_%s_%s' % (area,cues,side)
    if abs_dist:
        file_name += '_abs'
    if mis_angle:
        file_name += '_%d' % int(mis_angle)
    if match_cue:
        file_name += '_match.png'
    else:
        file_name += '_n-match.png'
    plt.savefig(file_name)
    print 'Done'

# end of FUNCTION distance_to_cue_scatter

def rotation_cue_scatter(area='CA1',side='start',mis_angle=None,abs_dist=False):
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    colors = { '45': 'b', '90': 'c', '135': 'grey', '180': 'm'}

    title_str = area
    if mis_angle is not None:
        title_str += ' (%d)' % mis_angle
    title_str += ' - relative rotation angle'

    query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area
    # (rat, day, tc, [ (m1,hm1,s1,e1), (m2,hm2,s2,e2)] )
    field_pairs = get_field_pair_list(query) 

    data = []
    for field_pair in field_pairs:
        rat, day, tc = field_pair[:3]
        if day > latest_qualified_day(rat,"DR"):
            continue
        param = 2 if side=='start' else 3
        angle_s, angle_m = field_pair[3][0][param], field_pair[3][1][param]
        rotate = angle_m - angle_s
        if abs(rotate) > 180:
            rotate -= np.sign(rotate) * 360
        cue_id, dist = nearest_cue(angle_s,field_pair[3][0][1],'local',cue_id=None)
        if abs_dist:
            dist = abs(dist)
        hm = field_pair[3][1][1]
        data.append( (dist,rotate/hm,colors[str(int(hm*2))]) )

    N = len(data)
    print '%d field pairs collected' % N

    x, y, c = np.array( zip(*data) )
    if mis_angle is None:
        idx = np.ones(N).astype(bool)
    else:
        idx = c == colors[str(mis_angle)]

    plt.scatter(x[idx],y[idx],color=list(c[idx]),s=2)
    plt.axhline(y=1,color='grey',lw=1,alpha=0.2)
    plt.axhline(y=0,color='k',lw=1.5,alpha=0.2)
    plt.axhline(y=-1,color='grey',lw=1,alpha=0.2)
    plt.axvline(x=15,color='k',lw=1,alpha=0.2)
    if abs_dist:
        plt.xlim( 0, 45 )
    else:
        plt.axvline(x=-15,color='k',lw=1,alpha=0.2)
        plt.xlim( -45, 45 )

    plt.xlabel('Distance to Local Cue Boundary (degree)',size='small')
    plt.ylabel('Relative Rotation Angle')
    plt.tick_params(axis='both',labelsize=7)
    plt.title(title_str)

    #plt.show()
    file_name = '%s_%s' % (area,side)
    if abs_dist:
        file_name += '_abs'
    if mis_angle:
        file_name += '_%d' % int(mis_angle)
    plt.savefig(file_name)
    print 'Done'

# end of FUNCTION rotation_cue_scatter

def rotation_orientation_single_panel(area='CA1',side='start'):
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    mis_angles = ['all', 45, 90, 135, 180]
    colors = { '45': 'b', '90': 'c', '135': 'grey', '180': 'm'}

    title_str = area + ' - ' + side

    query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area
    # (rat, day, tc, [ (m1,hm1,s1,e1), (m2,hm2,s2,e2)] )
    field_pairs = get_field_pair_list(query) 
    params = { 'start': [2,], 'end': [3,], 'both': [2,3] }

    data = []
    for field_pair in field_pairs:
        rat, day, tc = field_pair[:3]
        if day > latest_qualified_day(rat,"DR"):
            continue
        for param in params[side]:
            angle_s, angle_m = field_pair[3][0][param], field_pair[3][1][param]
            rotate = angle_m - angle_s
            if abs(rotate) > 180:
                rotate -= np.sign(rotate) * 360
            cue_id, dist = nearest_cue(\
                angle_s,field_pair[3][0][1],'local',cue_id=None)
            hm = field_pair[3][1][1]
            data.append( (dist,float(rotate)/hm,colors[str(int(hm*2))]) )

    N = len(data)
    print '%d field pairs collected' % N

    for i, angle in enumerate(mis_angles):
        #x, y, c = np.array( zip(*data) )
        x = np.array( zip(*data)[0] )
        y = np.array( zip(*data)[1] )
        c = np.array( zip(*data)[2] )
        if angle == 'all':
            idx = np.ones(N).astype(bool)
            lc = 'k'
            lw = 3
        else:
            idx = c == colors[str(angle)]
            lc = colors[ str(int(angle)) ]
            lw = 1

        x, y, c = x[idx], y[idx], c[idx]
        ratios = [ np.sum((x<=-15)*(y>0))/float(np.sum(x<=-15)), \
                   np.sum((x>-15)*(x<15)*(y>0))/float(np.sum((x>-15)*(x<15))), \
                   np.sum((x>=15)*(y>0))/float(np.sum(x>=15)) ]
        plt.plot(ratios,c=lc,lw=lw)
        plt.axhline(y=0.5,color='y',lw=1,alpha=0.2)
        plt.xlim( -0, 2 )
        plt.ylim( 0, 1 )

    plt.tick_params(axis='both',labelsize=7)
    plt.title(title_str)

    file_name = '%s_%s' % (area,side)
    plt.savefig(file_name)
    print 'Done'
# end of FUNCTION rotation_orientation_single_panel

def rotation_orientation_stats(area='CA1',side='start'):
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    mis_angles = ['all', 45, 90, 135, 180]
    colors = { '45': 'b', '90': 'c', '135': 'grey', '180': 'm'}
    params = { 'start': [2,], 'end': [3,], 'both': [2,3] }

    query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area
    # (rat, day, tc, [ (m1,hm1,s1,e1), (m2,hm2,s2,e2)] )
    field_pairs = get_field_pair_list(query) 

    data = []
    for field_pair in field_pairs:
        rat, day, tc = field_pair[:3]
        if day > latest_qualified_day(rat,"DR"):
            continue
        for param in params[side]:
            angle_s, angle_m = field_pair[3][0][param], field_pair[3][1][param]
            rotate = angle_m - angle_s
            if abs(rotate) > 180:
                rotate -= np.sign(rotate) * 360
            cue_id, dist = nearest_cue(\
                angle_s,field_pair[3][0][1],'local',cue_id=None)
            hm = field_pair[3][1][1]
            data.append( (dist,rotate/hm,colors[str(int(hm*2))]) )

    N = len(data)
    print '%d field pairs collected' % N

    for i, angle in enumerate(mis_angles):
        x = np.array( zip(*data)[0] )
        y = np.array( zip(*data)[1] )
        c = np.array( zip(*data)[2] )
        if angle == 'all':
            idx = np.ones(N).astype(bool)
        else:
            idx = c == colors[str(angle)]

        x, y, c = x[idx], y[idx], c[idx]
        nl = np.sum((x>-15)*(x<15)*(y>0))
        ng = np.sum((x>-15)*(x<15)*(y<0))
        fl = np.sum((x<=-15)*(y>0)) + np.sum((x>=15)*(y>0))
        fg = np.sum((x<=-15)*(y<0)) + np.sum((x>=15)*(y<0))
        chi2, p, dof, ex = chi2_contingency( [[nl,ng],[fl,fg]],correction=False )
        print 'For mismatch sessions (%s): chi2=%.6f, p=%.6f' % (str(angle),chi2,p)
        print '  (nl, ng, fl, fg) = %d, %d, %d, %d' % (nl,ng,fl,fg)

    print 'Done'
# end of FUNCTION rotation_orientation_stats

def rotation_angle_stats(area='CA1',norm=True,orientation='local'):
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    mis_angles = ['all', 45, 90, 135, 180]
    colors = { '45': 'b', '90': 'c', '135': 'grey', '180': 'm'}

    query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area
    # (rat, day, tc, [ (m1,hm1,s1,e1), (m2,hm2,s2,e2)] )
    field_pairs = get_field_pair_list(query) 

    data = [ [], [] ] # start, end
    for field_pair in field_pairs:
        rat, day, tc = field_pair[:3]
        if day > latest_qualified_day(rat,"DR"):
            continue
        for param in [2, 3]:
            angle_s, angle_m = field_pair[3][0][param], field_pair[3][1][param]
            rotate = angle_m - angle_s
            if abs(rotate) > 180:
                rotate -= np.sign(rotate) * 360
            cue_id, dist = nearest_cue(\
                angle_s,field_pair[3][0][1],'local',cue_id=None)
            hm = field_pair[3][1][1]
            if orientation=='local' and rotate<=0:
                continue
            if orientation=='global' and rotate>=0:
                continue
            if norm: 
                rotate = float(rotate) / hm
                spread = min( abs(rotate-1), abs(rotate+1) )
            else:
                spread = min( abs(rotate-hm), abs(rotate+hm) )
            data[param-2].append( (spread,colors[str(int(hm*2))]) )

    for i, angle in enumerate(mis_angles):
        side = [None, None]
        for j, d in enumerate(data):
            dcopy = d[:]
            x = np.array( zip(*dcopy)[0] )
            c = np.array( zip(*dcopy)[1] )
            if angle == 'all':
                idx = np.ones(len(dcopy)).astype(bool)
            else:
                idx = c == colors[str(angle)]
            side[j] = x[idx]
        u, p = mannwhitneyu( side[0],side[1] )
        print 'For mismatch sessions (%s): u=%.6f, p=%.6f' % (str(angle),u,p*2)
        print '  (start mean, end mean) = %.4f, %.4f' \
            % (np.mean(side[0]),np.mean(side[1]))

    print 'Done'
# end of FUNCTION rotation_angle_stats

def calc_angle_diff(x, target):
    pos = (x-target+360) % 360
    neg = (target-x+360) % 360
    if pos <= neg:
        return pos
    else:
        return -neg

def nearest_cue( x, half_mis, cues, cue_id=None ):
    if cues == 'local':
        x = (x-half_mis+360) % 360
    elif cues == 'global':
        x = (x+half_mis) % 360

    if cue_id is None: # search for whichever closest cue
        min_dist = 360
        for i, angle in enumerate(BD_ANGLES[cues]):
            dist = calc_angle_diff(x, angle)
            if abs(dist) < abs(min_dist):
                min_dist = dist
                cue_id = i
    else: # calculate the distance from the specified cue
        min_dist = calc_angle_diff(x, BD_ANGLES[cues][cue_id])

    return cue_id, min_dist 

@memoize
def get_field_pair_list(query):
    session_table = get_node('/metadata','sessions')
    table_file = tb.openFile('/home/paramecium/output/fields.h5', 'r')
    field_table = table_file.root.place_fields

    clusts = [ (field['rat'],field['day'],field['session'],field['tc']) \
        for field in field_table.where(query) ]
    single_field_clusts = [ x for x in clusts if clusts.count(x)==1 ]

    field_sessions = {}
    for clust in single_field_clusts:
        rat, day, maze, tc = clust
        data = (rat,day,tc)
        if data not in field_sessions.keys():
            field_sessions[data] = [maze,]
        else:
            field_sessions[data].append(maze)

    field_pairs = []
    for key in field_sessions.keys():
        rat, day, tc = key
        maze_list = np.sort( field_sessions[key] )
        if len(maze_list) == 1:
            continue
        for ma, mb in zip(maze_list[:-1],maze_list[1:]):
            # continuous sessions only
            if mb - ma > 1:
                continue

            # STD-MIS pairs
            fa = get_unique_row( field_table, \
                '(rat==%d)&(day==%d)&(session==%d)&(tc=="%s")' % (rat,day,ma,tc) )
            fb = get_unique_row( field_table, \
                '(rat==%d)&(day==%d)&(session==%d)&(tc=="%s")' % (rat,day,mb,tc) )
            if (fa['type']!='STD') or (fb['type']!='MIS'):
                continue

            field_edges = [] # (maze, half_mis, field start, field end)
            for field, maze in zip([fa, fb], [ma, mb]):
                hm = 0 if field['type']=='STD' else field['mis_angle']/2.0
                field_edges.append( \
                    (maze,hm,field['field_edges'][1],field['field_edges'][0]) )
            field_pairs.append( (rat,day,tc,field_edges) )

    table_file.close()

    return field_pairs
# end of method get_field_pair_list
