# encoding: utf-8
"""
global_effect.py
"""

import sys
import tables as tb
import numpy as np
import matplotlib.pylab as plt

from scipy.stats import mannwhitneyu

from core.analysis import AbstractAnalysis

from scanr.lib import *
from scanr.meta import get_session_metadata

from analysis_shan.data_filters import latest_qualified_day

from ldb.derivative_map import get_nonoverlap_cues, get_overlap_cues

LOCAL, GLOBAL, NOCUE = 0, 1, 2

# ================================================================================

class EdgeCountAnalysis(AbstractAnalysis):
    datadir = './EdgeCountAna'
    label = 'edge count'

    def collect_data(self, area='CA3', stype='STD', subdiv=None):
        self.results = {} 

        table_file = tb.openFile('/home/paramecium/output/fields.h5','r')
        field_table = table_file.root.place_fields
        table_query = \
            '(area=="%s")&(type=="%s")&(expt_type=="DR")&(exclude==False)' \
                % (area,stype)
        if subdiv is not None:
            table_query += '&(subdiv=="%s")' % subdiv

        win_types = [LOCAL, GLOBAL, NOCUE]
        
        edge_count, window_size = {}, {}
        rats = list(set([field['rat'] for field in field_table.where(table_query)]))
        rats.sort()
        for rat in rats[:]:
            field_cnt, count, ds_list, win_size = 0, {}, [], {}

            query = table_query + '&(rat==%d)' % rat
            for field in field_table.where(query):
                day = field['day']
                if day > latest_qualified_day(rat,'DR'):
                    continue
                
                mis_angle = 0 if stype=='STD' else field['mis_angle']
                maze = field['session']
                ds = (day, maze)
                if ds not in ds_list:
                    count[ds] = np.zeros((2,3)) 
                    win_size[ds] = np.zeros(3)
                    for win_type in win_types:
                        windows, total_size = \
                            self.get_windows( stype,mis_angle,win_type )
                        win_size[ds][win_type] = total_size
                    ds_list.append(ds)
                
                for i, edge in enumerate(field['field_edges']): # end, start
                    found = False
                    for win_type in win_types:
                        windows, total_size = \
                            self.get_windows( stype,mis_angle,win_type )
                        for s, e in windows:
                            if s < e:
                                if (edge<s) or (edge>e):
                                    count[ds][1-i][win_type] += 1 # start, end
                                    found = True
                                    break
                            else:
                                if (edge<s) and (edge>e):
                                    count[ds][1-i][win_type] += 1 # start, end
                                    found = True
                                    break
                        if found:
                            break
                field_cnt += 1
            if not field_cnt:
                rats.remove( rat )
            edge_count[rat] = count
            window_size[rat] = win_size
            self.out( '%d field(s) collected for rat %d (%s - %s)' \
                % (field_cnt,rat,area,stype) )

            #for edge in [0, 1]:
            #    check_sum = np.zeros(3)
            #    for key in count.keys():
            #        check_sum += np.array(count[key][edge])
            #    self.out( 'check: (%d, %d, %d)' % tuple(check_sum))

        table_file.close()

        self.results['edge_count'] = edge_count
        self.results['window_size'] = window_size
        self.results['rat_list'] = rats
        self.results['panel'] = '%s - %s' % (area,stype)
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, nbin=10):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        # LOCAL, GLOBAL, NOCUE = 0, 1, 2
        START, END = 0, 1 
        bins = np.linspace(-1,1,nbin+1)

        rats = self.results['rat_list']
        for i, norm_session in enumerate([True, False]):
            global_idx, local_idx = [[], []], [[], []] # start, end
            for rat in rats:
                count = self.results['edge_count'][rat]
                win_size = self.results['window_size'][rat]

                for edge in [START, END]:
                    data = []
                    for wtype in [LOCAL, GLOBAL, NOCUE]:
                        if norm_session:
                            freqs = []
                            for key in count.keys():
                                freqs.append( float(count[key][edge][wtype]) / \
                                    win_size[key][wtype] )
                            data.append( np.mean(freqs) )
                        else:
                            cnt, wsize = 0, 0
                            for key in count.keys():
                                cnt += count[key][edge][wtype]
                                wsize += win_size[key][wtype]
                            data.append( float(cnt)/wsize )

                    f_l, f_g, f_n = data
                    idx = 0 if f_l+f_n==0 else float(f_l-f_n)/(f_l+f_n)
                    local_idx[edge].append( idx )
                    idx = 0 if f_g+f_n==0 else float(f_g-f_n)/(f_g+f_n)
                    global_idx[edge].append( idx )

            for edge in [START, END]:
                title_str = 'Field Start' if edge==START else 'Field End'  
                if norm_session:
                    title_str += ' (avg w/in session)'
                else:
                    title_str += ' (combine session)'
                plt.subplot(2,2,2*edge+i+1)
                plt.hist([local_idx[edge], global_idx[edge]], \
                    bins=bins,color=['c','r'])
                plt.legend( ['local', 'global'], loc='best')
                plt.title(title_str)
                plt.xlim(-1, 1)
        plt.suptitle( self.results['panel'] )

        plt.show()
    # end of METHOD process_data

    def get_windows(self, stype, mis_angle, wtype):
        # LOCAL, GLOBAL, NOCUE = 0, 1, 2
        if wtype == NOCUE:
            set_1 = np.array( [(30,15), (120,60), (165,150), (210,195), \
                               (255,240), (300,285), (345,330)] )
            set_2 = np.array( [(30,15), (75,60), (120,105), (165,150), \
                               (255,195), (345,285)] )
            if (stype=='STD') or (mis_angle==90) or (mis_angle==180):
                windows = (set_1-mis_angle/2.0+360) % 360
                total_size = 150
            else:
                windows = (set_2-mis_angle/2.0+360) % 360
                total_size = 180
        else:
            cue_type = ['local', 'global']
            delta = mis_angle/2.0 if wtype==LOCAL else -mis_angle/2.0
            cues = delta + \
                get_nonoverlap_cues(cue_type[wtype], stype, mis_angle=mis_angle)
            # windows = [ (s1,e1), (s2,e2), ... ]
            windows = np.array( \
                [((cue+15+360)%360,(cue-15+360)%360) for cue in cues] )
            total_size = len(cues) * 30

        return windows, total_size
    # end of METHOD get_windows

# end of CLASS EdgeCountAnalysis

# ================================================================================

class CmpCleanCueAnalysis(AbstractAnalysis):
    datadir = './CmpCleanCues'
    label = 'compare clean cues analysis'

    def collect_data(self, N=1000):
        self.results = {} 

        table_file = tb.openFile('/home/paramecium/output/fields.h5','r')
        field_table = table_file.root.place_fields
        table_query = '(expt_type=="DR")&(exclude==False)'

        win_types = [LOCAL, GLOBAL, NOCUE]
        data = {}

        panel_list = []
        for area in ['CA1', 'CA3']:
            for ses in ['STD', 'MIS']:
                panel = area + '_' + ses

                rds_id, rds_list = -1, []
                old_rds = None
                edge_types, edge_rds, mis_angles = [], [], []
                query = table_query + '&(area=="%s")&(type=="%s")' % (area,ses)
                for field in field_table.where(query):
                    rat, day = field['rat'], field['day']
                    if day > latest_qualified_day(rat,'DR'):
                        continue

                    mis_angle = 0 if ses=='STD' else field['mis_angle']

                    maze = field['session']
                    rds = (rat, day, maze)
                    if rds != old_rds:
                        windows, total_sizes = {}, {}
                        for win_type in win_types:
                            windows[win_type], total_sizes[win_type] = \
                                self.get_windows( ses,mis_angle,win_type )
                        old_rds = rds

                    for i, edge in enumerate(field['field_edges']): # end, start
                        found = False
                        for win_type in win_types:
                            for s, e in windows[win_type]:
                                if s < e:
                                    if (edge<s) or (edge>e):
                                        edge_types.append( win_type )
                                        edge_rds.append( rds_id )
                                        mis_angles.append( mis_angle )
                                        found = True
                                        break
                                else:
                                    if (edge<s) and (edge>e):
                                        edge_types.append( win_type )
                                        edge_rds.append( rds_id )
                                        mis_angles.append( mis_angle )
                                        found = True
                                        break
                            if found:
                                if rds not in rds_list:
                                    rds_list.append( rds )
                                    rds_id += 1
                                break

                if not len(edge_rds):
                    continue

                edge_rds = np.array(edge_rds)
                edge_types = np.array(edge_types)
                mis_angles = np.array(mis_angles)
                if not len(edge_rds):
                    continue

                bootstrap, bootstrap_ix = [], []
                new_index = np.random.randint(rds_id+1, size=(N,rds_id+1))
                for i, idx in enumerate(new_index):
                    if np.mod(i,10) == 0:
                        sys.stdout.write('.')
                        sys.stdout.flush()
                    edge_count, sizes = [[], [], []], [[], [], []]
                    index = np.zeros( len(edge_rds) ).astype(bool)
                    for rid in np.array(range(rds_id+1))[idx]:
                        select_edge = edge_rds == rid
                        if not np.sum(select_edge):
                            edge
                        select_etype = edge_types[select_edge]
                        mis_angle = mis_angles[select_edge][0]
                        index += select_edge

                        for win_type in win_types:
                            windows, total_sizes = {}, {}
                            windows[win_type], total_sizes[win_type] = \
                                self.get_windows( ses, mis_angle, win_type )
                            edge_count[win_type].append( \
                                np.sum(select_etype==win_type) ) 
                            sizes[win_type].append( total_sizes[win_type] )
                    bootstrap_ix.append( index )
                    bootstrap.append( (edge_count, sizes) )
                self.out('')

                panel_list.append(panel)
                data[panel] = {}
                data[panel]['edge_rds'] = edge_rds
                data[panel]['edge_types'] = edge_types
                data[panel]['bootstrap'] = bootstrap
                data[panel]['bootstrap_ix'] = bootstrap_ix

        table_file.close()

        self.results['panel_list'] = panel_list
        self.results['data'] = data

        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, nbin=10):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        win_types = [ LOCAL, GLOBAL, NOCUE ]
        colors = ['r', 'g', 'b']
        panel_list = ['CA1_STD', 'CA1_MIS', 'CA3_STD', 'CA3_MIS']

        for i, panel in enumerate(panel_list):
            area, ses = panel.split('_')
            
            plt.subplot(2,2,i+1)
            bs = self.results['data'][panel]['bootstrap']
            N = len(bs)
            for win_type in win_types:
                freq = []
                for i in range(N):
                    counts = np.array( bs[i][0][win_type] ).astype(float)
                    sizes = np.array( bs[i][1][win_type] )
                    freq.append( np.mean(counts/sizes) )
                plt.hist( freq, bins=20+1, color=colors[win_type], alpha=0.5 )
            plt.legend( ['local', 'global', 'no-cue'] )
            plt.title( panel )

        plt.show()
    # end of METHOD process_data

    def difference_distribution(self, nbin=20):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        win_types = [ LOCAL, GLOBAL, NOCUE ]
        title_str = ['local-none', 'global-none', 'local-global']
        panel_list = ['CA1_STD', 'CA1_MIS', 'CA3_STD', 'CA3_MIS']

        for i, panel in enumerate(panel_list):
            area, ses = panel.split('_')
            
            bs = self.results['data'][panel]['bootstrap']
            N = len(bs)
            freq = [[], [], []]
            for win_type in win_types:
                for j in range(N):
                    counts = np.array( bs[j][0][win_type] ).astype(float)
                    sizes = np.array( bs[j][1][win_type] )
                    freq[win_type].append( np.mean(counts/sizes) )
                freq[win_type] = np.array(freq[win_type])

            hist_data = [ freq[LOCAL]-freq[NOCUE], \
                          freq[GLOBAL]-freq[NOCUE], \
                          freq[LOCAL]-freq[GLOBAL] ]
            for j, hdata in enumerate(hist_data):
                plt.subplot(4,3,3*i+j+1)
                bins = np.linspace(-0.015,0.025,nbin+1)
                plt.hist( hdata, bins=bins, color='grey')
                plt.axvline(x=0,lw=2,color='k')
                if j == 0:
                    plt.ylabel(panel)
                plt.title( title_str[j] )
                plt.xlim( [-0.015, 0.025] )

        plt.show()
    # end of METHOD process_data

    def get_windows(self, stype, mis_angle, wtype):
        # LOCAL, GLOBAL, NOCUE = 0, 1, 2
        if wtype == NOCUE:
            set_1 = np.array( [(30,15), (120,60), (165,150), (210,195), \
                               (255,240), (300,285), (345,330)] )
            set_2 = np.array( [(30,15), (75,60), (120,105), (165,150), \
                               (255,195), (345,285)] )
            if (stype=='STD') or (mis_angle==90) or (mis_angle==180):
                windows = (set_1-mis_angle/2.0+360) % 360
                total_size = 150
            else:
                windows = (set_2-mis_angle/2.0+360) % 360
                total_size = 180
        else:
            cue_type = ['local', 'global']
            delta = mis_angle/2.0 if wtype==LOCAL else -mis_angle/2.0
            cues = delta + \
                get_nonoverlap_cues(cue_type[wtype], stype, mis_angle=mis_angle)
            # windows = [ (s1,e1), (s2,e2), ... ]
            windows = np.array( \
                [((cue+15+360)%360,(cue-15+360)%360) for cue in cues] )
            total_size = len(cues) * 30

        return windows, total_size
    # end of METHOD get_windows

# end of CLASS EdgeCountAnalysis

# ================================================================================

class OverlapCueAnalysis(AbstractAnalysis):

    datadir = './OverlapCues'
    label = 'overlap cues analysis'

    def collect_data(self, N=1000):
        self.results = {} 

        table_file = tb.openFile('/home/paramecium/output/fields.h5','r')
        field_table = table_file.root.place_fields
        table_query = '(expt_type=="DR")&(exclude==False)'

        OVERLAP, CLEAN = 0, 1
        win_types = [OVERLAP, CLEAN]
        data = {}

        panel_list = []
        for area in ['CA1', 'CA3']:
            for ses in ['STD', 'MIS']:
                panel = area + '_' + ses

                rds_id, rds_list = -1, []
                old_rds = None
                edge_types, edge_rds, mis_angles = [], [], []
                query = table_query + '&(area=="%s")&(type=="%s")' % (area,ses)
                for field in field_table.where(query):
                    rat, day = field['rat'], field['day']
                    if day > latest_qualified_day(rat,'DR'):
                        continue

                    mis_angle = 0 if ses=='STD' else field['mis_angle']

                    maze = field['session']
                    rds = (rat, day, maze)
                    if rds != old_rds:
                        windows, total_sizes = {}, {}
                        for win_type in win_types:
                            windows[win_type], total_sizes[win_type] = \
                                self.get_windows( ses,mis_angle,win_type )
                        old_rds = rds

                    for i, edge in enumerate(field['field_edges']): # end, start
                        found = False
                        for win_type in win_types:
                            for s, e in windows[win_type]:
                                if s < e:
                                    if (edge<s) or (edge>e):
                                        edge_types.append( win_type )
                                        edge_rds.append( rds_id )
                                        mis_angles.append( mis_angle )
                                        found = True
                                        break
                                else:
                                    if (edge<s) and (edge>e):
                                        edge_types.append( win_type )
                                        edge_rds.append( rds_id )
                                        mis_angles.append( mis_angle )
                                        found = True
                                        break
                            if found:
                                if rds not in rds_list:
                                    rds_list.append( rds )
                                    rds_id += 1
                                break

                if not len(edge_rds):
                    continue

                edge_rds = np.array(edge_rds)
                edge_types = np.array(edge_types)
                mis_angles = np.array(mis_angles)
                if not len(edge_rds):
                    continue

                edge_count, sizes = [[], [], []], [[], [], []]
                for rid in np.array(range(rds_id+1)):
                    select_edge = edge_rds == rid
                    if not np.sum(select_edge):
                        edge
                    select_etype = edge_types[select_edge]
                    mis_angle = mis_angles[select_edge][0]

                    for win_type in win_types:
                        windows, total_sizes = {}, {}
                        windows[win_type], total_sizes[win_type] = \
                            self.get_windows( ses, mis_angle, win_type )
                        edge_count[win_type].append( \
                            np.sum(select_etype==win_type) ) 
                        sizes[win_type].append( total_sizes[win_type] )
                obs = (edge_count, sizes)

                bootstrap, bootstrap_ix = [], []
                new_index = np.random.randint(rds_id+1, size=(N,rds_id+1))
                for i, idx in enumerate(new_index):
                    if np.mod(i,10) == 0:
                        sys.stdout.write('.')
                        sys.stdout.flush()
                    edge_count, sizes = [[], [], []], [[], [], []]
                    index = np.zeros( len(edge_rds) ).astype(bool)
                    for rid in np.array(range(rds_id+1))[idx]:
                        select_edge = edge_rds == rid
                        if not np.sum(select_edge):
                            edge
                        select_etype = edge_types[select_edge]
                        mis_angle = mis_angles[select_edge][0]
                        index += select_edge

                        for win_type in win_types:
                            windows, total_sizes = {}, {}
                            windows[win_type], total_sizes[win_type] = \
                                self.get_windows( ses, mis_angle, win_type )
                            edge_count[win_type].append( \
                                np.sum(select_etype==win_type) ) 
                            sizes[win_type].append( total_sizes[win_type] )
                    bootstrap_ix.append( index )
                    bootstrap.append( (edge_count, sizes) )
                self.out('')

                panel_list.append(panel)
                data[panel] = {}
                data[panel]['obs'] = obs
                data[panel]['edge_rds'] = edge_rds
                data[panel]['edge_types'] = edge_types
                data[panel]['bootstrap'] = bootstrap
                data[panel]['bootstrap_ix'] = bootstrap_ix

        table_file.close()

        self.results['panel_list'] = panel_list
        self.results['data'] = data

        self.out('Done')
    # end of METHOD collect_data

    def plot_original(self, nbin=20):

        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        OVERLAP, CLEAN = 0, 1
        win_types = [OVERLAP, CLEAN]
        colors = ['b', 'c']
        panel_list = ['CA1_STD', 'CA1_MIS', 'CA3_STD', 'CA3_MIS']

        for i, panel in enumerate(panel_list):
            area, ses = panel.split('_')
            
            plt.subplot(2,2,i+1)
            obs = self.results['data'][panel]['obs']
            
            freq = [[], []]
            for win_type in win_types:
                counts = np.array( obs[0][win_type] ).astype(float)
                sizes = np.array( obs[1][win_type] )
                freq[win_type] = counts / sizes 
                #plt.hist( freq[win_type], bins=nbin+1, \
                #    color=colors[win_type], alpha=0.5 )
                plt.hist( freq[win_type], bins=np.linspace(0,0.45,nbin+1), \
                    color=colors[win_type], alpha=0.5 )
                plt.axvline(x=np.mean(freq[win_type]),lw=2,color=colors[win_type])
            plt.legend( ['overlap', 'clean'], loc='best' )
            plt.title( panel )

            print 'Mannwhitney test: u=%.4f, p=%.4f' % mannwhitneyu(freq[0],freq[1])

        plt.show()
    # end of METHOD process_data

    def process_data(self, nbin=10):

        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        OVERLAP, CLEAN = 0, 1
        win_types = [OVERLAP, CLEAN]
        colors = ['b', 'c']
        panel_list = ['CA1_STD', 'CA1_MIS', 'CA3_STD', 'CA3_MIS']

        for i, panel in enumerate(panel_list):
            area, ses = panel.split('_')
            
            plt.subplot(2,2,i+1)
            bs = self.results['data'][panel]['bootstrap']
            N = len(bs)
            freq = [[], []]
            for win_type in win_types:
                for i in range(N):
                    counts = np.array( bs[i][0][win_type] ).astype(float)
                    sizes = np.array( bs[i][1][win_type] )
                    freq[win_type].append( np.mean(counts/sizes) )
                #plt.hist( freq[win_type], bins=20+1, \
                #    color=colors[win_type], alpha=0.5 )
                plt.hist( freq[win_type], bins=np.linspace(0.025,0.075,21), \
                    color=colors[win_type], alpha=0.5 )
            plt.legend( ['overlap', 'clean'], loc='best' )
            plt.title( panel )

            print 'paired t-test: t=%.4f, p=%.4f' % ttest_rel(freq[0],freq[1])

        plt.show()
    # end of METHOD process_data

    def difference_distribution(self, nbin=20, norm=True):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        OVERLAP, CLEAN = 0, 1
        win_types = [OVERLAP, CLEAN]
        title_str = 'overlap-clean'
        panel_list = ['CA1_STD', 'CA1_MIS', 'CA3_STD', 'CA3_MIS']

        for i, panel in enumerate(panel_list):
            area, ses = panel.split('_')
            
            bs = self.results['data'][panel]['bootstrap']
            N = len(bs)
            freq = [[], [], []]
            for win_type in win_types:
                for j in range(N):
                    counts = np.array( bs[j][0][win_type] ).astype(float)
                    sizes = np.array( bs[j][1][win_type] )
                    freq[win_type].append( np.mean(counts/sizes) )
                freq[win_type] = np.array(freq[win_type])

            f_o = np.array(freq[OVERLAP]).astype(float)
            f_c = np.array(freq[CLEAN]).astype(float)
            if norm: # difference index
                valid = (f_o>0) * (f_c>0)
                hdata = (f_o[valid]-f_c[valid]) / (f_o[valid]+f_c[valid])
            else:
                hdata = f_o - f_c

            plt.subplot(2,2,i+1)
            #bins = np.linspace(-0.015,0.025,nbin+1)
            plt.hist( hdata, bins=nbin, color='grey')
            plt.axvline(x=0,lw=2,color='k')
            if j == 0:
                plt.ylabel(panel)
            plt.title( panel )
            #plt.xlim( [-0.015, 0.025] )

        plt.show()
    # end of METHOD process_data

    def get_windows(self, stype, mis_angle, wtype):
        OVERLAP, CLEAN = 0, 1
        delta = mis_angle / 2.0

        if wtype == OVERLAP:
            cues = delta + get_overlap_cues('local', stype, mis_angle=mis_angle)
        elif wtype == CLEAN:
            cues = delta + get_nonoverlap_cues('local', stype, mis_angle=mis_angle)
        # windows = [ (s1,e1), (s2,e2), ... ]
        windows = np.array( \
            [((cue+15+360)%360,(cue-15+360)%360) for cue in cues] )
        total_size = len(cues) * 30

        return windows, total_size
    # end of METHOD get_windows

# end of CLASS EdgeCountAnalysis
