# encoding: utf-8
"""
group_comparison.py

Collections of miscellaneous comparisons between areas, session types, etc.
"""

import tables as tb
import numpy as np
import matplotlib.pylab as plt
import os

from core.analysis import AbstractAnalysis

from scanr.lib import *

from analysis_shan.data_filters import latest_qualified_day

# Constants

FIELD_FILE = '/home/paramecium/output/fields.h5' # place field table file
#need to be modified 

# cue locations (track angle in degrees)
BD_ANGLES = { 'local': [45, 135, 225, 315], \
              'global': [0, 45, 135, 180, 270] }
MARGIN = 15 # half-window size for quadrant boundary windows (in degrees)

# ---------------------------------------------------------------------------------

class SessionTypeCompAna(AbstractAnalysis):
    """
    Compare the field edge in-window ratios between STD and MIS sessions for CA1 and
    CA3
    """

    datadir = './STDvsMIS'
    label = 'session type compare'

    def collect_data(self,ft='both',margin=MARGIN):
        """
        For each rat, count how many field edges fell within the quadrant boundary
        windows and how many fell outside.
        """
        self.out.timestamp = False
        self.results = {}

        table_file = tb.openFile(FIELD_FILE,'r')
        field_table = table_file.root.place_fields
        table_query = '(expt_type=="DR")&(exclude==False)'

        areas = ['CA1', 'CA3']
        sess_types = ['STD', 'MIS']
        bd_angles = BD_ANGLES['local']

        edge_counts = dict( [ (area,dict([(stype,{}) for stype in sess_types])) \
            for area in areas] )

        for area in areas:
            area_query = table_query + '&(area=="%s")' % area
            for stype in sess_types:
                self.out('Collecting %s-%s data:' % (area,stype))
                query = area_query + '&(type=="%s")' % stype
                rat_list = []
                cnt = dict() 
                for field in field_table.where(query):
                    rat, day = field['rat'], field['day']
                    if day > latest_qualified_day(rat,'DR'):
                        continue
                    if rat not in cnt.keys():
                        cnt[rat] = [0, 0] # edge_in, edge_out
                    rat_list.append(rat)

                    if ft == 'com':
                        data = [ field['COM'], ]
                    elif ft == 'start':
                        data = [ field['field_edges'][1], ]
                    elif ft == 'end':
                        data = [ field['field_edges'][0], ]
                    elif ft == 'both':
                        data = field['field_edges']
                    else:
                        print 'Wrong arg!'

                    for datum in data:
                        alpha = (datum-field['mis_angle']/2.0+360) % 360
                        out_flag = True
                        for bd in bd_angles:
                            if abs(alpha-bd) < margin: # within qb window
                                cnt[rat][0] += 1 # edge_in ++
                                out_flag = False
                                break
                        if out_flag:
                            cnt[rat][1] += 1 # edge_out ++
                edge_counts[area][stype].update( cnt )

                rat_list = sorted( list(set(rat_list)) )
                for rat in rat_list:
                    self.out('Rat %d has %d data points' % (rat,np.sum(cnt[rat])))
                self.out('\n')

        self.results['areas'] = areas
        self.results['sess_types'] = sess_types
        self.results['ft'] = ft
        self.results['edge_counts'] = edge_counts
    # end of METHOD collect_data

    def process_data(self,min_filter=None,nbin=15):
        """
        Plot the session type X area field edge in-window ratio distributions
        """
        areas = self.results['areas']
        sess_types = self.results['sess_types']
        ft = self.results['ft']
        edge_counts = self.results['edge_counts']

        self.init_graph()
        bins = np.linspace(0,1,nbin+1)

        for i, area in enumerate(areas):
            edges = edge_counts[area]
            rats = [ rat for rat in edges['STD'].keys() \
                if rat in edges['MIS'].keys() ]
            if min_filter is None:
                rat_list = rats[:]
            else:
                rat_list = []
                for rat in rats:
                    m = min( np.sum(edges['STD'][rat]), np.sum(edges['MIS'][rat]) )
                    if m >= min_filter:
                        rat_list.append( rat )
            self.out('%d %s rats selected' % (len(rat_list),area))
            for j, stype in enumerate(sess_types):
                plt.subplot(2,2,j*2+(i+1))
                total_in, total = 0, 0
                ratio = []
                for rat in rat_list:
                    e_cnt = edges[stype][rat]
                    total_in += e_cnt[0]
                    total += np.sum(e_cnt)
                    ratio.append( e_cnt[0]/float(e_cnt[0]+e_cnt[1]) )
                plt.axvline(x=np.mean(ratio),lw=1,color='r')
                plt.axvline(x=float(total_in)/total,lw=1,color='g')
                plt.legend( ['Mean','Overall'],fontsize='x-small' )
                plt.hist(ratio,bins=bins,color='c')
                if j == 1: # bottom row
                    plt.xlabel(area)
                if i == 0: # left-most column
                    plt.ylabel(stype)
                plt.tick_params(axis='both',labelsize=7)
                plt.xlim(0,1)
        plt.suptitle('Field Edge (%s) In-Window Ratio' % ft)

        file_str = ft + '_edge_ratio_hist'
        if min_filter is not None:
            file_str += '_min-%d' % min_filter
        file_str += '.pdf'
        plt.savefig( os.path.join(self.datadir,file_str) )

        self.out('Done')
    # end of METHOD process_data

    def init_graph(self):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)
    # end of METHOD init_graph

# end of class SessoinTypeCompAna
