# encoding: utf-8
"""
ld_gradient_map.py

Population spatial derivatives analyses
"""

import tables as tb
import numpy as np
import matplotlib.pylab as plt

from random import randint

from tools.misc import memoize

from analysis_shan.texture_boundary import ModifiedSessionData
from analysis_shan.data_filters import latest_qualified_day

# should be consistant with PlaceFieldReport
BINS = 360 # need to fix pop_grad_shuffle_test if not 360
BLUR_WIDTH = 4.3
BD_ANGLES = { 'local':[45, 135, 225, 315], 'global':[0, 45, 135, 180, 270] }
MARGIN = 15 # in degree

# --------------------------------------------------------------------------------

def population_rate_diff_map(area='CA1',norm=True,rec='abs',scale='one',rot='local'):
    """
    Plot the spatial derivative maps for STD and MIS sessions separately, and for
    each different MIS angle

    norm     - normalize individual cell map by maximum value if True
    rec      - ['abs', 'inc', 'dec']
               abs: collect both the rate increasing and decreasing parts (the 
                    absolute values of rate differences between neighboring bins)
               inc: collect only the rate increasing part
               dec: collect only the rate decreasing part
    scale    -  ['one', 'cell']
               one: rescale the population map to [0,1]
               cell: average by cell number
    rot      - ['local', 'global'] rotate the MIS data following the specified cues
    """
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    stypes = ['STD','MIS','MIS','MIS','MIS','MIS']
    mis_angles = [None, None, 45, 90, 135, 180]

    for i, panel in enumerate( zip(stypes,mis_angles) ):
        stype, mis_angle = panel
        title_str = stype if mis_angle is None else stype+' - '+str(mis_angle)

        pop_diff = get_population_derivative( area,stype,mis_angle=mis_angle, \
            scale=scale,norm=norm,rec=rec,rot=rot )

        r_min, r_max = 1.5, 2.0
        theta = np.linspace(0,360,BINS+1) * np.pi / 180.0 # in rad
        rad = np.array( [r_min, r_max] )
        X, Y = np.meshgrid(theta,rad)
        plt.subplot(3,2,i+1,projection='polar')
        plt.pcolormesh( theta, rad, np.tile(pop_diff,(2,1)),cmap='gray_r' )
        adj = 0 if mis_angle is None else mis_angle/2.0
        if rot=='local' or stype=='STD' or mis_angle!=None:
            for alpha in BD_ANGLES['local']:
                a = (alpha+adj) * np.pi / 180.0
                plt.plot( [a, a], [0.5,r_min], 'r' )
        if rot=='global' or stype=='STD' or mis_angle!=None:
            for alpha in BD_ANGLES['global']:
                a = (alpha-adj) * np.pi / 180.0
                plt.plot( [a, a], [1.0,r_min], 'g', lw=5, zorder=-1 )

        plt.colorbar()
        plt.yticks([])
        plt.ylim([0,r_max])
        plt.title(title_str)

    plt.show()
# end of FUNCTION population_rate_diff_map 

def pop_grad_shuffle_test(N=1000,nbin=40,window='both',rec='abs'):
    """
    Need to be modified
    """
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    select_in = np.zeros(BINS).astype(bool)
    for alpha in BD_ANGLES['local']:
        if window == 'both':
            select_in[alpha-MARGIN:alpha+MARGIN] = True
        elif window == 'pre':
            select_in[alpha:alpha+MARGIN] = True
        elif window == 'post':
            select_in[alpha-MARGIN:alpha] = True

    for i, stype in enumerate( ['STD','MIS'] ):
        cell_list = collect_cell_list("CA1",stype) 

        pop_diff = np.zeros(BINS)
        for rds, tc, mis in cell_list:
            D = get_1d_rate_diff_map(rds, tc, mis, rec=rec)
            pop_diff += D
        obs = calculate_avg_diff( pop_diff, select_in )

        shuffle = []
        for j in xrange(N):
            pop_diff = np.zeros(BINS)
            for rds, tc, mis in cell_list:
                D = get_1d_rate_diff_map(rds, tc, mis)
                pop_diff += np.roll(D, randint(0,BINS-1))
            shuffle.append( calculate_avg_diff( pop_diff, select_in ) )
        
        n = np.sum( shuffle>obs )
        p = float(n) / N
        print('p = %.4f (%d/%d)' % (p,n,N))

        x_min = min( obs, min(shuffle) )
        x_max = max( obs, max(shuffle) )
        bins = np.linspace(x_min,x_max,nbin+1)
        bin_size = bins[1] - bins[0]

        plt.subplot(1,2,i+1)
        plt.hist(shuffle, bins=bins)
        plt.axvline(x=obs,lw=2,color='r')
        plt.xlim( [x_min-bin_size, x_max+bin_size] )
        plt.tick_params(axis='both',labelsize=7)
        plt.title(stype,size='small')
        plt.suptitle('Average Difference Shuufle Test')

    plt.show()
# end of FUNCTION pop_grad_shuff

def rate_diff_PSTH(area='CA1',stype='STD',mis_angle=None,norm=True,rec='abs',half_win=45,clean=False):
    """
    Plot the averaged peri-cue spatial derivative curves for specified area and
    session type

    norm     - normalize individual cell map by maximum value if True
    rec      - ['abs', 'inc', 'dec']
                abs: collect both the rate increasing and decreasing parts (the 
                     absolute values of rate differences between neighboring bins)
                inc: collect only the rate increasing part
                dec: collect only the rate decreasing part
    half_win - half window width
    clean    - use only local cues not overlapped with global cues, or global cues
               not overlapped with local cues
    """
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    colors = ['r', 'g'] # local: red, global: green

    for i, panel in enumerate( ['local', 'global'] ):
        title_str = area + '-' + stype
        if mis_angle is not None:
            title_str += ' (%d)' % mis_angle
        cell_list = collect_cell_list(area,stype,mis_angle) 

        x = np.linspace(-half_win,half_win-1,2*half_win) + 0.5

        adj = 0 if mis_angle is None else mis_angle/2.0
        if panel == 'local':
            angles = (np.array(BD_ANGLES[panel])+adj)%360 + 360
        elif panel == 'global':
            angles = (np.array(BD_ANGLES[panel])-adj+360)%360 + 360

        acc_diff = np.zeros(2*half_win)
        count = 0
        for rds, tc, mis in cell_list:
            D = get_1d_rate_diff_map(rds, tc, mis, norm=norm, rec=rec, rot=panel)
            ext_D = np.tile(D,3)

            if clean:
                if panel == 'local':
                    if stype == 'STD':
                        sel_angles = angles[ np.array([2,3]) ]
                    elif stype == 'MIS':
                        if mis_angle==45 or mis==22.5:
                            sel_angles = np.array( [angles[0],] )
                        elif mis_angle==90 or mis==45:
                            sel_angles = angles[ np.array([1,2]) ]
                        elif mis_angle==135 or mis==67.5:
                            sel_angles = np.array( [angles[3],] )
                        elif mis_angle==180 or mis==90:
                            sel_angles = angles[ np.array([0,1]) ]
                elif panel == 'global':
                    if stype == 'STD':
                        sel_angles = angles[ np.array([0,3,4]) ]
                    elif stype == 'MIS':
                        if mis_angle==45 or mis==22.5:
                            sel_angles = angles[ np.array([1,2]) ]
                        elif mis_angle==90 or mis==45:
                            sel_angles = angles[ np.array([0,3,4]) ]
                        elif mis_angle==135 or mis==67.5:
                            sel_angles = angles[ np.array([1,2]) ]
                        elif mis_angle==180 or mis==90:
                            sel_angles = angles[ np.array([0,3,4]) ]
            else:
                sel_angles = angles.copy()

            for alpha in sel_angles:
                extract = ext_D[alpha-half_win:alpha+half_win]
                acc_diff += extract
                count += 1
        print '%d %s cues detected for %d cells...' % (count,panel,len(cell_list))
        acc_diff /= float(count)

        plt.plot(x,acc_diff,colors[i],lw=2,zorder=-100)
        plt.axvline(x=0,lw=1,color='k')
        plt.xlim( [x[0],x[-1]] )

    plt.title(title_str)

    plt.show()
# end of FUNCTION rate_diff_PSTH

# --------------------------------------------------------------------------------

def collect_cell_list(area, stype, mis_angle=None):
    table_file = tb.openFile('/home/paramecium/output/fields.h5','r')
    field_table = table_file.root.place_fields
    table_query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area

    cell_list = []
    for field in field_table.where(table_query+'&(type=="%s")'%stype):
        rat, day = field['rat'], field['day']
        if day > latest_qualified_day(rat,"DR"):
            continue
        if mis_angle is not None and field['mis_angle']!=mis_angle:
            continue
        mis = 0 if stype=='STD' or mis_angle!=None else field['mis_angle']/2.0
        clust = ((rat,day,field['session']),field['tc'],mis) # (rds,tc,half_mis)
        if clust not in cell_list:
            cell_list.append(clust)

    table_file.close()

    return cell_list
# end of FUNCTION collect_cell_list

def calculate_avg_diff( pd, select ):
    # normalize population rate difference
    pd_max, pd_min = np.max(pd), np.min(pd)
    norm_pd = (pd-pd_min) / float(pd_max-pd_min)

    avg_in = np.sum(norm_pd[select]) / float(np.sum(select))
    avg_out = np.sum(norm_pd[~select]) / float(np.sum(~select))

    return avg_in - avg_out
# end of FUNCTION calculate_avg_diff

def get_population_derivative(area,stype,mis_angle=None,scale='one',**traits):
    pop_diff = np.zeros(BINS)

    cell_list = collect_cell_list(area,stype,mis_angle) 
    for rds, tc, mis in cell_list:
        D = get_1d_rate_diff_map(rds, tc, mis, **traits)
        pop_diff += D

    if scale == 'one': # re=scale to [0,1]
        pd_max, pd_min = np.max(pop_diff), np.min(pop_diff)
        pop_diff = (pop_diff-pd_min) / float(pd_max-pd_min)
    if scale == 'cell': # average by cell number
        pop_diff /= len(cell_list)

    return pop_diff
# end of FUNCTION get_population_derivative

@memoize
def get_1d_rate_diff_map(rds, tc, mis, norm=True, rec=None, **traits):
    R = get_1d_ratemap(rds, tc, mis, **traits)
    D = np.roll(R,-1) - R

    if rec == 'abs': # collect the absolute value of diff
        D = abs(D)
    elif rec == 'inc':
        D[D>0] = 0 # collect when the firing rate is increasing  
        D *= -1
    elif rec == 'dec':
        D[D<0] = 0 # collect when the firing rate is decreasing  

    if norm:
        D = D / float(np.max(D))
    return D
# end of FUNCTION get_1d_rate_diff_map

def get_1d_ratemap(rds, tc, mis, ratemap_filter='running', rot='local'):
    session = ModifiedSessionData.get(rds)
    if ratemap_filter == 'running':
        R = session.get_cluster_ratemap( tc,bins=BINS,mis=mis,rot=rot, \
            blur_width=BLUR_WIDTH,**session.running_filter() )
    elif ratemap_filter == 'off_track':
        R = session.get_cluster_ratemap( tc,bins=BINS,mis=mis,rot=rot,\
            blur_width=BLUR_WIDTH,velocity_filter=False, exclude_off_track=True )
    elif ratemap_filter == 'none':
        R = session.get_cluster_ratemap( tc,bins=BINS,mis=mis,rot=rot, \
            blur_width=BLUR_WIDTH,velocity_filter=False )
    return R
# end of FUNCTION get_ratemap
