#encoding: utf-8
"""
per_lap.py
"""

import pdb
import os
import numpy as np
import matplotlib.pylab as plt
import tables as tb
import operator as op

from core.analysis import AbstractAnalysis
from core.report import BaseReport
from tools.radians import circle_diff_vec_deg
from tools.misc import contiguous_groups

from scanr.lib import *
from scanr.config import Config
from scanr.spike import TetrodeSelect
from scanr.data import get_node, get_unique_row
from scanr.session import SessionData
from scanr.tracking import get_tracking
from scanr.time import time_slice_sample, exclude_from, select_from, stamp_to_time
from scanr.cluster import get_min_quality_criterion, AND, PlaceCellCriteria
from scanr.field import mark_all_fields, field_extent

from analysis_shan.data_filters import latest_qualified_day

BINS = Config['placefield']['default_bins']
MIN_RATE = Config['placefield']['min_peak_rate']

# =================================================================================

class ScanLapFieldReport(BaseReport):
    datadir = 'ScanLap'
    label = 'scan lap report'

    def collect_data(self, area=None, stype=None):
        self.results = {}

        table_file = tb.openFile('/home/paramecium/output/fields.h5')
        field_table = table_file.root.place_fields
        qb = np.array( [45, 135, 225, 315] )

        table_query = '(expt_type=="DR")&(exclude==False)'
        if area is not None:
            table_query += '&(area=="%s")' % area
        if stype is not None:
            table_query += '&(type=="%s")' % stype 

        # delete clusters with multiple fields
        delete_list = []
        query = table_query + '&(field_num>1)'
        for field in field_table.where( query ):
            delete_list.append( \
                (field['rat'],field['day'],field['session'],field['tc']) )
        delete_set = set(delete_list)
        
        data_list = []
        query = table_query + '&(field_num==1)'
        for field in field_table.where( query ):
            rat, day, maze = field['rat'], field['day'], field['session']
            if day > latest_qualified_day(rat, "DR"):
                continue
            data = (rat, day, maze, field['tc'])
            data_list.append( data )
        data_set = set( data_list )
        
        starts, ends, mis_angles = [], [], []
        data_list = list( data_set.difference(delete_set) )
        for data in data_list:
            field = get_unique_row( field_table,\
                '(rat==%d)&(day==%d)&(session==%d)&(tc=="%s")' % data )
            rat, day, maze, tc = data
            com, mis = field['COM'], field['mis_angle']
            cut = com - 180 if com >= 180 else com + 180

            session_data = NewSessionData.get( (rat,day,maze) )
            session_data._compute_laps(cut=cut*np.pi/180.0)
            laps = session_data.laps
            
            end, start = field['field_edges']
            valid_laps = np.ones( session_data.N_laps ).astype(bool)
            for angle in (qb+mis/2.0)%360:
                edges = np.array( [start,end,start+360,end+360,start+720,end+720] )
                dist = abs( angle + 360 - edges )
                if dist.min() > 15: # field does not intersect with qb
                    continue

                for i, lap in enumerate( zip(laps[:-1],laps[1:]) ):
                    if not valid_laps[i]:
                        continue
                    lap_start, lap_end = lap
                    for s, e in session_data.extended_scan_and_pause_list:
                        if ((s>=lap_start) and (s<=lap_end)) or ((e>=lap_start) and (e<=lap_end)):
                            alpha = session_data.F_('alpha')(session_data.T_(s))
                            dist = abs( alpha + 360 - \
                                np.array([angle,angle+360,angle+720]) )
                            if dist.min() < 15: # scan/pause in qb window
                                valid_laps[i] = False
                                continue

                            alpha = session_data.F_('alpha')(session_data.T_(e))
                            dist = abs( alpha + 360 - \
                                np.array([angle,angle+360,angle+720]) )
                            if dist.min() < 15: # scan/pause in qb window
                                valid_laps[i] = False
                                continue
            if np.sum(valid_laps) < 5:
                continue
            lap_list = (np.array(range(session_data.N_laps))+1)[valid_laps]

            filter_kwds = dict( \
                velocity_filter=True, exclude_off_track=True, tlim=lap_list, \
                exclude=session_data.extended_scan_and_pause_list )
            R, valid = session_data.get_cluster_ratemap( \
                tc, return_valid=True, **filter_kwds )
            
            if R.max() < MIN_RATE:
                continue

            for i, new_field in enumerate(mark_all_fields(R)):
                new_end, new_start = field_extent(new_field)
                mis_angles.append( mis )
                starts.append( new_start )
                ends.append( new_end )

        field_table.close()

        self.results['starts'] = starts
        self.results['ends'] = ends
        self.results['mis_angles'] = mis_angles
        self.results['area'] = area
        self.results['type'] = stype

        self.out('%d fields collected' % len(starts))
        self.out('Done')
    # end of METHOD collect_data

    def plot_hist(self,nbins=36):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        bins = np.linspace(0, 360, nbins+1)

        title_str = '%s_%s' % (self.results['area'],self.results['type'])

        starts, ends = self.results['starts'], self.results['ends']
        mis = np.array(self.results['mis_angles']) / 2.0
        data = (np.array(starts+ends) - np.r_[(mis,mis)] + 360) % 360
        print data.shape

        plt.hist( data, bins=bins, color='b' )
        for angle in [45, 135, 225, 315]:
            plt.axvline(x=angle,lw=2,color='r')
        plt.xlim([0,360])
        plt.tick_params(axis='both',labelsize=7)
        plt.title(title_str, size='small')

        plt.savefig( os.path.join(self.datadir,'%s.png'%title_str) )
    # end of METHOD plot_hist
# end of CLASS ScanLapFieldReport

# =================================================================================

class ThetaSpikeCntAna(AbstractAnalysis):
    datadir = 'ThetaSpkCnt'
    label = 'theta spike count'

    def collect_data(self, area=None, stype=None, min_spk=2, max_ISI=0.5/6):
        self.results = {}

        table_file = tb.openFile('/home/paramecium/output/fields.h5')
        field_table = table_file.root.place_fields

        query = '(expt_type=="DR")&(exclude==False)'
        if area is not None:
            query += '&(area=="%s")' % area
        if stype is not None:
            query += '&(type=="%s")' % stype 

        data = []
        old_rds = None
        for field in field_table.where( query ):
            rat, day = field['rat'], field['day']
            if day > latest_qualified_day(rat, "DR"):
                continue

            maze = field['session']
            rds = (rat, day, maze)
            if rds != old_rds:
                session = NewSessionData.get( rds )

                ts, x, y, hd = get_tracking( *rds )
                run_idx = session.filter_tracking_data( \
                    ts, boolean_index=True, **session.running_filter() )
                groups = np.array( [(g[0],g[1]-1) \
                    for g in contiguous_groups(run_idx)] )
                N_group = len(groups)
                if not N_group:
                    print 'Error!'
                ts_groups = ts[groups] + np.c_[np.zeros(N_group),np.ones(N_group)]
                
                old_rds = rds

            spikes = session.clusts[field['tc']].spikes

            # select spikes in field
            end, start = field['field_edges']

            # turn on if want to exclude qb fields
            #exclude_field = False
            #for angle in [45, 135, 225, 315]:
            #    if abs(start-angle)<15 or abs(end-angle)<15:
            #        exclude_field = True
            #        break
            #if exclude_field:
            #    continue

            alpha = session.F_('alpha')(session.T_(spikes))
            if start > end: # in angle (degree)
                in_field = (alpha>=end) * (alpha<=start)
                field_size = start - end
            else:
                in_field = (alpha>=end) + (alpha<=start)
                field_size = start + 360 - end
            field_spk = spikes[in_field]
            field_alpha = alpha[in_field]

            com, mis = field['COM'], field['mis_angle']
            cut = com - 180 if com >= 180 else com + 180
            session._compute_laps(cut=cut*np.pi/180.0)
            laps = session.laps
            for lap_s, lap_e in zip(laps[:-1],laps[1:]):
                # select spikes in lap
                in_lap = (field_spk>=lap_s) * (field_spk<lap_e)
                if not np.sum(in_lap):
                    continue
                lap_spk = field_spk[in_lap]
                lap_alpha = field_alpha[in_lap]

                # calculate start, end and center of the field for each lap 
                idx = select_from( lap_spk, ts_groups )
                if np.sum(idx) < min_spk:
                    continue
                run_spk = lap_spk[idx]
                run_alpha = lap_alpha[idx]
                if start > end:
                    m, M = run_alpha.min(), run_alpha.max()
                else:
                    pass_idx = run_alpha < start
                    if np.sum(pass_idx):
                        m = run_alpha[pass_idx].min() + 360
                        M = run_alpha[pass_idx].max() + 360
                        if np.sum(~pass_idx):
                            m = run_alpha[~pass_idx].min()
                    else:
                        m = run_alpha.min()
                        M = run_alpha.max()
                lap_ctr = ((m+M)/2.0) % 360
                lap_range = (M-m) % 360
                #if lap_range < min_size:
                #    continue
                if lap_range > 359:
                    pdb.set_trace()

                # calculate spike number in each theta cycle within running groups
                spk_cnt = []
                for s, e in ts_groups:
                    idx = select_from( lap_spk, [(s,e),] )
                    if np.sum(idx) < 2:
                        continue
                    spk = lap_spk[idx]
                    spk_alpha = lap_alpha[idx]
                    ISI = stamp_to_time( spk[1:]-spk[:-1], zero_stamp=0 )
                    if np.sum(ISI<0):
                        pdb.set_trace()

                    n_intv = len(ISI)
                    idx = np.array(range(n_intv))[ISI>max_ISI] # large ISI
                    if not len(idx): # only one theta cycle detected
                        continue
                    dist = circle_diff_vec_deg( \
                        spk_alpha, lap_ctr*np.ones(len(spk_alpha)) )
                    theta_cycles = zip( np.r_[[0,],idx+1], np.r_[idx,[n_intv,]] )
                    spk_cnt += [ (field_size,lap_range,(tail-head+1),\
                       (dist[head]+dist[tail])/2.0) for head,tail in theta_cycles ]
                if len(spk_cnt):
                    data.append( spk_cnt )

        field_table.close()

        self.results['data'] = data

        self.out('Done')
    # end of METHOD collect_data

    def plot_scatter(self, x_norm='spk', y_norm=True):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        data = self.results['data']

        x, y = np.array([]), np.array([])
        for points in data:
            fs, lap_rg, cnt, dist = zip(*points)
            if x_norm == 'spk':
                x = np.r_[ (x, np.array(dist)/np.array(lap_rg).astype(float)) ]
            elif x_norm == 'field':
                x = np.r_[ (x, np.array(dist)/np.array(fs).astype(float)) ]
            cnt = np.array(cnt).astype(float)
            if y_norm == True:
                cnt /= float(max(cnt))
            y = np.r_[ (y, cnt) ]

        plt.scatter( x, y, s=1)
        bins = np.linspace(-0.5,0.5,51)
        digit_x = np.digitize(x, bins) - 1
        digit_y = []
        for i in range(len(bins)-1):
            digit_y.append( np.mean(y[digit_x==i]) )
        plt.plot((bins[:-1]+bins[1:])/2.0,digit_y,'r',lw=2)
        if x_norm == 'spk':
            plt.xlim( [-0.5, 0.5] )
        else:
            plt.axis('tight')
        plt.show()

    # end of METHOD plot_hist
# end of CLASS ThetaSpkCntAna

# =================================================================================

class SingleLapISI(AbstractAnalysis):
    datadir = 'LapISI'
    label = 'single lap isi'

    def collect_data(self, area=None, stype=None, min_spk=5, min_size=1):
        self.results = {}

        table_file = tb.openFile('/home/paramecium/output/fields.h5')
        field_table = table_file.root.place_fields
        qb = np.array( [45, 135, 225, 315] )

        query = '(expt_type=="DR")&(exclude==False)'
        if area is not None:
            query += '&(area=="%s")' % area
        if stype is not None:
            query += '&(type=="%s")' % stype 

        data = []
        old_rds = None
        for field in field_table.where( query ):
            rat, day = field['rat'], field['day']
            if day > latest_qualified_day(rat, "DR"):
                continue

            maze = field['session']
            rds = (rat, day, maze)
            if rds != old_rds:
                session = NewSessionData.get( rds )

                ts, x, y, hd = get_tracking( *rds )
                run_idx = session.filter_tracking_data( \
                    ts, boolean_index=True, **session.running_filter() )
                groups = np.array( [(g[0],g[1]-1) \
                    for g in contiguous_groups(run_idx)] )
                N_group = len(groups)
                if not N_group:
                    print 'Error!'
                ts_groups = ts[groups] + np.c_[np.zeros(N_group),np.ones(N_group)]
                
                old_rds = rds

            spikes = session.clusts[field['tc']].spikes

            # select spikes in field
            end, start = field['field_edges']
            alpha = session.F_('alpha')(session.T_(spikes))
            if start > end: # in angle (degree)
                in_field = (alpha>=end) * (alpha<=start)
                field_size = start - end
            else:
                in_field = (alpha>=end) + (alpha<=start)
                field_size = start + 360 - end
            field_spk = spikes[in_field]
            field_alpha = alpha[in_field]

            com, mis = field['COM'], field['mis_angle']
            cut = com - 180 if com >= 180 else com + 180
            session._compute_laps(cut=cut*np.pi/180.0)
            laps = session.laps
            for lap_s, lap_e in zip(laps[:-1],laps[1:]):
                # select spikes in lap
                in_lap = (field_spk>=lap_s) * (field_spk<lap_e)
                if not np.sum(in_lap):
                    continue
                lap_spk = field_spk[in_lap]
                lap_alpha = field_alpha[in_lap]

                # calculate start, end and center of the field for each lap 
                idx = select_from( lap_spk, ts_groups )
                if np.sum(idx) < min_spk:
                    continue
                run_spk = lap_spk[idx]
                run_alpha = lap_alpha[idx]
                if start > end:
                    m, M = run_alpha.min(), run_alpha.max()
                else:
                    pass_idx = run_alpha < start
                    if np.sum(pass_idx):
                        m = run_alpha[pass_idx].min() + 360
                        M = run_alpha[pass_idx].max() + 360
                        if np.sum(~pass_idx):
                            m = run_alpha[~pass_idx].min()
                    else:
                        m = run_alpha.min()
                        M = run_alpha.max()
                lap_ctr = ((m+M)/2.0) % 360
                lap_range = (M-m) % 360
                if lap_range < min_size:
                    continue
                if lap_range > 359:
                    pdb.set_trace()

                # calculate ISI within running groups
                for s, e in ts_groups:
                    idx = select_from( lap_spk, [(s,e),] )
                    if np.sum(idx) < 2:
                        continue
                    spk = lap_spk[idx]
                    spk_alpha = lap_alpha[idx]
                    ISI = stamp_to_time( spk[1:]-spk[:-1], zero_stamp=0 )
                    if np.sum(ISI<0):
                        pdb.set_trace()
                    dist = circle_diff_vec_deg( \
                        spk_alpha, lap_ctr*np.ones(len(spk_alpha)) )
                    dist = (dist[:-1]+dist[1:]) / 2.0
                    if np.sum(dist>lap_range):
                        pdb.set_trace()
                    data.append( zip(field_size*np.ones(len(dist)), \
                        lap_range*np.ones(len(dist)), ISI, dist) )

        field_table.close()

        self.results['data'] = data

        self.out('Done')
    # end of METHOD collect_data

    def plot_scatter(self, log_y=True, norm='spk'):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        data = self.results['data']

        x, y = np.array([]), np.array([])
        for points in data:
            fs, lap_rg, ISI, dist = zip(*points)
            if norm == 'spk':
                x = np.r_[ (x, np.array(dist)/np.array(lap_rg).astype(float)) ]
            elif norm == 'field':
                x = np.r_[ (x, np.array(dist)/np.array(fs).astype(float)) ]
            y = np.r_[ (y, ISI) ]

        plt.scatter( x, y, s=1)
        bins = np.linspace(-0.5,0.5,51)
        digit_x = np.digitize(x, bins) - 1
        digit_y = []
        for i in range(len(bins)-1):
            digit_y.append( np.mean(y[digit_x==i]) )
        plt.plot((bins[:-1]+bins[1:])/2.0,digit_y,'r',lw=2)
        if log_y:
            ax = plt.gca()
            ax.set_yscale('log')
        if norm == 'spk':
            plt.xlim( [-0.5, 0.5] )
        else:
            plt.axis('tight')
        plt.show()

        #title_str = '%s_%s' % (self.results['area'],self.results['type'])

        #starts, ends = self.results['starts'], self.results['ends']
        #mis = np.array(self.results['mis_angles']) / 2.0
        #data = (np.array(starts+ends) - np.r_[(mis,mis)] + 360) % 360
        #print data.shape

        #plt.hist( data, bins=bins, color='b' )
        #for angle in [45, 135, 225, 315]:
        #    plt.axvline(x=angle,lw=2,color='r')
        #plt.xlim([0,360])
        #plt.tick_params(axis='both',labelsize=7)
        #plt.title(title_str, size='small')

        #plt.savefig( os.path.join(self.datadir,'%s.png'%title_str) )
    # end of METHOD plot_hist
# end of CLASS ScanLapFieldReport

# =================================================================================

class NewSessionData(SessionData):
    def filter_tracking_data(self, ts, x=None, y=None, tlim=None, 
        exclude_off_track=False, exclude=None, select=None, 
        velocity_filter=True, speed_filter=False, boolean_index=False):
        """Retrieve filtered positional tracking data
    
        Required arguments:
        ts -- timestamp array for the position data
        
        Keyword arguments:
        x,y -- position tracking data to be filtered (by default, interpolated
            values are used to compute radius if exclude_off_track is True
            and for return values)
        tlim -- array of lap numbers
        exclude -- optional list of timestamp intervals from which to exclude 
            spike and trajectory data
        select -- optional list of timestamp intervals from which to select 
            spike and trajectory data
        exclude_off_track -- whether to exclude data occurring off the track
        velocity_filter -- optional minimum forward-running velocity threshold
        speed_filter -- optional minimum path-speed threshold
        boolean_index -- get a boolean index array instead of filtered data
        
        Returns (x, y) tuple of filtered positional data arrays or a boolean
        index array if boolean_index is set to True.
        """
        Cfg = Config['dr_track']
        if self.attrs['track'] == 'circle-97':
            Cfg = Config['ld_track']
        track_radius = Cfg['diameter'] / 2
        track_width = (Cfg['outer_diameter']-Cfg['inner_diameter']) / 2

        if not self.is_circle_track and type(tlim) is int:
            tlim = None

        xy_provided = not (x is None or y is None)
        if not xy_provided:
            t = self.T_(ts)
            x, y = self.F_('x')(t), self.F_('y')(t)
            
        # Get trajectories and spike positions for specified time bounds
        if tlim is not None:
            new_ts, _xy = np.array([]), None
            for lap in tlim:
                if lap < 1 or lap > self.N_laps:
                    raise ValueError, 'bad lap number specified: %d'%lap
                start, end = self.laps[lap-1], self.laps[lap]
                t_lap, _xy_lap = time_slice_sample(ts, np.c_[x, y], start, end)
                new_ts = np.r_[ new_ts, t_lap ]
                _xy = _xy_lap.copy() if _xy is None else np.r_[ _xy, _xy_lap ]
            x, y = _xy.T
            ts = new_ts
        
        filters = [np.ones_like(x, dtype='?')]
        
        if velocity_filter:
            filters.append(self.velocity_filter(ts))
            
        if speed_filter:
            filters.append(self.path_speed_filter(ts))
        
        if exclude is not None:
            filters.append(exclude_from(ts, exclude))
        
        if select is not None:
            filters.append(select_from(ts, select))
        
        if exclude_off_track:
            if xy_provided:
                radius = np.sqrt(x**2 + y**2) - track_radius
            else:
                radius = self.F_('radius')(t)
            filters.append(
                np.logical_and(radius <= track_width/2,
                    radius >= -track_width/2))
        
        ix = reduce(op.mul, filters)
        if boolean_index:
            return ix
            
        return x[ix], y[ix]
# end of CLASS NewSessionData
