# encoding: utf-8
"""
population_map.py

Population field edges/derivatives distributions
"""

import tables as tb
import numpy as np
import matplotlib.pylab as plt

from scipy.interpolate import interp1d

from random import randint

from core.analysis import AbstractAnalysis
from tools.misc import memoize

from analysis_shan.texture_boundary import ModifiedSessionData
from analysis_shan.data_filters import latest_qualified_day

# should be consistant with PlaceFieldReport
BINS = 360 # need to fix pop_grad_shuffle_test if not 360
BLUR_WIDTH = 4.3
BD_ANGLES = { 'local':[45, 135, 225, 315], 'global':[0, 45, 135, 180, 270] }
MARGIN = 15 # in degree

# --------------------------------------------------------------------------------

def check_in_field_popmap(area=None,stype=None,ext=10,n=400,**traits):
    """
    ext - extend %ext percent of the field on both sides
          ex: ext=10, ext_start = start + 10% of field width
                      ext_end = end - 10% of field width
              the extended field would be 20% larger wider
    """
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    table_file = tb.openFile('/home/paramecium/output/fields.h5','r')
    field_table = table_file.root.place_fields

    table_query = '(expt_type=="DR")&(exclude==False)'
    fname_str = 'check_features'
    if area is not None:
        table_query += '&(area=="%s")' % area
        fname_str += '_%s' % area
    if stype is not None:
        table_query += '&(type=="%s")' % stype
        fname_str += '_%s' % stype 

    rate, d1, onset, offset = np.zeros(n), np.zeros(n), np.zeros(n), np.zeros(n)

    cnt = 0
    for field in field_table.where( table_query ):
        rat, day, maze = field['rat'], field['day'], field['session']
        if day > latest_qualified_day(rat,"DR"):
            continue

        end, start = field['field_edges']
        ext_start = (start + ext/100.0 * ((start-end+360)%360)) % 360
        ext_end = (end - ext/100.0 * ((start-end+360)%360) + 360) % 360
        if ext_start < ext_end: # wrap around if field cross 0/360
            ext_start += 360

        buff_start = ext_start + 1 
        buff_end = (ext_end-1+360) % 360
        if ext_start < buff_end:
            ext_start += 360
            buff_start += 360
        
        R = get_1d_ratemap( (rat,day,maze), field['tc'], 0) # ratemap not rotated
        R = np.r_[R, R] # for field crossing 0/360

        r = R[buff_end:buff_start]
        dr = r[:-1] - r[1:]
        ddr = dr[:-1] - dr[1:]

        r = r[1:-1]
        dr = dr[:-1]

        field_on = ((dr>0) * (ddr>0)).astype(int)
        field_off = ((dr<0) * (ddr>0)).astype(int)

        rate += norm_feature( r, n=n, kind='linear', **traits )
        d1 += norm_feature( dr, n=n, kind='linear', **traits )
        onset += norm_feature( field_on, n=n, kind='nearest', **traits )
        offset += norm_feature( field_off, n=n, kind='nearest', **traits )
        cnt += 1

    features = [rate, d1, onset, offset]
    titles = [ 'Firing Rate (r)', 'r\'', \
        'Increasing firing rate increment', 'Decreasing firing rate decrement' ]
    fig, axes = plt.subplots(4,1,figsize=(4,8))
    for ax, feature, title in zip( axes, features, titles):
        ax.plot(feature/float(cnt))
        if title == 'Firing Rate (r)':
            ax.axvline(x=n*ext/100.0,c='r',lw=1)
            ax.axvline(x=n-n*ext/100.0,c='r',lw=1)
        if title == 'r\'':
            ax.axhline(y=0,c='k',lw=1)
        ax.xaxis.set_ticklabels([])
        ax.tick_params(axis='y',labelsize=7)
        ax.set_xlim( [0, n] )
        ax.set_title(title)
    fig.suptitle('Population Map Features',size='large')

    plt.savefig('./%s.png' % fname_str)

    table_file.close()

# end of FUNCTION check_in_field_popmap

def norm_feature( v, n=360, kind='linear', norm=True):
    v_interp = interp1d( np.arange(v.size), v, kind=kind )
    norm_v = v_interp( np.linspace(0,v.size-1,n) )
    if norm:
        if norm_v.max() > 0:
            norm_v /= norm_v.max()

    return norm_v

# end of FUNCTION norm_feature

# --------------------------------------------------------------------------------

def population_heat_map(area='CA1',norm=True,rot='local',ft='onset',set_vmin=False,**ft_params):
    """
    Plot the population heat maps for STD and MIS sessions separately, and for
    each different MIS angle

    norm     - normalize individual cell map by maximum value if True
    rec      - ['abs', 'inc', 'dec']
               abs: collect both the rate increasing and decreasing parts (the 
                    absolute values of rate differences between neighboring bins)
               inc: collect only the rate increasing part
               dec: collect only the rate decreasing part
    scale    -  ['one', 'cell']
               one: rescale the population map to [0,1]
               cell: average by cell number
    rot      - ['local', 'global'] rotate the MIS data following the specified cues
    """
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    stypes = ['STD','MIS','MIS','MIS','MIS','MIS']
    mis_angles = [None, None, 45, 90, 135, 180]

    for i, panel in enumerate( zip(stypes,mis_angles) ):
        stype, mis_angle = panel
        title_str = stype if mis_angle is None else stype+' - '+str(mis_angle)

        #pop_diff = get_population_derivative( area,stype,mis_angle=mis_angle, \
        #    scale=scale,norm=norm,rec=rec,rot=rot )
        pop_ft = get_population_feature( area,stype,mis_angle=mis_angle, \
            norm=norm,rot=rot,ft=ft,**ft_params )

        r_min, r_max = 1.5, 2.0
        theta = np.linspace(0,360,BINS+1) * np.pi / 180.0 # in rad
        rad = np.array( [r_min, r_max] )
        X, Y = np.meshgrid(theta,rad)
        plt.subplot(3,2,i+1,projection='polar')
        if set_vmin:
            plt.pcolormesh( theta, rad, np.tile(pop_ft,(2,1)),cmap='gray_r',vmin=0 )
        else:
            plt.pcolormesh( theta, rad, np.tile(pop_ft,(2,1)),cmap='gray_r' )
        adj = 0 if mis_angle is None else mis_angle/2.0
        if rot=='local' or stype=='STD' or mis_angle!=None:
            for alpha in BD_ANGLES['local']:
                a = (alpha+adj) * np.pi / 180.0
                plt.plot( [a, a], [0.5,r_min], 'r' )
        if rot=='global' or stype=='STD' or mis_angle!=None:
            for alpha in BD_ANGLES['global']:
                a = (alpha-adj) * np.pi / 180.0
                plt.plot( [a, a], [1.0,r_min], 'g', lw=5, zorder=-1 )

        plt.colorbar()
        plt.yticks([])
        plt.ylim([0,r_max])
        plt.title(title_str)

    plt.show()
# end of FUNCTION population_rate_diff_map 

def pop_grad_shuffle_test(N=1000,nbin=40,window='both',rec='abs'):
    """
    Need to be modified
    """
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    select_in = np.zeros(BINS).astype(bool)
    for alpha in BD_ANGLES['local']:
        if window == 'both':
            select_in[alpha-MARGIN:alpha+MARGIN] = True
        elif window == 'pre':
            select_in[alpha:alpha+MARGIN] = True
        elif window == 'post':
            select_in[alpha-MARGIN:alpha] = True

    for i, stype in enumerate( ['STD','MIS'] ):
        cell_list = collect_cell_list("CA1",stype) 

        pop_diff = np.zeros(BINS)
        for rds, tc, mis in cell_list:
            D = get_1d_rate_diff_map(rds, tc, mis, rec=rec)
            pop_diff += D
        obs = calculate_avg_diff( pop_diff, select_in )

        shuffle = []
        for j in xrange(N):
            pop_diff = np.zeros(BINS)
            for rds, tc, mis in cell_list:
                D = get_1d_rate_diff_map(rds, tc, mis)
                pop_diff += np.roll(D, randint(0,BINS-1))
            shuffle.append( calculate_avg_diff( pop_diff, select_in ) )
        
        n = np.sum( shuffle>obs )
        p = float(n) / N
        print('p = %.4f (%d/%d)' % (p,n,N))

        x_min = min( obs, min(shuffle) )
        x_max = max( obs, max(shuffle) )
        bins = np.linspace(x_min,x_max,nbin+1)
        bin_size = bins[1] - bins[0]

        plt.subplot(1,2,i+1)
        plt.hist(shuffle, bins=bins)
        plt.axvline(x=obs,lw=2,color='r')
        plt.xlim( [x_min-bin_size, x_max+bin_size] )
        plt.tick_params(axis='both',labelsize=7)
        plt.title(stype,size='small')
        plt.suptitle('Average Difference Shuufle Test')

    plt.show()
# end of FUNCTION pop_grad_shuff

def pop_ft_PSTH(area='CA1',stype='STD',ft='onset',mis_angle=None,norm=True,rec='abs',half_win=45,clean=False,exclude_field=False,fd_hw=20):
    """
    Plot the averaged peri-cue spatial derivative curves for specified area and
    session type

    norm     - normalize individual cell map by maximum value if True
    rec      - ['abs', 'inc', 'dec']
                abs: collect both the rate increasing and decreasing parts (the 
                     absolute values of rate differences between neighboring bins)
                inc: collect only the rate increasing part
                dec: collect only the rate decreasing part
    half_win - half window width
    clean    - use only local cues not overlapped with global cues, or global cues
               not overlapped with local cues
    """
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    colors = ['r', 'g'] # local: red, global: green

    table_file = tb.openFile('/home/paramecium/output/fields.h5','r')
    field_table = table_file.root.place_fields

    for i, panel in enumerate( ['local', 'global'] ):
        title_str = area + '-' + stype
        if mis_angle is not None:
            title_str += ' (%d)' % mis_angle
        cell_list = collect_cell_list(area,stype,mis_angle) 

        adj = 0 if mis_angle is None else mis_angle/2.0
        if panel == 'local':
            angles = (np.array(BD_ANGLES[panel])+adj)%360 + 360
        elif panel == 'global':
            angles = (np.array(BD_ANGLES[panel])-adj+360)%360 + 360

        acc_ft = np.zeros(2*half_win)
        bin_cnt = np.zeros(2*half_win)
        count = 0
        for rds, tc, mis in cell_list:
            if ft == 'derivative':
                data = get_1d_rate_diff_map(rds,tc,mis,norm=norm,rec=rec,rot=panel)
            elif ft == 'onset':
                data = get_onset_map(rds,tc,mis,rot=panel)
            elif ft == 'offset':
                data = get_offset_map(rds,tc,mis,rot=panel)
            if exclude_field:
                query = '(rat==%d)&(day==%d)&(session==%d)&(tc=="%s")' \
                    % (rds+(tc,))
                cnt = np.ones( len(data) )
                for field in field_table.where(query):
                    if ft == 'onset':
                        edge = field['field_edges'][1] # start
                    elif ft == 'offset':
                        edge = field['field_edges'][0] # end
                    if panel == 'local':
                        edge = (edge-mis+360) % 360
                    elif panel == 'global':
                        edge = (edge+mis) % 360
                    e = (edge-fd_hw+360) % 360
                    s = (edge+fd_hw) % 360
                    if s < e:
                        data[:s] = 0
                        data[e:] = 0
                        cnt[:s] = 0
                        cnt[e:] = 0
                    else:
                        data[e:s] = 0
                        cnt[e:s] = 0
                ext_cnt = np.tile(cnt,3)
            ext_D = np.tile(data,3)

            if clean:
                if panel == 'local':
                    if stype == 'STD':
                        sel_angles = angles[ np.array([2,3]) ]
                    elif stype == 'MIS':
                        if mis_angle==45 or mis==22.5:
                            sel_angles = np.array( [angles[0],] )
                        elif mis_angle==90 or mis==45:
                            sel_angles = angles[ np.array([1,2]) ]
                        elif mis_angle==135 or mis==67.5:
                            sel_angles = np.array( [angles[3],] )
                        elif mis_angle==180 or mis==90:
                            sel_angles = angles[ np.array([0,1]) ]
                elif panel == 'global':
                    if stype == 'STD':
                        sel_angles = angles[ np.array([0,3,4]) ]
                    elif stype == 'MIS':
                        if mis_angle==45 or mis==22.5:
                            sel_angles = angles[ np.array([1,2]) ]
                        elif mis_angle==90 or mis==45:
                            sel_angles = angles[ np.array([0,3,4]) ]
                        elif mis_angle==135 or mis==67.5:
                            sel_angles = angles[ np.array([1,2]) ]
                        elif mis_angle==180 or mis==90:
                            sel_angles = angles[ np.array([0,3,4]) ]
            else:
                sel_angles = angles.copy()

            for alpha in sel_angles:
                extract = ext_D[alpha-half_win:alpha+half_win]
                acc_ft += extract
                if exclude_field:
                    bin_cnt += ext_cnt[alpha-half_win:alpha+half_win]
                count += 1
        print '%d %s cues detected for %d cells...' % (count,panel,len(cell_list))
        if exclude_field:
            acc_ft /= bin_cnt.astype(float)
        else:
            acc_ft /= float(count)

        x = np.linspace(-half_win,half_win-1,2*half_win) + 0.5
        plt.plot(x,acc_ft,colors[i],lw=2,zorder=-100)
        plt.axvline(x=0,lw=1,color='k')
        plt.xlim( [x[0],x[-1]] )
        plt.ylim( [0,0.20] )
        #plt.gca().set_ylim(bottom=0)

    plt.title(title_str)

    plt.show()

    table_file.close()
# end of FUNCTION pop_ft_PSTH

# --------------------------------------------------------------------------------

def pop_ft_PSTH_field(area='CA1',stype='STD',ft='onset',mis_angle=None,norm=True,rec='abs',half_win=45,clean=False,fd_hw=20):
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    colors = ['r', 'g'] # local: red, global: green

    table_file = tb.openFile('/home/paramecium/output/fields.h5','r')
    field_table = table_file.root.place_fields

    title_str = area + '-' + stype
    if mis_angle is not None:
        title_str += ' (%d)' % mis_angle
    cell_list = collect_cell_list(area,stype,mis_angle) 
    adj = 0 if mis_angle is None else mis_angle/2.0

    for i, panel in enumerate( ['local', 'global'] ):
        if panel == 'local':
            angles = (np.array(BD_ANGLES[panel])+adj)%360 + 360
        elif panel == 'global':
            angles = (np.array(BD_ANGLES[panel])-adj+360)%360 + 360

        for j, fd_rel in enumerate( ['keep', 'exclude', 'only', 'overlap'] ):
            acc_ft = np.zeros(2*half_win)
            bin_cnt = np.zeros(2*half_win)
            count = 0
            for rds, tc, mis in cell_list:
                if ft == 'derivative':
                    data = get_1d_rate_diff_map( \
                        rds,tc,mis,norm=norm,rec=rec,rot=panel)
                elif ft == 'onset':
                    data = get_onset_map(rds,tc,mis,rot=panel)
                elif ft == 'offset':
                    data = get_offset_map(rds,tc,mis,rot=panel)

                if fd_rel != 'keep':
                    query = '(rat==%d)&(day==%d)&(session==%d)&(tc=="%s")' \
                        % (rds+(tc,))
                    cnt = np.zeros(len(data))
                    for field in field_table.where(query):
                        end, start = field['field_edges']
                        if panel == 'local':
                            start = (start-mis+360) % 360
                            end = (end-mis+360) % 360
                        elif panel == 'global':
                            start = (start+mis) % 360
                            end = (end+mis) % 360

                        if fd_rel == 'overlap':
                            s = start if ft=='offset' else (start-fd_hw+360)%360
                            e = end if ft=='onset' else (end+fd_hw)%360
                        else:
                            edge = start if ft=='onset' else end
                            s = (edge+fd_hw) % 360
                            e = (edge-fd_hw+360) % 360

                        # overlap & only
                        if s < e:
                            cnt[:s] = 1 
                            cnt[e:] = 1 
                        else:
                            cnt[e:s] = 1 
                    if fd_rel == 'exclude':
                        cnt = 1 - cnt # reverse only
                    data[ ~cnt.astype(bool) ] = 0
                    ext_cnt = np.tile(cnt,3)
                ext_D = np.tile(data,3)

                if clean:
                    if panel == 'local':
                        if stype == 'STD':
                            sel_angles = angles[ np.array([2,3]) ]
                        elif stype == 'MIS':
                            if mis_angle==45 or mis==22.5:
                                sel_angles = np.array( [angles[0],] )
                            elif mis_angle==90 or mis==45:
                                sel_angles = angles[ np.array([1,2]) ]
                            elif mis_angle==135 or mis==67.5:
                                sel_angles = np.array( [angles[3],] )
                            elif mis_angle==180 or mis==90:
                                sel_angles = angles[ np.array([0,1]) ]
                    elif panel == 'global':
                        if stype == 'STD':
                            sel_angles = angles[ np.array([0,3,4]) ]
                        elif stype == 'MIS':
                            if mis_angle==45 or mis==22.5:
                                sel_angles = angles[ np.array([1,2]) ]
                            elif mis_angle==90 or mis==45:
                                sel_angles = angles[ np.array([0,3,4]) ]
                            elif mis_angle==135 or mis==67.5:
                                sel_angles = angles[ np.array([1,2]) ]
                            elif mis_angle==180 or mis==90:
                                sel_angles = angles[ np.array([0,3,4]) ]
                else:
                    sel_angles = angles.copy()

                for alpha in sel_angles:
                    extract = ext_D[alpha-half_win:alpha+half_win]
                    acc_ft += extract
                    if fd_rel != 'keep':
                        bin_cnt += ext_cnt[alpha-half_win:alpha+half_win]
                    count += 1
            print '%d %s cues detected for %d cells...' \
                % (count,panel,len(cell_list))
            if fd_rel != 'keep':
                acc_ft /= bin_cnt.astype(float)
            else:
                acc_ft /= float(count)

            x = np.linspace(-half_win,half_win-1,2*half_win) + 0.5
            plt.subplot(2,2,j+1)
            plt.plot(x,acc_ft,colors[i],lw=2,zorder=-100)
            plt.axvline(x=0,lw=1,color='k')
            plt.xlim( [x[0],x[-1]] )
            #plt.ylim( [0,0.20] )
            plt.gca().set_ylim(bottom=0)
            plt.title(fd_rel)

    plt.suptitle(title_str)
    fig = plt.gcf()
    fig.set_size_inches(8,12,forward=True)

    plt.show()

    table_file.close()
# end of FUNCTION pop_ft_field

# --------------------------------------------------------------------------------

def check_ft_rmapstack(area=None,stype=None,mis_angle=None,ft='onset',window=[-10,20],clean=False,rot='local',norm=True):
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)
    
    table_file = tb.openFile('/home/paramecium/output/fields.h5','r')
    field_table = table_file.root.place_fields

    adj = 0 if mis_angle is None else mis_angle/2.0
    if rot == 'local':
        angles = (np.array(BD_ANGLES[rot])+adj)%360 + 360
    elif rot == 'global':
        angles = (np.array(BD_ANGLES[rot])-adj+360)%360 + 360

    clusts = []
    cell_list = collect_cell_list(area,stype,mis_angle) 
    for rds, tc, mis in cell_list:
        if ft == 'onset':
            data = get_onset_map(rds,tc,mis,rot=rot)
            threshold = 0
        elif ft == 'offset':
            data = get_offset_map(rds,tc,mis,rot=rot)
            threshold = 0
        # Not defined 
        #elif ft == 'derivative':
        #    data = get_1d_rate_diff_map(rds,tc,mis,norm=norm,rec=rec,rot=rot)
        ext_D = np.tile(data,3)

        if clean:
            if rot == 'local':
                if stype == 'STD':
                    sel_angles = angles[ np.array([2,3]) ]
                elif stype == 'MIS':
                    if mis_angle==45 or mis==22.5:
                        sel_angles = np.array( [angles[0],] )
                    elif mis_angle==90 or mis==45:
                        sel_angles = angles[ np.array([1,2]) ]
                    elif mis_angle==135 or mis==67.5:
                        sel_angles = np.array( [angles[3],] )
                    elif mis_angle==180 or mis==90:
                        sel_angles = angles[ np.array([0,1]) ]
            elif rot == 'global':
                if stype == 'STD':
                    sel_angles = angles[ np.array([0,3,4]) ]
                elif stype == 'MIS':
                    if mis_angle==45 or mis==22.5:
                        sel_angles = angles[ np.array([1,2]) ]
                    elif mis_angle==90 or mis==45:
                        sel_angles = angles[ np.array([0,3,4]) ]
                    elif mis_angle==135 or mis==67.5:
                        sel_angles = angles[ np.array([1,2]) ]
                    elif mis_angle==180 or mis==90:
                        sel_angles = angles[ np.array([0,3,4]) ]
        else:
            sel_angles = angles.copy()

        for alpha in sel_angles:
            extract = ext_D[alpha+window[0]:alpha+window[1]]
            if np.sum(extract) > threshold: # contributing to specified window 
                clusts.append( (rds,tc,mis,alpha) )

    not_sort, to_be_sort = [], []
    for rds, tc, mis, alpha in clusts:
        rat, day, maze = rds
        query = '(rat==%d)&(day==%d)&(session==%d)&(tc=="%s")' % \
            (rat, day, maze, tc)
        min_dist, nearest_edge = np.inf, None
        for field in field_table.where(query):
            start, end = field['field_edges'][1], field['field_edges'][0]
            if rot == 'local':
                start = (start-mis+360) % 360
                end = (end-mis+360) % 360
                hw = 45
            elif rot == 'global':
                start = (start+mis) % 360
                end = (end+mis) % 360
                hw = 22.5

            if ft == 'onset':
                edges = [ start, start+360, start+720 ]
            elif ft == 'offset':
                edges = [ end, end+360, end+720 ]

            dist = edges[ np.argmin(np.abs(np.array(edges)-alpha)) ] - alpha
            if abs(dist) > 360: # detect error
                print dist

            if abs(dist) < abs(min_dist): # if current field closer to cue 
                if (dist>=-hw) and (dist<hw): # field edge within window
                    min_dist = dist
                    nearest_edge = edges[1]
                else: # check whether field overlap with window 
                    e = end
                    s = start+360 if start<end else start
                    win_e = (alpha+window[0]+360) % 360
                    win_s = (alpha+window[1]+360) % 360
                    if win_s < win_e:
                        win_s += 360
                    if (s>win_e) and (e<win_s):
                        min_dist = dist
                        nearest_edge = edges[1]
            
        if np.isinf(min_dist): # no field in window or overlap w window
            not_sort.append( ((rat,day,maze,tc,mis),alpha%360,-1,-1) )
        else:
            to_be_sort.append( \
                ((rat,day,maze,tc,mis),alpha%360,min_dist,nearest_edge%360) )

    distances = np.array( zip(*to_be_sort)[2] )
    index = np.argsort(distances)
    combined = []
    for i in index:
        combined.append( to_be_sort[i] )
    combined += not_sort

    ratemat = None
    for clust, angle, dist, edge in combined:
        rat, day, maze, tc, mis = clust
        session = ModifiedSessionData.get( (rat,day,maze) )
        R = session.get_cluster_ratemap( tc,bins=BINS,mis=mis,rot=rot,\
            blur_width=BLUR_WIDTH,**session.running_filter() )
        R = np.roll( R, -( int(angle+window[0]-1+360+180)%360 ) )
        if norm:
            R /= np.max(R)
        if ratemat is None:
            ratemat = R.copy()
        else:
            ratemat = np.vstack( [ratemat, R] )
    plt.imshow(ratemat,interpolation='none',aspect='auto',cmap='gray_r', \
        origin='lower',extent=[0,360,0,len(combined)-1])
    plt.axvline(x=180,lw=4,color='r',alpha=0.1) # 0
    plt.axvline(x=180+window[0],lw=2,color='m',alpha=0.1) # window[0] 
    plt.axvline(x=180+window[1],lw=2,color='m',alpha=0.1) # window[1]
    plt.axvline(x=180+hw,lw=1,color='b',alpha=0.1)
    plt.axvline(x=180-hw,lw=1,color='b',alpha=0.1)

    plt.tick_params(axis='both',labelsize=7)
    plt.axis('tight')
    plt.show()

    table_file.close()
# end of FUNCTION check_ft_rmapstack

# --------------------------------------------------------------------------------

def collect_cell_list(area, stype, mis_angle=None):
    table_file = tb.openFile('/home/paramecium/output/fields.h5','r')
    field_table = table_file.root.place_fields
    table_query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area

    cell_list = []
    for field in field_table.where(table_query+'&(type=="%s")'%stype):
        rat, day = field['rat'], field['day']
        if day > latest_qualified_day(rat,"DR"):
            continue
        if mis_angle is not None and field['mis_angle']!=mis_angle:
            continue
        mis = 0 if stype=='STD' or mis_angle!=None else field['mis_angle']/2.0
        clust = ((rat,day,field['session']),field['tc'],mis) # (rds,tc,half_mis)
        if clust not in cell_list:
            cell_list.append(clust)

    table_file.close()

    return cell_list
# end of FUNCTION collect_cell_list

def calculate_avg_diff( pd, select ):
    # normalize population rate difference
    pd_max, pd_min = np.max(pd), np.min(pd)
    norm_pd = (pd-pd_min) / float(pd_max-pd_min)

    avg_in = np.sum(norm_pd[select]) / float(np.sum(select))
    avg_out = np.sum(norm_pd[~select]) / float(np.sum(~select))

    return avg_in - avg_out
# end of FUNCTION calculate_avg_diff

def get_population_feature(area,stype,mis_angle=None,ft='onset',scale='one',norm=True,**traits):
    pop_ft = np.zeros(BINS)

    cell_list = collect_cell_list(area,stype,mis_angle) 
    for rds, tc, mis in cell_list:
        if ft == 'derivative':
            data = get_1d_rate_diff_map(rds, tc, mis, norm=norm, **traits)
        elif ft == 'onset':
            data = get_onset_map(rds, tc, mis, **traits)
        elif ft == 'offset':
            data = get_offset_map(rds, tc, mis, **traits)
        pop_ft += data

    if scale == 'one': # re-scale to [0,1]
        pd_max, pd_min = np.max(pop_ft), np.min(pop_ft)
        pop_ft = (pop_ft-pd_min) / float(pd_max-pd_min)
    if scale == 'cell': # average by cell number
        pop_ft /= len(cell_list)

    return pop_ft
# end of FUNCTION get_population_derivative

def get_population_derivative(area,stype,mis_angle=None,scale='one',**traits):
    pop_diff = np.zeros(BINS)

    cell_list = collect_cell_list(area,stype,mis_angle) 
    for rds, tc, mis in cell_list:
        D = get_1d_rate_diff_map(rds, tc, mis, **traits)
        pop_diff += D

    if scale == 'one': # re=scale to [0,1]
        pd_max, pd_min = np.max(pop_diff), np.min(pop_diff)
        pop_diff = (pop_diff-pd_min) / float(pd_max-pd_min)
    if scale == 'cell': # average by cell number
        pop_diff /= len(cell_list)

    return pop_diff
# end of FUNCTION get_population_derivative

def get_onset_map(rds, tc, mis, **traits):
    R = get_1d_ratemap(rds, tc, mis, **traits)
    D = np.roll(R,1) - R
    DD = D - np.roll(D,-1)

    field_on = ( (D>0) * (DD>0) ).astype(int)

    return field_on 
# end of FUNCTION get_onset_map

def get_offset_map(rds, tc, mis, **traits):
    R = get_1d_ratemap(rds, tc, mis, **traits)
    D = np.roll(R,1) - R
    DD = D - np.roll(D,-1)

    field_off = ( (D<0) * (DD>0) ).astype(int)

    return field_off 
# end of FUNCTION get_offset_map

@memoize
def get_1d_rate_diff_map(rds, tc, mis, norm=True, rec=None, **traits):
    R = get_1d_ratemap(rds, tc, mis, **traits)
    D = np.roll(R,-1) - R

    if rec == 'abs': # collect the absolute value of diff
        D = abs(D)
    elif rec == 'inc':
        D[D>0] = 0 # collect when the firing rate is increasing  
        D *= -1
    elif rec == 'dec':
        D[D<0] = 0 # collect when the firing rate is decreasing  

    if norm:
        D = D / float(np.max(D))
    return D
# end of FUNCTION get_1d_rate_diff_map

def get_1d_ratemap(rds, tc, mis, ratemap_filter='running', rot='local'):
    session = ModifiedSessionData.get(rds)
    if ratemap_filter == 'running':
        R = session.get_cluster_ratemap( tc,bins=BINS,mis=mis,rot=rot, \
            blur_width=BLUR_WIDTH,**session.running_filter() )
    elif ratemap_filter == 'off_track':
        R = session.get_cluster_ratemap( tc,bins=BINS,mis=mis,rot=rot,\
            blur_width=BLUR_WIDTH,velocity_filter=False, exclude_off_track=True )
    elif ratemap_filter == 'none':
        R = session.get_cluster_ratemap( tc,bins=BINS,mis=mis,rot=rot, \
            blur_width=BLUR_WIDTH,velocity_filter=False )
    return R
# end of FUNCTION get_ratemap
