# encoding: utf-8
"""
stack_ratemaps.py

Plot out the stack of individual cell ratemaps to see the distributions of fields
"""

import tables as tb
import numpy as np
import matplotlib.pylab as plt

from scanr.lib import *
from scanr.data import get_node, get_unique_row

from core.analysis import AbstractAnalysis

from analysis_shan.data_filters import latest_qualified_day
from analysis_shan.shared_tables import get_shared_table
from analysis_shan.texture_boundary import ModifiedSessionData

#from ldb.population_map import get_derivative_map

BINS = 360
BLUR_WIDTH = 4.3
BD_ANGLES = { 'local':[45,135,225,315], 'global':[0,45,135,180,270] }

def plot_ratemap_stack(area='CA1',subarea=None,ft='com',rotate='local',norm=False,label=True):
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    session_table = get_node('/metadata','sessions')
    field_file = tb.openFile('/home/paramecium/output/fields.h5','r')
    field_table = field_file.root.place_fields
    table_query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area
    if subarea is not None:
        table_query += '&(subdiv=="%s")' % subarea

    area_str = area if subarea is None else area + ' ' + subarea

    for plt_idx, stype in enumerate( ['All','STD','MIS'] ):
        if stype is not 'All':
            query = table_query + '&(type=="%s")' % stype
        else:
            query = table_query

        print 'Collecting field %s...' % ft
        data = []
        for field in field_table.where(query):
            rat, day, maze, tc = \
                field['rat'], field['day'], field['session'], field['tc']
            if day > latest_qualified_day(rat,'DR'):
                continue
            if stype is not 'All' and stype != field['type']:
                continue

            start, end = field['field_edges'][1], field['field_edges'][0]
            com = field['COM']

            mis = 0
            if field['type'] == 'MIS':
                session = get_unique_row( session_table, \
                    '(rat==%d)&(day==%d)&(session==%d)' %(rat,day,maze) )
                mis = session['mis_angle'] / 2.0
                if rotate == 'local':
                    start = (start-mis+360) % 360
                    end = (end-mis+360) % 360
                    com = (com-mis+360) % 360
                    #angle = (angle-mis+360) % 360
                elif rotate == 'global':
                    start = (start+mis) % 360
                    end = (end+mis) % 360
                    com = (com+mis) % 360
                    #angle = (angle+mis) % 360
            data.append( ((rat,day,maze,tc), start, end, com, mis) )
        N = len(data)
        print '%d fields collected' % N

        if ft == 'start':
            angles = np.array( zip(*data)[1])
        elif ft == 'end':
            angles = np.array( zip(*data)[2])
        elif ft == 'com':
            angles = np.array( zip(*data)[3])
        index = np.argsort(angles)

        ratemat = None 
        for i in index:
            rat, day, maze, tc = data[i][0]
            session = ModifiedSessionData.get( (rat,day,maze) )
            R = session.get_cluster_ratemap(tc,bins=BINS,mis=data[i][-1], \
                rot=rotate,blur_width=BLUR_WIDTH,**session.running_filter() )
            if norm:
                R /= np.max(R)
            if ratemat is None:
                ratemat = R.copy()
            else:
                ratemat = np.vstack( [ratemat, R] )

        plt.subplot(1,3,plt_idx+1)
        plt.imshow(ratemat,interpolation='none',aspect='auto',cmap='gray_r', \
            origin='lower',extent=[0,360,0,N-1])
        if label:
            for cue in BD_ANGLES[rotate]:
                plt.axvline(x=cue,lw=2,color='r',alpha=0.1)
            plt.plot( np.array(zip(*data)[1])[index], range(N), \
                'b.',markersize=3,alpha=0.1)
            plt.plot( np.array(zip(*data)[2])[index], range(N), \
                'b.',markersize=3,alpha=0.1)
            plt.plot( np.array(zip(*data)[3])[index], range(N), \
                'b.',markersize=1,alpha=0.1)
        plt.title(stype)
        plt.suptitle(area_str)
        plt.tick_params(axis='both',labelsize=7)
        plt.axis('tight')
    plt.show()

    field_file.close()
    print 'Done!'

# end of function plot_ratemap_stack

def panels():
    for area in ['CA3', 'CA1']:
        for ft in ['com','start','end']:
            for rotate in ['local', 'global']:
                for norm in [True, False]:
                    for label in [True, False]:
                        file_name = '%s_%s_%s' % (area,ft,rotate)
                        if norm:
                            file_name += '_norm'
                        if label:
                            file_name += '_label'
                        file_name += '.png'
                        plot_ratemap_stack(area=area,ft=ft,rotate=rotate, \
                            norm=norm,label=label)
                        plt.savefig(file_name)
# end of function panels

def check_field_com():
    """
    check if the center of mass stored in place field table are correct
    """
    field_file = tb.openFile('/home/paramecium/output/fields.h5','r')
    field_table = field_file.root.place_fields

    count = 0
    for field in field_table[:]:
        start, end = field['field_edges'][1], field['field_edges'][0]
        com = field['COM']

        if start < end:
            if (com>start) and (com<end): 
                print 'Warning: start= %f, end= %f, com= %f' % (start,end,com)
        else:
            if (com>start) or (com<end):
                print 'Warning: start= %f, end= %f, com= %f' % (start,end,com)
        count += 1
    print '%d fields checked' % count

    field_file.close()
# end of function check_field_com

# =================================================================================

class RmapStackAna(AbstractAnalysis):

    datadir = './RmapStack'
    label = 'ratemap stack'

    def collect_data(self,area='CA1',subarea=None,stype='All',rotate='local',window=None,cond='onset'):
        self.results = {}

        area_str = area if subarea is None else area + ' ' + subarea

        if window is not None:
            N, data = self.collect_window( area=area,subarea=subarea,stype=stype, \
                rotate=rotate,window=window,cond=cond )
        else:
            N, data = self.collect_all( area=area,subarea=subarea,stype=stype, \
                rotate=rotate)

        self.results['area_str'] = area_str
        self.results['stype'] = stype
        self.results['rotate'] = rotate
        self.results['N'] = N
        self.results['data'] = data

        self.out('Done')
    # end of METHOD collect_data

    def process_data(self,ft='com',hist=None,norm=True,label=True,nbin=36,**traits):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        area_str = self.results['area_str']
        stype = self.results['stype']
        rotate = self.results['rotate']
        N = self.results['N']
        data = self.results['data']

        mapping = { 'start':1, 'end':2, 'com':3 }

        angles = np.array( zip(*data)[mapping[ft]] )
        index = np.argsort(angles)

        ratemat = None 
        for i in index:
            rat, day, maze, tc = data[i][0]
            R = self.build_map(rat,day,maze,tc,data[i][-1],rotate,**traits)
            if norm:
                R /= np.max(R)
            if ratemat is None:
                ratemat = R.copy()
            else:
                ratemat = np.vstack( [ratemat, R] )

        if hist is not None:
            ax2 = plt.subplot2grid( (4,1), (3,0) )
            distr = np.array( zip(*data)[mapping[hist]] )
            bins = np.linspace(0, 360, nbin+1)
            ax2.hist( distr, bins=bins )
            for cue in BD_ANGLES[rotate]:
                plt.axvline(x=cue,lw=2,color='r',alpha=0.1)
            plt.xlim(0, 360)
            plt.tick_params(axis='both',labelsize=7)

            ax1 = plt.subplot2grid( (4,1), (0,0), rowspan=3 )
            plt.sca(ax1)
        plt.imshow(ratemat,interpolation='none',aspect='auto',cmap='gray_r', \
            origin='lower',extent=[0,360,0,N-1])
        if label:
            for cue in BD_ANGLES[rotate]:
                plt.axvline(x=cue,lw=2,color='r',alpha=0.1)
            plt.plot( np.array(zip(*data)[1])[index], range(N), \
                'b.',markersize=3,alpha=0.1)
            plt.plot( np.array(zip(*data)[2])[index], range(N), \
                'b.',markersize=3,alpha=0.1)
            plt.plot( np.array(zip(*data)[3])[index], range(N), \
                'b.',markersize=1,alpha=0.1)
        plt.tick_params(axis='both',labelsize=7)
        plt.axis('tight')

        plt.suptitle( "%s (%s)" % (area_str,stype) )

        plt.savefig(self.datadir+'/out.png')

        self.out('Done!')
    # end of METHOD process_data

    def collect_all(self,area='CA1',subarea=None,stype='All',rotate='local'):
        session_table = get_node('/metadata','sessions')
        field_file = tb.openFile('/home/paramecium/output/fields.h5','r')
        field_table = field_file.root.place_fields

        query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area
        if stype != 'All':
            query += '&(type=="%s")' % stype
        if subarea is not None:
            query += '&(subdiv=="%s")' % subarea

        area_str = area if subarea is None else area + ' ' + subarea

        self.out('Collecting field data...')
        data = []
        for field in field_table.where(query):
            rat, day, maze, tc = \
                field['rat'], field['day'], field['session'], field['tc']
            if day > latest_qualified_day(rat,'DR'):
                continue
            if stype is not 'All' and stype != field['type']:
                continue

            start, end = field['field_edges'][1], field['field_edges'][0]
            com = field['COM']

            mis = 0
            if field['type'] == 'MIS':
                session = get_unique_row( session_table, \
                    '(rat==%d)&(day==%d)&(session==%d)' %(rat,day,maze) )
                mis = session['mis_angle'] / 2.0
                if rotate == 'local':
                    start = (start-mis+360) % 360
                    end = (end-mis+360) % 360
                    com = (com-mis+360) % 360
                elif rotate == 'global':
                    start = (start+mis) % 360
                    end = (end+mis) % 360
                    com = (com+mis) % 360
            data.append( ((rat,day,maze,tc), start, end, com, mis) )
        N = len(data)
        self.out('%d fields collected' % N)

        field_file.close()

        return N, data
    # end of METHOD collect_all

    def collect_window(self,area='CA1',subarea=None,stype='All',rotate='local',window=[-10,20],cond='onset'):
        session_table = get_node('/metadata','sessions')
        field_file = tb.openFile('/home/paramecium/output/fields.h5','r')
        field_table = field_file.root.place_fields

        query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area
        if stype != 'All':
            query += '&(type=="%s")' % stype
        if subarea is not None:
            query += '&(subdiv=="%s")' % subarea

        area_str = area if subarea is None else area + ' ' + subarea
        angles = np.array([45, 135, 225, 315]) + 360

        self.out('Collecting field data...')
        data, grad_sum = [], []
        for field in field_table.where(query):
            rat, day, maze, tc = \
                field['rat'], field['day'], field['session'], field['tc']
            if day > latest_qualified_day(rat,'DR'):
                continue
            if stype is not 'All' and stype != field['type']:
                continue

            end, start = field['field_edges']
            com = field['COM']
            if field['type'] == 'MIS':
                if rotate == 'local':
                    start = np.floor(start-mis+360) % 360
                    end = np.floor(end-mis+360) % 360
                    com = np.floor(com-mis+360) % 360
                elif rotate == 'global':
                    start = np.floor(start+mis) % 360
                    end = np.floor(end+mis) % 360
                    com = np.floor(com+mis) % 360
            else:
                start = np.floor(start)
                end = np.floor(end)
            
            mis = 0 if stype=='STD' else field['mis_angle']/2.0
            if cond == 'field_start':
                D = np.zeros(BINS)
                D[start] = 1
            elif cond == 'field_end':
                D = np.zeros(BINS)
                D[end] = 1
            elif (cond=='onset') or (cond=='offset'):
                D = get_derivative_map( (rat,day,maze),tc,ft=cond,mis=mis, \
                    rot=rotate,map_type=map_type)

                if start < end:
                    D[start:end] = 0
                else:
                    D[:end] = 0
                    D[start:] = 0
            threshold = 0

            ext_D = np.tile(D, 3)

            for alpha in angles:
                extract = ext_D[alpha+window[0]:alpha+window[1]]
                gs = np.sum(extract)
                grad_sum.append(gs)
                # contributing to specified window
                if gs>threshold:
                    if (cond=='onset') or (cond=='field_start'):
                        edges = [start, start+360, start+720]
                    elif (cond=='offset') or (cond=='field_end'):
                        edges = [end, end+360, end+720]
                    dist = edges[ np.argmin(np.abs(np.array(edges)-alpha)) ] - alpha
                    # assume windows do not overlap, and each field included once
                    if (dist>=window[0]) and (dist<window[1]):
                        data.append( ((rat,day,maze,tc), start, end, com, mis) )
        N = len(data)
        self.out('%d fields selected' % N)

        #plt.hist(grad_sum)
        #plt.show()

        field_file.close()

        return N, data
    # end of METHOD collect_window

    # to be overwritten
    def build_map(self,rat,day,maze,tc,mis,rotate):
        session = ModifiedSessionData.get( (rat,day,maze) )
        R = session.get_cluster_ratemap(tc,bins=BINS,mis=mis, \
            rot=rotate,blur_width=BLUR_WIDTH,**session.running_filter() )
        return R
    # end of METHOD stack_map

# end of function plot_ratemap_stack


# =================================================================================

def field_size_vs_qb_distance_panels(**traits):
    for area in ['CA1', 'CA3']:
        for stype in ['All', 'STD', 'MIS']:
            for fy in ['CoM_dist', 'mid_dist', 'edge_dist']:
                file_name = '%s_%s_%s' % (area,stype,fy[:3])
                file_name += '.png'
                field_size_vs_qb_distance(area=area,stype=stype,fy=fy,**traits)
                plt.savefig(file_name)

def field_size_vs_qb_distance(area='CA1',stype='All',subarea=None,**traits):
    plt.close('all')
    plt.rcdefaults()
    plt.hold(True)

    title_str = area if subarea is None else area + ' ' + subarea
    title_str += ' - %s' % stype

    query = '(area=="%s")&(expt_type=="DR")&(exclude==False)' % area
    if stype != 'All':
        query += '&(type=="%s")' % stype
    if subarea is not None:
        query += '&(subdiv=="%s")' % subarea

    ax = plt.subplot(2,2,1)
    panel(query,stype,ax=ax,abs_dist=True,**traits)
    plt.ylabel('Distance')
    ax = plt.subplot(2,2,2)
    panel(query,stype,ax=ax,abs_dist=True,small_fd=True,**traits)
    ax = plt.subplot(2,2,3)
    panel(query,stype,ax=ax,abs_dist=False,**traits)
    plt.xlabel('Field Size')
    plt.ylabel('Distance')
    ax = plt.subplot(2,2,4)
    panel(query,stype,ax=ax,abs_dist=False,small_fd=True,**traits)
    plt.xlabel('Field Size')

    plt.suptitle(title_str)
# end of FUNCTION

def panel(query,stype,ax=None,fx='field_size',fy='CoM_dist',abs_dist=True,small_fd=False,ptype='heatmap',cmap='gray',label=False,norm_col=False):

    x, y = collect_field_features(fx,fy,query,stype,abs_dist=abs_dist)

    ymin = 0 if abs_dist else -45
    if small_fd:
        xmin, xmax = 10, 110 
        xbins = 10 
        ybins = 9 if abs_dist else 15
    else:
        xmin, xmax = 0, 360
        xbins= 18
        ybins = 9 if abs_dist else 15
    x_edges = np.linspace(xmin,xmax,xbins+1)
    y_edges = np.linspace(ymin,45,ybins+1)

    if ax is None:
        ax = plt.gca()
    if ptype == 'scatter':
        ax.plot(x,y,'.')
    elif ptype == 'heatmap':
        h2d = ax.hist2d( x,y,bins=[x_edges,y_edges], \
            range=[[xmin,xmax],[ymin,45]],cmap=cmap )
        if norm_col:
            col_max = np.max(h2d[0],1) # the returned counts are transposed
            # digitize doex not include right bin edge, need to manually fix the
            # the right most bin 
            bin_x = np.digitize(x,x_edges) - 1
            bin_x[ x==x_edges[-1] ] = xbins-1
            w = np.array( [ 1.0/col_max[bx] if (0<=bx) and (bx<len(col_max)) \
                else 1.0 for bx in bin_x] )

            plt.hold(False)
            ax.hist2d( x,y,weights=w,bins=[x_edges,y_edges], \
                range=[[xmin,xmax],[ymin,45]],cmap=cmap )
            plt.hold(True)
        else:
            cbar = plt.colorbar(h2d[3])
            cbar.ax.tick_params(labelsize=8)

    if label:
        line_c = 'white' if cmap=='jet' else 'r'
        plt.plot([0,90],[0,45],c=line_c)
        if not abs_dist:
            plt.plot([0,90],[0,-45],c=line_c)

    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, 45)

    title = 'Absolute Distance to QB' if abs_dist else 'Distance to QB'
    if small_fd:
        title += ' (enlarged)'
    plt.title(title,fontdict={'fontsize':9})
    plt.tick_params(axis='both',labelsize=8)
# end of FUNCTION

def collect_field_features(fx, fy, query, stype, abs_dist=False):
    field_file = tb.openFile('/home/paramecium/output/fields.h5','r')
    field_table = field_file.root.place_fields

    features = { 'field_size': [],
                 'CoM_dist': [],
                 'mid_dist': [],
                 'edge_dist': [] }

    def wrap_dist(angle, adj):
        cues = np.array( BD_ANGLES['local'] )
        rot_cues = (cues+adj) % 360 # rotate local cues if MIS session
        ext_cues = np.r_[rot_cues, rot_cues+360, rot_cues+720]
        return angle + 360 - ext_cues[np.argmin(np.abs(angle+360-ext_cues))]

    for field in field_table.where(query):
        rat, day, maze, tc = \
            field['rat'], field['day'], field['session'], field['tc']
        if day > latest_qualified_day(rat,'DR'):
            continue
        if stype is not 'All' and stype != field['type']:
            continue
        
        adj = field['mis_angle']/2.0 if field['type']=='MIS' else 0
        start, end = field['field_edges'][1], field['field_edges'][0]

        if 'field_size' in [fx, fy]:
            fd_size = (start+360-end) % 360
            features['field_size'].append( fd_size )

        if 'CoM_dist' in [fx, fy]:
            com_dist = wrap_dist( field['COM'], adj )
            if abs_dist:
                com_dist = abs(com_dist)
            features['CoM_dist'].append( com_dist )

        if 'mid_dist' in [fx, fy]:
            mid = (start+end)/2.0 if start>end else ((start+360+end)/2.0)%360
            mid_dist = wrap_dist( mid, adj )
            if abs_dist:
                mid_dist = abs(mid_dist)
            features['mid_dist'].append( mid_dist )

        if 'edge_dist' in [fx, fy]:
            s_dist, e_dist = wrap_dist(start,adj), wrap_dist(end,adj)
            edge_dist = s_dist if abs(s_dist)<abs(e_dist) else e_dist
            if abs_dist:
                edge_dist = abs(edge_dist)
            features['edge_dist'].append( edge_dist )

    field_file.close()

    return features[fx], features[fy]
# end of FUNCTION collect_field_features


# =================================================================================

def check_fd_size_hist():
    field_file = tb.openFile('/home/paramecium/output/fields.h5','r')
    field_table = field_file.root.place_fields

    i = 1
    for area in ['CA1', 'CA3']:
        for stype in ['STD', 'MIS']:
            query = '(area=="%s")&(expt_type=="DR")&(type=="%s")&(exclude==False)'\
                % (area, stype)

            data = []
            for field in field_table.where(query):
                rat, day, maze, tc = \
                    field['rat'], field['day'], field['session'], field['tc']
                if day > latest_qualified_day(rat,'DR'):
                    continue
                if stype != field['type']:
                    continue
                
                start, end = field['field_edges'][1], field['field_edges'][0]
                data.append( (start+360-end)%360 )
            
            for small_fd in [False, True]:
                if small_fd:
                    xmin, xmax, nbins = 10, 110, 10
                else:
                    xmin, xmax, nbins = 0, 360, 18
                bins = np.linspace(xmin,xmax,nbins+1)

                plt.subplot(4,2,i)
                plt.hist(data, bins=bins)
                plt.xlim(xmin, xmax)
                if i < 3:
                    if small_fd:
                        plt.title('Field Size (enlarged)',fontdict={'fontsize':8})
                    else:
                        plt.title('Field Size',fontdict={'fontsize':8})
                if not small_fd:
                    plt.ylabel('%s - %s' % (area,stype))
                plt.tick_params(axis='both',labelsize=8)
                i += 1
    plt.suptitle('Field Size Distribution')
    plt.show()

    field_file.close()
# end of FUNCTION check_fd_size_hist
