# encoding: utf-8
"""
behavior_effect.py
"""


import tables as tb
import numpy as np
import matplotlib.pyplot as plt

import itertools
import warnings

from matplotlib.collections import LineCollection
from shapely.geometry import Point
from scipy.interpolate import interp1d, griddata

from core.report import BaseReport
from tools.misc import contiguous_groups
from tools.plot import quicktitle
from tools.images import masked_array_to_rgba

from scanr.cluster import string_to_quality

from tdb.config import Config
from tdb.data import get_tdb_file, get_group
from tdb.meta import stamp_to_time, walk_mazes
from tdb.ratemap import compute_velocity, compute_heading, smooth_ratemap
from tdb.tdb_misc import get_panel_session_list
from tdb.td_maps import get_boundary_line, plot_boundaries, bin_edges

FieldCfg = Config['placefield']
OCC_THRD = FieldCfg['occ_thrd'] # occupancy threshold, 0.06 sec per bin
MIN_VELOCITY = 5.76 # cm/s

#==================================================================================

class HDPathReport(BaseReport):
    datadir = './HDPath'
    label = 'direction path'

    ncols = 4
    nrows = 4

    def collect_data(self, min_quality='Fair', panel='leather_STD'):
        tdb_file = get_tdb_file()
        field_table = tdb_file.root.spatial.fields

        self.results = {}

        data_list = []
        rds_list = get_panel_session_list(panel)
        for rds in rds_list:
            query = \
                '(area=="CA1")&(field_no==1)&(rat==%d)&(day==%d)&(session==%d)'\
                    % rds
            for field in field_table.where(query):
                quality = string_to_quality( field['quality'] )
                if quality < string_to_quality(min_quality):
                    continue
                tc = 't%dc%d' % (field['tt'],field['cl'])
                data_list.append( (rds,tc,panel) )
        tdb_file.close()

        data_list = [ data+(orientation,) for data, orientation in \
            itertools.product(data_list,['NE','NW','SW','SE']) ]

        self.results['data_list'] = data_list
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, color_map='jet', hw=10, heading_window=45, min_velocity=MIN_VELOCITY):
        data_list = self.results['data_list']
        cmap = plt.get_cmap( color_map )

        def plot_data(ax, groups, pos_x, pos_y, norm_rate):
            for epoch in ['run', 'n_run']:
                for group in groups[epoch]:
                    s, e = group
                    points = np.array([pos_x[s:e],pos_y[s:e]]).T.reshape(-1,1,2)
                    segments = np.concatenate( [points[:-1],points[1:]], axis=1 )

                    if epoch == 'run':
                        lc = LineCollection(segments,linewidths=1,cmap=cmap)
                        lc.set_array( norm_rate[s:e] )
                        lc.set_alpha( 0.7 )
                    else:
                        lc = LineCollection(segments,linewidths=1,colors='k')

                    ax.add_collection(lc)
            return

        old_rds, old_clust = None, None
        for data, ax in self.get_plot( data_list ):
            rds, tc, panel, orientation = data
            cur_clust = (rds, tc)

            if rds != old_rds:
                rat, day, maze = rds
                session = get_group(rds=rds)
                fs = session._v_attrs['tracking_sample_rate']
                pos_x, pos_y, ts = [ \
                    session._v_children[key] for key in ['x','y','t_pos'] ]

                t_sec = stamp_to_time(ts, zero_stamp=ts[0])
                dt = np.diff( t_sec )

                v = compute_velocity(t_sec, pos_x, pos_y)
                running = v >= min_velocity

                bd_line, theta = get_boundary_line( pattern=panel )
                xy = np.array([pos_x,pos_y]).T.reshape(-1,2)
                dist = np.array( [Point(x,y).distance(bd_line) for x,y in xy] )
                inside = dist <= hw

                heading = compute_heading(pos_x, pos_y)
                direction = { 'NE': theta, 'NW': theta+90, \
                              'SW': theta+180, 'SE': theta+270 }
                old_rds = rds

            if cur_clust != old_clust:
                spikes = session._v_children[tc]
                idx = np.digitize( spikes, ts )
                rate = np.array( \
                    [ np.sum(idx==i+1)/float(dt[i]) for i in range(len(ts)-1) ] )
                max_rate = float( rate.max() )
                norm_rate = rate / max_rate 
                old_clust = cur_clust

            groups = {}
            angle = direction[orientation]
            sigma = np.min( \
                np.c_[abs(heading-angle),abs(heading+360-angle)], axis=1 )
            along = sigma < heading_window
            groups['run'] = contiguous_groups( running*inside*along )
            groups['n_run'] = contiguous_groups( ~running*inside*along )

            plot_boundaries(*rds,**{'c':'gray','alpha':0.5,'lw':1})
            plot_data( ax, groups, pos_x, pos_y, norm_rate )
            ax.get_xaxis().set_ticks([])
            ax.get_yaxis().set_ticks([])
            if self.firstcol:
                quicktitle( ax, '(%d, %d, %d) %s - %s'%(rat,day,maze,tc,panel) )
            if panel == 'leather_STD':
                plt.axis([-15, 40, -50, 50])
            elif panel == 'leather_SHF':
                plt.axis([-35, 20, -50, 50])
            elif panel == 'tape_STD':
                plt.axis([-50, 50, -40, 15])
            elif panel == 'tape_SHF':
                plt.axis([-50, 50, -20, 35])
                    
    # end of METHOD process_data

    def plot_ratemaps(self, color_map='jet', hw=10, heading_window=45, min_velocity=MIN_VELOCITY,**traits):
        data_list = self.results['data_list']
        cmap = plt.get_cmap( color_map )

        old_rds, old_clust = None, None
        for data, ax in self.get_plot( data_list ):
            rds, tc, panel, orientation = data
            cur_clust = (rds, tc)

            if rds != old_rds:
                rat, day, maze = rds
                session = get_group(rds=rds)
                fs = session._v_attrs['tracking_sample_rate']

                x, y, ts = [ session._v_children[key] for key in ['x','y','t_pos'] ]
                t_sec = stamp_to_time(ts, zero_stamp=ts[0])
                v = compute_velocity(t_sec, x, y)
                heading = compute_heading(x, y)

                old_rds = rds

            if cur_clust != old_clust:
                spikes = session._v_children[tc]
                spike_sec = stamp_to_time(spikes, zero_stamp=ts[0])
                
                spk_v = interp1d(t_sec,v,fill_value=0.0,kind='nearest', \
                    bounds_error=False)(spike_sec)
                spk_hd = interp1d(t_sec,heading,fill_value=0.0,kind='nearest', \
                    bounds_error=False)(spike_sec)
                old_clust = cur_clust

            bd_line, theta = get_boundary_line( pattern=panel )
            direction = { 'NE': theta, 'NW': theta+90, \
                          'SW': theta+180, 'SE': theta+270 }
            win_sm = (direction[orientation]-heading_window+360) % 360
            win_lg = (direction[orientation]+heading_window) % 360

            if win_lg > win_sm:
                valid = (v>=min_velocity) * (heading<win_lg) * (heading>win_sm)
                spk_valid = (spk_v>=min_velocity)*(spk_hd<win_lg)*(spk_hd>win_sm)
                spk_sec = spike_sec[ spk_valid ]
            else:
                valid = (v>=min_velocity) * ((heading<win_lg)+(heading>win_sm))
                spk_valid = (spk_v>=min_velocity)*((spk_hd<win_lg)+(spk_hd>win_sm))
                spk_sec = spike_sec[ spk_valid ]
            H_occ, ybins, xbins = np.histogram2d( \
                y[valid], x[valid], bins=bin_edges(rds,**traits) )
            H_occ = H_occ /float(fs) # in second
            spk_x = interp1d(t_sec,x,fill_value=0.0,kind='nearest', \
                bounds_error=False)(spk_sec)
            spk_y = interp1d(t_sec,y,fill_value=0.0,kind='nearest', \
                bounds_error=False)(spk_sec)
            H_spk, ybins, xbins = np.histogram2d( \
                spk_y, spk_x, bins=bin_edges(rds,**traits) )

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                H_rate = 1.0 * H_spk / H_occ
            if np.sum( np.isinf(H_rate) ) > 0:
                H_rate[ np.isinf(H_rate) ] = 0.0
            H_rate[ np.isnan(H_rate) ] = 0.0

            mask = H_occ < OCC_THRD
            bin_y, bin_x = H_occ.shape
            gy, gx = np.mgrid[ 0:bin_y, 0:bin_x ]
            H_rate[mask] = griddata( zip(gy[~mask],gx[~mask]), \
                H_rate[~mask].ravel(), (gy[mask],gx[mask]), method='nearest' )
            H_rate = smooth_ratemap( H_rate, mask )
            H_rate[mask] = 0

            [y_edges, x_edges] = bin_edges(rds,**traits)
            img = masked_array_to_rgba( \
                H_rate,mask=mask,mask_color='b',cmap='gray_r' )
            ax.imshow( img,origin='lower',interpolation='none',aspect='equal',\
                extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )
            plot_boundaries(*rds,**{'c':'r','alpha':0.5,'lw':1})
            ax.get_xaxis().set_ticks([])
            ax.get_yaxis().set_ticks([])
            if self.firstcol:
                quicktitle( ax, '(%d, %d, %d) %s - %s'%(rat,day,maze,tc,panel) )
    # end of METHOD plot_ratemap

# end of CLASS HDPathReport
