#encoding: utf-8
"""
tdb.cluster -- Module for cluster data manipulation
"""

# Library imports
import sys
from traits.api import HasTraits, Int, Float, String, Array, Instance

# Package imports
from .config import Config
from tools.enum import Enum

# Constants
MIN_I_POS = Config['placefield']['min_positional_info']
MIN_I_SPIKE = Config['placefield']['min_skaggs_info']
MIN_SPIKE_COUNT = Config['placefield']['min_spike_count']
PLACE_SIGLEVEL = Config['placefield']['info_significance']

# Cluster quality ratings enumeration
ClusterQuality = Enum("None", "Poor", "Marginal", "Fair", "PrettyGood", "Good")


class ClusterData(HasTraits):
    
    """
    Handle to a single cluster
    """
    
    out = Instance(CPrint)

    # Automatically calculated
    N = Int(desc='number of spikes')
    mean_firing_rate = Float
    name = String
    
    # These should be set on instantiation
    rat = Int
    day = Int
    session = Int
    rds = Trait( Tuple(int,int,int) )
    tt = Int(desc='tetrode')
    cl = Int(desc='cluster')
    spikes = Array(desc='timestamp spike train')
    x = Array
    y = Array
    T = Float(desc='total session duration')
    spike_width = Float
    quality = Instance(ClusterQuality.None, ClusterQuality.None)
    comment = String
    N_running = Int(desc='total forward-running spikes')
    I_pos = Float
    I_spike = Float
    pos_p_value = Float
    spike_p_value = Float
    
    # Traits defaults

    def __init__(self, rds=None, tc=None, **traits):
        HasTraits.__init__(self, **traits)
        assert type(rds) is tuple and len(rds)==3, \
            'bad session specification: %s' % str(rds)
        assert type(tc) is tuple and len(tc)==2, \
            'bad cluster specification: %s' % str(tc)

        self.out = CPrint(prefix=self.__class__.__name__, color='purple')
        
        self.rds = tuple( int(v) for v in rds )
        self.name = 't%dc%d' % tc
        sgroup = get_group(rds=self.rds)
        self.node = sgroup._v_children[ self.name ]

        self._load_data()

    def _load_data(self):
        self.rat, self.day, self.session = self.rds
        self.tt = int( tc[0] )
        self.cl = int( tc[1] )
        
    def _N_default(self):
        return self.spikes.size
    
    def _mean_firing_rate_default(self):
        return self.N / self.T
    
    def _name_default(self):
        return 't%dc%d'%(self.tt, self.cl)
    

def string_to_quality(rating):
    if type(rating) is type(ClusterQuality.None):
        return rating
    try:
        rating = str(rating)
    except TypeError:
        sys.stderr.write("string_to_quality: requires string argument\n")
        raise
    q = ClusterQuality.None
    if rating != "":
        if "pretty" in rating.lower():
            rating = 'Pretty Good'
        try:
            q = getattr(ClusterQuality, rating.title().replace(" ", ""))
        except AttributeError:
            sys.stderr.write('string_to_quality: bad rating: \"%s\"\n'%rating)
    return q
