# encoding: utf-8
"""
cut_off_fields.py
"""

import os
import re
import itertools

import tables as tb
import numpy as np
import matplotlib.pyplot as plt

from scipy.stats import ks_2samp, chi2_contingency, chi2, mannwhitneyu
from scipy.misc import comb
from matplotlib import cm

from core.analysis import AbstractAnalysis
from core.report import BaseReport
from tools.misc import memoize
from tools.images import masked_array_to_rgba

from scanr.cluster import string_to_quality

from tdb.data import get_tdb_file, get_group
from tdb.meta import walk_mazes
from tdb.spike import get_session_clusters
from tdb.td_maps import get_coordinate_range, plot_boundaries, bin_edges, get_boundary_line_params
from tdb.tdb_misc import get_panel_session_list
from tdb.ratemap import get_ratemap
from tdb.field import label_place_fields

#=================================================================================

class CoMDistrAna(AbstractAnalysis):
    datadir = './CoMDistr'
    label = 'com distribution analysis'

    def collect_data(self,quality='Fair',repeat_filter='first',N=1000):

        self.results = {}

        platforms = ['leather', 'tape']
        stypes = ['STD', 'SHF']
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]
        for panel in panels:
            self.out('Collecting data for %s sessions...' % panel)

            cluster_list = repeat_cluster_filtered_list( \
                panel,quality=quality,repeat_filter=repeat_filter )

            tdb_file = get_tdb_file()
            field_table = tdb_file.root.spatial.fields

            CoMs, dist, cut_off = [], [], []
            n = 0
            for rat, day, maze, tt, cl in cluster_list:
                query = '(area=="CA1")&(rat==%d)&(day==%d)&(session==%d)&(tt==%d)&(cl==%d)' % (rat,day,maze,tt,cl)
                for field in field_table.where(query):
                    CoMs.append( (field['CoM_x'], field['CoM_y']) )
                    cut_off_field, r, d = detect_cut_off_field( \
                        panel,(rat,day,maze),'t%dc%d'%(tt,cl), \
                            field['threshold'],field['field_no'],params_out=True )
                    cut_off.append( cut_off_field )
                    dist.append( d )
                    n += 1
            cut_off = np.array(cut_off)
            dist = np.array(dist)

            # generate bootstrapped index
            x, y = ( np.array(arr) for arr in zip(*CoMs) )
            on_board = (abs(x)<50) * (abs(y)<50) # CoMs within apparatus
            n = np.sum(on_board)
            bs_index = np.random.randint(n, size=(N,n))

            tdb_file.close()
            self.results[panel] = {}
            self.results[panel]['data'] = (CoMs, dist, cut_off)
            self.results[panel]['bootstrap'] = bs_index
            self.results[panel]['on_board'] = on_board
            self.out('%d fields collected for %s sessions' % (len(CoMs),panel))

        self.results['panels'] = panels
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, CoM_filter='dist', **traits):
        self._reset_plot()

        panels = self.results['panels']
        for i, panel in enumerate( panels ):
            CoMs, dist, cut_off = self.results[panel]['data']
            xmin, xmax, ymin, ymax = get_coordinate_range( \
                pattern=panel, **traits)

            plt.subplot(2,2,i+1)

            if CoM_filter is None:
                plt.scatter( *zip(*CoMs),s=20,c='k' )
            elif CoM_filter == 'cut-off':
                plt.scatter( *zip(*CoMs),s=20,c=cut_off.astype(float), \
                    cmap=cm.binary )
            elif CoM_filter == 'dist':
                plt.scatter( *zip(*CoMs),s=20,c=dist,cmap=cm.bwr,vmin=0,vmax=75 )

            plot_boundaries(pattern=panel)
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.axis([-80, 80, -80, 80])
            plt.title(panel,fontsize=32,va='bottom')
        if CoM_filter is None:
            plt.suptitle('Field Center of Mass Distribution',fontsize=40)
        else:
            plt.suptitle('Field Center of Mass Distribution (%s)' % CoM_filter, \
                fontsize=40)

        f = plt.gcf()
        f.set_size_inches(15,15,forward=True)
        if CoM_filter is None:
            f.savefig( os.path.join(self.datadir,'CoM_distr.png'))
        else:
            f.savefig( os.path.join(self.datadir,'CoM_distr_%s.png'%CoM_filter))

        self.out('Done')
    # end of METHOD process_data

    def CoM_hist(self,nbin=25,CoM_filter=None,extent=None,min_dist=10,**traits):
        self._reset_plot()

        panels = self.results['panels']
        for i, panel in enumerate( panels ):
            if extent is None: # All fields
                if 'STD' in panel:
                    bins = np.linspace(-75,50,nbin+1)
                elif 'SHF' in panel:
                    bins = np.linspace(-55,70,nbin+1)
            elif extent == 'far': # to further vertices, default nbin 19
                if 'STD' in panel:
                    bins = np.linspace(-72.77,48.51,nbin+1)
                elif 'SHF' in panel:
                    bins = np.linspace(-53.36,67.91,nbin+1)
            elif extent == 'near': # to nearer vertices, default nbin 16
                if 'STD' in panel:
                    bins = np.linspace(-48.51,24.26,nbin+1)
                elif 'SHF' in panel:
                    bins = np.linspace(-29.11,43.66,nbin+1)

            CoMs, dist, cut_off = self.results[panel]['data']
            if CoM_filter is None:
                data = CoMs[:]
            if CoM_filter == 'cut-off': 
                data = np.array(CoMs)[cut_off.astype(bool)]
            elif CoM_filter == 'dist': 
                data = np.array(CoMs)[dist<min_dist]

            a, b, c = get_boundary_line_params(pattern=panel)
            CoM_dist = [ float(a*x+b*y+c)/np.sqrt(a*a+b*b) for x, y in data ]

            plt.subplot(2,2,i+1)
            plt.hist(CoM_dist,bins=bins)
            plt.axvline(x=0,lw=2,c='r')
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Field Center of Mass Histograms',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(15,15,forward=True)
        f.savefig( os.path.join(self.datadir,'CoM_hist.png'))
    # end of METHOD CoM_hist

    def bootstrap_hist(self,nbin=20,sig='***'):
        self._reset_plot()

        #bins = nbin
        bins = np.linspace(-0.1,1.0,nbin+1)

        exp = np.sqrt(17) / 20.0 # area ratio
        if sig == '*':
            s = 0.1
        elif sig == '**':
            s = 0.05
        elif sig == '***':
            s = 0.01

        panels = self.results['panels']
        for i, panel in enumerate( panels ):
            a, b, c = get_boundary_line_params(pattern=panel)

            CoMs, dist, cut_off = self.results[panel]['data']
            bs_idx = self.results[panel]['bootstrap']
            data = np.array( CoMs[:] )

            CoM_dist = np.array( [abs(float(a*x+b*y+c)/np.sqrt(a*a+b*b)) \
                for x, y in data] )[self.results[panel]['on_board']]
            n = len(CoM_dist)
            print n, max(CoM_dist), min(CoM_dist)
            obs = np.sum(CoM_dist<10) / float(n)
            bs = [ np.sum(CoM_dist[idx]<10)/float(n) for idx in bs_idx ]

            plt.subplot(2,2,i+1)
            plt.hist(bs,bins=bins)

            plt.axvline(x=exp,lw=2,c='r')
            left_tail = 2*obs - np.percentile(bs,100-100*(s/2.0))
            right_tail = 2*obs - np.percentile(bs,100*(s/2.0))
            for CI in [left_tail, right_tail]:
                plt.axvline( x=CI,alpha=0.5,ls='--',lw=1,color='k')
            plt.axvline(x=obs,lw=2,color='k')

            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Bootstrap',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(15,15,forward=True)
        f.savefig( os.path.join(self.datadir,'bootstrap_hist.png'))
    # end of METHOD bootstrap_hist

    def _reset_plot(self):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

# end of CLASS CoMDistrAna

#=================================================================================

class CountCutoffFieldAna(AbstractAnalysis):
    datadir = './CountCutoff'
    label = 'count cut off analysis'

    def collect_data(self,quality='Fair',repeat_filter='first',control='all',N=None):
        self.results = {}

        platforms = ['leather', 'tape']
        stypes = ['STD', 'SHF']
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        ctrl_list = repeat_cluster_filtered_list( \
            'plain-%s'%control,quality=quality,repeat_filter=repeat_filter)

        for panel in panels:
            self.out('Collecting data for %s sessions...' % panel)
            obs_list = repeat_cluster_filtered_list( \
                panel,quality=quality,repeat_filter=repeat_filter )

            tdb_file = get_tdb_file()
            field_table = tdb_file.root.spatial.fields

            self.results[panel] = {}
            for control_data in [True, False]:
                cut_off = []
                cluster_list = ctrl_list if control_data else obs_list
                for rat, day, maze, tt, cl in cluster_list:
                    query = '(area=="CA1")&(rat==%d)&(day==%d)&(session==%d)&(tt==%d)&(cl==%d)' % (rat,day,maze,tt,cl)
                    for field in field_table.where(query):
                        # check contribution
                        #cut_off.append( detect_cut_off_field(panel,(rat,day,maze),\
                        #    't%dc%d'%(tt,cl),field['threshold'],field['field_no'],\
                        #        min_r=0.8,min_d=10) )
                        # detect_not_cross
                        cut_off.append( detect_cut_off_field(panel,(rat,day,maze),\
                            't%dc%d'%(tt,cl),field['threshold'],field['field_no'],\
                                min_r=0.75,min_d=10000) )
                        #cut_off.append( detect_cut_off_field(panel,(rat,day,maze),\
                        #   't%dc%d'%(tt,cl),field['threshold'],field['field_no']) )

                if control_data:
                    ctrl = np.array(cut_off)
                    n_ctrl = len(cut_off)
                    data_type = 'control'
                else:
                    obs = np.array(cut_off)
                    n_obs = len(cut_off)
                    data_type = 'observation'

                n_cut_off, n_all = np.sum(cut_off), len(cut_off)
                self.out( '%d out of %d fields cut-off at boundary (%s)' \
                    % (n_cut_off,n_all,data_type) )
                self.results[panel][data_type] = (n_cut_off, n_all)

            if N is not None: # do permutation test
                self.out('start permutaion...')
                cat = np.r_[ctrl, obs]
                perm = []
                for i in range(N):
                    index = np.random.permutation(n_ctrl+n_obs)
                    # ((p-ctrl_cut_off,p-ctrl_all), (p-obs_cut_off,p-obs_all))
                    perm.append( ((np.sum(cat[index][:n_ctrl]),n_ctrl), \
                        (np.sum(cat[index][n_ctrl:]),n_obs)) )
                self.results[panel]['perm'] = perm

            tdb_file.close()

        self.results['panels'] = panels
        self.results['control'] = 'plain-%s' % control
        self.results['N'] = N
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, nbin=10):
        self._reset_plot()

        bins = np.linspace(-0.5,0.5,nbin+1)

        for i, panel in enumerate(self.results['panels']):
            data = self.results[panel]
            perm_diff = [ float(obs[0])/obs[1] - float(ctrl[0])/ctrl[1] \
                for ctrl, obs in data['perm'] ]
            real_diff = float(data['observation'][0])/data['observation'][1] - \
                float(data['control'][0])/data['control'][1]

            p = np.sum(np.array(perm_diff)>real_diff) / float(len(perm_diff))
            p = min(p,1-p-1.0/len(perm_diff)) * 2
            self.out('p = %.4f for %s sessions' % (p,panel))

            plt.subplot(2,2,i+1)
            plt.hist(perm_diff,color='grey',bins=bins)
            plt.axvline(x=0,lw=1,c='k')
            plt.axvline(x=real_diff,lw=2,c='r')
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Cut-Off Ratio Differences',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(20,15,forward=True)
        f.savefig( os.path.join(self.datadir,'permutaion_test.png'))
    # end of METHOD process_data

    def stats(self):
        self._reset_plot()

        for i, panel in enumerate(self.results['panels']):
            data = self.results[panel]
            perm_diff = [ float(obs[0])/obs[1] - float(ctrl[0])/ctrl[1] \
                for ctrl, obs in data['perm'] ]
            real_diff = float(data['observation'][0])/data['observation'][1] - \
                float(data['control'][0])/data['control'][1]

            plt.subplot(2,2,i+1)
            plt.hist(perm_diff,color='grey',bins=bins)
            plt.axvline(x=0,lw=1,c='k')
            plt.axvline(x=real_diff,lw=2,c='r')
            plt.tick_params(axis='both',labelsize=20,pad=10)
            #plt.xlim( (bins[0], bins[-1]) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Cut-Off Ratio Differences',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(20,15,forward=True)
        f.savefig( os.path.join(self.datadir,'permutaion_test.png'))
    # end of METHOD process_data

    def _reset_plot(self):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)
# end of CLASS CountCutoffFieldAna

#=================================================================================

class CountFarSideAna(AbstractAnalysis):
    datadir = './CountFarSide'
    label = 'count far side analysis'

    def collect_data(self,quality='Fair',repeat_filter='first',control='all',N=None):
        self.results = {}

        platforms = ['leather', 'tape']
        stypes = ['STD', 'SHF']
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        ctrl_list = repeat_cluster_filtered_list( \
            'plain-%s'%control,quality=quality,repeat_filter=repeat_filter)

        for panel in panels:
            self.out('Collecting data for %s sessions...' % panel)
            obs_list = repeat_cluster_filtered_list( \
                panel,quality=quality,repeat_filter=repeat_filter )

            tdb_file = get_tdb_file()
            field_table = tdb_file.root.spatial.fields

            self.results[panel] = {}
            for control_data in [True, False]:
                farsides = []
                cluster_list = ctrl_list if control_data else obs_list
                for rat, day, maze, tt, cl in cluster_list:
                    query = '(area=="CA1")&(rat==%d)&(day==%d)&(session==%d)&(tt==%d)&(cl==%d)' % (rat,day,maze,tt,cl)
                    for field in field_table.where(query):
                        farsides.append( count_far_side(panel,(rat,day,maze),\
                            't%dc%d'%(tt,cl),field['threshold'], \
                                field['field_no'],d=10) )

                if control_data:
                    ctrl = np.array(farsides)
                    n_ctrl = len(farsides)
                    data_type = 'control'
                else:
                    obs = np.array(farsides)
                    n_obs = len(farsides)
                    data_type = 'observation'

                n_field = len(farsides)
                avg = np.sum(farsides) / float(n_field)
                self.out( 'average %.2f far sides for %d fields (%s)' \
                    % (avg,n_field,data_type) )
                self.results[panel][data_type] = farsides

            u, p = mannwhitneyu(obs, ctrl)
            self.out('%s: U=%.4f, p=%.4f' % (panel,u,p))

            if N is not None: # do permutation test
                self.out('start permutaion...')
                cat = np.r_[ctrl, obs]
                perm = []
                for i in range(N):
                    index = np.random.permutation(n_ctrl+n_obs)
                    perm.append( (np.sum(cat[index][:n_ctrl])/float(n_ctrl), \
                        np.sum(cat[index][n_ctrl:])/float(n_obs)) )
                self.results[panel]['perm'] = perm

            tdb_file.close()

        self.results['panels'] = panels
        self.results['control'] = 'plain-%s' % control
        self.results['N'] = N
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, control=False):
        self._reset_plot()

        bins = np.linspace(-0.5,2.5,4) # 0, 1, 2

        for i, panel in enumerate(self.results['panels']):
            data = self.results[panel]['control'] if control else \
                self.results[panel]['observation']
            
            plt.subplot(2,2,i+1)
            plt.hist(data,color='b',bins=bins)
            plt.xlim( [-0.5, 2.5] )
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Far Side Counts',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(20,15,forward=True)
        if control:
            f.savefig( os.path.join(self.datadir,'fs_hist_ctrl.png'))
        else:
            f.savefig( os.path.join(self.datadir,'fs_hist_obs.png'))
    # end of METHOD process_data

    def permutation_mean(self, nbin=10):
        self._reset_plot()

        #bins = np.linspace(-0.4,0.4,nbin+1)
        bins = nbin

        for i, panel in enumerate(self.results['panels']):
            data = self.results[panel]
            perm_diff = [ obs-ctrl for ctrl, obs in data['perm'] ]
            real_diff = np.mean(data['observation']) - np.mean(data['control'])

            p = np.sum(np.array(perm_diff)<real_diff) / float(len(perm_diff))
            if p > 0.5:
                p = 1 - p
            p *= 2
            self.out('p = %.4f for %s sessions' % (p,panel))

            plt.subplot(2,2,i+1)
            plt.hist(perm_diff,color='grey',bins=bins)
            plt.axvline(x=0,lw=2,c='k')
            plt.axvline(x=real_diff,lw=2,c='r')
            plt.tick_params(axis='both',labelsize=20,pad=10)
            #plt.xlim( (bins[0], bins[-1]) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Far Side Counts Differences',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(20,15,forward=True)
        f.savefig( os.path.join(self.datadir,'permutaion_mean.png'))
    # end of METHOD permutation_mean 

    def _reset_plot(self):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)
# end of CLASS CountCutoffFieldAna

#=================================================================================

class OverlapContourAna(AbstractAnalysis):
    datadir = './OverlapContour'
    label = 'overlap contour analysis'

    def collect_data(self,quality='Fair',repeat_filter='first',control='all',N=None,filter_zero=False):
        self.results = {}

        platforms = ['leather', 'tape']
        stypes = ['STD', 'SHF']
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        ctrl_list = repeat_cluster_filtered_list( \
            'plain-%s'%control,quality=quality,repeat_filter=repeat_filter)

        for panel in panels:
            self.out('Collecting data for %s sessions...' % panel)
            obs_list = repeat_cluster_filtered_list( \
                panel,quality=quality,repeat_filter=repeat_filter )

            tdb_file = get_tdb_file()
            field_table = tdb_file.root.spatial.fields

            self.results[panel] = {}
            for control_data in [True, False]:
                oc = []
                cluster_list = ctrl_list if control_data else obs_list
                for rat, day, maze, tt, cl in cluster_list:
                    query = '(area=="CA1")&(rat==%d)&(day==%d)&(session==%d)&(tt==%d)&(cl==%d)' % (rat,day,maze,tt,cl)
                    for field in field_table.where(query):
                        oc.append( overlap_contour(panel,(rat,day,maze),\
                            't%dc%d'%(tt,cl),field['threshold'],field['field_no']) )
                data_type = 'control' if control_data else 'observation'
                oc = np.array(oc)
                if filter_zero:
                    oc = oc[oc>0]
                self.results[panel][data_type] = oc
                self.out( '%d fields collected (%s)' % (len(oc),data_type) )

            if N is not None: # do permutation test
                self.out('start permutaion...')

                ctrl = self.results[panel]['control']
                obs = self.results[panel]['observation']
                cat = np.r_[ctrl, obs]

                perm = []
                for i in range(N):
                    index = np.random.permutation( len(cat) )
                    perm.append( (cat[index[:len(ctrl)]],cat[index[len(ctrl):]]))
                self.results[panel]['perm'] = perm

            tdb_file.close()

        self.results['panels'] = panels
        self.results['control'] = 'plain-%s' % control
        self.results['N'] = N
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, nbin=10, control=False):
        self._reset_plot()

        #bins = nbin
        bins = np.linspace(0,70,nbin+1)

        for i, panel in enumerate(self.results['panels']):
            data = self.results[panel]['control'] if control else \
                self.results[panel]['observation']
            
            plt.subplot(2,2,i+1)
            plt.hist(data,color='b',bins=bins)
            plt.xlim( [0, 70] )
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Overlap Contour Bins',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(20,15,forward=True)
        f.savefig( os.path.join(self.datadir,'oc_hist.png'))
    # end of METHOD process_data

    def permutation_mean(self, nbin=10):
        self._reset_plot()

        #bins = np.linspace(-0.4,0.4,nbin+1)
        bins = nbin

        for i, panel in enumerate(self.results['panels']):
            data = self.results[panel]
            perm_diff = [ np.mean(obs)-np.mean(ctrl) for ctrl, obs in data['perm'] ]
            real_diff = np.mean(data['observation']) - np.mean(data['control'])

            p = np.sum(np.array(perm_diff)>real_diff) / float(len(perm_diff)) * 2
            self.out('p = %.4f for %s sessions' % (p,panel))

            plt.subplot(2,2,i+1)
            plt.hist(perm_diff,color='grey',bins=bins)
            plt.axvline(x=0,lw=2,c='k')
            plt.axvline(x=real_diff,lw=2,c='r')
            plt.tick_params(axis='both',labelsize=20,pad=10)
            #plt.xlim( (bins[0], bins[-1]) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Overlap Bin Counts Differences',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(20,15,forward=True)
        f.savefig( os.path.join(self.datadir,'permutaion_mean.png'))
    # end of METHOD process_data

    def permutation_ratio(self, thrd=20, nbin=10):
        self._reset_plot()

        #bins = np.linspace(-0.4,0.4,nbin+1)
        bins = nbin

        for i, panel in enumerate(self.results['panels']):
            data = self.results[panel]
            perm_diff = [ np.sum(ctrl>thrd)/float(len(ctrl)) - \
                np.sum(obs>thrd)/float(len(obs)) for ctrl, obs in data['perm'] ]
            real_diff = \
                np.sum(data['observation']>thrd)/float(len(data['observation'])) -\
                    np.sum(data['control']>thrd)/float(len(data['control']))

            p = np.sum(np.array(perm_diff)>real_diff) / float(len(perm_diff)) * 2
            self.out('p = %.4f for %s sessions' % (p,panel))

            plt.subplot(2,2,i+1)
            plt.hist(perm_diff,color='grey',bins=bins)
            plt.axvline(x=0,lw=2,c='k')
            plt.axvline(x=real_diff,lw=2,c='r')
            plt.tick_params(axis='both',labelsize=20,pad=10)
            #plt.xlim( (bins[0], bins[-1]) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Strong Overlap Counts Differences',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(20,15,forward=True)
        f.savefig( os.path.join(self.datadir,'permutaion_ratio.png'))
    # end of METHOD process_data

    def long_overlap_ratio(self, control=False, thrd=25):
        for i, panel in enumerate(self.results['panels']):
            data = self.results[panel]['control'] if control else \
                self.results[panel]['observation']
            data = np.array(data)

            self.out( '%.4f %d %d' % (np.mean(data),np.sum(data>thrd),len(data)) )
    # end of METHOD process_data

    def stats(self):
        for i, panel in enumerate(self.results['panels']):
            obs = self.results[panel]['observation']
            ctrl = self.results[panel]['control']
            u, p = mannwhitneyu(obs, ctrl)
            self.out('%s: U=%.4f, p=%.4f' % (panel,u,p))
    # end of METHOD process_data

    def _reset_plot(self):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)
# end of CLASS CountCutoffFieldAna


#=================================================================================

class ChkCutoffFieldReport(BaseReport):

    datadir = './ChkCutOff'
    label = 'check cut off'

    ncols = 4
    nrows = 5
    figwidth = 8.0
    fighight = 11.0

    def collect_data(self,quality='Fair',min_r=0.8,min_d=5,**traits):
        self.results = {}

        tdb_file = get_tdb_file()
        field_table = tdb_file.root.spatial.fields

        quality_query = '(quality!="Poor")'
        if quality == 'Fair':
            quality_query += '&(quality!="Marginal")'
        query = '(area=="CA1")&(field_no==1)&' + quality_query

        data_list = []
        for field in field_table.where(query):
            rds = (field['rat'], field['day'], field['session'])
            if rds == (710, 17, 3): # wrong board placement
                continue
            tc = 't%dc%d' % (field['tt'],field['cl'])

            session = get_group(rds=rds)
            platform = session._v_attrs['platform']
            if platform not in ['leather', 'tape']:
                continue
            ptype = session._v_attrs['type']
            if ptype not in ['STD', 'SHF']:
                continue
            panel = platform + '_' + ptype

            data_list.append( (rds,panel,tc,field['threshold']) )
        data_list = list( set(data_list) )
        tdb_file.close()

        def plot_data(ax, rds, panel, tc, threshold):
            [y_edges, x_edges] = bin_edges(rds,**traits)

            session = get_group(rds=rds)
            spikes = session._v_children[tc]
            H_rate, mask = get_ratemap(session,spikes,rds,**traits)

            field_num, labels, contours, n_edges, thrd = \
                label_place_fields(rds,session,spikes,threshold=threshold,**traits)
            H_out = contours.copy()
            for i in range(field_num):
                cut_off, r, d = detect_cut_off_field( panel,rds,tc,threshold,i+1, \
                    min_r=min_r,min_d=min_d,params_out=True,**traits )
                if cut_off:
                    H_out[ labels==i+1 ] = i+1
                self.out( '(%d,%d,%d) %s field %d: r=%.2f, d=%.2f' \
                    % (rds+(tc,i+1,r,d)) )

            img = masked_array_to_rgba(H_out,mask=mask, \
                mask_color='b',cmap='cubehelix_r')
            ax.imshow( img,origin='lower',interpolation='none',aspect='equal',\
                extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )
            plot_boundaries(*rds)
            ax.set_title('(%d,%d,%d) %s' %(rds+(tc,)), size='xx-small')
            plt.tick_params(axis='both',labelsize=7)
        # end of FUNCTION plot_data

        for data, ax in self.get_plot( data_list ):
            rds, panel, tc, threshold = data 
            plot_data(ax, rds, panel, tc, threshold)

    # end of METHOD collect_data
# end of CLASS ChkCutoffFieldReport

#=================================================================================

def tdb_default_panels():
    # single line
    platforms = ['leather', 'tape']
    stypes = ['STD', 'SHF']
    sl = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]
    # single line + deStijl
    tex, pla = 'TEX', 'square_PLA'
    panels = sl + ['deStijl', 'plain-all', 'plain-before', 'plain-after']
    patterns = sl + [tex, pla, pla, pla] # platform_stype

    return panels, patterns

def get_first_qualified_maze(rat, day, tc, quality='Fair'):
    tdb_file = get_tdb_file()
    field_table = tdb_file.root.spatial.fields

    # list the sessions where the cells passed the quality criteria and had field
    tt, cl = [ int(i) for i in re.findall(r'\d+',tc) ] # tXcY --> [X, Y]
    quality_query = '(quality!="Poor")'
    if quality == 'Fair':
        quality_query += '&(quality!="Marginal")'
    query = '(rat==%d)&(day==%d)&(tt==%d)&(cl==%d)' % (rat,day,tt,cl) \
        + '&' + quality_query
    maze_list = [ field['session'] for field in field_table.where(query) ]

    tdb_file.close()

    first_maze = {}
    for maze in maze_list:
        session = get_group( rds=(rat,day,maze) )
        stype = session._v_attrs['type']
        if (stype not in first_maze.keys()) or (first_maze[stype]>maze):
            first_maze[stype] = int( maze )

    return first_maze.values() 
# end of FUNCTION get_first_qualified_maze

def repeat_cluster_filtered_list(panel,quality='Fair',repeat_filter='highest_mean'):
    tdb_file = get_tdb_file()
    field_table = tdb_file.root.spatial.fields

    maze_list, highest_mean = {}, {}
    rds_list = get_panel_session_list(panel)
    for rds in rds_list:
        rat, day, maze = rds
        session = get_group(rds=rds)
        tt_cl_list = get_session_clusters(session)
        for tt, cl in tt_cl_list:
            clust = session._v_children['t%dc%d'%(tt,cl)]

            clust_quality = string_to_quality( clust._v_attrs['quality'] )
            if clust_quality < string_to_quality(quality):
                continue

            query = '(area=="CA1")&(rat==%d)&(day==%d)&(session==%d)&(tt==%d)&(cl==%d)' % (rat,day,maze,tt,cl)
            field_idx = field_table.getWhereList(query)
            if not len(field_idx):
                continue
            mr = max( field_table[field_idx]['mean_rate'] )

            key = (rat, day, tt, cl)
            if key not in maze_list:
                maze_list[key] = [maze,]
                highest_mean[key] = mr
            else:
                if repeat_filter == 'first':
                    if maze < maze_list[key][0]:
                        maze_list[key] = [maze,]
                elif repeat_filter == 'highest_mean':
                    if mr > highest_mean[key]:
                        maze_list[key] = [maze,]
                        highest_mean[key] = mr 
                else:
                    maze_list[key].append( maze )

    tdb_file.close()
    if repeat_filter is not None:
        cluster_list = [ (rat,day,maze_list[(rat,day,tt,cl)][0],tt,cl) \
            for rat, day, tt, cl in maze_list.keys() ]
    else:
        clust_list = []
        for rat, day, tt, cl in maze_list.keys():
            for maze in maze_list[(rat,day,tt,cl)]:
                cluster_list.append( (rat,day,maze,tt,cl) )

    return cluster_list
# end of FUNCTION repeat_cluster_filtered_list

def detect_cut_off_field(panel,rds,tc,threshold,field_no,min_r=0.8,min_d=5,params_out=False,debug=False,**traits):
    a, b, c = get_boundary_line_params(pattern=panel)

    session = get_group(rds=rds)
    spikes = session._v_children[tc]

    [y_edges, x_edges] = bin_edges(rds,**traits)
    x_centers = (x_edges[:-1]+x_edges[1:]) / 2.0
    y_centers = (y_edges[:-1]+y_edges[1:]) / 2.0
    Hx, Hy = np.meshgrid(x_centers, y_centers)

    field_num, labels, contours, n_edges, threshold = \
        label_place_fields(rds,session,spikes,threshold=threshold,**traits)

    select = labels==field_no
    xy = zip( Hx[select], Hy[select] )
    values = np.array( [a*x+b*y+c for x, y in xy] )
    pos = np.sum(values>0)+np.sum(values==0)/2.0 # no. of pixels on positive side
    ratio = float(pos) / len(xy) 
    ratio = max( ratio, 1-ratio) # reverse ratio if pos is the smaller side

    select = contours==field_no
    min_dist = min( [abs(a*x+b*y+c)/float(np.sqrt(a*a+b*b)) \
        for x, y in zip(Hx[select],Hy[select])] )

    if debug:
        return ratio, min_dist, labels, contours

    if (min_dist<=min_d) and (ratio>=min_r):
        if params_out:
            return True, ratio, min_dist
        else:
            return True
    else:
        if params_out:
            return False, ratio, min_dist
        else:
            return False
# end of FUNCTION detect_cut_off_field

def count_far_side(panel,rds,tc,threshold,field_no,d=10,params_out=False,debug=False,**traits):
    a, b, c = get_boundary_line_params(pattern=panel)

    session = get_group(rds=rds)
    spikes = session._v_children[tc]

    [y_edges, x_edges] = bin_edges(rds,**traits)
    x_centers = (x_edges[:-1]+x_edges[1:]) / 2.0
    y_centers = (y_edges[:-1]+y_edges[1:]) / 2.0
    Hx, Hy = np.meshgrid(x_centers, y_centers)

    field_num, labels, contours, n_edges, threshold = \
        label_place_fields(rds,session,spikes,threshold=threshold,**traits)

    far_side = 0
    select = contours==field_no
    dist = [ (a*x+b*y+c)/float(np.sqrt(a*a+b*b)) \
        for x, y in zip(Hx[select],Hy[select]) ]
    if min(dist) <= -10:
        far_side += 1
    if max(dist) >= 10:
        far_side += 1

    return far_side
# end of FUNCTION count_far_side

def overlap_contour(panel,rds,tc,threshold,field_no,hw=10,**traits):
    a, b, c = get_boundary_line_params(pattern=panel)

    session = get_group(rds=rds)
    spikes = session._v_children[tc]

    [y_edges, x_edges] = bin_edges(rds,**traits)
    x_centers = (x_edges[:-1]+x_edges[1:]) / 2.0
    y_centers = (y_edges[:-1]+y_edges[1:]) / 2.0
    Hx, Hy = np.meshgrid(x_centers, y_centers)

    field_num, labels, contours, n_edges, threshold = \
        label_place_fields(rds,session,spikes,threshold=threshold,**traits)

    select = contours==field_no
    xy = zip( Hx[select], Hy[select] )
    values = np.array( [a*x+b*y+c for x, y in xy] )
    pos = np.sum(values>0)+np.sum(values==0)/2.0 # no. of pixels on positive side
    ratio = float(pos) / len(xy) 
    ratio = max( ratio, 1-ratio) # reverse ratio if pos is the smaller side

    select = contours==field_no
    dist = np.array( [abs(a*x+b*y+c)/float(np.sqrt(a*a+b*b)) \
        for x, y in zip(Hx[select],Hy[select])] )

    return np.sum(dist<hw)
# end of FUNCTION detect_cut_off_field
