# encoding: utf-8
"""
tdb.data.py -- Interfaces, schema definitions and functions for handling HDF data 
Modified from scanr.data.py
"""

import sys
import tables as tb
import numpy as np

from os.path import exists as there

from tools.misc import DataSpreadsheet

from tdb.config import Config
from tdb.paths import get_data_file_path, get_group_path
from tdb.meta import metadata, get_rat_list, get_day_list, get_maze_list

# Constants
CfgData = Config['h5']

# Top-level pointer to the H5 data file object
if 'tdb_file' not in locals():
    tdb_file = None

# ================================================================================
# HDF file operations
# ================================================================================

def _open_file():
    return bool(type(tdb_file) is tb.File and tdb_file.isopen)

def get_tdb_file(readonly=False):
    """Return an open H5 file object linked to the main data file
    """
    global tdb_file

    # Close the data file if it is open and there is a mode mismatch
    if _open_file():
        if      (tdb_file._isWritable() and readonly) or \
            not (tdb_file._isWritable() or readonly):
            close_file()

    # If the file is closed, open it with the specified mode ('r' or 'a')
    if not _open_file():
        tdb_data_path = get_data_file_path('data', search=False)
        mode = (readonly and there(tdb_data_path)) and 'r' or 'a'
        tdb_file = tb.openFile(tdb_data_path, mode=mode, title='TDBData')
        sys.stdout.write('tdb_file: opened data file (mode=\'%s\')\n' % tdb_file.mode)

    return tdb_file
        
def close_file():
    """Close the H5 file if it is still open
    """
    if _open_file():
        tdb_file.close()
        sys.stdout.write('tdb_file: closed data file\n')

def flush_file():
    """Flush the H5 file if it is open
    """
    if _open_file():
        tdb_file.flush()
        sys.stdout.write('tdb_file: flushed data file\n')

# ================================================================================
# Modify data tree
# ================================================================================

def _node_exists(h5file, where, name=None):
    """Test whether a given node exists in the specified H5 file
    """
    try:
        h5file.getNode(where, name)
    except tb.NoSuchNodeError:
        return False
    return True

def initialize_groups(overwrite=True):
    """Create the primary data group structure of the H5 file based on the
    master metadata file. Specify overwrite=False to preserve the existing
    tree and only create new groups.
    
    Note: run update_group_metadata after this to populate group attributes
    with the appropriate metadata.
    """
    d = dict(rat=None, day=None, session=None)
    f = get_tdb_file(False) # readonly=False
    
    def create_group(where, name, **kw):
        if _node_exists(f, where, name):
            g = f.getNode(where, name)
            sys.stdout.write('initialize_groups: found node %s\n'%g)
        else:
            g = f.createGroup(where, name, **kw)
            sys.stdout.write('initialize_groups: created node %s\n'%g)
        return g
    
    # Create the hierarchical root group (if necessary)
    if _node_exists(f, '/tdb_data') and overwrite:
        f.removeNode('/tdb_data', recursive=True)
    if _node_exists(f, '/tdb_data'):
        tdb_data = f.getNode('/tdb_data')
    else:        
        tdb_data = create_group('/', 'tdb_data', title='2D Boundary Data')
    
    # Walk the sessions, creating group nodes for each session
    for rat in get_rat_list():
        d['rat'] = rat
        rat_group = create_group(tdb_data, CfgData['rat']%d, title='Rat %d'%rat)
        for day in get_day_list(rat):
            d['day'] = day
            day_group = create_group(rat_group, CfgData['day']%d, 
                title='Rat %d, Day %d'%(rat, day))
            for maze in get_maze_list(rat, day):
                d['session'] = maze
                maze_group = create_group(day_group, CfgData['session']%d, 
                    title='Rat %d, Day %d, Maze %d'%(rat, day, maze))
            d['session'] = None
        d['day'] = None
    d['rat'] = None
    
    f.flush()
    f.close()
    
def update_group_metadata(write=True):
    """Update the group metadata attributes of main data tree in the H5 file
    
    Metadata is available as the _v_attrs dictionary attribute of any group.
    
    Note: this used to be done as part of initialize_groups(), but the metadata
    setting has been refactored out to allow for easy updating of existing
    data trees without destructively initializing the tree structure.
    """
    def update_attr(grp, key, value):
        attrs = grp._v_attrs
        if key in attrs:
            if attrs[key] != value:
                if write:
                    attrs[key] = value
                sys.stdout.write('%s: updated %s to %s\n' % (
                    grp._v_pathname, key, str(value)))
        else:
            if write:
                attrs[key] = value
            sys.stdout.write('%s: added key %s with value %s\n' % (
                grp._v_pathname, key, str(value)))
    
    # Walk the sessions, updating metadata attributes of group nodes
    for rat in get_rat_list():
        rat_group = get_group(rat=rat)
        for day in get_day_list(rat):
            day_group = get_group(rat=rat, day=day)

            type_text = metadata[rat]['days'][day]['type']
            if type_text == "single line":
                expt_type = "SL"
            elif type_text == "deStijl":
                expt_type = "DS"
            
            for maze in get_maze_list(rat, day):
                maze_group = get_group(rds=(rat, day, maze))
                update_attr(maze_group, 'rat', rat)
                update_attr(maze_group, 'day', day)
                update_attr(maze_group, 'session', maze)
                update_attr(maze_group, 'expt_type', expt_type)
                session_meta = metadata[rat]['days'][day]['sessions'][maze]
                for k in session_meta.keys():
                    update_attr(maze_group, k, session_meta[k])
                maze_group._f_flush()

            day_group._f_flush()
        rat_group._f_flush()
    sys.stdout.write('Finished!\n')    

def new_array(where, name, obj, title='', force=False):
    """Create a new Array leaf in the tree, erasing an old one if it exists
    
    Arguments:
    where -- group path or node where array should be created
    name -- array name
    obj -- array object to store in the data file
    force -- force erase previous array without asking
    
    Returns new Array node object.
    """
    data_file = get_tdb_file()
        
    array = None
    try:
        array = data_file.getNode(where, name)
    except tb.NodeError:
        pass
    else:
        if force:
            sys.stdout.write('new_array: Erasing previous \'%s\' array.\n'%name)
            do_erase = 'y'
        else:
            do_erase = raw_input('Found previous \'%s\' array. Erase? (y/N) '%name)
        if do_erase.lower().strip().startswith('y'):
            data_file.removeNode(array)
            array = None
    finally:
        if not array:
            array = data_file.createArray(where, name, obj, title=title)
        data_file.flush()

    return array

def new_table(where, name, descr, title='', force=False):
    """Create a new Table leaf in the tree, erasing an old one if it exists
    
    Arguments:
    where -- group path or node where table should be created
    name -- key value from config 'h5' section for a table name
    descr -- table column descriptor
    title -- optional title for the new table
    force -- force erasure of a pre-existing table without asking
    
    Returns new Table node object.
    """
    data_file = get_tdb_file()
    table = None
    
    # Check for configured table name first, check with user otherwise
    try:
        table_name = CfgData[name]
    except KeyError:
        try_name = name.lower().strip()
        do_create = \
            raw_input('Not a configured table.\nCreate \'%s\' under \'%s\'? (y/N) '%
                (try_name, str(where)))
        if do_create.lower().strip().startswith('y'):
            table_name = try_name
        else:
            return None
    
    # Check for pre-existing table and query user for deletion if not forced
    try:
        table = data_file.getNode(where, table_name)
    except tb.NodeError:
        pass
    else:
        if force:
            sys.stdout.write('new_table: Erasing previous \'%s\' table.\n'%
                table_name)
            do_erase = 'y'
        else:
            do_erase = raw_input('Found previous \'%s\' table. Erase? (y/N) '%
                table_name)
        if do_erase.lower().strip().startswith('y'):
            data_file.removeNode(table)
            table = None
    finally:
        if not table:
            table = data_file.createTable(where, table_name, descr, title=title)
        data_file.flush()
        
    return table

# ================================================================================
# Access tree nodes
# ================================================================================

def get_root_group():
    """Return root group for the H5 data set
    """
    return get_tdb_file(True).root
    
def get_group(**kwds):
    """Return the group node corresponding the specified group
    
    Specify alternative pytable File object *data_file* with data tree that is
    isomorphic to the standard tdb data File. 
    
    Keyword arguments same as for get_group_path().
    """
    # Allow an rds keyword with tuple value to be passed in
    rds = kwds.pop('rds', None)
    if rds:
        kwds.update(tree='tdb', rat=rds[0], day=rds[1], session=rds[2])
    data_file = get_tdb_file()
    
    # Get the path and return the group node if it exists
    gpath = get_group_path(**kwds)
    group = None
    try:
        group = data_file.getNode(gpath)
    except tb.NodeError:
        sys.stderr.write('get_group: group node %s does not exist\n'%gpath)
    return group

def get_node(where, name=None, raise_on_fail=True):
    """Get the specified Table or Array node object from the H5 tree

    Arguments:
    where -- group pathname containing node, or Node object for pass-through
    name -- node name required if where is a pathname
    raise_on_fail -- throw exception if table not found; if False, return None
    
    Returns Table node object.
    """
    node = None
    if issubclass(where.__class__, tb.Node):
        node = where
    elif type(where) is str and type(name) is str:
        data_file = get_tdb_file()
        try:
            node = data_file.getNode(where, CfgData[name])
        except KeyError:
            try:
                node = data_file.getNode(where, name)
            except tb.NodeError:
                sys.stderr.write(
                    'get_node: node \'%s\' is not a child of %s\n'%(name, where))
        except tb.NodeError:
            sys.stderr.write(
                'get_node: table \'%s\' does not exist\n'%CfgData[name])
        if not node and raise_on_fail:
            raise tb.NodeError, 'could not find node %s'%name
    else:
        raise ValueError, 'node object or name is required'
    return node

# ================================================================================
# Misc. 
# ================================================================================

def dump_table(where, name=None, condn=None, filename=None, sep=','):
    """Dump data table out to a spreadsheet
    
    Optionally use a table search query by setting *condn* keyword to either
    a query string or callable that maps row->Bool.
    """
    if type(where) is tb.Table:
        table = where
    else:
        table = get_node(where, name)
    
    if filename is None:
        filename = "%s_table.csv"%table.name
    
    colnames = table[0].dtype.names
    coltypes = []
    for col in colnames:
        typename = table.coltypes[col]
        if bool(len(table.coldtypes[col].shape)): coltypes.append('s')
        elif typename.find('int') >= 0: coltypes.append('d')
        elif typename.find('bool') >= 0: coltypes.append('d')
        elif typename.find('float') >= 0: coltypes.append('f')
        elif typename.find('str') >= 0: coltypes.append('s')   
        else: coltypes.append('s')       
    spreadsheet = DataSpreadsheet(filename, zip(colnames, coltypes))
    record = spreadsheet.get_record()
    
    def save_row(row):
        for name, coltype in zip(colnames, coltypes):
            if coltype == 's' or type(row[name]) is np.ndarray:
                record[name] = str(row[name]).replace(',', ';')
            elif coltype == 'd':
                record[name] = int(row[name])
            elif coltype == 'f':
                record[name] = float(row[name])
        spreadsheet.write_record(record)
    
    if condn is None:
        for row in table.iterrows():
            save_row(row)
    elif type(condn) is str:
        for row in table.where(condn):
            save_row(row)
    elif callable(condn):
        for row in table.iterrows():
            if condn(row):
                save_row(row)
    
    sys.stdout.write('Dumped %s to %s.\n'%(table._v_pathname, filename))
    spreadsheet.close()

def get_unique_row(table, query, raise_on_fail=True):
    """Get the singular, unique record in the table that matches the query
    
    If raise_on_fail is set to False, then multiple record results will return
    the first record, and empty results will return None. Otherwise, an
    exception will be raised in those cases.
    """
    rec = None
    where = table.getWhereList(query)
    if len(where):
        rec = table[where[0]]
        if len(where) > 1 and raise_on_fail:
            raise RuntimeError, '%s: found >1 record: \"%s\"'%(table._v_pathname, 
                query)
    elif raise_on_fail:
        raise RuntimeError, '%s: found no records: \"%s\"'%(table._v_pathname, 
            query)
    return rec
