# encoding: utf-8
"""
ensumble_corr.py
"""

import os
import sys
import itertools

import tables as tb
import numpy as np
import matplotlib.pyplot as plt

from matplotlib.colors import rgb_to_hsv, hsv_to_rgb
from scipy.ndimage.filters import gaussian_filter
from scipy.interpolate import interp2d

from core.analysis import AbstractAnalysis
from core.report import BaseReport
from tools.images import masked_array_to_rgba

from scanr.cluster import string_to_quality

from tdb.meta import stamp_to_time
from tdb.data import get_tdb_file, get_group
from tdb.field import label_place_fields
from tdb.tdb_misc import get_panel_session_list, highest_mean_filter
from tdb.ratemap import get_ratemap, smooth_ratemap, get_occ_rim, compute_velocity
from tdb.td_maps import bin_edges, get_boundary_line, plot_boundaries

MIN_VELOCITY = 10.0

#==================================================================================

class NeuralSpaceGradReport(BaseReport):
    datadir = './NSpaceGrad'
    label = 'neural space grad'

    ncols = 4
    nrows = 6

    def collect_data(self, min_quality='Fair', repeat_filter='highest_mean', norm=True, **traits):
        tdb_file = get_tdb_file()
        field_table = tdb_file.root.spatial.fields

        self.results = {}

        platforms = [ 'leather', 'tape' ]
        stypes = [ 'STD', 'SHF' ]
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        data_list = []
        for panel in panels:
            self.out('Collecting data for %s...' % panel)

            maze_list, highest_mean = {}, {}
            rds_list = get_panel_session_list( panel )
            for rds in rds_list:
                rat, day, maze = rds
                query = '(area=="CA1")&(rat==%d)&(day==%d)&(session==%d)' % rds
                for field in field_table.where(query):
                    quality = string_to_quality( field['quality'] )
                    if quality < string_to_quality(min_quality):
                        continue
                    tc = 't%dc%d' % (field['tt'], field['cl'])
                    rate = field['mean_rate']

                    clust = (rat, day, tc)
                    if clust not in maze_list.keys():
                        if (repeat_filter is None) and (field['field_no']==1):
                            maze_list[clust] = [maze, ]
                        elif repeat_filter == 'highest_mean':
                            maze_list[clust] = [maze, ]
                            highest_mean[clust] = rate
                    else:
                        if (repeat_filter is None) and (field['field_no']==1):
                            maze_list[clust].append( maze )
                        elif repeat_filter == 'highest_mean':
                            if rate > highest_mean[clust]:
                                highest_mean[clust] = rate
                                maze_list[clust] = [maze, ]

            clust_list = []
            for key in maze_list.keys():
                for maze in maze_list[key]:
                    rat, day, tc = key
                    clust_list.append( (rat,day,maze,tc) )

            H, head, tail = None, 0, 0
            cnt = 0
            for rds in rds_list:
                for clust in clust_list:
                    if np.mod(cnt,10) == 0:
                        sys.stdout.write('.')
                        sys.stdout.flush()
                    rat, day, maze, tc = clust
                    if rds != (rat,day,maze):
                        continue

                    session = get_group(rds=rds)
                    spikes = session._v_children[tc]

                    H_rate, mask = get_ratemap(session, spikes, rds, **traits)
                    H_rate = smooth_ratemap(H_rate, mask)
                    rim = get_occ_rim(session, **traits)
                    H_rate[rim+mask] = 0 # set mask==True or rim==True to zero
                    if norm:
                        H_rate /= H_rate.max()

                    H = H_rate.copy() if H is None else np.dstack( (H,H_rate) )
                    tail += 1
                    cnt += 1

                data_list.append( (rds,panel,head,tail) )
                head = tail

            self.results[panel] = H.copy()

            sys.stdout.write('\n')
            rats = list( set([data[0] for data in clust_list]) )
            days = list( set([data[0:2] for data in clust_list]) )
            mazes = list( set([data[0:3] for data in clust_list]) )
            clusts = list( set([data[0:4] for data in clust_list]) )
            self.out( '%d rats, %d days, %d mazes, %d cells collected' \
                %(len(rats), len(days), len(mazes), len(clusts)) )

        tdb_file.close()

        self.results['data_list'] = data_list
        self.out('Done')
    # end of METHOD collect_data

    def collect_data_old(self, min_quality='Fair', norm=True, **traits):
        tdb_file = get_tdb_file()
        field_table = tdb_file.root.spatial.fields

        self.results = {}

        platforms = [ 'leather', 'tape' ]
        stypes = [ 'STD', 'SHF' ]
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        data_list = []
        for panel in panels:
            self.out('Collecting data for %s...' % panel)
            H, head, tail = None, 0, 0
            rds_list = get_panel_session_list( panel )
            maze_cnt, cell_cnt = 0, 0
            for rds in rds_list:
                session = get_group(rds=rds)
                query = \
                    '(area=="CA1")&(field_no==1)&(rat==%d)&(day==%d)&(session==%d)'\
                        % rds
                for i, field in enumerate(field_table.where(query)):
                    if np.mod(i,10) == 0:
                        sys.stdout.write('.')
                        sys.stdout.flush()
                    quality = string_to_quality( field['quality'] )
                    if quality < string_to_quality(min_quality):
                        continue
                    tc = 't%dc%d' % (field['tt'], field['cl'])
                    spikes = session._v_children[tc]

                    H_rate, mask = get_ratemap(session, spikes, rds, **traits)
                    H_rate = smooth_ratemap(H_rate, mask)
                    rim = get_occ_rim(session, **traits)
                    H_rate[rim+mask] = 0 # set mask==True or rim==True to zero
                    if norm:
                        H_rate /= H_rate.max()

                    H = H_rate.copy() if H is None else np.dstack( (H,H_rate) )
                    tail += 1
                    cell_cnt += 1

                data_list.append( (rds,panel,head,tail) )
                head = tail
                maze_cnt += 1

            self.results[panel] = H.copy()
            sys.stdout.write('\n')
            self.out('%d sessions, %d cells collected' % (maze_cnt, cell_cnt))
            self.out('')

        tdb_file.close()

        self.results['data_list'] = data_list
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, bd=False):
        plt_types = ['Phase', 'Amplitude', 'Amp - Smooth', 'Phase - Amp filter']
        data_list = [ (m, n) for m, n in itertools.product( \
            self.results['data_list'],plt_types)]

        bd_colors = ['k', 'k', 'k', 'w']

        for item, ax in self.get_plot(data_list):
            data, plt_type = item
            rds, panel, head, tail = data
            if plt_type == 'Phase': # first encounter
                H = self.results[panel][:,:,head:tail]
                bd_line, theta = get_boundary_line( pattern=panel )
                [y_edges, x_edges] = bin_edges(pattern=panel)
                extent = [ x_edges[0], x_edges[-1], y_edges[0], y_edges[-1] ]
                im_arr = self.plot_default_subplots( \
                    H, theta, extent, plot_out=False ) # [gim,aim,s_aim,gaim]
                im_idx = 0

            plt.imshow( im_arr[im_idx],origin='lower',interpolation='nearest', \
                aspect=1,extent=extent )
            if bd:
                plot_boundaries( pattern=panel, **{'c':bd_colors[im_idx],'lw':1} )
            if self.firstcol:
                plt.ylabel( '(%d, %d, %d) %s'%(rds+(panel,)), fontsize=7 )
            plt.tick_params(axis='both',labelsize=5)

            im_idx += 1

        self.out('Done')
    # end of METHOD process_data

    def plot_panel(self, panel='leather_STD', bd=True):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        bd_line, theta = get_boundary_line( pattern=panel )
        [y_edges, x_edges] = bin_edges(pattern=panel)
        extent = [ x_edges[0], x_edges[-1], y_edges[0], y_edges[-1] ]

        plot_bd = panel if bd else None
        im_arr = self.plot_default_subplots( \
            self.results[panel], theta, extent, 1, 4, 1, plot_panel_bd=plot_bd )

        fname = panel[:]
        if bd:
            fname += '_bd'
        fname += '.png'

        #plt.show()
        fig = plt.gcf()
        fig.set_size_inches(24,5.5,forward=True)
        fig.savefig( os.path.join(self.datadir,fname) )
    # end of METHOD plot_panel

    def debug(self, set_no=1):
        kwds = {}

        if set_no == 1:
            H1 = np.zeros( (100,100) )
            H1[10:71,50:91] = 1

            H2 = np.zeros( (100,100) )
            H2[25:61,30:50] = 1

            H = np.dstack( (H1,H2) )
            theta = 90

            row, col = 3, 2
            kwds['mask'] = H1 < 0
            kwds['mask_color'] = 'k'
            kwds['cmap'] = 'jet'
        elif set_no == 2:
            H1 = np.zeros( (100,100) )
            H1[30:51,65:76] = 1
            H1 = gaussian_filter(H1,sigma=2.5)

            H2 = np.zeros( (100,100) )
            H2[35:51,35:45] = 1
            H2 = gaussian_filter(H2,sigma=2.5)

            H = np.dstack( (H1,H2) )
            theta = 90

            row, col = 3, 2
            kwds['mask'] = H1 < 0
            kwds['mask_color'] = 'k'
            kwds['cmap'] = 'jet'

        m, n, d = H.shape
        extent = [0, n, 0, m]
        for i in range(d):
            plt.subplot(row, col, i+1)
            img = masked_array_to_rgba(H[:,:,i],**kwds)
            plt.imshow( img,origin='lower',interpolation='none',aspect='auto',\
                extent=extent )
        im_arr = self.plot_default_subplots( \
            H, theta, extent, row, col, start_id=i+2)
        plt.show()
    # end of METHOD debug

    #------------------------------------------------------------------------------

    def plot_default_subplots(self, H, theta, extent, row=1, col=4, start_id=1, plot_panel_bd=None, plot_out=True):
        gamma, amp = self.get_nspace_grad( H, theta )
        mask = gamma < 0

        gim = masked_array_to_rgba( gamma,mask=mask,mask_color='k',cmap='hsv', \
            norm=False,cmin=theta,cmax=theta+180 )

        aim = masked_array_to_rgba( amp,mask=np.zeros(amp.shape).astype(bool), \
            mask_color='k',cmap='jet' ) 

        s_amp = gaussian_filter( amp, sigma=0.8 )
        s_aim = masked_array_to_rgba( s_amp,mask=np.zeros(amp.shape).astype(bool),\
            mask_color='k',cmap='jet' ) 

        hsv = rgb_to_hsv( gim[:,:,:-1] )
        M, m = s_amp[~mask].max(), s_amp[~mask].min()
        amp_fill = s_amp.copy()
        amp_fill[mask] = m
        x = (amp_fill-m) / float(M-m)
        hsv[:,:,-1] = 100.0 / (1+np.exp(-10*(x-0.5)))
        gaim = gim.copy()
        gaim[:,:,:-1] = hsv_to_rgb( hsv ) 
        gaim[mask] = gim[mask]

        img_array = [ gim, aim, s_aim, gaim ]
        if not plot_out:
            return img_array

        colors = ['k', 'k', 'k', 'w']
        for i, data in enumerate( zip(img_array,colors) ):
            im, color = data
            plt.subplot(row,col,start_id+i)
            plt.imshow( im,origin='lower',interpolation='nearest', \
                aspect='auto',extent=extent )
            if plot_panel_bd is not None:
                plot_boundaries( pattern=plot_panel_bd, **{'c':color,'lw':2} )

        return img_array
    # end of METHOD plot_default_subplots

    def get_nspace_grad(self, H, theta):
        m, n, d = H.shape

        Dx, Dy = None, None
        for i in range(d):
            dy, dx = np.gradient( H[:,:,i] ) # derivatives of the i_th cluster
            Dx = dx if Dx is None else np.dstack( (Dx,dx) )
            Dy = dy if Dy is None else np.dstack( (Dy,dy) )
        # Jacobian[x_0, y_0] = [ Dx[x_0,y_0,:] Dy[x_0,y_0,:] ]

        gamma, amp = np.zeros((m,n)), np.zeros((m,n))
        for x in range(n):
            for y in range(m):
                j_x, j_y = Dx[y,x,:], Dy[y,x,:]
                j_x_square, j_y_square = np.dot(j_x,j_x), np.dot(j_y,j_y)
                A = ( j_x_square - j_y_square ) * 0.5
                B = np.dot( j_x, j_y )
                if A==0 and B==0:
                    # amp == sqrt(jx^2) == sqrt(jy^2)
                    gamma[y,x], amp[y,x] = -1, np.sqrt(j_x_square)
                else:
                    sin_alpha = float(A) / np.sqrt(A**2+B**2)
                    beta = np.arcsin(sin_alpha) # alpha = beta or pi-beta
                    angles = [ np.pi/4.0-beta/2.0, np.pi*5/4.0-beta/2.0, \
                               np.pi*3/4.0+beta/2.0, np.pi*7/4.0+beta/2.0 ]
                    max_dist, omega = 0, -1
                    for angle in angles:
                        dist = np.linalg.norm( np.cos(angle)*j_x+np.sin(angle)*j_y )
                        if dist > max_dist:
                            omega = (angle*180.0/np.pi+360) % 360 # 0 <= omega < 360
                            max_dist = dist
                    if (omega<theta) or (omega>theta+180):
                        omega = (omega+180) % 360
                    gamma[y,x], amp[y,x] = omega, max_dist
        return gamma, amp
    # end of METHOD get_nspace_grad
# end of CLASS NeuralSpaceGradReport

#==================================================================================

class SLCorrAna(AbstractAnalysis):
    datadir = './SLCorr'
    label = 'single line corr'

    def collect_data(self, min_quality='Fair', repeat_filter=None, grid_type='ortho', norm=True, **traits):
        tdb_file = get_tdb_file()
        field_table = tdb_file.root.spatial.fields

        self.results = {}

        platforms = [ 'leather', 'tape' ]
        stypes = [ 'STD', 'SHF' ]
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        for panel in panels:
            self.out('Collecting data for %s...' % panel)

            maze_list, highest_mean = {}, {} 
            rds_list = get_panel_session_list( panel )
            for rds in rds_list:
                rat, day, maze = rds
                query = '(area=="CA1")&(rat==%d)&(day==%d)&(session==%d)' % rds
                for field in field_table.where(query):
                    quality = string_to_quality( field['quality'] )
                    if quality < string_to_quality(min_quality):
                        continue
                    tc = 't%dc%d' % (field['tt'], field['cl'])
                    rate = field['mean_rate']

                    clust = (rat, day, tc)
                    if clust not in maze_list.keys():
                        if (repeat_filter is None) and (field['field_no']==1):
                            maze_list[clust] = [maze, ]
                        elif repeat_filter == 'highest_mean':
                            maze_list[clust] = [maze, ]
                            highest_mean[clust] = rate
                    else:
                        if (repeat_filter is None) and (field['field_no']==1):
                            maze_list[clust].append( maze )
                        elif repeat_filter == 'highest_mean':
                            if rate > highest_mean[clust]:
                                highest_mean[clust] = rate
                                maze_list[clust] = [maze, ]

            clust_list = []
            for key in maze_list.keys():
                for maze in maze_list[key]:
                    rat, day, tc = key
                    clust_list.append( (rat,day,maze,tc) )

            grids = self.create_ratemap_grids(panel, grid_type=grid_type)
            H = None
            for clust in clust_list:
                rat, day, maze, tc = clust
                rds = (rat, day, maze)
                session = get_group(rds=rds)
                spikes = session._v_children[tc]

                H_rate, mask = get_ratemap( \
                    session, spikes, rds, grids=grids, **traits)
                H_rate = smooth_ratemap(H_rate, mask)
                H_rate[mask] = 0
                if norm and H_rate.max()>0: # field might be out of sample window
                    H_rate = H_rate / float(H_rate.max())

                if 'leather' in panel:
                    H = H_rate.copy() if H is None else np.vstack( (H,H_rate) )
                else:
                    H = H_rate.copy() if H is None else np.hstack( (H,H_rate) )

            self.results[panel] = H if 'leather' in panel else H.transpose()

            rats = list( set([data[0] for data in clust_list]) )
            days = list( set([data[0:2] for data in clust_list]) )
            mazes = list( set([data[0:3] for data in clust_list]) )
            clusts = list( set([data[0:4] for data in clust_list]) )
            self.out( '%d rats, %d days, %d mazes, %d cells collected' \
                %(len(rats), len(days), len(mazes), len(clusts)) )

        for value in self.results.values():
            print value.shape

        tdb_file.close()
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, label=True, grid_type='ortho'):
        plt_cnt = 1

        for platform in ['leather', 'tape']:
            for stype in ['STD', 'SHF']:
                panel = '%s_%s' % (platform,stype)
                H = self.results[panel]
                C = np.corrcoef(H, rowvar=0)
                print C.shape

                plt.subplot(2,2,plt_cnt)
                plt.imshow( C,interpolation='none',aspect=1.0,origin='lower' )
                if grid_type == 'ortho':
                    bd = 15 if stype=='STD' else 25
                else:
                    bd = 12 if stype=='STD' else 22
                if label:
                    plt.axvline(bd,color='white',ls='--',lw=2)
                    plt.axhline(bd,color='white',ls='--',lw=2)
                plt.colorbar()
                plt.title('%s' % panel)
                plt.tick_params(axis='both',labelsize=7)

                plt_cnt += 1

        plt.show()

        self.out('Done')
    # end of METHOD process_data

    def debug(self, rds, tc, **traits):
        session = get_group(rds=rds)
        spikes = session._v_children[tc]
        attrs = session._v_attrs
        panel = '%s_%s' % (attrs['platform'],attrs['type'])
        grids = self.create_ratemap_grids(panel)

        nrow = len(grids[3]) - 1
        ncol = len(grids[1]) - 1

        H_rate, mask = get_ratemap(session, spikes, rds, grids=grids, **traits)
        H_rate = smooth_ratemap(H_rate, mask)
        H_rate[mask] = 0

        im = masked_array_to_rgba(H_rate,mask=mask,mask_color='k',cmap='jet')
        plt.imshow( im,origin='lower',interpolation='nearest',aspect=1, \
            extent=[0,ncol,0,nrow] )

        plt.show()
    # end of METHOD debug

    #------------------------------------------------------------------------------

    def create_ratemap_grids(self, panel, grid_type='ortho'):
        # Defined by the equation of the boundaries

        if grid_type == 'ortho':
            consts = tuple( np.linspace(-170,166,43) )
            x_consts = consts[:]
            y_consts = consts[:]

            if 'leather' in panel:
                x_coef = (-4, 1) # -4x + y = k
                y_coef = (1, 4)  # x + 4y = k
            elif 'tape' in panel:
                x_coef = (4, 1)  # 4x + y = k
                y_coef = (-1, 4) # -x + 4y = k
        else:
            if 'leather' in panel:
                x_consts = tuple( np.linspace(-146,150,38) )
                if grid_type == 'slant':
                    y_consts = tuple( np.linspace(-50,50,51) )
                elif grid_type == 'slant_rim':
                    y_consts = tuple( np.linspace(-40,40,41) )
                x_coef = (-4, 1) # -4x + y = k
                y_coef = (0, 1)  # y = k
            elif 'tape' in panel:
                if grid_type == 'slant':
                    x_consts = tuple( np.linspace(-50,50,51) )
                elif grid_type == 'slant_rim':
                    x_consts = tuple( np.linspace(-40,40,41) )
                y_consts = tuple( np.linspace(-146,150,38) )
                x_coef = (1, 0)  # x = k
                y_coef = (-1, 4) # -x + 4y = k

        return (x_coef, x_consts, y_coef, y_consts) 
    # end of METHOD create_ratemap_grids
# end of CLASS SLCorrAna

#==================================================================================

class CollapsedMapAna(AbstractAnalysis):
    datadir = './CollapsedMap'
    label = 'collapsed map ana'

    def collect_data(self, min_quality='Fair', **traits):
        self.results = {}

        platforms = [ 'leather', 'tape' ]
        stypes = [ 'STD', 'SHF' ]
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        for panel in panels:
            self.out('Collecting data for %s...' % panel)

            clust_list = highest_mean_filter( \
                panel=panel, min_quality=min_quality, col=('threshold',float) )
            grids = self.create_ratemap_grids(panel)
            xbins = grids[1] if 'leather' in panel else grids[3]
            xbins = np.array( xbins )
            centers = (xbins[:-1]+xbins[1:]) / 2.0

            R = None
            for clust in clust_list:
                rat, day, maze, tc, threshold = clust
                rds = (rat, day, maze)
                session = get_group(rds=rds)
                spikes = session._v_children[tc]

                H_rate, mask = get_ratemap( \
                    session, spikes, rds, grids=grids, **traits)
                H_rate = smooth_ratemap(H_rate, mask)
                H_rate[mask] = np.nan

                field_num, labels, contours, n_edges, threshold = \
                    label_place_fields( rds, session, spikes, turn_off_rim=False, \
                        threshold=threshold, grids=grids, **traits )
                for i in range(field_num):
                    H = H_rate.copy()
                    H[ labels!=(i+1) ] = np.nan
                    axis = 0 if 'leather' in panel else 1
                    ratemap = np.nanmean( H, axis=axis )
                    if R is None:
                        R = ratemap.copy()
                    else:
                        R = np.vstack( (R,ratemap) )

            self.results[panel] = (R, centers)

            rats = list( set([data[0] for data in clust_list]) )
            days = list( set([ tuple([d for d in data][0:2]) \
                for data in clust_list]) )
            mazes = list( set([ tuple([d for d in data][0:3]) \
                for data in clust_list]) )
            clusts = list( set([ tuple([d for d in data][0:4]) \
                for data in clust_list]) )
            self.out( '%d rats, %d days, %d mazes, %d cells collected' \
                %(len(rats), len(days), len(mazes), len(clusts)) )

        self.results['panels'] = panels

        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, norm=True, intersect_filter=True):
        for i, panel in enumerate(self.results['panels']):
            R, x = self.results[panel]
            R_f = None

            bd_c = -50 if 'STD' in panel else 30
            bd_x = (bd_c-x[0]) / float(x[1]-x[0]) + 1
            print bd_x

            plt.subplot(2, 2, i+1)
            for j in range( R.shape[0] ):
                ratemap = R[j]
                if intersect_filter:
                    if np.isnan(ratemap[np.floor(bd_x-1)]) and np.isnan(ratemap[np.ceil(bd_x-1)]): 
                        continue
                if norm:
                    r = ratemap[ ~np.isnan(ratemap) ]
                    ratemap = (ratemap-r.min())/float(r.max()-r.min())
                R_f = ratemap.copy() if R_f is None else \
                    np.vstack( [R_f,ratemap] )
                plt.plot( ratemap,'k',lw=1, alpha=0.4 )

            avg = np.nanmean( R_f, axis=0 )
            plt.plot( avg,'k',lw=2 )

            plt.axvline(bd_x,color='r',lw=2)
            plt.xlim( [1, len(x)] )
            if norm:
                plt.ylim( [0, 1.1] )
            plt.title('%s' % panel)
            plt.tick_params(axis='both',labelsize=7)

        plt.show()

        self.out('Done')
    # end of METHOD process_data

    #------------------------------------------------------------------------------

    def create_ratemap_grids(self, panel):
        consts = tuple( np.linspace(-250,250,61) )

        if 'leather' in panel:
            x_coef = (-4, 1) # -4x + y = k
            y_coef = (1, 4)  # x + 4y = k
        elif 'tape' in panel:
            x_coef = (4, 1)  # 4x + y = k
            y_coef = (-1, 4) # -x + 4y = k

        return (x_coef, consts, y_coef, consts) 
    # end of METHOD create_ratemap_grids

# end of CLASS SLCorrAna
