# encoding: utf-8
"""
field.py
"""

import tables as tb
import numpy as np
import warnings
import matplotlib.pyplot as plt

from itertools import product
from scipy import ndimage
from scipy.ndimage.filters import gaussian_filter1d
from scipy.optimize import minimize_scalar

from core.analysis import AbstractAnalysis
from core.report import BaseReport
from tools.images import masked_array_to_rgba
from tools.plot import quicktitle
from tools.misc import memoize

from scanr.cluster import string_to_quality

from tdb.config import Config
from tdb.meta import walk_mazes, stamp_to_time
from tdb.data import get_tdb_file, new_table, get_group, get_node, get_unique_row
from tdb.spike import get_session_clusters, parse_cell_name
from tdb.td_maps import bin_edges, plot_boundaries
from tdb.ratemap import get_ratemap, smooth_ratemap, get_session_occupancy, get_spike_heat_map, get_session_interpolant, get_occ_rim, get_spike_position
from tdb.plot_rate_map import plot_traj_and_spike, plot_ratemap

# parameters selected by Shan on 2016/7/28
FieldCfg = Config['placefield']
BIN_SIZE = FieldCfg['bin_size'] # width of the spatial bin, 2.0 cm 
SKAGGS_THRD = FieldCfg['skaggs_thrd'] # minimum Skaggs information score, 0.6
SKAGGS_P = FieldCfg['skaggs_p'] # Skaggs score significance level, 0.01
MIN_SIZE = FieldCfg['min_size'] # minimum field size, 35 bin
MIN_RATE = FieldCfg['min_rate'] # minimum peak firing rate of field, 0.1 Hz
MAX_RATE = FieldCfg['max_rate'] # maximum averag rate of pyramidal cell, 10 Hz
MIN_SPIKE = FieldCfg['min_spike'] # minimum spike count in field, 30 spikes
MIN_SESSION_SPIKE = FieldCfg['min_session_spike'] # minimum spike count per session, 50 spikes

# ================================================================================
#BIN_SIZE = 2.0 
#OCC_THRD = 0.06
#SKAGGS_THRD = 0.6
#SKAGGS_P = 0.01
##   MIN_SIZE = 20
#MIN_SIZE = 35
#MIN_RATE = 0.1
#MAX_RATE = 10
##   MIN_SPIKE = 10
#MIN_SPIKE = 30
#MIN_SESSION_SPIKE = 50
# ================================================================================

# ================================================================================
# Create place field table
# ================================================================================

FieldColDescr = { 'rat'      : tb.UInt16Col(pos=1),
                  'day'      : tb.UInt8Col(pos=2),
                  'session'  : tb.UInt8Col(pos=3),
                  'tt'       : tb.UInt8Col(pos=4),
                  'cl'       : tb.UInt8Col(pos=5),
                  'area'     : tb.StringCol(itemsize=16,pos=6),
                  'quality'  : tb.StringCol(itemsize=16,pos=7),
                  'N_run'    : tb.UInt64Col(pos=8), # running spike number
                  'Skaggs'   : tb.FloatCol(pos=9),
                  'p_value'  : tb.FloatCol(pos=10),
                  'field_no' : tb.Int8Col(pos=11),
                  'mean_rate': tb.FloatCol(pos=12),
                  'max_rate' : tb.FloatCol(pos=13),
                  'CoM_x'    : tb.FloatCol(pos=14),
                  'CoM_y'    : tb.FloatCol(pos=15),
                  'size'     : tb.FloatCol(pos=16), # in cm^2
                  'perimeter': tb.FloatCol(pos=17), # in cm
                  'rate'     : tb.FloatCol(pos=18), # average session firing rate
                  'N_in'     : tb.UInt64Col(pos=19), # in field running spike number
                  'threshold': tb.FloatCol(pos=20), # do not rerun fitting
                  'exclude'  : tb.UInt8Col(pos=21) } # for debug

class FieldTable(AbstractAnalysis):
    """
    Detect place fields for each session-cluster and create the field table
    Depend on the Skaggs table, created by SkaggsInfoTable
    """
    datadir = './FieldTable'
    label = 'create field table'

    def collect_data(self, table='fields', tune_params=False,**traits):

        tdb_file = get_tdb_file() 
        skaggs_table = tdb_file.root.spatial.skaggs
        field_table = new_table( \
            '/spatial',table,FieldColDescr,title='Place Fields')
        rec = field_table.row

        # Turn on 'break_cnt' when tuning parameters
        #break_cnt = 0
        for rds in walk_mazes():
            rat, day, maze = rds

            dgroup = get_group(rat=rat, day=day)
            areas = dgroup.tt_area[:]

            [y_edges, x_edges] = bin_edges(rds, **traits)
            x_centers = (x_edges[:-1]+x_edges[1:]) / 2.0
            y_centers = (y_edges[:-1]+y_edges[1:]) / 2.0
            Hx, Hy = np.meshgrid( x_centers, y_centers )

            session = get_group(rds=rds)
            start, end = session._v_attrs['start'], session._v_attrs['end']
            ts = session._v_children['t_pos']
            sec = stamp_to_time(ts, zero_stamp=ts[0]) 

            H, valid = get_session_occupancy(session,**traits)
            H_occ = H.copy()

            tc_list = get_session_clusters(session)
            for tt, cl in tc_list:
                exclude = 0

                self.out( \
                  'Finding place fields for rat %d, day %d, session %d, t%dc%d'\
                    % (rat,day,maze,tt,cl) )
                
                query = '(rat==%d)&(day==%d)&(session==%d)&(tt==%d)&(cl==%d)' \
                    % (rat, day, maze, tt, cl)
                skaggs_data = get_unique_row(skaggs_table,query)
                if skaggs_data['N_run'] < MIN_SESSION_SPIKE:
                    self.out( 'Not enough running spikes in session (N_run=%d)' \
                        % skaggs_data['N_run'] )
                    if not tune_params:
                        continue
                    exclude = 1
                #if skaggs_data['rate'] > MAX_RATE: # putative interneuron
                clust = session._v_children['t%dc%d'%(tt,cl)]
                lc = 20*clust._v_attrs['maxwidth'] - 0.6*skaggs_data['rate'] + 1
                if lc < 0: # putative interneuron
                    self.out( 'Average firing rate too high (r=%f)' \
                        % skaggs_data['rate'] )
                    if not tune_params:
                        continue
                    exclude = 2
                if (skaggs_data['Skaggs']<SKAGGS_THRD) or \
                   (skaggs_data['p_value']>SKAGGS_P):
                    self.out( 'Does not pass Skaggs SI test (SI=%.2f, p=%.3f)' \
                        % (skaggs_data['Skaggs'], skaggs_data['p_value']) )
                    if not tune_params:
                        continue
                    exclude = 3 

                spikes = session._v_children['t%dc%d' % (tt,cl)]

                H_spk = get_spike_heat_map(rds, session, spikes, **traits)
                H_rate, mask = get_ratemap(session, spikes, rds, **traits)
                H_rate = smooth_ratemap( H_rate, mask, smooth='gaussian' )
                #H_rate = smooth_ratemap( H_rate, mask )
                if H_rate.max() < MIN_RATE:
                    self.out( 'Firing rate too low (max rate=%f)' % H_rate.max() )
                    if not tune_params:
                        continue
                    exclude = 4 

                if exclude > 0: # when tuning params
                    field_num = 0
                    threshold = -1 
                else:
                    #field_num, labels, contours, n_edges, threshold  = \
                    #    label_place_fields( rds,session,spikes,tune=tune_params, \
                    #        smooth=True,**traits)
                    field_num, labels, contours, n_edges, threshold  = \
                        label_place_fields( rds,session,spikes,tune=tune_params, \
                            smooth=True,smooth_kernel='gaussian',**traits)

                fields = []
                for i in range(field_num):
                    selected = labels == (i+1)
                    in_field = selected * ~mask
                    f_size = np.sum(selected)
                    perimeter = np.sum(n_edges[contours==i+1]) * BIN_SIZE
                    N_in = np.sum(H_spk[in_field])

                    H_field = H_rate.copy()
                    H_field[labels!=(i+1)] = 0
                    H_occ[mask] = 0
                    mean_rate = float(N_in) / np.sum(H_occ[in_field])
                    max_rate = H_field.max()
                    CoM_x = float( np.sum(Hx*H_field) ) / np.sum(H_field)
                    CoM_y = float( np.sum(Hy*H_field) ) / np.sum(H_field)
                    #CoM = np.unravel_index(H_field.argmax(), H_field.shape)

                    if tune_params:
                        # check field size
                        if f_size < MIN_SIZE:
                            self.out( 'Field too small (size=%d)' % f_size )
                            exclude = 5 
                        # check spike number
                        if N_in < MIN_SPIKE:
                            self.out( 'Not enough spikes in field (N=%d)' % N_in )
                            exclude = 6 
                        # check maximum firing rate
                        if max_rate < MIN_RATE:
                            self.out( 'Firing rate too low (max rate=%f)'%max_rate )
                            exclude = 4 

                    fields.append( [ i+1, mean_rate, max_rate, CoM_x, CoM_y, \
                        f_size*BIN_SIZE*BIN_SIZE, perimeter, N_in, exclude ] )
                    exclude = 0

                self.out('%d potential fields found' % field_num)
                if field_num == 0:
                    if not tune_params:
                        continue
                    fields.append( [-1,-1,-1,-1,-1,-1,-1,-1,exclude] )

                for field in fields:
                    rec['rat'] = rat
                    rec['day'] = day
                    rec['session'] = maze
                    rec['tt'] = tt
                    rec['cl'] = cl
                    rec['area'] = areas[tt]
                    rec['quality'] = skaggs_data['quality'] 
                    rec['N_run'] = skaggs_data['N_run']
                    rec['Skaggs'] = skaggs_data['Skaggs'] 
                    rec['p_value'] = skaggs_data['p_value'] 
                    rec['field_no'] = field[0] 
                    rec['mean_rate'] = field[1]
                    rec['max_rate'] = field[2]
                    rec['CoM_x'] = field[3] 
                    rec['CoM_y'] = field[4]
                    rec['size'] = field[5]
                    rec['perimeter'] = field[6]
                    rec['rate'] = skaggs_data['rate']
                    rec['N_in'] = field[7]
                    rec['threshold'] = threshold
                    rec['exclude'] = field[8]
                    rec.append()

                field_table.flush()

                #break_cnt += 1
                #if break_cnt > 10:
                #    tdb_file.close()
                #    return

        # end of FOR LOOP rds

        tdb_file.close()
    # end of METHOD collect_data

# end of CLASS FieldTable


# ================================================================================
# Check place field detection
# ================================================================================


class ChkFieldDetectReport(BaseReport):
    """
    For each cluster, plot the trajectory and spike map, smoothed rate map, filled
    place fields and place field contours and combined to PDF
    """

    datadir = './ChkFieldDetect'
    label = 'check field detection'

    ncols = 4
    nrows = 5
    figwidth = 8.0
    figheight = 11.0

    def collect_data(self,min_quality=None,area=None,with_field_only=False,table='fields'):
        """
        Collect session and cluster list
        """
        self.results = {}

        if min_quality is not None:
            quality = string_to_quality(min_quality)
        else:
            quality = string_to_quality('Poor')

        plt_types = [ 'Trajectory and Spike', 'Heat Map', \
                      'Field Map', 'Field Contour' ]

        data_list = []
        if with_field_only:
            field_table = get_node('/spatial',table)
            query = '(field_no==1)'
            if area is not None:
                query += '&(area=="%s")' % area

            clust_list = []
            for field in field_table.where(query):
                if string_to_quality(field['quality']) < quality:
                    continue
                rds = (field['rat'], field['day'], field['session'])
                cl_name = 't%dc%d' % (field['tt'], field['cl']) 
                clust_list.append( (rds,cl_name) )
            data_list = [ x+(y,) for x, y in product(clust_list,plt_types) ]
            field_table.close()
        else:
            for rds in walk_mazes():
                rat, day, maze = rds
                session = get_group(rds=rds)
                tc_list = get_session_clusters(session)
                for tt, cl in tc_list:
                    cl_name = 't%dc%d' % (tt, cl)
                    if area is not None:
                        tt_area = get_group(rat=rat,day=day).tt_area
                        if tt_area[tt] != area:
                            continue
                    if min_quality is not None:
                        clust = session._v_children[cl_name]
                        clust_quality = string_to_quality(clust._v_attrs['quality'])
                        if clust_quality < quality:
                            continue
                    for plt_type in plt_types:
                        data_list.append( (rds,cl_name,plt_type) )

        self.results['plt_types'] = plt_types
        self.results['data_list'] = data_list
        self.results['table'] = table
    # end of METHOD collect_data

    def process_data(self,rotate_bd=False,plot_bd=False,smooth=True,turn_off_rim=False,scale_plain=False,**traits):
        """
        Create the PDF
        """
        data_list = self.results['data_list']
        plt_types = self.results['plt_types']

        skaggs_table = get_node('/spatial','skaggs')
        field_table = get_node('/spatial',self.results['table'],raise_on_fail=False)

        def plot_data(rds, tc, ax, plt_type):
            session = get_group(rds=rds)
            fs = session._v_attrs['tracking_sample_rate']
            spikes = session._v_children[tc]
            [y_edges, x_edges] = bin_edges(rds,scale_plain=scale_plain,**traits)
            y_lim = [y_edges[0], y_edges[-1]]
            x_lim = [x_edges[0], x_edges[-1]]

            tt, cl = parse_cell_name(tc)
            query = '(rat==%d)&(day==%d)&(session==%d)&(tt==%d)&(cl==%d)' \
                % (rds+(tt,cl))
            rec = get_unique_row(skaggs_table,query)
            #idx = skaggs_table.getWhereList(query)
            #rec = skaggs_table[idx[0]]
            I, p = rec['Skaggs'], rec['p_value']
            if (I<SKAGGS_THRD) or (p>SKAGGS_P):
                mask_c = 'm'
            else:
                mask_c = 'b'

            if plt_type == 'Trajectory and Spike':
                x, y, ts = [ session._v_children[key] for key in ['x','y','t_pos'] ]
                attrs = session._v_attrs
                if scale_plain:
                    if attrs['platform']=='square' and attrs['type']=='PLA':
                        x = np.array(x) / 110.0 * 100
                        y = np.array(y) / 110.0 * 100
                #sec = stamp_to_time(ts, zero_stamp=ts[0])
                #spike_sec = stamp_to_time(spikes, zero_stamp=ts[0])
                #x_spk = get_session_interpolant(session,'x')(spike_sec)
                #y_spk = get_session_interpolant(session,'y')(spike_sec)
                x_spk, y_spk = get_spike_position( \
                    session, spikes, min_velocity=0, scale_plain=scale_plain )
                plot_traj_and_spike(ax,rds,x,y,x_spk,y_spk,**traits)
            elif plt_type == 'Heat Map':
                plot_ratemap( ax,session,spikes,rds,outlier=False,\
                    scale_plain=scale_plain,**traits )
            else:
                if field_table is None: # field table does not exist
                   thrd = None
                else:
                    rec = get_unique_row( field_table,query+'&(field_no==1)', \
                        raise_on_fail=False )
                    thrd = None if rec is None else rec['threshold']
                #field_num, labels, contours, n_edges, threshold = \
                #    label_place_fields( rds,session,spikes,threshold=thrd, \
                #        turn_off_rim=turn_off_rim,smooth=smooth, \
                #            scale_plain=scale_plain,**traits )
                field_num, labels, contours, n_edges, threshold = \
                    label_place_fields( rds,session,spikes,threshold=thrd, \
                        turn_off_rim=turn_off_rim,smooth=smooth, \
                            smooth_kernel='gaussian', \
                                scale_plain=scale_plain,**traits )
                H_rate, mask = get_ratemap( \
                    session,spikes,rds,scale_plain=scale_plain,**traits )
                if plt_type == 'Field Map':
                    img = masked_array_to_rgba(labels,mask=mask, \
                        mask_color=mask_c,cmap='cubehelix_r')
                    ax.imshow(img,origin='lower',interpolation='none', \
                      aspect='auto',\
                        extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )
                elif plt_type == 'Field Contour':
                    img = masked_array_to_rgba(contours,mask=mask, \
                        mask_color=mask_c,cmap='cubehelix_r')
                    ax.imshow(img,origin='lower',interpolation='none', \
                      aspect='auto',\
                        extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )

            if plot_bd:
                plot_boundaries( *rds,rotate=rotate_bd, \
                    scale_plain=scale_plain,**{'c':'b','alpha':0.5} )

            ax.set( xlim=x_lim, ylim=y_lim )
                
        # end of METHOD plot_data

        old_rds = None

        for item, ax in self.get_plot(data_list):
            rds, tc, plt_type = item
            if old_rds != rds:
                self.out('Generating plots for rat %d day %d session %d...' % rds)
                old_rds = rds
            self.out('%s - %s' % (tc,plt_type))
            plot_data(rds,tc,ax,plt_type)
            if self.col == 0:
                quicktitle( ax,'(%d, %d, %d) %s' % \
                    (rds+(tc,)),size='xx-small' )
            elif self.firstrow:
                ax.set_title(plt_type,size='small')
            plt.tick_params(axis='both',labelsize=7)

        skaggs_table.close()
    # end of METHOD process_data

# end of CLASS ChkFieldDetectReport

@memoize
def label_place_fields(rds,session,spikes,tune=False,smooth=True,threshold=None,turn_off_rim=False,smooth_kernel='kuwahara',**traits):
    H_rate, mask = get_ratemap(session, spikes, rds, **traits)
    H_rate = smooth_ratemap( H_rate, mask, smooth=smooth_kernel )
    if turn_off_rim:
        rim = get_occ_rim(session, **traits)
        H_rate[rim] = 0
    H_mask = H_rate.copy()
    H_mask[mask] = 0
    if not tune:
        H_spk = get_spike_heat_map(rds, session, spikes, **traits)

    if threshold is None:
        threshold = define_threshold(H_mask, method='baseline', \
            alpha=1.6, beta=0.7, gamma=1.0 ) # gaussian 
        #threshold = define_threshold(H_mask, method='baseline', \
        #    alpha=1.3, beta=1.0, gamma=1.0 ) # kuwahara
        # check standard field cut-off criteria
        #threshold = define_threshold( H_mask, method='max_rate', alpha=0.10 )

    if np.isnan(threshold):
        return 0, 0*H_rate, 0*H_rate, 0*H_rate, 0
    H = H_rate > threshold 
    labels, num = ndimage.label(H)
    field_num = num + 1
    
    # smooth field contours if smoothing enabled
    if smooth and num>0:
        field_num = 1 
        new_labels = np.zeros( labels.shape )
        for i in range(num):
            selected = labels == (i+1)
            selected = ndimage.binary_opening( \
                selected,structure=np.ones((3,3))).astype(bool)
            selected = ndimage.binary_closing( \
                selected,structure=np.ones((3,3))).astype(bool)
            kernel = np.ones((4,4))
            kernel[0,0], kernel[-1,-1], kernel[0,-1], kernel[-1,0] = 0, 0, 0, 0
            selected = ndimage.binary_opening( \
                selected,structure=kernel).astype(bool)
            smoothed, n = ndimage.label( selected )
            for j in range(n):
                new_labels[ smoothed==(j+1) ] = field_num
                field_num += 1
        labels = new_labels.copy()
        num = field_num - 1

    # filtering potential place fields, do nothing when tuning parameters
    if not tune:
        field_num = 1 
        for i in range(num):
            selected = labels == (i+1)
            in_field = selected * (~mask)
            if (np.sum(selected)<MIN_SIZE) or (np.sum(H_spk[in_field])<=MIN_SPIKE) \
               or (np.max(H_rate[in_field])<MIN_RATE):
                labels[selected] = 0
            else:
                labels[selected] = field_num
                field_num += 1

    contours = labels.copy()
    h, w = contours.shape
    c = (labels > 0).astype('d')
    n_edges = np.zeros( (h,w) )
    for i in range(h):
        for j in range(w):
            if (i==0) or (i==h-1) or (j==0) or (j==w-1):
                continue
            sur = c[i-1][j] + c[i+1][j] + c[i][j-1] + c[i][j+1]
            n_edges[i][j] = 4 - sur
            if sur == 4: # not boundary
                contours[i][j] = 0

    return field_num-1, labels, contours, n_edges, threshold
# end of FUNCTION label_place_fields

def define_threshold(H, method='half_rate', alpha=None, beta=None, gamma=None):
    def error_func(x,a,b,c):
        lower, upper = v[(v>0)*(v<=x)], v[v>x]
        if (len(lower)==0) or (len(upper)==0):
            return np.inf
        lower_var = np.var(lower)
        lower_area_ratio = float(len(lower)) / np.sum(v>0)
        lower_mean = np.mean(lower)
        upper_mean = np.mean(upper)
        return 10.0*lower_var - a*lower_area_ratio + b*lower_mean - 0.1*c*upper_mean

    if H.max() == 0:
        return np.nan 

    if method == 'half_rate':
        half_rate = np.percentile(H[H>0],50)
        threshold = half_rate  * alpha
    elif method == 'max_rate':
        threshold = H.max() * alpha
    elif method == 'std':
        threshold = max(0, H.max()-alpha*np.std(H[H>0]) )
    elif method == 'baseline':
        v = H.flatten()
        index = v > 0
        if np.sum(index) <= 1:
            return v.max() # no bin pass threshold 
        scale, shift = np.std(v[index]), np.mean(v[index])
        if scale == 0:
            scale = 1
        # check if needs to truncate outliers
        Q1, Q3 = np.percentile(v[index],25), np.percentile(v[index],75)
        outlier_min = Q1 - (Q3-Q1)*1.5
        outlier_max = Q3 + (Q3-Q1)*1.5
        v[ v>outlier_max ] = outlier_max 
        v[ (v>0)*(v<outlier_min)] = outlier_min

        v[index] = (v[index]-shift) / float(scale) + 5.0
        v_min = v[index].min()
        threshold = minimize_scalar( error_func,method='bounded', \
            bounds=(v_min,v.max()),args=(alpha,beta,gamma), \
                options={'disp':True,'xatol':1e-07}).x
        if error_func(threshold,alpha,beta,gamma) > error_func(0.99*v_min,alpha,beta,gamma):
            threshold = 0 
        else:
            threshold = (threshold-5) * scale + shift

    return threshold
# end of FUNCTION define_threshold

# ================================================================================
# Extract and smooth field contour 
# ================================================================================

# smooth contour with gaussian filtering, not used
#if smooth:
#    C = np.zeros( contours.shape )
#    L = np.zeros( labels.shape )
#    for i in range(1,field_num):             
#        contour_tuple = tuple( tuple(row) for row in contours==i )
#        label_tuple = tuple( tuple(row) for row in labels==i )
#        poly = field_polygon( contour_tuple, label_tuple, k)
#        C_tmp = np.zeros( contours.shape )
#        for x, y in poly:
#            C_tmp[y][x] = i
#        C += C_tmp
#        L += binary_fill_holes( C_tmp ) 
#    contours = C.copy()
#    labels = L.copy()

@memoize
def field_polygon(C, F, k):
    """
    Polygonize and smooth place field contour

    C: contour map in tuple
    F: field map in tuple
    k: smooth level
    """

    H_c = np.array(C).astype(bool)
    H = np.array(F).astype(bool)
    [y_edges, x_edges] = bin_edges( pattern='deStijl' )
    m, n = len(x_edges), len(y_edges)
    directions = [ (1,0), (0,1), (-1,0), (0,-1) ] # in bins

    # starting intersect point (search starts from bottom left)
    sy, sx = next( iter( \
        [ (i,list(row).index(True)) for i,row in enumerate(H_c) if True in row ]))

    def walk(cur_x, cur_y, del_x, del_y):
        if (cur_x,cur_y,cur_x+del_x,cur_y+del_y) in visited:
            return False

        x1, x2, y1, y2 = cur_x, cur_x, cur_y, cur_y
        if del_y == 0: # walk horizontally
            y1 = cur_y-1
            if del_x < 0:
                x1, x2 = cur_x-1, cur_x-1
        elif del_x == 0: # walk vertically
            x1 = cur_x-1
            if del_y < 0:
                y1, y2 = cur_y-1, cur_y-1

        if (x1<0) or (y1<0) or (x2==m) or (y2==n): # out of boundary
            return True
        return H[y1][x1]!=H[y2][x2] # 1 True 1 False?

    poly = [ (x_edges[sx],y_edges[sy]), ]
    visited = []
    ex, ey = sx, sy
    legal_moves = directions[:]
    while(True):
        for dx, dy in legal_moves:
            if walk(ex, ey, dx, dy):
                visited.append( (ex,ey,ex+dx,ey+dy) )
                visited.append( (ex+dx,ey+dy,ex,ey) )
                ex, ey = ex+dx, ey+dy
                no_going_back = (-dx,-dy)
                break
        poly.append( (x_edges[ex],y_edges[ey]) )
        if (sx==ex) and (sy==ey):
            break
        legal_moves = directions[:]
        legal_moves.remove( no_going_back )

    #poly = Chaikin_smooth(poly,k)
    poly = gaussian_smooth(poly,k)

    return poly
# end of FUNCTION field_polygon

def Chaikin_smooth( pre, K ):
    post = []
    for seq in zip(*pre): # loop through x and y sequences
        cur = seq[:]
        k = K
        while k > 0:
            A, B = cur[:-1], cur[1:]
            cur = [ (3*m+n)/4.0 for m, n in zip( \
                [u for v in zip(A,B) for u in v], \
                [u for v in zip(B,A) for u in v]) ]
            k -= 1
        post.append( cur )
    post = zip(*post)

    return post
# end of FUNCTION Chaikin_smooth

def gaussian_smooth( pre, K ):
    post = []
    for seq in zip(*pre):
        cur = gaussian_filter1d(seq,K,mode='wrap')
        post.append(cur)
    post = zip(*post)

    return post
# end of FUNCTION gaussian_smooth
