# encoding: utf-8
"""
field_alignment.py
"""

import warnings
import copy
import timeit
import pdb
import itertools
import random
import os
import sys
import tables as tb
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import cPickle as pickle

from scipy.ndimage.filters import gaussian_filter, gaussian_filter1d
from scipy.stats import chisquare, chi2_contingency
from cv2 import bilateralFilter
from itertools import product, chain, combinations
from PIL import Image, ImageDraw
from datetime import datetime
from shapely.geometry import MultiLineString, Polygon

from core.analysis import AbstractAnalysis
from core.report import BaseReport
from tools.plot import quicktitle
from tools.images import masked_array_to_rgba
from tools.misc import memoize

from tdb.config import Config
from tdb.meta import walk_rats, walk_mazes, get_day_list, get_maze_list, stamp_to_time
from tdb.data import get_tdb_file, new_table, get_group, get_node, get_unique_row
from tdb.spike import get_session_clusters
from tdb.spatial_info import get_spike_heat_map, get_session_occupancy, label_place_fields, bin_edges, compute_velocity, get_spike_position
from tdb.tracking import plot_boundaries
from tdb.surface import DeStijl
from tdb.field import field_polygon

BIN_SIZE = 1.5 # in cm, 2.2/cm^2 square bins
MIN_VELOCITY = 10.0 # 10.0 cm/s
FT_BIN_SIZE = 0.1 # in cm
BD_NUM = 7 # number of texture boundaries
PATCH_NUM = 8 # number of texture patches

QUERY = '(area=="CA1")&(field_no==1)&(quality!="Poor")&(Skaggs>=0.6)&(p_value<=0.01)'

# ================================================================================
# shared
# ================================================================================

class AlignmentAnalysis(AbstractAnalysis):

    def collect_data(self,dtype,batch):
        """
        Return the cluster list [ (rat,day,session,tc), ... ]
        """
        if batch == 'all':
            return self._collect_all_sessions(dtype)
        elif batch == 'first': 
            return self._collect_first_batch(dtype)
        else:
            raise ValueError('Incorrect batch content')
    # end of METHOD collect_data()

    def _collect_first_batch(self, dtype):
        """
        Called by collect_data

        Specify the test (control) sessions before any control (test) session and
        return the cluster list
        """
        field_table = get_node('/spatial','fields')

        if dtype == 'test':
            rds_list = [ (539,6,1), (539,6,2),
                         (576,4,1), (576,4,2), 
                         (577,7,1),
                         (579,28,1), (579,28,2),
                         (580,21,1), (580,21,2) ]
        elif dtype == 'control':
            rds_list = [ (637,5,1), (637,5,2),
                         (740,3,1), (740,3,2) ]

        # Collect cluster list
        self.out('Collecting cluster list...')
        cluster_list = []
        for rds in rds_list:
            query = '(rat==%d)&(day==%d)&(session==%d)&' % rds + QUERY
            for rec in field_table.where(query):
                tc = 't%dc%d' % (rec['tt'], rec['cl'])
                cluster_list.append( rds+(tc,) )

        return cluster_list

    # end of METHOD _collect_first_batch()

    def _collect_all_sessions(self, dtype):
        """
        Called by collect_data

        Return the cluster list containing all sessions
        """
        field_table = get_node('/spatial','fields')

        if dtype == 'test':
            pattern = 'deStijl'
        elif dtype == 'control':
            pattern  = 'plain-black'

        # Collect cluster list
        self.out('Collecting cluster list...')
        cluster_list = []
        for rec in field_table.where(QUERY):
            rds = (rec['rat'], rec['day'], rec['session'])
            session = get_group(rds=rds)
            attrs = session._v_attrs
            if session._v_attrs['pattern'] != pattern:
                continue
            tc = 't%dc%d' % (rec['tt'], rec['cl'])
            cluster_list.append( rds+(tc,) )

        return cluster_list
    # end of METHOD collect_data

    def _init_output_folder(self, data_dir):
        """
        Create output folder if not already exist
        """
        output_dir = os.path.join( data_dir, 'h5_output' )
        if not os.path.exists( output_dir ):
            os.mkdir( output_dir )
        return output_dir

    def _init_display(self):
        """
        Initialize display environment
        """
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)
    # end of METHOD _init_display

# end of CLASS AlignmentAnalysis

# ================================================================================
# FillPatch 
# ================================================================================

class FillPatchAnalysis(AlignmentAnalysis):
    """
    Report the coverage index distribution.
    
    The coverage index is the highest coverage rate (the proportion of a subarea
    occupied by a field) among all subareas.  A subarea can be defined by a
    single floor patch or a group of connected patches.
    """
    datadir = './FillPatch'
    label = 'fill patch analysis'

    def collect_data(self,dtype='test',batch='all',debug=True):
        """
        Create and fill in the Coverage Table

        The Coverage Table records the coverage index and whether each floor patch
        belongs to the best subarea
        """

        clusters = super( FillPatchAnalysis, self ).collect_data( \
            dtype=dtype, batch=batch )

        self._init_table_cols() 
        self.output_dir = self._init_output_folder( self.datadir )
        self.tb_name = 'coverage'
        file_name = self._get_file_name(dtype,batch)

        with tb.openFile(os.path.join(self.output_dir,file_name),'w') as tb_file:
            data_table = tb_file.createTable( \
                '/',self.tb_name, self.Cols, title='Coverage')
            self._fill_in_table( clusters,data_table,debug=debug )
            data_table.flush()
        tb_file.close()

        self.results = {} # No pickled information

        self.out('Done')
    # end of METHOD collect_data

    def read_coverage(self,dtype,batch):
        file_name = _get_file_name(dtype,batch)
        with tb.openFile(os.path.join(self.output_dir,file_name),'r') as tb_file:
            data_table = get_node(tb_file.root,self.tb_name)
            coverage = data_table.col('coverage')
        tb_file.close()

        return coverage

    def plot_coverage_pdf(self,dtype,batch,thrd=0.7,hist_wd=0.05,rg=[0,1]):
        """
        Plot the distribution of the coverage index
        """
        super( FillPatchAnalysis, self )._init_display()

        bins = np.linspace(rg[0],rg[1],(rg[1]-rg[0])/hist_wd+1)
        #print bins

        coverage = self.read_coverage(dtype,batch)

        N = len(coverage)
        n = np.sum( coverage >= thrd )
        self.out('%d/%d fields have coverage index >= %.2f' % (n,N,thrd))

        f = plt.figure()
        plt.hist(coverage,bins=bins,range=rg)
        plt.title('Coverage Distribution')
        plt.tick_params(axis='both',labelsize=7)
        #plt.axis('tight')
        plt.axis([0,1,0,50])

        f = plt.gcf()
        f.savefig( self.datadir + '/coverage_distr.pdf')
    # end of METHOD plot_coverage_distr

    def plot_coverage_CDF(self,batch,thrd=0.7,hist_wd=0.01,rg=[0,1]):
        """
        Plot the distribution of the coverage index
        """
        super( FillPatchAnalysis, self )._init_display()

        bins = np.linspace(rg[0],rg[1],(rg[1]-rg[0])/hist_wd+1)
        #print bins

        dtypes = ['test', 'control']
        for dtype in dtypes:
            coverage = self.read_coverage(dtype,batch)
            counts, cdf_bins = np.histogram(coverage,bins=bins,normed=True)
            cdf = np.cumsum(counts)
            plt.plot(cdf_bins[1:],cdf)
        plt.legend( dtypes )
        plt.title('Coverage Index CDF')
        plt.tick_params(axis='both',labelsize=7)
        plt.axis('tight')

        f = plt.gcf()
        f.savefig( self.datadir + '/coverage_cdf.pdf')
    # end of METHOD plot_coverage_distr

    def _init_table_cols(self):
        patch_cols = dict( \
            [('patch_%d'%i, tb.BoolCol(pos=i+1)) for i in range(PATCH_NUM)] )
        start = PATCH_NUM + 1

        cols = { 'ID'        : tb.UInt64Col(pos=start),
                 'rat'       : tb.UInt16Col(pos=start+1),
                 'day'       : tb.UInt8Col(pos=start+2),
                 'session'   : tb.UInt8Col(pos=start+3),
                 'tc'        : tb.StringCol(itemsize=10,pos=start+4),
                 'field_no'  : tb.Int8Col(pos=start+5),
                 'coverage'  : tb.FloatCol(pos=start+7),
                 'area'      : tb.FloatCol(pos=start+8) }
        cols.update( patch_cols )

        self.Cols = cols 
    # end of METHOD _init_table_cols

    def _get_file_name(self,dtype,batch):
        return '%s_%s.h5' % (dtype,batch) 

    # should modify this to use shapely
    def _fill_in_table(self,clusters,data_table,debug=False):
        row = data_table.row

        self.out('Collecting data from each field...')

        id_cnt = 0
        for cluster in clusters:
            rat, day, maze, tc = cluster
            rds = (rat, day, maze)

            session = get_group(rds=rds)
            spikes = session._v_children[tc]
            field_num, labels, contours = label_place_fields(rds,session,spikes)
            
            for i in range(field_num):
                self.out( 'Rat %d, day %d, session %d, %s, field %d' \
                    % (rat,day,maze,tc,i+1) )

                # individual place field map (2D Boolean array)
                selected = labels == (i+1)

                total_field, coverage, patch_set = \
                    self._find_best_coverage(selected)

                row['ID'] = id_cnt
                row['rat'] = rat
                row['day'] = day
                row['session'] = maze
                row['tc'] = tc
                row['field_no'] = i + 1
                row['coverage'] = coverage
                row['area'] = total_field
                for j in range(PATCH_NUM):
                    row['patch_%d'%j] = j in patch_set
                row.append()
                data_table.flush()

                id_cnt += 1
        self.out('%d field(s) collected from %d clusters' % (id_cnt,len(clusters)))
    # end of METHOD _fill_in_table

    def _find_best_coverage(self,field_map,thrd=0.83):
        """
        Find the combination of floor patches with the highest coverage rate
        Also return the combination itself
        """
        def calc_coverage(idx):
            field_sum = np.sum( areas_field[idx] )
            patch_sum = np.sum( areas_patch[idx] )
            return field_sum / float(patch_sum)

        areas_field = self._get_field_area_vector(field_map) # in cm^2
        areas_patch = self._get_patch_area_vector() # in cm^2
        total_field = float( np.sum(areas_field) ) # in cm^2
        if total_field == 0:
            return 0, 0, []

        frac = areas_field / total_field
        test_set = np.array(range(PATCH_NUM))[ (frac>0)*(frac<=(1-thrd)) ]
        keep_set = np.array(range(PATCH_NUM))[ frac>(1-thrd) ]

        max_coverage = 0.0
        best_set = np.array([])

        test_subsets = chain.from_iterable( \
            combinations(test_set,n) for n in range(len(test_set)+1))
        for subset in test_subsets:
            patch_set = np.r_[ keep_set, np.array(subset) ].astype(int)
            if not self._check_connectivity(patch_set): # patches not connected
                continue
            area = np.sum( areas_field[patch_set] )
            if area/total_field < thrd: # not enough field included
                continue
            coverage = calc_coverage(patch_set)
            if coverage > max_coverage:
                max_coverage = coverage
                best_set = patch_set.copy()

        return total_field, max_coverage, best_set
    # end of METHOD _find_best_coverage

    def _check_connectivity(self,patches):
        return check_connectivity( tuple(patches) )
    # end of METHOD _check_connectivity

    def _get_patch_area_vector(self):
        return get_patch_area_vector()
    # end of METHOD _get_patch_area_vector

    def _get_field_area_vector_shapely(self, field_map):
        surf = DeStijl()
        vertices = np.array( surf.vertices )

        [yedges, xedges] = bin_edges( pattern='deStijl' )
        y_bin_size = yedges[1] - yedges[0]
        x_bin_size = xedges[1] - xedges[0]

        areas = []
        for i in range(PATCH_NUM):
            pv = np.array( surf.patch_v[i] )
            x_vec, y_vec = zip(*vertices[pv])

            if len(pv) == 3: # triangle patches
                m, M = min(y_vec), max(y_vec)
                f = lambda y: (y,max(x_vec)) if (i==2)or(i==3) else (min(x_vec),y)
            elif len(pv) == 4: # rectangle patches
                m, M = min(y_vec), max(y_vec)
                f = lambda y: ( min(x_vec), max(x_vec) )

            y_range = np.digitize( (m,M), yedges[1:-1] )
            if M != yedges[y_range[1]]: # +1 for range except ending at grid
                y_range[1] += 1

            first_bin_h = float(yedges[y_range[0]+1]-m) / y_bin_size
            last_bin_h = float(M-yedges[y_range[1]-1]) / y_bin_size

            n = y_range[1] - y_range[0]
            field_area = 0 # unit is number of bin * bin
            for y_bin in range(*y_range):
                bd_min, bd_max = f((yedges[y_bin]+yedges[y_bin+1])/2.0) # cm

                y = y_bin
                bin_width = self._count_bins( field_map,y,bd_min,bd_max,xedges )

                if y_bin == y_range[0]:
                    bin_height = first_bin_h
                elif y_bin == (y_range[1]-1):
                    bin_height = last_bin_h
                else:
                    bin_height = 1
                field_area += (bin_width*bin_height*x_bin_size*y_bin_size) # cm
            areas.append(field_area)

        areas = np.array( areas )

        return areas
    # end of METHOD _get_field_area_vector

    def _get_field_area_vector(self, field_map):
        surf = DeStijl()
        vertices = np.array( surf.vertices )

        [yedges, xedges] = bin_edges( pattern='deStijl' )
        y_bin_size = yedges[1] - yedges[0]
        x_bin_size = xedges[1] - xedges[0]

        areas = []
        for i in range(PATCH_NUM):
            pv = np.array( surf.patch_v[i] )
            x_vec, y_vec = zip(*vertices[pv])

            if len(pv) == 3: # triangle patches
                m, M = min(y_vec), max(y_vec)
                f = lambda y: (y,max(x_vec)) if (i==2)or(i==3) else (min(x_vec),y)
            elif len(pv) == 4: # rectangle patches
                m, M = min(y_vec), max(y_vec)
                f = lambda y: ( min(x_vec), max(x_vec) )

            y_range = np.digitize( (m,M), yedges[1:-1] )
            if M != yedges[y_range[1]]: # +1 for range except ending at grid
                y_range[1] += 1

            first_bin_h = float(yedges[y_range[0]+1]-m) / y_bin_size
            last_bin_h = float(M-yedges[y_range[1]-1]) / y_bin_size

            n = y_range[1] - y_range[0]
            field_area = 0 # unit is number of bin * bin
            for y_bin in range(*y_range):
                bd_min, bd_max = f((yedges[y_bin]+yedges[y_bin+1])/2.0) # cm

                y = y_bin
                bin_width = self._count_bins( field_map,y,bd_min,bd_max,xedges )

                if y_bin == y_range[0]:
                    bin_height = first_bin_h
                elif y_bin == (y_range[1]-1):
                    bin_height = last_bin_h
                else:
                    bin_height = 1
                field_area += (bin_width*bin_height*x_bin_size*y_bin_size) # cm
            areas.append(field_area)

        areas = np.array( areas )

        return areas
    # end of METHOD _get_field_area_vector

    def _count_bins(self, H, y, m, M, edges):
        """
        Count how many bins the field occupies at y=y
        """
        bin_size = edges[1] - edges[0]

        h, w = H.shape
        if y<0 or y>=h: # index out of range 
            return 0 

        v = H[y]
        left, right = np.digitize( (m,M), edges[1:-1] )
        if (left==right) and v[left]:
            return float(M-m)/bin_size

        width = np.sum( v[left+1:right] ) # number of bins
        if v[left]: 
            width += ( float(edges[left+1]-m) / bin_size )
        if v[right]:
            width += ( float(M-edges[right]) / bin_size )

        return width

# end of CLASS FillPatchAnalysis

@memoize
def check_connectivity( patches ):
    if not len(patches):
        print 'No input patches to be tested for connectivity!'
        return False

    surf = DeStijl()
    adj = surf.adjacent.copy()

    # Remove not selected vertex from adjacency map
    excluded = list( set(range(surf.Np)).difference(set(patches)) )
    for i in excluded:
        adj[i,:] = False
        adj[:,i] = False

    def dfs(n): # n is patch number
        left.remove(n)
        for i in left:
            if adj[n][i]:
                dfs(i)
    # end of dfs

    if len(patches) == 2:
        return adj[patches[0]][patches[1]]
    left = list( patches )
    dfs( patches[0] )

    if len(left): # points not connected
        return False
    else:
        return True
# end of FUNCTION check_connectivity

@memoize
def get_patch_area_vector():
    surf = DeStijl()
    vertices = np.array( surf.vertices )

    areas = []
    for i in range(PATCH_NUM):
        pv = np.array( surf.patch_v[i] )
        x_vec, y_vec = zip(*vertices[pv])
        mx, Mx, my, My = min(x_vec), max(x_vec), min(y_vec), max(y_vec)

        if len(pv) == 3: # triangle patches
            area = (Mx-mx) * (My-my) / 2.0
        elif len(pv) == 4: # rectangle patches
            area = (Mx-mx) * (My-my)
        areas.append(area)
    areas = np.array( areas )

    return areas
# end of METHOD _get_field_area_vector

# ================================================================================
# Crossing
# ================================================================================

class CrossingAnalysis(AlignmentAnalysis):
    """
    Quantize the fraction of place field crossing the texture boundaries,
    and test for significance by comparing the observed value to control
    """
    datadir = './Crossing'
    label = 'crosssing analysis'

    def collect_data(self,dtype='test',batch='all',debug=True):
        self.results = {}
        self.debug = debug
        self.fill_in_debug = debug

        # collect cluster list
        clusters = super( CrossingAnalysis, self ).collect_data( \
            dtype=dtype, batch=batch )

        # initialize output pytable parameters
        self._init_table_cols()

        # initialize output folder
        output_dir = os.path.join( self.datadir, 'h5_output' )
        if not os.path.exists( output_dir ):
            os.mkdir( output_dir )

        with tb.openFile(os.path.join(output_dir,'%s_%s.h5'%(dtype,batch)),'w') \
            as tb_file:
            data_table = tb_file.createTable( \
                '/','crossing', self.Cols, title='Crossing')
            self._fill_in_table( clusters,data_table )
            data_table.flush()
        tb_file.close()

        self.out('Done')
    # end of METHOD collect_data

    def overall_test(self,batch='all',thrd=0.8,hist_wd=0.1,rg=[0,1]):

        # open specified data file
        h5_dir = os.path.join( self.datadir, 'h5_output')
        dtypes = ['test', 'control']

        not_crossings = dict( [(dtype,np.zeros(BD_NUM+1)) for dtype in dtypes] ) 
        for dtype in dtypes:
            file_name = '%s_%s.h5' % (dtype,batch)
            file_path = os.path.join( h5_dir, file_name)
            data_file = tb.openFile(file_path,'r')
            data_table = data_file.root.crossing

            for rec in data_table[:]:
                bd_pass = False
                for j in range(BD_NUM):
                    ul, lr = rec['ul_%d'%j], rec['lr_%d'%j]
                    area_ratio = -1 if (ul+lr)<0 else max(ul,lr)/float(ul+lr)
                    if np.isnan( area_ratio ):
                        area_ratio = -1
                    protr = rec['protr_ul_%d'%j] if ul<lr else rec['protr_lr_%d'%j]

                    if rec['parl_%d'%j] < 3: # bins
                        cross_bd = False
                    elif (area_ratio>=0.5) and (area_ratio<0.8):
                        cross_bd = True 
                    elif protr >= 3.5: # bins
                        cross_bd = True
                    else:
                        cross_bd = False
                    if not cross_bd:
                        not_crossings[dtype][j] += 1
                    else:
                        bd_pass = True
                if not bd_pass:
                    not_crossings[dtype][BD_NUM] += 1

            N = len(data_table)
            self.out( '%d fields found for %s:' % (len(data_table),dtype) )
            print not_crossings[dtype]
            print not_crossings[dtype]/float(N)

            data_file.close()
        self.out('Done')

    # end of METHOD overall_test

    def _init_table_cols(self):
        """
        Initialize output h5 tables, called by collect_data
        """
        ul_cols = dict( \
            [('ul_%d'%i, tb.FloatCol(pos=i+1)) for i in range(BD_NUM)] )
        lr_cols = dict( \
            [('lr_%d'%i, tb.FloatCol(pos=i+BD_NUM+1)) for i in range(BD_NUM)] )
        parl_cols = dict( [('parl_%d'%i, \
            tb.FloatCol(pos=i+2*BD_NUM+1)) for i in range(BD_NUM)] )
        protr_ul_cols = dict( [('protr_ul_%d'%i, \
            tb.FloatCol(pos=i+3*BD_NUM+1)) for i in range(BD_NUM)] )
        protr_lr_cols = dict( [('protr_lr_%d'%i, \
            tb.FloatCol(pos=i+4*BD_NUM+1)) for i in range(BD_NUM)] )
        start = 5 * BD_NUM + 1

        cols = { 'ID'        : tb.UInt64Col(pos=start),
                 'rat'       : tb.UInt16Col(pos=start+1),
                 'day'       : tb.UInt8Col(pos=start+2),
                 'session'   : tb.UInt8Col(pos=start+3),
                 'tc'        : tb.StringCol(itemsize=10,pos=start+4),
                 'field_no'  : tb.Int8Col(pos=start+5) }
        cols.update( ul_cols )
        cols.update( lr_cols )
        cols.update( parl_cols )
        cols.update( protr_ul_cols )
        cols.update( protr_lr_cols )

        self.Cols = cols
    # end of METHOD init_pytable

    def _fill_in_table(self, clusters, data_table):

        # place field/coarse binning
        [field_y_edges, field_x_edges] = bin_edges( pattern='deStijl' )

        row = data_table.row

        self.out('Collecting data from each field...')
        id_cnt = 0
        for cluster in clusters:
            rat, day, maze, tc = cluster
            rds = (rat, day, maze)

            session = get_group(rds=rds)
            spikes = session._v_children[tc]
            field_num, labels, contours = label_place_fields(rds,session,spikes)
            
            for i in range(field_num):
                self.out( 'Rat %d, day %d, session %d, %s, field %d' \
                    % (rat,day,maze,tc,i+1) )

                # individual place field map (2D Boolean array)
                selected = labels == (i+1)
                if self.debug:
                    self.H_debug = np.zeros(selected.shape).astype(bool)

                UL, LR = 0, 1
                ftypes = ['ul', 'lr']
                in_areas = [ [], [] ]
                protrusions = [ [], [] ]
                parl = []

                for j in range(BD_NUM):

                    # Check if field near boundary
                    if not self._in_proximity(selected,j,field_x_edges, \
                      field_y_edges,bd_wd=3.0):
                        in_areas[0].append(-1)
                        in_areas[1].append(-1)
                        protrusions[0].append(-1)
                        protrusions[1].append(-1)
                        if len(parl) <= j:
                            parl.append(-1)
                        continue

                    for k in [UL, LR]:
                        area, parallel, protrusion = \
                            self._count_area( selected,j,ftypes[k], \
                                field_x_edges,field_y_edges)
                        in_areas[k].append( area ) 
                        protrusions[k].append( protrusion )
                        if k == UL:
                            parl.append(parallel)

                        if self.debug:
                            print 'boundary: %d' % j
                            print 'area: %f' % in_areas[k][-1]

                            plt.close('all')
                            plt.rcdefaults()
                            plt.hold(True)

                            [y_edges, x_edges] = bin_edges( pattern='deStijl' )

                            plt.subplot(1,2,1)
                            H_shuffle = selected.copy()
                            plt.imshow(H_shuffle,origin='lower', \
                                interpolation='none', aspect='auto', \
                                extent=[x_edges[0],x_edges[-1], \
                                        y_edges[0],y_edges[-1]] )
                            plot_boundaries( pattern='deStijl' \
                                **{'c':'r','lw':0.5,'alpha':0.5})
                            plt.axis('equal')

                            plt.subplot(1,2,2)
                            plt.imshow(self.H_debug,origin='lower', \
                                interpolation='none', aspect='auto', \
                                extent=[x_edges[0],x_edges[-1], \
                                        y_edges[0],y_edges[-1]] )
                            plot_boundaries( pattern='deStijl' \
                                **{'c':'r','lw':0.5,'alpha':0.5})
                            plt.axis('equal')

                            self.H_debug = np.zeros(selected.shape).astype(bool)

                            plt.show()
                    # end of (k) [UL/LR]
                # end of (j) [boundary]
                
                row['ID'] = id_cnt
                row['rat'] = rat
                row['day'] = day
                row['session'] = maze
                row['tc'] = tc
                row['field_no'] = i + 1
                for j in range(BD_NUM):
                    row['ul_%d'%j] = in_areas[UL][j]
                    row['lr_%d'%j] = in_areas[LR][j]
                    row['protr_ul_%d'%j] = protrusions[UL][j]
                    row['protr_lr_%d'%j] = protrusions[LR][j]
                    row['parl_%d'%j] = parl[j]
                row.append()
                data_table.flush()

                id_cnt += 1

    # end of METHOD fill_in_table 

    def _count_bins(self, H, ax, cur, m, M, edges):
        """
        Count how many bins the field occupies at y=y

        cur: bin
        m, M: cm
        """
        H_shuffle = H.copy()

        if ax == 'diag':
            bin_size = edges[0][1] - edges[0][0]

            # from cm to bin number
            x_min, x_max = np.digitize( [m[0],M[0]], edges[0][1:-1])
            y_min, y_max = np.digitize( [m[1],M[1]], edges[1][1:-1])

            if M[0] != edges[0][x_max]:
                x_max += 1
            if M[1] != edges[1][y_max]:
                y_max += 1

            x_range, y_range = x_max-x_min, y_max-y_min
            x_iter = np.array( range(x_min,x_max) )
            y_iter = np.array( range(y_min,y_max) )
            if x_range > y_range:
                r = x_range / y_range # assume int
                y_iter = np.array( [int(i) for i in itertools.chain( \
                    *[[j,]*r for j in range(y_min,y_max)])] )
            elif x_range < y_range:
                r = y_range / x_range
                x_iter = np.array( [int(i) for i in itertools.chain( \
                    *[[j,]*r for j in range(x_min,x_max)])] )
            y_iter = y_iter[::-1]

            v = H_shuffle[ y_iter, x_iter ]

            width = np.sum(v)
            if v[0]:
                width -= ( float(m[0]-edges[0][x_iter[0]]) / bin_size )
            if v[-1]:
                width -= ( float(edges[0][x_iter[-1]+1]-M[0]) / bin_size )
                
            if self.fill_in_debug:
                self.H_debug[y_iter,x_iter] = v 

            return width

        if ax == 'x':
            v = H_shuffle[:,cur]
            edge = edges[1]
        elif ax == 'y':
            v = H_shuffle[cur]
            edge = edges[0]

        bin_size = edge[1] - edge[0]
        s, e = np.digitize( [m,M], edge[1:-1] )
        if (s==e) and v[s]:
            return float(M-m)/bin_size

        width = np.sum( v[s+1:e] ) # number of bins
        if v[s]: 
            width += ( float(edge[s+1]-m) / bin_size )
        if v[e]:
            width += ( float(M-edge[e]) / bin_size )

        if self.fill_in_debug:
            if ax == 'x':
                self.H_debug[s:e+1,cur] = v[s:e+1] 
            elif ax == 'y':
                self.H_debug[cur,s:e+1] = v[s:e+1]

        return width

    def _in_proximity(self, field_map, bd, xedges, yedges, bd_wd=3.0):
        """
        Check if the field is close enough to the boundary 
        """
        edges = [ xedges, yedges ]
        bin_sizes = [ xedges[1]-xedges[0], yedges[1]-yedges[0] ]

        surface = DeStijl()
        x1, y1, x2, y2 = surface.vertices[ surface.edges[bd][0] ] + \
            surface.vertices[ surface.edges[bd][1] ]
        if x1 == x2: # vertical lines
            ax, PARL, PERP = 'y', 1, 0 # y, x
            m, M = min(y1,y2), max(y1,y2)
            f = lambda k: (x1-bd_wd, x1+bd_wd)
        elif y1 == y2: # horizontal lines
            ax, PARL, PERP = 'x', 0, 1 # x, y
            m, M = min(x1,x2), max(x1,x2)
            f = lambda k: (y1-bd_wd, y1+bd_wd)
        else: # diagonal lines
            ax, PARL, PERP = 'diag', 1, 0 # y, x
            m, M = min(y1,y2), max(y1,y2) # y can be replaced by x
            w = float(bd_wd) / np.sqrt(2)
            f = lambda k: ((k-w,k-w), (k+w,k+w)) # k in cm, stands for point (k,k)

        s, e = np.digitize( [m,M], edges[PARL][1:-1] )
        if M != edges[PARL][e]:
            e += 1

        cnt = 0
        for i in range(s,e):
            coord = (edges[PARL][i]+edges[PARL][i+1]) / 2.0 # cm
            bd_min, bd_max = np.array(f(coord))
            bins = self._count_bins(field_map,ax,i,bd_min,bd_max,edges)
            if bins > 0: # intersect
                return True
                #cnt += 1

            if (ax=='diag') and (i<e-1):
                if edges[PARL][i+1] == edges[PERP][i+1]: # need to fix if not equal
                    coord = edges[PARL][i+1] # cm
                    bd_min, bd_max = np.array(f(coord))
                    bins = self._count_bins( \
                        field_map,ax,i,bd_min,bd_max,edges)
                    if bins > 0: # intersect
                        return True
            #            cnt += 1

            #if cnt >= 3:
            #    return True

        return False
    # end of METHOD in_proximity

    def _count_area(self, field_map, bd, side, xedges, yedges, bd_wd=3.0):
        """
        Calculate the field area on each side of the boundary
        """

        edges = [ xedges, yedges ]
        bin_sizes = [ xedges[1]-xedges[0], yedges[1]-yedges[0] ]

        surface = DeStijl()
        hh, hw = surface.height/2.0, surface.width/2.0
        x1, y1, x2, y2 = surface.vertices[ surface.edges[bd][0] ] + \
            surface.vertices[ surface.edges[bd][1] ]
        if x1 == x2: # vertical lines
            ax, PARL, PERP = 'y', 1, 0 # y, x
            m, M = min(y1,y2), max(y1,y2)
            f = lambda k: (x1-bd_wd, x1+bd_wd)
            g = lambda k: (-hw,x1) if side=='ul' else (x1,hw)
        elif y1 == y2: # horizontal lines
            ax, PARL, PERP = 'x', 0, 1 # x, y
            m, M = min(x1,x2), max(x1,x2)
            f = lambda k: (y1-bd_wd, y1+bd_wd)
            g = lambda k: (y1,hh) if side=='ul' else (-hh,y1)
        else: # diagonal lines
            ax, PARL, PERP = 'diag', 1, 0 # y, x
            m, M = min(y1,y2), max(y1,y2) # y can be replaced by x
            w = float(bd_wd) / np.sqrt(2)
            f = lambda k: ((k-w,k-w), (k+w,k+w)) # k in cm, stands for point (k,k)
            g = lambda k: (((2*k-hh,k),(k,hh)) if k>=0 else ((-hw,k),(k,2*k+hw))) \
                if side=='ul' else \
                    (((k,2*k-hw),(hw,k)) if k>=0 else ((k,-hh),(2*k+hh,k))) 

        s, e = np.digitize( [m,M], edges[PARL][1:-1] )
        if M != edges[PARL][e]:
            e += 1

        index = []
        for i in range(s,e):
            coord = (edges[PARL][i]+edges[PARL][i+1]) / 2.0 # cm
            bd_min, bd_max = np.array(f(coord))
            bins = self._count_bins(field_map,ax,i,bd_min,bd_max,edges)
            if bins > 0: # intersect
                index.append( i )

            if (ax=='diag') and (i<e-1):
                if edges[PARL][i+1] == edges[PERP][i+1]:
                    coord = edges[PARL][i+1] # cm
                    bd_min, bd_max = np.array(f(coord))
                    bins = self._count_bins( \
                        field_map,ax,i,bd_min,bd_max,edges)
                    if bins > 0: # intersect
                        index.append( -i )

        if self.debug:
            self.fill_in_debug = True

        first_bin_h = float(edges[PARL][s+1]-m) / bin_sizes[PARL]
        last_bin_h = float(M-edges[PARL][e-1]) / bin_sizes[PARL]

        area, total_w, total_h = 0, 0, 0
        for i in index:
            if i < 0:
                coord = edges[PARL][-i+1] # cm
            else:
                coord = (edges[PARL][i]+edges[PARL][i+1]) / 2.0 # cm
            bd_min, bd_max = np.array(g(coord))
            bins = self._count_bins(field_map,ax,i,bd_min,bd_max,edges)

            if i == s:
                bin_height = first_bin_h
            elif i == (e-1):
                bin_height = last_bin_h
            else:
                bin_height = 1
            area += (bins * bin_height)
            total_w += bins
            total_h += bin_height
        protrusion = float(total_w) / total_h

        if self.debug:
            self.fill_in_debug = False 

        return area, total_h, protrusion 

    # end of METHOD _count_area

    def _init_display(self):
        """
        Initialize display environment
        """
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        self.selected_mask = { 'boundary': True, 'rim': True, 'interior': False }
    # end of METHOD _init_display

    def _plot_selected_bd(self, i, ticks=True):
        selected = np.zeros(BD_NUM).astype(bool)
        selected[i] = True
        F = ratemap_filters(mask=self.selected_mask,bd=selected,bd_wd=3.0)
        plt.imshow(F,origin='lower',cmap='gray')

        if ticks:
            y_edges,x_edges = bin_edges( pattern='deStijl', bin_size=FT_BIN_SIZE )
            n = len(x_edges)
            x_index = np.append( np.array(range(n))[0:n-1:np.ceil(n/3.0)], [n-1,])
            x_ticks = ['%.1f' % f for f in x_edges[x_index]]
            n = len(y_edges)
            y_index = np.append( np.array(range(n))[0:n-1:np.ceil(n/3.0)], [n-1,])
            y_ticks = ['%.1f' % f for f in y_edges[y_index]]

            plt.xticks(x_index, x_ticks)
            plt.yticks(y_index, y_ticks)
            plt.tick_params(axis='both',labelsize=7)
        else:
            plt.xticks([])
            plt.yticks([])
    # end of METHOD _plot_selected_bd

    def output_csv(self):
        file_name = 'h5_output/obs.h5'
        file_path = os.path.join( self.datadir, file_name)
        data_file = tb.openFile(file_path,'r')
        data_table = data_file.root.areas

        output = os.path.join(self.datadir,'obs.csv')

        # (rat,day,session,tc,ul_rates,lr_rates)
        cols = ['Rat','Day','Session','Cluster']
        cols = cols + ['ratio_%d' %i for i in range(BD_NUM)]
        cols = cols + ['parl_%d' %i for i in range(BD_NUM)]
        cols = cols + ['protr_%d' %i for i in range(BD_NUM)]

        with open(output,'w') as f:
            f.write( ';'.join(cols) + '\n')
            for data in data_table[:]:
                f.write( ';'.join( [str(rec) for rec in \
                    [data['rat'],data['day'],data['session'],data['tc']]] ) + ';' )
                ratio = [ -1 if (m+n)<=0 else max(m,n)/float(m+n) \
                    for m,n in zip( [data['ul_%d'%i] for i in range(BD_NUM)], \
                                    [data['lr_%d'%i] for i in range(BD_NUM)] ) ]
                f.write( ';'.join( ['%.2f'%i for i in ratio] ) + ';')
                parl = [ data['parl_%d'%i] for i in range(BD_NUM) ]
                f.write( ';'.join( ['%.2f'%i for i in parl] ) + ';')
                protr = [ data['protr_ul_%d'%i] if data['ul_%d'%i]<data['lr_%d'%i]\
                    else data['protr_lr_%d'%i] for i in range(BD_NUM) ]
                f.write( ';'.join( ['%.2f'%i for i in protr] ) + '\n')
        f.close()
        data_file.close()
        self.out('Done!')
    # end of METHOD output_csv

# end of CutOffAnalysis

@memoize
def params_for_shuffle_space():
    """
    Collect necessary information for field shifting
    Called by CutOffAnalysis
    """
    y_edges, x_edges = bin_edges( pattern='deStijl' ) # cm

    w, h = len(x_edges)-1, len(y_edges)-1
    H_x, H_y = np.meshgrid( range(w), range(h) ) # bin number

    surface = DeStijl()
    hw, hh = surface.width/2.0, surface.height/2.0 # cm
    app_xbin_range = np.digitize( [-hw-10,hw+10], x_edges[1:-1] ) # bin number
    app_ybin_range = np.digitize( [-hh-10,hh+10], y_edges[1:-1] )

    return H_x, H_y, app_xbin_range, app_ybin_range
# end of FUNCTION params_for_shuffle_space


def ratemap_filters(filter_type=None,mask=None,bd=None,patch=None,**kw):
    if bd is None:
        bd = np.zeros(BD_NUM).astype(bool) 
    if patch is None:
        patch = np.zeros(PATCH_NUM).astype(bool) 
    default_mask = { 'default'  : False,
                     'patch'    : np.nan,
                     'boundary' : np.nan,
                     'rim'      : np.nan,
                     'ul_band'  : np.nan,   # uppler-left band
                     'lr_band'  : np.nan,   # lower-right band
                     'ul_area'  : np.nan,
                     'lr_area'  : np.nan } 
    
    if filter_type == 'bd': # mark selected texture boundaries
        mask = default_mask.copy()
        mask['boundary'] = True
        mask['rim'] = False

        F = texture_boundary_filter( \
            tuple(mask.iteritems()), selected_bd=tuple(bd), **kw ) 
    elif filter_type == 'all_bd': # mark all texture boundaries
        mask = default_mask.copy()
        mask['boundary'] = True
        mask['rim'] = False

        bd = np.ones(BD_NUM).astype(bool)

        F = texture_boundary_filter( \
            tuple(mask.iteritems()), selected_bd=tuple(bd), **kw ) 
    elif filter_type == 'not_bd': # mark areas far from texture boundaries
        mask = default_mask.copy()
        mask['default'] = True 
        mask['boundary'] = False 
        mask['rim'] = False

        bd = np.ones(BD_NUM).astype(bool)

        F = texture_boundary_filter( \
            tuple(mask.iteritems()), selected_bd=tuple(bd), **kw ) 
    elif filter_type == 'ul_band':
        mask = default_mask.copy()
        mask['ul_band'] = True 
        mask['rim'] = False

        F = texture_boundary_filter( \
            tuple(mask.iteritems()), selected_bd=tuple(bd), **kw ) 
    elif filter_type == 'lr_band':
        mask = default_mask.copy()
        mask['lr_band'] = True 
        mask['rim'] = False

        F = texture_boundary_filter( \
            tuple(mask.iteritems()), selected_bd=tuple(bd), **kw ) 
    elif filter_type == 'ul_area':
        mask = default_mask.copy()
        mask['ul_area'] = True 

        F = texture_boundary_filter( \
            tuple(mask.iteritems()), selected_bd=tuple(bd), **kw ) 
    elif filter_type == 'lr_area':
        mask = default_mask.copy()
        mask['lr_area'] = True 

        F = texture_boundary_filter( \
            tuple(mask.iteritems()), selected_bd=tuple(bd), **kw ) 
    elif filter_type is None:
        new_mask = default_mask.copy()
        new_mask.update(mask)
        
        F = texture_boundary_filter( tuple(new_mask.iteritems()), \
            selected_bd=tuple(bd),selected_patch=tuple(patch),**kw)

    return F
# end of FUNCTION ratemap_filters


@memoize
def texture_boundary_filter(mask, selected_bd=None, selected_patch=None, **kwargs):
    """
    Generate a 2D Bool matrix which can be used to filter the selected part of the
    specified texture surface pattern.

    Applicable to: deStijl

    Input
        mask - a tuple specifying which parts should pass the filter

    Output
        H_filter - a 2D Bool numpy array filter

    Keywords
        disp - True/False, display debug information
        selected_bd - Specify the selected boundaries
        selected_patch - Specify the selected patches 
        bd_wd - the width of boundary band, in cm.
        band_wd - the width of upper-right/lower-left band, in cm
        buffer_wd - the width of the buffer zone flanking the boundaries
        rim - the width of extra display area outside of the apparatus
        rim_margin - the width of the buffer zone surrounding the apparatus
    """

    # default keyword arguments
    dt_kwargs = { 'disp'            : False,
                  'bd_wd'           : 7.5,
                  'band_wd'         : 2.5,
                  'buffer_wd'       : 2.5,
                  'rim'             : 30,
                  'rim_margin'      : 0.0 }
    kw = dt_kwargs.copy()
    kw.update( kwargs )

    # for memoize purpose
    mask = dict(mask) 
    if selected_bd is not None:
        selected_bd = np.array( selected_bd )
    if selected_patch is not None:
        selected_patch = np.array( selected_patch )

    surface = DeStijl()
    vertices = surface.vertices
    edges = surface.edges
    hw, hh = surface.width/2.0, surface.height/2.0

    def xy_to_bin( coords ):
        bins = []
        for x, y in coords:
            x_bin = int( next( k[0]-1-(x<x_centers[k[0]-1]) \
                for k in enumerate(x_edges) if x<k[1] ) )
            y_bin = int( next( k[0]-1-(y<y_centers[k[0]-1]) \
                for k in enumerate(y_edges) if y<k[1] ) )
            bins.append( (x_bin,y_bin) )
        return bins

    def fill(target):
        im = Image.new('1',H_filter.shape, 'black')    
        draw = ImageDraw.Draw(im)

        if target == 'patch':
            selected_patches = np.array(surface.patches)[selected_patch]
            for patch in selected_patches:
                v = [ vertices[i] for i in patch ]
                v_bin = xy_to_bin(v)
                draw.polygon( v_bin, fill='white' )
        elif (target=='ul_band') or (target=='lr_band'):
            band_lw = int( np.floor(2.0*kw['band_wd']/FT_BIN_SIZE) )
            delta = kw['buffer_wd'] + kw['band_wd']

            sign = 1 if target=='ul_band' else -1
            selected_edges = np.array(edges)[selected_bd]
            for edge in selected_edges:
                x1, y1, x2, y2 =  vertices[edge[0]] + vertices[edge[1]]
                if x1 == x2: # vertical lines
                    shift_x = sign * -delta
                    shift_y = 0
                elif y1 == y2: # horizontal lines
                    shift_x = 0
                    shift_y = sign * delta
                else: # diagonal
                    shift_x = sign * -delta/np.sqrt(2) # delta * cos(45) 
                    shift_y = sign *  delta/np.sqrt(2)  # delta * sin(45)
                v = [ (x1+shift_x,y1+shift_y), (x2+shift_x,y2+shift_y) ]
                v_bin = xy_to_bin(v)
                draw.line( v_bin, fill='white', width=band_lw )
        elif (target=='ul_area') or (target=='lr_area'):
            rx, ry = (-hw, hh) if target=='ul_area' else (hw,-hh)

            selected_edges = np.array(edges)[selected_bd]
            for edge in selected_edges:
                x1, y1, x2, y2 =  vertices[edge[0]] + vertices[edge[1]]
                if x1 == x2: # vertical lines
                    mx, my, Mx, My = (x1,y1,x2,y2) if y1<y2 else (x2,y2,x1,y1)
                    v = [ (mx,my), (Mx,My), (rx,My), (rx,my) ]
                elif y1 == y2: # horizontal lines
                    mx, my, Mx, My = (x1,y1,x2,y2) if x1<x2 else (x2,y2,x1,y1)
                    v = [ (mx,my), (Mx,My), (Mx,ry), (mx,ry) ]
                else: # diagonal
                    v = [ (x1,y1), (x2,y2), (rx,ry) ]
                v_bin = xy_to_bin(v)
                draw.polygon( v_bin, fill='white' )
        elif target == 'boundary':
            lw = int( np.floor(2.0*kw['bd_wd']/FT_BIN_SIZE) )

            selected_edges = np.array(edges)[selected_bd]
            for edge in selected_edges:
                v = list( (vertices[edge[0]],) + (vertices[edge[1]],) )
                v_bin = xy_to_bin(v)
                draw.line( v_bin, fill='white', width=lw )
        # end IF( target )

        # update filter
        w, h = im.size
        for i in range(h):
            for j in range(w):
                if im.getpixel((j,i)) > 0:
                    H_filter[i][j] = mask[target] 

        if target == 'rim':
            rim_wd = kw['rim'] + kw['rim_margin']
            H_x, H_y = np.meshgrid(x_centers,y_centers)
            H_rim = np.logical_or( \
                np.logical_or( H_x<(x_min+rim_wd),H_x>(x_max-rim_wd) ),
                np.logical_or( H_y<(y_min+rim_wd),H_y>(y_max-rim_wd) ) )
            H_filter[ H_rim ] = mask['rim']
    # end of FUNCTION fill

    y_edges,x_edges = bin_edges( pattern=surface.code, bin_size=FT_BIN_SIZE )
    x_min, x_max = x_edges[0], x_edges[-1]
    y_min, y_max = y_edges[0], y_edges[-1]
    x_centers = (x_edges[:-1]+x_edges[1:]) / 2.0
    y_centers = (y_edges[:-1]+y_edges[1:]) / 2.0
    x_nbin, y_nbin = len(x_centers), len(y_centers)

    H_filter = mask['default'] * np.ones((y_nbin,x_nbin)).astype(bool)

    if not np.isnan(mask['patch']): #!= mask['default']:
        fill('patch')
    if not np.isnan(mask['ul_band']): #!= mask['default']:
        fill('ul_band')
    if not np.isnan(mask['lr_band']): #!= mask['default']:
        fill('lr_band')
    if not np.isnan(mask['ul_area']): #!= mask['default']:
        fill('ul_area')
    if not np.isnan(mask['lr_area']): #!= mask['default']:
        fill('lr_area')
    if not np.isnan(mask['boundary']):
        fill('boundary')
    if not np.isnan(mask['rim']):
        fill('rim')

    # Visualize the filter: True --> white, False --> black
    if kw['disp']:
        print 'mask:'
        print mask
        print 'selected_bd: '
        print selected_bd
        print 'selected_patch'
        print selected_patch

        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        plt.imshow(H_filter,origin='lower',cmap='gray')

        n = len(x_edges)
        x_index = np.append( np.array(range(n))[0:n-1:np.ceil(n/10.0)], [n-1,])
        x_ticks = ['%.1f' % f for f in x_edges[x_index]]
        plt.xticks(x_index, x_ticks)

        n = len(y_edges)
        y_index = np.append( np.array(range(n))[0:n-1:np.ceil(n/10.0)], [n-1,])
        y_ticks = ['%.1f' % f for f in y_edges[y_index]]
        plt.yticks(y_index, y_ticks)

        # label boundaries
        for edge in edges:
            v = list( (vertices[edge[0]],) + (vertices[edge[1]],) )
            v_bin = zip( *xy_to_bin(v) )
            plt.plot(v_bin[0],v_bin[1],'r')
        v = [ (hw,-hh), (hw,hh), (-hw,hh), (-hw,-hh), (hw,-hh) ]
        v_bin = zip( *xy_to_bin(v) )
        plt.plot(v_bin[0],v_bin[1],'r')

    return H_filter

# end of FUNCTION texture_boundary_filter


# ================================================================================
# Check place field edge smoothing
# ================================================================================

class ChkFieldEdgeSmoothReport(BaseReport):
    """
    Plot the gaussian-smoothed field contour with different sigma
    """
    datadir = './ChkFieldSmooth'
    label = 'check field smooth'

    ncols = 4
    nrows = 5
    figwidth = 8.0
    figheight = 11.0

    def collect_data(self, debug=False):
        field_table = get_node('/spatial','fields')

        self.out('Collecting cluster list...')
        cluster_list = []
        for rec in field_table.where(QUERY):
            rds = (rec['rat'], rec['day'], rec['session'])
            session = get_group(rds=rds)
            attrs = session._v_attrs
            if (attrs['plt_shape']!='square') or (attrs['surface']!='textured'):
                continue
            tc = 't%dc%d' % (rec['tt'], rec['cl'])
            cluster_list.append( (rds,)+(tc,) )
        if debug:
            cluster_list = cluster_list[:3]

        kvals = [0, 1, 2, 3]
        data_list = [ a+(b,) for a,b in itertools.product(cluster_list,kvals) ] 

        self.results = {}
        self.results['kvals'] = kvals
        self.results['data_list'] = data_list

        self.out('Done')
    # end of METHOD collect_data

    def process_data(self,pattern='DeStijl'):
        data_list = self.results['data_list']
        kvals = self.results['kvals']

        field_table = get_node('/spatial','fields')
        # For future expansion
        if pattern == 'DeStijl':
            surf = DeStijl()
        else:
            print 'Error!'

        def plot_data(ax, rds, tc, k):
            session = get_group( rds=rds )
            spikes = session._v_children[tc]
            field_num, labels, contours = \
                    label_place_fields(rds, session, spikes)
            for f in range(field_num):
                contour_tuple = tuple( tuple(row) for row in contours==f+1 )
                ratemap_tuple = tuple( tuple(row) for row in labels==f+1 )
                poly = field_polygon( contour_tuple, ratemap_tuple, k )
                plt.plot( *zip(*poly), c='r' )
            # plot platform edge
            surf.plot_fringe(c='grey',zorder=-1)
            #plt.plot( *zip(*[(-50,-50),(50,-50),(50,50),(-50,50),(-50,-50)]), \
            #    c='grey',zorder=-1 )
        # end of METHOD plot_data

        old_rds = None
        for item, ax in self.get_plot(data_list):
            rds, tc, k = item
            if old_rds != rds:
                self.out('Generating plots for rat %d day %d session %d...' % rds)
                old_rds = rds
            self.out('%s: k = %d' % (tc,k))
            plot_data( ax, rds, tc, k )
            if self.col == 0:
                quicktitle(ax,'(%d, %d, %d) %s' % (rds+(tc,)), size='xx-small')
            elif self.firstrow:
                ax.set_title('k = %d'%k,size='small')
            plt.tick_params(axis='both',labelsize=7)

        field_table.close()
    # end of METHOD process_data
# end of CLASS ChkFieldEdgeSmoothReport

# ================================================================================
# Parallel edge 
# ================================================================================

class ChkSlopeReport(BaseReport):
    """
    Check which part of the field contour would be recognized as 'parallel' with
    different parameters
    """
    datadir = './ChkSlope'
    label = 'check slope'

    ncols = 4
    nrows = 5
    figwidth = 8.0
    figheight = 11.0

    def collect_data(self, debug=False):
        field_table = get_node('/spatial','fields')

        self.out('Collecting cluster list...')
        data_list = []
        for rec in field_table.where(QUERY):
            rds = (rec['rat'], rec['day'], rec['session'])
            session = get_group(rds=rds)
            attrs = session._v_attrs
            if (attrs['plt_shape']!='square') or (attrs['surface']!='textured'):
                continue
            tc = 't%dc%d' % (rec['tt'], rec['cl'])
            data_list.append( (rds,)+(tc,) )
        if debug:
            data_list = data_list[:3]

        self.results = {}
        self.results['data_list'] = data_list

        field_table.close()
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, k=2, w=7.5, thrd=15):
        data_list = self.results['data_list']

        field_table = get_node('/spatial','fields')

        surface = DeStijl()
        edges, vertices = surface.edges, surface.vertices
        bd_zones, bd_slopes = [], []
        for edge in edges:
            x1, y1, x2, y2 =  vertices[edge[0]] + vertices[edge[1]]
            bd_slopes.append( float(y2-y1)/(x2-x1) if x1!=x2 else np.inf )
            if x1 == x2: # vertical lines
                coords = [ (x1-w,y1), (x1+w,y1), (x2+w,y2), (x2-w,y2) ]
            elif y1 == y2: # horizontal lines
                coords = [ (x1,y1-w), (x1,y1+w), (x2,y2+w), (x2,y2-w) ]
            else: # diagonal lines
                w_p = np.sqrt(2) * w
                coords = [ (x1,y1), (x1-w_p,y1), (x2,y2+w_p), \
                           (x2,y2), (x2+w_p,y2), (x1,y1-w_p) ]
            bd_zones.append( Polygon(coords) )
        colors = ['r','b','g','c','y','m','orange','grey']

        def plot_data(ax, rds, tc):
            session = get_group( rds=rds )
            spikes = session._v_children[tc]
            field_num, labels, contours = \
                    label_place_fields(rds, session, spikes)
            for f in range(field_num):
                contour_tuple = tuple( tuple(row) for row in contours==f+1 )
                ratemap_tuple = tuple( tuple(row) for row in labels==f+1 )
                poly = field_polygon( contour_tuple, ratemap_tuple, k )
                lines = MultiLineString( zip(poly[:-1],poly[1:]) )

                parallel_len = np.zeros(BD_NUM)
                for line in lines:
                    x, y = line.xy
                    slope = float(y[1]-y[0])/(x[1]-x[0]) if x[1]!=x[0] else np.inf
                    plt.plot( x, y, c=colors[BD_NUM], lw=0.2 )
                    for i, bd_zone in enumerate(bd_zones):
                        M, m = max(slope,bd_slopes[i]), min(slope,bd_slopes[i])
                        if np.isinf(M) and np.isinf(m):
                            theta = 0
                        elif np.isinf(M):
                            theta = 90 - abs(np.rad2deg(np.arctan(float(m))))
                        elif M*m == -1: # perpendicular
                            theta = 90.0
                        else:
                            theta = abs(np.rad2deg(np.arctan(float(M-m)/(1+M*m))))
                        if theta > thrd:
                            continue

                        elif bd_zone.contains(line):
                            parallel_len[i] += line.length
                            plt.plot( x, y, c=colors[i], alpha=0.5, )
                        else:
                            intersect = bd_zone.intersection(line)
                            line_len = intersect.length
                            if line_len > 0:
                                parallel_len[i] += line_len
                                plt.plot( intersect.xy[0], intersect.xy[1], \
                                    c=colors[i], alpha=0.5 )
                self.out( ' '.join( ['%.2f'%l for l in parallel_len] ) )
                self.out( '(%.2f /%.2f)' % (sum(parallel_len), \
                    sum([line.length for line in lines])) )
            # plot boundaries 
            plot_boundaries(539,6,1,**{'c':'k','lw':1,'zorder':-1})
            plt.axis( [-80,80,-80,80] )
        # end of METHOD plot_data

        old_rds = None
        for item, ax in self.get_plot(data_list):
            rds, tc = item
            if old_rds != rds:
                self.out('Generating plots for rat %d day %d session %d...' % rds)
                old_rds = rds
            self.out('%s' % tc)
            plot_data( ax, rds, tc )
            quicktitle(ax,'(%d, %d, %d) %s' % (rds+(tc,)), size='xx-small')
            plt.tick_params(axis='both',labelsize=7)

        field_table.close()
    # end of METHOD process_data

# end of CLASS ChkSlopeReport

# ================================================================================

class ParallelEdgeAnalysis(AbstractAnalysis):
    datadir = './ParlEdge'
    label = 'parallel edge analysis'

    def collect_data(self, shuffle=None, debug=False, start_no=None):
        self.results = {}
        self.debug = debug

        field_table = get_node('/spatial','fields')

        # Collect cluster list
        self.out('Collecting cluster list...')
        cluster_list = []
        for rec in field_table.where(QUERY):
            rds = (rec['rat'], rec['day'], rec['session'])
            session = get_group(rds=rds)
            attrs = session._v_attrs
            if (attrs['plt_shape']!='square') or (attrs['surface']!='textured'):
                continue
            tc = 't%dc%d' % (rec['tt'], rec['cl'])
            cluster_list.append( rds+(tc,) )
        if self.debug:
            cluster_list = cluster_list[:10]

        # Initialize output pytable parameters
        self._init_table_cols()
        if start_no is not None:
            self.file_start_no = start_no

        # Initialize output folder
        output_dir = os.path.join( self.datadir, 'h5_output' )
        if not os.path.exists( output_dir ):
            os.mkdir( output_dir )

        self._init_bd_geos()

        if shuffle is None: # fill in observation table
            with tb.openFile(os.path.join(output_dir,'obs.h5'),'w') as tb_file:
                obs_table = tb_file.createTable( \
                    '/','length', self.ObsCols, title='Length')
                self._fill_in_table( \
                    cluster_list,obs_table,False )
                obs_table.flush()
            tb_file.close()
        else: # shuffle = shuffle number, fill in shuffle tables
            self.out('Creating shuffling h5 files...')
            for i in range(shuffle):
                N = self.file_start_no + i
                self.out( 'Shuffle #%d (%d/%d) %s' \
                    % (N,i+1,shuffle,datetime.now()))
                with tb.openFile( os.path.join(output_dir,'shuffle.%04d.h5'%N), \
                  'w') as tb_file:
                    shuffle_table = tb_file.createTable( '/','length', \
                        self.ShuffleCols, title='Length' )
                    self._fill_in_table( \
                        cluster_list,shuffle_table,True )
                    shuffle_table.flush()
                tb_file.close()

        self.out('Done')

    # end of METHOD collect_data

    def _init_table_cols(self):
        p_ul_cols = dict( \
            [('p_ul_%d'%i, tb.FloatCol(pos=i+1)) for i in range(BD_NUM)] )
        p_lr_cols = dict( \
            [('p_lr_%d'%i, tb.FloatCol(pos=i+1)) for i in range(BD_NUM)] )
        n_ul_cols = dict( \
            [('n_ul_%d'%i, tb.FloatCol(pos=i+1)) for i in range(BD_NUM)] )
        n_lr_cols = dict( \
            [('n_lr_%d'%i, tb.FloatCol(pos=i+1)) for i in range(BD_NUM)] )
        start = 4 * BD_NUM + 1

        shuffle = { 'ID'        : tb.UInt64Col(pos=start),
                    'var'       : tb.UInt64Col(pos=start+1),
                    'shift_x'   : tb.Int16Col(pos=start+2),
                    'shift_y'   : tb.Int16Col(pos=start+3),
                    'total'     : tb.FloatCol(pos=start+4) }
        shuffle.update( p_ul_cols )
        shuffle.update( p_lr_cols )
        shuffle.update( n_ul_cols )
        shuffle.update( n_lr_cols )

        obs = { 'ID'        : tb.UInt64Col(pos=start),
                'rat'       : tb.UInt16Col(pos=start+1),
                'day'       : tb.UInt8Col(pos=start+2),
                'session'   : tb.UInt8Col(pos=start+3),
                'tc'        : tb.StringCol(itemsize=10,pos=start+4),
                'field_no'  : tb.Int8Col(pos=start+5),
                'var'       : tb.UInt64Col(pos=start+6),
                'total'     : tb.FloatCol(pos=start+7) }
        obs.update( p_ul_cols )
        obs.update( p_lr_cols )
        obs.update( n_ul_cols )
        obs.update( n_lr_cols )

        self.ShuffleCols = shuffle
        self.ObsCols = obs
        self.file_start_no = 1
    # end of METHOD init_table_cols

    def _fill_in_table(self, clusters, data_table, shuffle):
        row = data_table.row

        if not shuffle:
            self.out('Collecting data from each field...')

        id_cnt, shuffle_cnt = 0, 1
        for cluster in clusters:
            rat, day, maze, tc = cluster
            rds = (rat, day, maze)

            session = get_group(rds=rds)
            spikes = session._v_children[tc]
            field_num, labels, contours = label_place_fields(rds,session,spikes)

            for i in range(field_num):
                if not shuffle:
                    self.out( 'Rat %d, day %d, session %d, %s, field %d' \
                        % (rat,day,maze,tc,i+1) )

                contour_tuple = tuple( tuple(row) for row in contours==i+1 )
                ratemap_tuple = tuple( tuple(row) for row in labels==i+1 )
                poly = field_polygon( contour_tuple, ratemap_tuple, k=2 )

                var, x_shift_range, y_shift_range = self._shuffle_space( poly )
                if shuffle:
                    shift_x, shift_y = self._random_shift( \
                        x_shift_range, y_shift_range )
                    shift = (shift_x, shift_y)
                else:
                    shift = None

                total, parl, nparl = self._calc_parallel_length(poly,shift)

                if shuffle:
                    row['ID'] = id_cnt
                    row['var'] = var
                    row['shift_x'] = shift_x
                    row['shift_y'] = shift_y
                    row['total'] = total
                    for j in range(BD_NUM):
                        for z in ['ul','lr']:
                            row['p_%s_%d'%(z,j)] = parl[z][j]
                    for j in range(BD_NUM):
                        for z in ['ul','lr']:
                            row['n_%s_%d'%(z,j)] = nparl[z][j]
                else:
                    row['ID'] = id_cnt
                    row['rat'] = rat
                    row['day'] = day
                    row['session'] = maze
                    row['tc'] = tc
                    row['field_no'] = i + 1
                    row['var'] = var
                    row['total'] = total
                    for j in range(BD_NUM):
                        for z in ['ul','lr']:
                            row['p_%s_%d'%(z,j)] = parl[z][j]
                    for j in range(BD_NUM):
                        for z in ['ul','lr']:
                            row['n_%s_%d'%(z,j)] = nparl[z][j]
                row.append()
                data_table.flush()

                id_cnt += 1
    # end of METHOD _fill_in_table

    def _init_bd_geos(self,w=5):
        surface = DeStijl()
        edges, vertices = surface.edges, surface.vertices
        bd_zones, bd_slopes = {'ul':[], 'lr':[]}, []
        for edge in edges:
            x1, y1, x2, y2 =  vertices[edge[0]] + vertices[edge[1]]
            bd_slopes.append( float(y2-y1)/(x2-x1) if x1!=x2 else np.inf )
            for zone in ['ul','lr']:
                if x1 == x2: # vertical lines
                    if zone == 'ul':
                        coords = [ (x1-w,y1), (x1,y1), (x2,y2), (x2-w,y2) ]
                    elif zone == 'lr':
                        coords = [ (x1,y1), (x1+w,y1), (x2+w,y2), (x2,y2) ]
                elif y1 == y2: # horizontal lines
                    if zone == 'ul':
                        coords = [ (x1,y1), (x1,y1+w), (x2,y2+w), (x2,y2) ]
                    elif zone == 'lr':
                        coords = [ (x1,y1-w), (x1,y1), (x2,y2), (x2,y2-w) ]
                else: # diagonal lines
                    w_p = np.sqrt(2) * w
                    if zone == 'ul':
                        coords = [ (x1,y1), (x1-w_p,y1), (x2,y2+w_p), (x2,y2) ]
                    elif zone == 'lr':
                        coords = [ (x1,y1), (x2,y2), (x2+w_p,y2), (x1,y1-w_p) ]
                bd_zones[zone].append( Polygon(coords) )

        self.bd_zones = bd_zones
        self.bd_slopes = bd_slopes
        self.surface = surface
    # end of METHOD _init_bd_geos

    def _calc_parallel_length(self, poly, shift, thrd=20):
        if shift is not None:
            xy_array = np.array( zip(*poly) )
            poly = list( zip(xy_array[0]+shift[0], xy_array[1]+shift[1]) )
        lines = MultiLineString( zip(poly,poly[1:]+[poly[0],]) )
        if self.debug:
            colors = ['r','b','g','c','y','m','orange','grey']
            plt.close('all')
            plt.rcdefaults()
            plt.hold(True)
            plot_boundaries(539,6,1,**{'c':'k','lw':1,'zorder':-1})
            plt.axis( [-80,80,-80,80] )

        parallel_len = { 'ul':np.zeros(BD_NUM), 'lr':np.zeros(BD_NUM) }
        not_len = { 'ul':np.zeros(BD_NUM), 'lr':np.zeros(BD_NUM) }
        total_len = sum( [line.length for line in lines] )

        for line in lines:
            x, y = line.xy
            slope = float(y[1]-y[0])/(x[1]-x[0]) if x[1]!=x[0] else np.inf
            if self.debug:
                plt.plot( x, y, c=colors[BD_NUM], lw=0.2 )
            for zone in ['ul', 'lr']:
                for i, bd_zone in enumerate(self.bd_zones[zone]):
                    M = max(slope,self.bd_slopes[i])
                    m = min(slope,self.bd_slopes[i])
                    if np.isinf(M) and np.isinf(m):
                        theta = 0
                    elif np.isinf(M):
                        theta = 90 - abs(np.rad2deg(np.arctan(float(m))))
                    elif M*m == -1: # perpendicular
                        theta = 90.0
                    else:
                        theta = abs(np.rad2deg(np.arctan(float(M-m)/(1+M*m))))

                    if bd_zone.contains(line):
                        if theta > thrd: # not parallel
                            not_len[zone][i] += line.length
                        else: # parallel
                            parallel_len[zone][i] += line.length
                            if self.debug:
                                plt.plot( x, y, c=colors[i], lw=2, alpha=0.5 )
                    else:
                        intersect = bd_zone.intersection(line)
                        line_len = intersect.length
                        if line_len > 0:
                            if theta > thrd:
                                not_len[zone][i] += line_len
                            else:
                                parallel_len[zone][i] += line_len
                                if self.debug:
                                    plt.plot( intersect.xy[0], intersect.xy[1], \
                                        c=colors[i], lw=2, alpha=0.5 )

        if self.debug:
            plt.show()

        return total_len, parallel_len, not_len
    # end of METHOD _calc_parallel_length

    def _shuffle_space(self, polygon, margin=10):
        hw = self.surface.width / 2.0 + margin
        hh = self.surface.height / 2.0 + margin

        x_list, y_list = zip(*polygon)
        field_x_range = np.array( [ max(min(x_list),-hw), min(max(x_list),hw) ] )
        field_y_range = np.array( [ max(min(y_list),-hh), min(max(y_list),hh) ] )

        x_shift_range = np.array([-hw,hw]) - field_x_range
        y_shift_range = np.array([-hh,hh]) - field_y_range
        variants = np.diff(x_shift_range) * np.diff(y_shift_range)

        return variants, x_shift_range, y_shift_range
    # end of METHOD shuffle_space

    def _random_shift(self, x_shift_range, y_shift_range):
        shift_x = random.uniform( *x_shift_range )
        shift_y = random.uniform( *y_shift_range )

        return shift_x, shift_y
    # end of METHOD random_shift

    def shuffle_test(self,shuffle_rg=[1,1001],hist_wd=0.1,rg=[0,1]):
        # open specified data file
        h5_dir = os.path.join( self.datadir, 'h5_output')

        length = { 'obs':[], 'shuffle':[] }
        counts = { 'obs':[], 'shuffle':[] }
        for i in [-1,] + range(*shuffle_rg): # shuffle number
            if i == -1: # real data
                file_name = 'obs.h5'
                dtype = 'obs'
            else:
                file_name = 'shuffle.%04d.h5' % i
                dtype = 'shuffle'
            file_path = os.path.join( h5_dir, file_name)
            data_file = tb.openFile(file_path,'r')
            data_table = data_file.root.length
            N = len(data_table[:])

            par_len, count = 0, 0
            for rec in data_table[:]:
                summation = sum( [rec['bd_%d'%j] for j in range(BD_NUM)] )
                par_len += summation
                if summation > 0:
                    count += 1
            self.out('%d/%d fields aligned with the texture boundaries'%(count,N))
            length[dtype].append( par_len )
            counts[dtype].append( count/float(N) )

            data_file.close()

        obs = length['obs'][0]
        print obs
        shuffle = np.array( length['shuffle'] )
        p = np.sum(shuffle>=obs) / float(len(shuffle)) 

        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)
            
        if rg is None:
            plt.hist(shuffle)
        else:
            bins = np.linspace(rg[0],rg[1],(rg[1]-rg[0])/hist_wd+1)
            print bins
            plt.hist(shuffle,bins=bins,range=rg)
        plt.axvline(x=obs,color='r')
        plt.axis('tight')
        plt.tick_params(axis='both',labelsize=7)
        plt.title('Parallel Length (p=%.3f)' % p)

        f = plt.gcf()
        f.savefig( self.datadir + '/shuffle_field.png')
        f.savefig( self.datadir + '/shuffle_field.pdf')

        obs = counts['obs'][0]
        print obs
        shuffle = np.array( counts['shuffle'] )
        p = np.sum(shuffle>=obs) / float(len(shuffle)) 
        bins = np.linspace(0,1,20)
        plt.hist(shuffle,bins=bins,range=[0,1])
        plt.axvline(x=obs,color='r')
        plt.tick_params(axis='both',labelsize=7)
        plt.title('Aligned field (p=%.3f)' % p)

        f = plt.gcf()
        f.savefig( self.datadir + '/shuffle_field_cnt.png')
        f.savefig( self.datadir + '/shuffle_field_cnt.pdf')
        
        self.out('Done')

    def shuffle_test_hist(self,shuffle_rg=[1,1001],hist_wd=0.1,rg=[0,1]):
        """
        Plot the shuffle distribution
        """

        # open specified data file
        h5_dir = os.path.join( self.datadir, 'h5_output')

        counts = { 'obs':[], 'shuffle':[] }
        counts['obs'] = [ [] for i in range(BD_NUM) ]
        counts['shuffle'] = [ [] for i in range(BD_NUM) ]

        # collect cross/cut/far ratio
        nan_cnt = 0
        for i in [-1,] + range(*shuffle_rg): # shuffle number
            if i == -1: # real data
                file_name = 'obs.h5'
                dtype = 'obs'
            else:
                file_name = 'shuffle.%04d.h5' % i
                dtype = 'shuffle'
            file_path = os.path.join( h5_dir, file_name)
            data_file = tb.openFile(file_path,'r')
            data_table = data_file.root.length

            count = np.zeros(BD_NUM)
            near = np.zeros(BD_NUM)
            for rec in data_table[:]:
                for j in range(BD_NUM):
                    p_ul, n_ul, p_lr, n_lr = rec['p_ul_%d'%j], rec['n_ul_%d'%j], \
                                             rec['p_lr_%d'%j], rec['n_lr_%d'%j]
                    ul = p_ul/float(p_ul+n_ul) if p_ul+n_ul>0 else -1
                    lr = p_ul/float(p_lr+n_lr) if p_lr+n_lr>0 else -1

                    if (ul>=0) or (lr>=0):
                        near[j] += 1
                    if max(ul,lr) >= 0.7:
                        count[j] += 1

            for j in range(BD_NUM):
                counts[dtype][j].append( float(count[j])/near[j] )
            data_file.close()
        # end of FOR LOOP( shuffle number )

        row, col = 4, 4
        bins = np.linspace(rg[0],rg[1],(rg[1]-rg[0])/hist_wd+1)
        print bins

        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        for i in range(BD_NUM):
            data = np.array( counts['shuffle'][i] )
            obs = counts['obs'][i][0]

            N = len(data)
            p = np.sum(data>=obs)/float(N)

            plt.subplot( row, col, i*2+1)
            self._plot_selected_bd(i,ticks=False)
            plt.title('p=%.3f'%p,size='small')

            plt.subplot( row, col, (i+1)*2)
            plt.hist(data,bins=bins,range=rg)
            plt.tick_params(axis='both',labelsize=7)
            plt.axis('tight')
            #plt.axis( [rg[0], rg[1], 0, 270] )
            plt.axvline(x=obs,color='r')
            plt.tick_params(axis='both',labelsize=5)

        title = 'Observed' if dtype=='obs' else 'Shuffle'
        plt.suptitle(title)
        f = plt.gcf()
        f.savefig( self.datadir + '/ind_shuffle_test.png')
        f.savefig( self.datadir + '/ind_shuffle_test.pdf')

        #for dtype in ['obs','shuffle']:
        #    plt.figure()
        #    for i in range(BD_NUM):
        #        data = np.array( ratio[dtype][i] )
        #
        #        plt.subplot( row, col, i*2+1)
        #        self._plot_selected_bd(i,ticks=False)
        #
        #        plt.subplot( row, col, (i+1)*2)
        #        plt.hist(data,bins=bins,range=rg)
        #        plt.tick_params(axis='both',labelsize=7)
        #        plt.axis('tight')
        #        #plt.axis( [rg[0], rg[1], 0, 270] )
        #        plt.tick_params(axis='both',labelsize=5)
        #    title = 'Observed' if dtype=='obs' else 'Shuffle'
        #    plt.suptitle(title)
        #    f = plt.gcf()
        #    f.savefig( self.datadir + '/bd_%s.png'%dtype)
        #    f.savefig( self.datadir + '/bd_%s.pdf'%dtype)
        
        self.out('Done')
    # end of METHOD shuffle_test_hist 

    def _plot_selected_bd(self, i, ticks=True):
        selected = np.zeros(BD_NUM).astype(bool)
        selected[i] = True
        selected_mask = { 'boundary': True, 'rim': True, 'interior': False }
        F = ratemap_filters(mask=selected_mask,bd=selected,bd_wd=3.0)
        plt.imshow(F,origin='lower',cmap='gray')

        if ticks:
            y_edges,x_edges = bin_edges( pattern='deStijl', bin_size=FT_BIN_SIZE )
            n = len(x_edges)
            x_index = np.append( np.array(range(n))[0:n-1:np.ceil(n/3.0)], [n-1,])
            x_ticks = ['%.1f' % f for f in x_edges[x_index]]
            n = len(y_edges)
            y_index = np.append( np.array(range(n))[0:n-1:np.ceil(n/3.0)], [n-1,])
            y_ticks = ['%.1f' % f for f in y_edges[y_index]]

            plt.xticks(x_index, x_ticks)
            plt.yticks(y_index, y_ticks)
            plt.tick_params(axis='both',labelsize=7)
        else:
            plt.xticks([])
            plt.yticks([])
    # end of METHOD _plot_selected_bd

# end of CLASS ParallelEdgeAnalysis
