# encoding: utf-8
"""
field_properties.py
"""

import os
import re
import itertools

import tables as tb
import numpy as np
import matplotlib.pyplot as plt

from scipy.stats import ks_2samp, chi2_contingency, chi2
from scipy.misc import comb
from matplotlib import cm

from core.analysis import AbstractAnalysis
from tools.misc import memoize

from scanr.cluster import string_to_quality

from tdb.data import get_tdb_file, get_group
from tdb.meta import walk_mazes
from tdb.spike import get_session_clusters
from tdb.td_maps import get_coordinate_range, plot_boundaries, bin_edges
from tdb.tdb_misc import get_panel_session_list
from tdb.field import label_place_fields

#=================================================================================

class FieldPropAna(AbstractAnalysis):
    datadir = './FieldProp'
    label = 'field properties'

    def collect_data(self,quality='Fair',repeat_filter='first'):
        tdb_file = get_tdb_file()
        field_table = tdb_file.root.spatial.fields

        self.results = {}

        panels, patterns = tdb_default_panels()
        for panel, pattern in zip(panels, patterns):
            self.out('Collecting data for %s sessions...' % panel)

            maze_list, highest_mean = {}, {}
            rds_list = get_panel_session_list(panel)
            for rds in rds_list:
                rat, day, maze = rds
                session = get_group(rds=rds)
                tt_cl_list = get_session_clusters(session)
                for tt, cl in tt_cl_list:
                    clust = session._v_children['t%dc%d'%(tt,cl)]

                    clust_quality = string_to_quality( clust._v_attrs['quality'] )
                    if clust_quality < string_to_quality(quality):
                        continue

                    query = '(area=="CA1")&(rat==%d)&(day==%d)&(session==%d)&(tt==%d)&(cl==%d)' % (rat,day,maze,tt,cl)
                    field_idx = field_table.getWhereList(query)
                    if not len(field_idx):
                        continue
                    mr = max( field_table[field_idx]['mean_rate'] )

                    key = (rat, day, tt, cl)
                    if key not in maze_list:
                        maze_list[key] = [maze,]
                        highest_mean[key] = mr
                    else:
                        if repeat_filter == 'first':
                            if maze < maze_list[key][0]:
                                maze_list[key] = [maze,]
                        elif repeat_filter == 'highest_mean':
                            if mr > highest_mean[key]:
                                maze_list[key] = [maze,]
                                highest_mean[key] = mr 
                        else:
                            maze_list[key].append( maze )

            N_field, clust_cnt = 0, 0
            D_e, D_c, field_num, area, pa_ratio = [], [], [], [], []
            for key in maze_list.keys():
                dist_e_list, dist_c_list, fs_list, pa_list, fn_list = \
                    [], [], [], [], []

                for maze in maze_list[key]:
                    rat, day, tt, cl = key
                    num = 0 

                    query = '(area=="CA1")&(rat==%d)&(day==%d)&(session==%d)&(tt==%d)&(cl==%d)' % (rat,day,maze,tt,cl)
                    field_idx = field_table.getWhereList(query)
                    if repeat_filter=='average':
                        fn_list.append( len(field_idx) )
                        if len(field_idx) > 1: # multiple fields
                            continue

                    for field in field_table[field_idx]:
                        cx, cy = field['CoM_x'], field['CoM_y']
                        dist_e = np.sqrt(cx**2+cy**2)
                        dist_c = max(np.abs([cx,cy])) # Chebyshev distance
                        fs = field['size'] # in cm^2
                        pa = field['perimeter']**2/float(fs)
                        num += 1
                        if repeat_filter == 'average':
                            dist_e_list.append( dist_e )
                            dist_c_list.append( dist_c )
                            fs_list.append( fs )
                            pa_list.append( pa )
                        else:
                            D_e.append( dist_e ) 
                            D_c.append( dist_c )
                            area.append( fs )
                            pa_ratio.append( pa )

                    if repeat_filter != 'average':
                        field_num.append( num )
                        N_field += num
                        clust_cnt += 1
                # end of FOR LOOP maze

                if repeat_filter == 'average':
                    if len(fn_list):
                        field_num.append( np.mean(fn_list) )
                    if len(fs_list):
                        D_e.append( np.mean(dist_e_list) )
                        D_c.append( np.mean(dist_c_list) )
                        area.append( np.mean(fs_list) )
                        pa_ratio.append( np.mean(pa_list) )

                        N_field += 1 
                        clust_cnt += 1

            self.results[panel] = {}
            self.results[panel]['Euclidean_distance'] = D_e
            self.results[panel]['Chebyshev_distance'] = D_c
            self.results[panel]['area'] = area
            self.results[panel]['perimeter_area_ratio'] = pa_ratio
            self.results[panel]['field_number'] = field_num
            self.out('%d clusters, %d fields collected for %s sessions' \
                % (clust_cnt,N_field,panel))
            
        tdb_file.close()

        self.results['panels'] = panels
        self.results['patterns'] = patterns
        self.out('Done')
    # end of METHOD collect_data

    def plot_field_number_hist(self):
        self._reset_plot()

        max_num = 0
        for i, panel in enumerate( self.results['panels'] ):
            field_num = self.results[panel]['field_number']
            if max(field_num) > max_num:
                max_num = max(field_num)
        bins = np.linspace(0.5,max_num+0.5,max_num+1)

        for i, panel in enumerate( self.results['panels'] ):
            field_num = np.array( self.results[panel]['field_number'] )
            field_num = field_num[field_num>0]
            plt.subplot(2,4,i+1)
            plt.hist( field_num, bins=bins, rwidth=0.4 )
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Field Number Distribution',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(30,20,forward=True)
        f.savefig( os.path.join(self.datadir,'field_num.png'))
    # end of METHOD plot_field_number_hist

    def compare_group_field_num(self, plain='plain-all'):
        self._reset_plot()
        self.out.timestamp = False
        self.log = False
        self.log = True

        groups = ['control', 'deStijl', 'STD', 'SHF']
        group_panels = { "control": [plain, ], 
                         "STD": ['leather_STD', 'tape_STD'], 
                         "SHF": ['leather_SHF', 'leather_SHF'],
                         "deStijl": ['deStijl', ] }
        n_g = len(groups)

        data = {}
        max_num = 0
        for group in groups:
            for panel in group_panels[group]:
                field_num = np.array( self.results[panel]['field_number'] )
                field_num = field_num[field_num>0]
                if group not in data.keys():
                    data[group] = field_num.copy()
                else:
                    data[group] = np.hstack( (data[group], field_num) )
            if max(field_num) > max_num:
                max_num = max(field_num)
        bins = np.linspace(0.5,max_num+0.5,max_num+1)

        obs = None
        for i, group in enumerate( groups ):
            fn = data[group]
            cnt = np.array( [np.sum(fn==j+1) for j in range(max_num)] )
            if obs is None:
                obs = cnt.copy()
            else:
                obs = np.vstack( (obs, cnt) )
            plt.subplot(2,2,i+1)
            plt.hist( fn, bins=bins, rwidth=0.4 )
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.title(group,fontsize=28,va='bottom')
        plt.suptitle('Group Field Number Distribution',fontsize=40)

        chi, p, dof, expected = chi2_contingency(obs,correction=True)
        chi_table = (obs-expected)**2 / (expected*1.0) 
        chi_row = np.sum(chi_table,1)

        self.out( 'chi square test (with Yates correction): chi = %f, p = %f' \
            % (chi,p) )
        self.out( 'Post hoc analyses')
        self.out('- compare between board types')
        for i, r_chi in enumerate(chi_row):
            r_p = 1 - chi2.cdf( r_chi, n_g-1 )
            self.out("%s: chi = %f, p = %f, corrected p = %f" \
                % (groups[i], r_chi, r_p, r_p*comb(n_g,2)) )

        f = plt.gcf()
        f.set_size_inches(12,12,forward=True)
        f.savefig( os.path.join(self.datadir,'group_field_num.png'))

        return obs, expected, chi_table
    # end of METHOD compare_group_CDF 


    def plot_distance_hist(self, dist_type='E', nbin=15):
        self._reset_plot()

        key = 'Euclidean_distance' if dist_type=='E' else 'Chebyshev_distance'

        min_d, max_d = np.inf, 0
        for i, panel in enumerate( self.results['panels'] ):
            hw = 55 if 'plain' in panel else 50
            dist = np.array( self.results[panel][key] ) / float(hw)
            if max(dist) > max_d:
                max_d = max(dist)
            if min(dist) < min_d:
                min_d = min(dist)
        bins = np.linspace(min_d,max_d,nbin+1)

        for i, panel in enumerate( self.results['panels'] ):
            hw = 55 if 'plain' in panel else 50
            dist = np.array( self.results[panel][key] ) / float(hw)
            plt.subplot(2,4,i+1)
            plt.hist( dist, bins=bins )
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Field CoM to Center Distance',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(30,20,forward=True)
        f.savefig( os.path.join(self.datadir,'field_dist.png'))
    # end of METHOD plot_distance_hist

    def compare_distance_CDF(self, plain='plain-all', dist_type='E', nbin=200):
        self._reset_plot()
        self.out.timestamp = False
        self.log = False
        self.log = True

        key = 'Euclidean_distance' if dist_type=='E' else 'Chebyshev_distance'
        sel_panels = \
            [plain,'leather_STD','leather_SHF','tape_STD','tape_SHF','deStijl']
        colors = ['k', 'r', 'm', 'b', 'c', 'g']
        legend_text = ['Plain Board Control', 'Leather (STD)', 'Leather (SHIFT)',
                       'Tape (STD)', 'Tape (SHIFT)', 'deStijl' ]
        #sel_panels = \
        #    [plain,'leather_STD','tape_STD']
        #colors = ['k', 'r', 'b']
        #legend_text = ['Plain Board Control', 'Leather (STD)', 'Tape (STD)']

        min_d, max_d = np.inf, 0
        for i, panel in enumerate( sel_panels ):
            hw = 55 if 'plain' in panel else 50
            dist = np.array( self.results[panel][key] ) / float(hw)
            if max(dist) > max_d:
                max_d = max(dist)
            if min(dist) < min_d:
                min_d = min(dist)
        bins = np.linspace(min_d,max_d,nbin+1)

        dist_ctrl = np.array( self.results[plain][key] ) / 55.0 

        for i, panel in enumerate( sel_panels ):
            hw = 55 if 'plain' in panel else 50
            dist = np.array( self.results[panel][key] ) / float(hw)
            lw = 3 #1 if i==0 else 3
            plt.hist( dist,bins=bins,cumulative=True,normed=True,color=colors[i],\
                histtype='step',lw=lw,alpha=0.7)
            plt.legend( legend_text,loc='upper left',fontsize=20,handlelength=0.7 )
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.ylim( (0, 1) )
            D, p = ks_2samp( dist, dist_ctrl )
            self.out('KS test (%s): D = %f, p = %f' % (panel,D,p))
        plt.title('Field CoM to Center Distance CDF',fontsize=35,va='bottom')

        f = plt.gcf()
        f.set_size_inches(15,12,forward=True)
        f.savefig( os.path.join(self.datadir,'field_dist_cdf.png'))
    # end of METHOD compare_distance_CDF

    def compare_group_dist_CDF(self, plain='plain-all', dist_type='E', nbin=200):
        self._reset_plot()
        self.out.timestamp = False
        self.log = False
        self.log = True

        key = 'Euclidean_distance' if dist_type=='E' else 'Chebyshev_distance'

        groups = ['control', 'STD', 'SHF', 'deStijl']
        group_panels = { "control": [plain, ], 
                         "STD": ['leather_STD', 'tape_STD'], 
                         "SHF": ['leather_SHF', 'leather_SHF'],
                         "deStijl": ['deStijl', ] }
        colors = ['k', 'r', 'b', 'g']
        legend_text = ['Plain Board Control', 'Standard', 'Shift', 'deStijl']
        n_g = len(groups)

        data = {}
        min_d, max_d = np.inf, 0
        for group in groups:
            for panel in group_panels[group]:
                hw = 55 if 'plain' in panel else 50
                dist = np.array( self.results[panel][key] ) / float(hw)
                if group not in data.keys():
                    data[group] = dist.copy()
                else:
                    data[group] = np.hstack( (data[group], dist) )
            m, M = data[group].min(), data[group].max()
            if m < min_d:
                min_d = m
            if M > max_d:
                max_d = M
        bins = np.linspace(min_d,max_d,nbin+1)

        for i, group in enumerate( groups ):
            dist = data[group]
            lw = 3 #1 if i==0 else 3
            plt.hist( dist,bins=bins,cumulative=True,normed=True,color=colors[i],\
                histtype='step',lw=lw,alpha=0.7)
            plt.legend( legend_text,loc='upper left',fontsize=20,handlelength=0.7 )
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.ylim( (0, 1) )

            for j in range(i+1,n_g):
                dist_2 = data[groups[j]]
                D, p = ks_2samp( dist, dist_2 )
                self.out( 'KS test (%s vs. %s): D = %f, p = %f, corrected p = %f' \
                    % (group,groups[j],D,p,p*comb(n_g,2)) )
        plt.title('Field CoM to Center Distance CDF',fontsize=35,va='bottom')

        f = plt.gcf()
        f.set_size_inches(15,12,forward=True)
        f.savefig( os.path.join(self.datadir,'group_dist_cdf.png'))
    # end of METHOD compare_group_CDF 

    def plot_field_size_hist(self, nbin=15):
        self._reset_plot()

        min_a, max_a = np.inf, 0
        for i, panel in enumerate( self.results['panels'] ):
            area = np.array( self.results[panel]['area'] ) / 10000.0 # to m^2
            if max(area) > max_a:
                max_a = max(area)
            if min(area) < min_a:
                min_a = min(area)
        bins = np.linspace(min_a,max_a,nbin+1)

        for i, panel in enumerate( self.results['panels'] ):
            area = np.array( self.results[panel]['area'] ) / 10000.0 # to m^2
            plt.subplot(2,4,i+1)
            plt.hist( area, bins=bins )
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Field Size (m^2)',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(30,20,forward=True)
        f.savefig( os.path.join(self.datadir,'field_size.png'))
    # end of METHOD plot_field_size_hist

    def compare_group_area_CDF(self, plain='plain-all', nbin=200):
        self._reset_plot()
        self.out.timestamp = False
        self.log = False
        self.log = True

        groups = ['control', 'STD', 'SHF', 'deStijl']
        group_panels = { "control": [plain, ], 
                         "STD": ['leather_STD', 'tape_STD'], 
                         "SHF": ['leather_SHF', 'leather_SHF'],
                         "deStijl": ['deStijl', ] }
        colors = ['k', 'r', 'b', 'g']
        legend_text = ['Plain Board Control', 'Standard', 'Shift', 'deStijl']
        n_g = len(groups)

        data = {}
        min_a, max_a = np.inf, 0
        for group in groups:
            for panel in group_panels[group]:
                area = np.array( self.results[panel]['area'] ) / 10000.0 # to m^2
                if group not in data.keys():
                    data[group] = area.copy()
                else:
                    data[group] = np.hstack( (data[group], area) )
            m, M = data[group].min(), data[group].max()
            if m < min_a:
                min_a = m
            if M > max_a:
                max_a = M
        bins = np.linspace(min_a,max_a,nbin+1)

        area_ctrl = data['control']

        for i, group in enumerate( groups ):
            area = data[group]
            lw = 3 #1 if i==0 else 3
            plt.hist( area,bins=bins,cumulative=True,normed=True,color=colors[i],\
                histtype='step',lw=lw,alpha=0.7)
            plt.legend( legend_text,loc='lower right',fontsize=20,handlelength=0.7 )
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.ylim( (0, 1) )

            for j in range(i+1,n_g):
                area_2 = data[groups[j]]
                D, p = ks_2samp( area, area_2 )
                self.out( 'KS test (%s vs. %s): D = %f, p = %f, corrected p = %f' \
                    % (group,groups[j],D,p,p*comb(n_g,2)) )

        plt.title('Field CoM to Center Distance CDF',fontsize=35,va='bottom')
        D, p = ks_2samp( area, area_ctrl )
        self.out( 'KS test (STD vs. SHF): D = %f, p = %f' % \
            ks_2samp(data['STD'],data['SHF']) )

        f = plt.gcf()
        f.set_size_inches(15,12,forward=True)
        f.savefig( os.path.join(self.datadir,'group_area_cdf.png'))
    # end of METHOD compare_group_area_CDF 

    def compare_size_CDF(self, plain='plain-all', nbin=200):
        self._reset_plot()
        self.out.timestamp = False
        self.log = False
        self.log = True

        sel_panels = \
            [plain,'leather_STD','leather_SHF','tape_STD','tape_SHF','deStijl']
        colors = ['k', 'r', 'm', 'b', 'c', 'g']
        legend_text = ['Plain Board Control', 'Leather (STD)', 'Leather (SHIFT)',
                       'Tape (STD)', 'Tape (SHIFT)', 'deStijl' ]

        min_a, max_a = np.inf, 0
        for i, panel in enumerate( self.results['panels'] ):
            area = np.array( self.results[panel]['area'] ) / 10000.0 # to m^2
            if max(area) > max_a:
                max_a = max(area)
            if min(area) < min_a:
                min_a = min(area)
        bins = np.linspace(min_a,max_a,nbin+1)

        area_ctrl = np.array( self.results[plain]['area'] ) / 10000.0 # to m^2

        for i, panel in enumerate( sel_panels ):
            area = np.array( self.results[panel]['area'] ) / 10000.0 # to m^2
            lw = 3 #1 if i==0 else 3
            plt.hist( area,bins=bins,cumulative=True,normed=True,color=colors[i],\
                histtype='step',lw=lw,alpha=0.7)
            plt.legend( legend_text,loc='lower right',fontsize=20,handlelength=0.7 )
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.ylim( (0, 1) )
            D, p = ks_2samp( area, area_ctrl )
            self.out('KS test (%s): D = %f, p = %f' % (panel,D,p))
        plt.title('Field Size CDF',fontsize=35,va='bottom')

        f = plt.gcf()
        f.set_size_inches(15,12,forward=True)
        f.savefig( os.path.join(self.datadir,'field_size_cdf.png'))
    # end of METHOD compare_distance_CDF

    def plot_pa_ratio_hist(self, nbin=30):
        self._reset_plot()

        min_r, max_r = np.inf, 0
        for i, panel in enumerate( self.results['panels'] ):
            ratio = self.results[panel]['perimeter_area_ratio']
            if max(ratio) > max_r:
                max_r = max(ratio)
            if min(ratio) < min_r:
                min_r = min(ratio)
        bins = np.linspace(min_r,max_r,nbin+1)

        for i, panel in enumerate( self.results['panels'] ):
            ratio = self.results[panel]['perimeter_area_ratio']
            plt.subplot(2,4,i+1)
            plt.hist( ratio, bins=bins )
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Perimeter to Area Ratio',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(30,20,forward=True)
        f.savefig( os.path.join(self.datadir,'pa_ratio.png'))
    # end of METHOD plot_field_number_hist

    def plot_dist_pa_scatter(self, dist_type='E'):
        self._reset_plot()

        key = 'Euclidean_distance' if dist_type=='E' else 'Chebyshev_distance'

        min_d, max_d = np.inf, 0
        for i, panel in enumerate( self.results['panels'] ):
            hw = 55 if 'plain' in panel else 50
            dist = np.array( self.results[panel][key] ) / float(hw)
            if max(dist) > max_d:
                max_d = max(dist)
            if min(dist) < min_d:
                min_d = min(dist)

        min_r, max_r = np.inf, 0
        for i, panel in enumerate( self.results['panels'] ):
            ratio = self.results[panel]['perimeter_area_ratio']
            if max(ratio) > max_r:
                max_r = max(ratio)
            if min(ratio) < min_r:
                min_r = min(ratio)

        for i, panel in enumerate( self.results['panels'] ):
            hw = 55 if 'plain' in panel else 50
            dist = np.array( self.results[panel][key] ) / float(hw)
            ratio = self.results[panel]['perimeter_area_ratio']
            plt.subplot(2,4,i+1)
            plt.scatter( dist, ratio )
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.axvline( x=1.0 )
            plt.xlim( (min_d, max_d) )
            plt.ylim( (min_r, max_r) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Distance vs. PA-Ratio Scatter',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(30,20,forward=True)
        f.savefig( os.path.join(self.datadir,'dist_pa_scatter.png'))
    # end of METHOD plot_field_number_hist

    def process_data(self):
        self._reset_plot()

        for i, panel in enumerate(self.results['panels']):
            data = self.results[panel]
            plt.subplot(2,4,i+1)
            plt.hist( data['perimeter_area_ratio'] )
            plt.scatter( data['Chebyshev_distance'], data['perimeter_area_ratio'] )

        plt.show()
    # end of METHOD process_data

    def _reset_plot(self):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

# end of CLASS FieldPropAna

#---------------------------------------------------------------------------------

class FieldCenterDistrAna_old(AbstractAnalysis):
    datadir = './FieldCoMDistr'
    label = 'field com distribution'

    def collect_data(self,quality='Fair',repeat_filter='first'):
        tdb_file = get_tdb_file()
        field_table = tdb_file.root.spatial.fields

        self.results = {}

        panels, patterns = tdb_default_panels()
        for panel, pattern in zip(panels, patterns):
            self.out('Collecting data for %s sessions...' % panel)

            maze_list, highest_mean = {}, {}
            rds_list = get_panel_session_list(panel)
            for rds in rds_list:
                rat, day, maze = rds
                session = get_group(rds=rds)
                tt_cl_list = get_session_clusters(session)
                for tt, cl in tt_cl_list:
                    clust = session._v_children['t%dc%d'%(tt,cl)]

                    clust_quality = string_to_quality( clust._v_attrs['quality'] )
                    if clust_quality < string_to_quality(quality):
                        continue

                    query = '(area=="CA1")&(rat==%d)&(day==%d)&(session==%d)&(tt==%d)&(cl==%d)' % (rat,day,maze,tt,cl)
                    field_idx = field_table.getWhereList(query)
                    if not len(field_idx):
                        continue
                    mr = max( field_table[field_idx]['mean_rate'] )

                    key = (rat, day, tt, cl)
                    if key not in maze_list:
                        maze_list[key] = [maze,]
                        highest_mean[key] = mr
                    else:
                        if repeat_filter == 'first':
                            if maze < maze_list[key][0]:
                                maze_list[key] = [maze,]
                        elif repeat_filter == 'highest_mean':
                            if mr > highest_mean[key]:
                                maze_list[key] = [maze,]
                                highest_mean[key] = mr 
                        else:
                            maze_list[key].append( maze )

            CoMs, dist = [], []
            for key in maze_list.keys():
                for maze in maze_list[key]:
                    rat, day, tt, cl = key
                    query = '(area=="CA1")&(rat==%d)&(day==%d)&(session==%d)&(tt==%d)&(cl==%d)' % (rat,day,maze,tt,cl)
                    #CoMs += [ (field['CoM_x'],field['CoM_y']) \
                    #    for field in field_table.where(query) ]
                    for field in field_table.where(query):
                        CoMs.append( (field['CoM_x'], field['CoM_y']) )
                        dist.append( dist_to_boundary( (rat,day,field['session']),\
                            't%dc%d'%(tt,cl),field['threshold'],panel, \
                                field['field_no'] ) )

            self.results[panel] = (CoMs, dist)
            self.out('%d fields collected for %s sessions' % (len(CoMs),panel))
            
        tdb_file.close()

        self.results['panels'] = panels
        self.results['patterns'] = patterns
        self.out('Done')
    # end of METHOD collect_data

    def collect_data_old(self,quality='Fair'):
        tdb_file = get_tdb_file()
        field_table = tdb_file.root.spatial.fields

        self.results = {}

        panels, patterns = tdb_default_panels()
        for panel, pattern in zip(panels, patterns):
            self.out('Collecting data for %s sessions...' % panel)
            CoMs = [] 

            rds_list = get_panel_session_list(panel)
            for rds in rds_list:
                rat, day, maze = rds
                session = get_group(rds=rds)
                tt_cl_list = get_session_clusters(session)
                for tt, cl in tt_cl_list:
                    tc = 't%dc%d' % (tt,cl)
                    clust = session._v_children[tc]
                    clust_quality = string_to_quality( clust._v_attrs['quality'] )
                    if clust_quality < string_to_quality(quality):
                        continue

                    query = '(area=="CA1")&(rat==%d)&(day==%d)&(session==%d)&(tt==%d)&(cl==%d)' % (rat,day,maze,tt,cl)
                    CoMs += [ (field['CoM_x'],field['CoM_y']) \
                        for field in field_table.where(query) ]

            self.results[panel] = CoMs
            self.out('%d fields collected for %s sessions' % (len(CoMs),panel))

        tdb_file.close()

        self.results['panels'] = panels
        self.results['patterns'] = patterns
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, **traits):
        self._reset_plot()
        patterns = self.results['patterns']

        for i, panel in enumerate( self.results['panels'] ):
            CoMs, dist = self.results[panel]
            xmin, xmax, ymin, ymax = get_coordinate_range( \
                pattern=patterns[i], **traits)
            plt.subplot(2,4,i+1)
            plt.scatter( *zip(*CoMs),s=20,c=dist,cmap=cm.bwr,vmin=0,vmax=75 )
            plot_boundaries(pattern=patterns[i])
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.axis([-80, 80, -80, 80])
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Field Center of Mass Distribution',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(30,15,forward=True)
        f.savefig( os.path.join(self.datadir,'CoM_distr.png'))
    # end of METHOD process_data

    def CoM_hist(self, nbin=25, dist_filter=None, **traits):
        self._reset_plot()

        cm_const_ratio = 500.0 / np.sqrt(17) / 500.0
        #bins = np.linspace(-250,250,nbin+1) # all fields
        bins = np.linspace(-150,150,nbin+1) # cut off at vertices

        platforms = ['leather', 'tape']
        stypes = ['STD', 'SHF']
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        for i, panel in enumerate( panels ):
            CoMs, dist = self.results[panel]
            dist = np.array(dist) / 75.0
            if dist_filter is not None:
                CoMs = np.array(CoMs)[dist<dist_filter]

            consts = []
            for x, y in CoMs:
                if 'leather' in panel:
                    consts.append( -4*x + y )
                elif 'tape' in panel:
                    consts.append( -x + 4*y )

            plt.subplot(2,2,i+1)
            plt.hist(consts,bins=bins)
            if 'STD' in panel:
                plt.axvline(x=-50,lw=2,c='r')
                x_label = (bins+50) * cm_const_ratio
            elif 'SHF' in panel:
                plt.axvline(x=30,lw=2,c='r')
                x_label = (bins-30) * cm_const_ratio
            plt.xticks(bins[::2],['%.2f' % xl for xl in x_label[::2]])
            plt.tick_params(axis='both',labelsize=20,pad=10)
            plt.xlim( (bins[0], bins[-1]) )
            plt.title(panel,fontsize=32,va='bottom')
        plt.suptitle('Field Center of Mass Histograms',fontsize=40)

        f = plt.gcf()
        f.set_size_inches(20,15,forward=True)
        f.savefig( os.path.join(self.datadir,'CoM_hist.png'))
    # end of METHOD CoM_hist

    def _reset_plot(self):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

# end of CLASS FieldCenterDistrAna

#=================================================================================

def tdb_default_panels():
    # single line
    platforms = ['leather', 'tape']
    stypes = ['STD', 'SHF']
    sl = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]
    # single line + deStijl
    tex, pla = 'TEX', 'square_PLA'
    panels = sl + ['deStijl', 'plain-all', 'plain-before', 'plain-after']
    patterns = sl + [tex, pla, pla, pla] # platform_stype

    return panels, patterns

def get_first_qualified_maze(rat, day, tc, quality='Fair'):
    tdb_file = get_tdb_file()
    field_table = tdb_file.root.spatial.fields

    # list the sessions where the cells passed the quality criteria and had field
    tt, cl = [ int(i) for i in re.findall(r'\d+',tc) ] # tXcY --> [X, Y]
    quality_query = '(quality!="Poor")'
    if quality == 'Fair':
        quality_query += '&(quality!="Marginal")'
    query = '(rat==%d)&(day==%d)&(tt==%d)&(cl==%d)' % (rat,day,tt,cl) \
        + '&' + quality_query
    maze_list = [ field['session'] for field in field_table.where(query) ]

    tdb_file.close()

    first_maze = {}
    for maze in maze_list:
        session = get_group( rds=(rat,day,maze) )
        stype = session._v_attrs['type']
        if (stype not in first_maze.keys()) or (first_maze[stype]>maze):
            first_maze[stype] = int( maze )

    return first_maze.values() 
# end of FUNCTION get_first_qualified_maze

def dist_to_boundary(rds, tc, threshold, panel, field_no, **traits):
    if panel == 'leather_STD': # 4x-y-50=0
        a, b, c = 4, -1, -50
    elif panel == 'leather_SHF': # 10x-3y+100=0
        a, b, c = 10, -3, 100
    elif panel == 'tape_STD': # x-4y-50=0
        a, b, c = 1, -4, -50
    elif panel == 'tape_SHF': # 3x-10y+100=0
        a, b, c = 3, -10, 100
    else:
        return -1 # no valid boundary, do not calculate distance

    session = get_group(rds=rds)
    spikes = session._v_children[tc]

    [y_edges, x_edges] = bin_edges(rds,**traits)
    x_centers = (x_edges[:-1]+x_edges[1:]) / 2.0
    y_centers = (y_edges[:-1]+y_edges[1:]) / 2.0
    Hx, Hy = np.meshgrid(x_centers, y_centers)

    field_num, labels, contours, n_edges, threshold = \
        label_place_fields(rds,session,spikes,threshold=threshold)
    select = contours==field_no
    xy = zip( Hx[select], Hy[select] )

    min_dist = np.inf
    keep = None
    for x, y in zip( Hx[select], Hy[select] ):
        d = abs(a*x+b*y+c) / float(np.sqrt(a*a+b*b))
        if d < min_dist:
            min_dist = d
            keep = (x,y)

    print panel, keep, min_dist

    return min_dist 
# end of FUNCTION dist_to_boundary
