# encoding: utf-8
"""
gradient_map.py -- firing rate gradient map analyses 
"""

import warnings
import sys
import pdb
import numpy as np
import matplotlib.pylab as plt

from core.report import BaseReport
from tools.plot import quicktitle
from tools.images import masked_array_to_rgba
from tools.misc import memoize

from scanr.time import elapsed
from scanr.cluster import string_to_quality
from scipy.stats import ttest_ind

from tdb.data import get_group
from tdb.meta import walk_mazes, stamp_to_time
from tdb.spike import get_session_clusters, parse_cell_name
from tdb.tracking import get_coordinate_range, plot_boundaries
from tdb.td_maps import get_ratemap, smooth_ratemap, bin_edges, zone_filter, get_shuffled_zone_filter
from tdb.surface import get_norm_vector

class GradientMapReport(BaseReport):
    """
    For each filtered session-cell plot its gradient (amplitude) map, and then
    merge the plots into one pdf report.
    """
    datadir = './GradMap'
    label = 'gradient map'

    ncols = 5
    nrows = 5
    figwidth = 10.0
    figheight = 11.0

    def collect_data(self):
        """
        Collect session-cluster list
        """
        self.results = {}

        data_list = []
        for rds in walk_mazes():
            session = get_group(rds=rds)
            tc_list = get_session_clusters(session)
            for tc in tc_list:
                cl_name = 't%dc%d' % tc
                t_spk = session._v_children[cl_name]
                n = len(t_spk)

                # remove low firing clusters
                if n < 50:
                    continue

                # remove putative interneurons
                start, end = session._v_attrs['start'], session._v_attrs['end']
                duration = elapsed(start, end)
                if float(n)/duration > 10.0: # in Hz
                    continue

                # remove Poor clusters
                quality = string_to_quality( t_spk._v_attrs['quality'] )
                if quality <= string_to_quality('Poor'):
                    continue

                data_list.append( (rds,cl_name) )
        self.results['data_list'] = data_list
    # end of METHOD collect_data

    def process_data(self,bd=False,grad='amp'):
        """
        For each cell-session generate the plots

        grad: ['amp', 'unit_dot', 'dot']
        """
        data_list = self.results['data_list']

        def plot_data(rds, cl, ax):
            session = get_group(rds=rds)
            t_spk = session._v_children[cl]
            H, mask = get_gradient_map(rds, session, t_spk, grad=grad)

            [y_edges, x_edges] = bin_edges(rds)
            img = masked_array_to_rgba(H,mask=mask,mask_color='b',cmap='gray_r')
            ax.imshow( img,origin='lower',interpolation='none',aspect='auto',\
                extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )

            if bd:
                plot_boundaries(*rds,**{'c':'r'})
            plt.tick_params(axis='both',labelsize=7)
        # end of METHOD plot_data

        old_rds = None
        for item, ax in self.get_plot(data_list):
            rds, cl = item
            if old_rds != rds:
                self.out('Generating plots for rat %d day %d session %d...' % rds)
                old_rds = rds
                dgroup = get_group(rat=rds[0], day=rds[1])
                tt_area = dgroup.tt_area
            self.out('%s' % cl)

            tt = parse_cell_name(cl)[0]
            area = tt_area[int(tt)]

            plot_data(rds,cl,ax)
            quicktitle(ax,'(%d, %d, %d) %s - %s'%(rds+(cl,area)),size='xx-small')
            #plt.subplots_adjust(hspace=1)
    # end of METHOD process_data

# end of CLASS GradientMapReport


class PopulationGradMapReport(BaseReport):
    """
    For each session, plot the summation gradient map of filtered CA1 clusters, and
    then merge the plots into one pdf report.
    """
    datadir = './PopGradMap'
    label = 'population gradient'

    figwidth = 10.0
    figheight = 11.0

    def _set_layout(self, layout):
        if layout == 'sess_type':
            self.ncols = 2 
            self.nrows = 2
        elif layout == 'session':
            self.ncols = 4
            self.nrows = 4

        self.layout = layout
    # end of METHOD set_layout
    
    def collect_per_session(self,rds,ext_mask,norm,grad,rot,H_sum=None,mask=None):
        dgroup = get_group(rat=rds[0],day=rds[1])
        tt_area = dgroup.tt_area

        session = get_group(rds=rds)
        tc_list = get_session_clusters(session)
        cl_cnt = 0
        for tc in tc_list:
            cl_name = 't%dc%d' % tc
            t_spk = session._v_children[cl_name]
            n = len(t_spk)

            # CA1 only
            if 'CA1' not in tt_area[ int(tc[0]) ]:
                continue
            #if 'subiculum' in tt_area[ int(tc[0]) ]:
            #    continue

            # remove low firing clusters
            if n < 50:
                continue

            # remove putative interneurons
            start, end = session._v_attrs['start'], session._v_attrs['end']
            duration = elapsed(start, end)
            if float(n)/duration > 10.0: # in Hz
                continue

            # remove Poor clusters
            quality = string_to_quality( t_spk._v_attrs['quality'] )
            if quality <= string_to_quality('Poor'):
                continue

            H_amp, rds_mask = get_gradient_map(rds,session,t_spk,grad=grad,rot=rot)

            M = rds_mask.copy() if not ext_mask else rds_mask + \
                np.pad(rds_mask,((0,0),(1,0)),'constant')[:,:-1] + \
                np.pad(rds_mask,((0,0),(0,1)),'constant')[:,1:] + \
                np.pad(rds_mask,((1,0),(0,0)),'constant')[:-1,:] + \
                np.pad(rds_mask,((0,1),(0,0)),'constant')[1:,:] + \
                np.pad(rds_mask,((0,1),(0,1)),'constant')[1:,1:] + \
                np.pad(rds_mask,((0,1),(1,0)),'constant')[1:,:-1] + \
                np.pad(rds_mask,((1,0),(0,1)),'constant')[:-1,1:] + \
                np.pad(rds_mask,((1,0),(1,0)),'constant')[:-1,:-1]
            mask = M.copy() if mask is None else mask*M # *==AND

            H_amp[M] = 0
            if norm:
                H_amp = H_amp / float( np.max(H_amp) )
            H_sum = H_amp if H_sum is None else (H_sum+H_amp)

            cl_cnt += 1
        self.out('%d clusters collected from session (%d,%d,%d)'%((cl_cnt,)+rds) )

        return H_sum, mask
    # end of METHOD collect_per_session

    def collect_data(self,grad='amp',rot=None,norm=False,ext_mask=False,layout='sess_type'):
        """
        Collect gradient map of each cluster and then sum within session 
        """
        self.results = {}

        self._set_layout( layout )

        data_list = []
        if self.layout == 'session':
            for rds in walk_mazes():
                self.out('collecting data from rat %d day %d session %d...' % rds)
                H_sum, mask = self.collect_per_session( \
                    rds,ext_mask,norm,grad,rot,H_sum=None,mask=None)
                if H_sum is None:
                    continue
                data_list.append( (rds,H_sum,mask) )
        elif self.layout == 'sess_type':
            for platform in ['leather', 'tape']:
                for exp_type in ['STD', 'SHF']:
                    self.out('collecting data for %s - %s...' % (platform,exp_type))
                    session_cnt = 0
                    H_sum, mask = None, None
                    for rds in walk_mazes():
                        session = get_group(rds=rds)
                        if session._v_attrs['platform'] != platform:
                            continue
                        if session._v_attrs['type'] != exp_type:
                            continue
                        H_sum, mask = self.collect_per_session( \
                            rds,ext_mask,norm,grad,rot,H_sum=H_sum,mask=mask)
                        session_cnt += 1
                    self.out('data collected from %d sessions...' % session_cnt)
                    if H_sum is None:
                        continue
                    data_list.append( ((platform,exp_type),H_sum,mask) )

        self.results['data_list'] = data_list
    # end of METHOD collect_data
    
    def process_data(self,bd=False,smooth='none'):
        """
        For each session generate the plots
        """
        data_list = self.results['data_list']
        self.layout = 'session'
        self.ncols = 4 
        self.nrows = 4
        #self.layout = 'sess_type'
        #self.ncols = 2 
        #self.nrows = 2

        for data, ax in self.get_plot(data_list):
            if self.layout == 'session':
                rds, H, mask = data
                [y_edges, x_edges] = bin_edges(rds)
                self.out('Generating plot for rat %d day %d session %d...' % rds)
                if bd:
                    plot_boundaries(*rds,**{'c':'r'})
                quicktitle(ax,'(%d, %d, %d)'%rds,size='small')
            elif self.layout == 'sess_type':
                pattern_tuple, H, mask = data
                pattern = pattern_tuple[0] + '_' + pattern_tuple[1]
                [y_edges, x_edges] = bin_edges(pattern=pattern)
                self.out('Generating plot for %s...' % pattern)
                if bd:
                    plot_boundaries(pattern=pattern,**{'c':'r'})
                quicktitle(ax,'%s - %s'%pattern_tuple,size='small')

            img = masked_array_to_rgba(H,mask=mask,mask_color='b',cmap='gray_r')
            ax.imshow( img,origin='lower',interpolation=smooth,aspect='auto',\
                extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )

            plt.tick_params(axis='both',labelsize=7)
            #plt.subplots_adjust(hspace=1)
    # end of METHOD process_data

    def compare_zone(self,control=False):
        data_list = self.results['data_list']
        for data in data_list:
            pattern_tuple, H, mask = data
            pattern = pattern_tuple[0] + '_' + pattern_tuple[1]
            if pattern_tuple[1] == 'SHF':
                continue
            self.out('%s:' % pattern)

            H /= np.max(H)

            grads = [[], []]
            for bd in [True, False]:
                Z = zone_filter(pattern=pattern,bd=bd,control=control)
                F = Z * ~mask
                M = H * F
                grads[bd] = M[F>0]
                avg = np.sum(grads[bd]) / np.float(np.sum(F))
                event_str = 'Within boudary window' if bd else \
                    'Outside boudary window'
                self.out('%s Avg: %.2f' % (event_str,avg))
    # end of METHOD compare_zone

    def shuffle_bd_test(self,nbin=40,N=1000,debug=False):
        plot_cnt = 1

        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        data_list = self.results['data_list']
        for data in data_list:
            pattern_tuple, H, mask = data
            pattern = pattern_tuple[0] + '_' + pattern_tuple[1]
            if pattern_tuple[1] == 'SHF':
                continue
            self.out('%s:' % pattern)

            H /= np.max(H)

            # calculate real gradient difference
            for bd in [True, False]:
                Z = zone_filter(pattern=pattern,bd=bd)
                F = Z * ~mask
                M = H * F
                avg = np.sum(M) / np.float(np.sum(F))
                real_avg_diff = avg if bd else real_avg_diff-avg
            self.out('Observed gradient mean difference = %.2f' % real_avg_diff)

            # calculate shuffled gradient difference
            self.out('Shuffling data...')
            avg_diff = []
            for i in xrange(N):
                if np.mod(i,10) == 0:
                    sys.stdout.write('.')
                    sys.stdout.flush()
                Z_i, Z_o = get_shuffled_zone_filter( pattern=pattern )
                F_i, F_o = Z_i*~mask, Z_o*~mask
                M_i, M_o = H*F_i, H*F_o
                if debug:
                    selected = H.copy()
                    selected[F_i==0] = 0
                    [y_edges, x_edges] = bin_edges( pattern=pattern )
                    plt.imshow( selected,origin='lower', \
                        interpolation='none', aspect='auto', \
                            extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )
                    plt.show()
                avg_i = np.sum(M_i) / np.float(np.sum(F_i))
                avg_o = np.sum(M_o) / np.float(np.sum(F_o))
                avg_diff.append( avg_i - avg_o )

            avg_diff = np.array(avg_diff)
            print('\n')

            n = np.sum( avg_diff>real_avg_diff )
            p = float(n) / N
            self.out('p = %.4f (%d/%d)' % (p,n,N))

            x_min = min( real_avg_diff, min(avg_diff) )
            x_max = max( real_avg_diff, max(avg_diff) )
            bins = np.linspace(x_min,x_max,nbin+1)
            bin_size = bins[1] - bins[0]

            plt.subplot(1,2,plot_cnt)
            plt.hist(avg_diff,bins=bins)
            plt.axvline(x=real_avg_diff,lw=2,color='r')
            plt.xlim( [x_min-bin_size, x_max+bin_size] )
            plt.tick_params(axis='both',labelsize=7)
            plt.title(pattern,size='small')
            plt.suptitle('Average Difference Shuffle Test')

            plot_cnt += 1

        plt.show()

    # end of METHOD shuffle_bd_test 

    def bootstrap_test(self,N=1000,nbin=40,grad='dot',rot=None,norm=True,ext_mask=True):
        plot_cnt = 1

        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        for platform in ['leather', 'tape']:
            pattern = platform + '_STD'
            self.out('%s:' % pattern)

            # observed
            clust_list = np.array( self.collect_clust_list( \
                platform=platform,exp_type='STD') )
            obs = self.calculate_avg_diff( \
                pattern,clust_list,grad,rot,ext_mask,norm )

            # individual cell rate change difference distribution
            ind_diff = []
            for clust in clust_list:
                ind_diff.append( self.calculate_avg_diff( \
                    pattern,[clust,],grad,rot,ext_mask,norm) )

            x_min = min(ind_diff)
            x_max = max(ind_diff)
            bins = np.linspace(x_min,x_max,nbin+1)
            bin_size = bins[1] - bins[0]

            plt.subplot(2,2,plot_cnt)
            plt.hist(ind_diff,bins=bins)
            plt.axvline(x=0,lw=2,color='g')
            plt.xlim( [x_min-bin_size, x_max+bin_size] )
            plt.tick_params(axis='both',labelsize=7)
            plt.title(pattern+' individual cells',size='small')
            plot_cnt += 1

            # bootstrap
            self.out('Bootstrapping data...')
            bootstrap = []
            n = len(clust_list)
            bs_index = np.random.randint( n, size=(N,n) )
            for i, idx in enumerate(bs_index):
                if np.mod(i,10) == 0:
                    sys.stdout.write('.')
                    sys.stdout.flush()
                bs_list = clust_list[idx]
                bootstrap.append( self.calculate_avg_diff( \
                    pattern,bs_list,grad,rot,ext_mask,norm) )
            bootstrap = np.array( bootstrap )
            print('\n')

            b5, b25, b975, b995 = \
                np.percentile(bootstrap,0.5), np.percentile(bootstrap,2.5), \
                np.percentile(bootstrap,97.5), np.percentile(bootstrap,99.5)
            self.out('Percentiles: B_5=%.2f, B_25=%.2f, B_975=%.2f, B_995=%.2f' \
                % (b5,b25,b975,b995))
            self.out('Observed: %.4f' % obs)
            self.out('\n')

            # centered percentile bootstrap
            self.out('==== Centered Percentile Bootstrap ====')
            self.out('CI(0.05): (%.2f, %.2f)' % (2*obs-b975,2*obs-b25))
            self.out('CI(0.01): (%.2f, %.2f)' % (2*obs-b995,2*obs-b5))
            self.out('\n')

            # percentile bootstrap
            self.out('==== Percentile Bootstrap ====')
            self.out('CI(0.05): (%.2f, %.2f)' % (b25,b975))
            self.out('CI(0.01): (%.2f, %.2f)' % (b5,b995))
            self.out('\n')

            x_min = min( 0, obs, min(bootstrap) )
            x_max = max( 0, obs, max(bootstrap) )
            bins = np.linspace(x_min,x_max,nbin+1)
            bin_size = bins[1] - bins[0]

            plt.subplot(2,2,plot_cnt)
            plt.hist(bootstrap,bins=bins)
            plt.axvline(x=obs,lw=2,color='r')
            plt.axvline(x=0,lw=2,color='g')
            plt.xlim( [x_min-bin_size, x_max+bin_size] )
            plt.tick_params(axis='both',labelsize=7)
            plt.title(pattern+' bootstrap',size='small')
            plt.suptitle('Average Difference Bootstrap Test')
            plot_cnt += 1

        plt.show()
    # end of METHOD bootstrap_test

    def filter_clust( self, rds ):
        dgroup = get_group(rat=rds[0],day=rds[1])
        tt_area = dgroup.tt_area

        session = get_group(rds=rds)
        tc_list = get_session_clusters(session)
        cl_cnt = 0

        ft_tc_list = []
        for tc in tc_list:
            cl_name = 't%dc%d' % tc
            t_spk = session._v_children[cl_name]
            n = len(t_spk)

            # CA1 only
            if 'CA1' not in tt_area[ int(tc[0]) ]:
                continue

            # remove low firing clusters
            if n < 50:
                continue

            # remove putative interneurons
            start, end = session._v_attrs['start'], session._v_attrs['end']
            duration = elapsed(start, end)
            if float(n)/duration > 10.0: # in Hz
                continue

            # remove Poor clusters
            quality = string_to_quality( t_spk._v_attrs['quality'] )
            if quality <= string_to_quality('Poor'):
                continue

            ft_tc_list.append( (rds,tc) )

        return ft_tc_list
    # end of METHOD filter_clust

    def collect_clust_list(self, platform=None, exp_type=None):
        clust_list = []

        self.out('collecting data for %s - %s...' % (platform,exp_type))
        session_cnt = 0
        for rds in walk_mazes():
            session = get_group(rds=rds)
            if session._v_attrs['platform'] != platform:
                continue
            if session._v_attrs['type'] != exp_type:
                continue
            clust_list += self.filter_clust( rds )
            session_cnt += 1
        self.out('data collected from %d sessions...' % session_cnt)

        return clust_list
    # end of METHOD collect_clust_list

    def calculate_avg_diff( self,pattern,clusts,grad,rot,ext_mask,norm ):
        H_sum, mask = None, None
        for rds, tc in clusts:
            dgroup = get_group(rat=rds[0],day=rds[1])
            session = get_group(rds=rds)

            cl_name = 't%dc%d' % tc
            t_spk = session._v_children[cl_name]
            n = len(t_spk)

            H_amp, rds_mask = get_gradient_map(rds,session,t_spk,grad=grad,rot=rot)

            M = rds_mask.copy() if not ext_mask else rds_mask + \
                np.pad(rds_mask,((0,0),(1,0)),'constant')[:,:-1] + \
                np.pad(rds_mask,((0,0),(0,1)),'constant')[:,1:] + \
                np.pad(rds_mask,((1,0),(0,0)),'constant')[:-1,:] + \
                np.pad(rds_mask,((0,1),(0,0)),'constant')[1:,:]
            mask = M.copy() if mask is None else mask*M # *==AND

            H_amp[M] = 0
            if norm:
                H_amp = H_amp / float( np.max(H_amp) )
            H_sum = H_amp if H_sum is None else (H_sum+H_amp)

        H = H_sum.copy()
        H /= np.max(H)

        avg = [0, 0]
        for bd in [True, False]:
            Z = zone_filter(pattern=pattern,bd=bd,control=False)
            F = Z * ~mask
            M = H * F
            avg[bd] = np.sum(M[F>0]) / np.float(np.sum(F))

        return avg[True]-avg[False]
    # end of METHOD calculate_avg_diff

# end of CLASS PopulationGradMapReport

def get_gradient_map(rds, session, spk, grad='amp', rot=None, vel_filter=True):
    """
    grad: ['amp', 'dot', 'rot_dot', 'vector']
    """
    if vel_filter:
        H_rate, mask = get_ratemap(session,spk,rds)
    else:
        H_rate, mask = get_ratemap(session,spk,rds,min_velocity=0)
    H_rate = smooth_ratemap( H_rate )
    dy, dx = np.gradient(H_rate)

    if grad == 'amp':
        # gradient amplitude
        H = np.sqrt( dx**2 + dy**2 )
    elif grad == 'dot':
        # absolute value of dot product between gradient direction and Single Line
        # boundary norm vector
        v = get_norm_vector(rds)
        H = abs(dx*v[0] + dy*v[1])
    elif grad == 'rot_dot':
        # absolute value of dot product between gradient direction and a control
        # vector which is the Single Line boundary norm vector rotated [rot] degree
        v = get_norm_vector(rds)
        theta = np.deg2rad(rot)
        rot_v = [ v[0]*np.cos(theta)-v[1]*np.sin(theta), \
                  v[0]*np.sin(theta)+v[1]*np.cos(theta) ]
        H = abs(dx*rot_v[0] + dy*rot_v[1])
    elif grad == 'vector':
        # not sure if meaningful, not done yet
        v = get_norm_vector(rds)
        H_x, H_y = dx.copy(), dy.copy()
        #idx = (dx*v[0]+dy*v[1]) < 0
        #H_x[idx] = -H_x[idx]
        #H_y[idx] = -H_y[idx]
        H = ( np.sqrt(dx**2+dy**2), H_x, H_y)

    return H, mask
