# encoding: utf-8
"""
tdb.meta -- Load experiment condition and tetrode location metadata
Modified from scanr.meta
"""

# Library imports
import os
import sys
import yaml
import numpy as np

from os.path import exists as there

from tools.misc import AutoVivification
from tools.path import unique_path

from tdb.config import Config
from tdb.paths import get_data_file_path

STAMP_RATE = Config['sample_rate']['time']

if 'metadata' not in locals():
    master_fn = get_data_file_path('master')
    try:
        fd = file(master_fn, 'r')
    except IOError, e:
        raise e, 'could not open master file: %s' % master_fn
    else:
        metadata = yaml.load(fd)
        fd.close()

def get_start_end(*rds):
    """Return the start and end time-stamps for the specified recording session
    """
    meta = get_session_metadata(*rds)
    return meta['start'], meta['end']

def _elapsed(start, end):
    """Get elapsed time in seconds between two timestamps
    """
    return float(end - start) / STAMP_RATE
    
def stamp_to_time(ts, zero_stamp=None):
    ts = np.asarray(ts, dtype='i8')
    if zero_stamp is None:
        if ts.size == 0:
            return np.array([], 'd')
        t0 = ts[0]
    else:
        t0 = long(zero_stamp)
    return (ts - t0) / STAMP_RATE

def get_duration(*rds):
    """Return the total duration of the specified session
    """
    return _elapsed(*get_start_end(*rds))

def get_session_metadata(*rds):
    """Return the session metadata dictionary for the specified recording session
    """
    if None in rds or len(rds) != 3:
        raise ValueError, 'incomplete session specification: %s' % str(rds)
    rat, day, session = rds
    try:
        the_session = metadata[rat]['days'][day]['sessions'][session]
    except KeyError, e:
        raise e, 'session not found'
    return the_session

def get_rat_list():
    """Sorted list of rat id numbers
    """
    return sorted(metadata.keys())

def walk_rats():
    """Generator for walking rat nodes in the metadata tree
    """
    for rat in get_rat_list():
        yield rat

def get_day_list(rat):
    """Sorted list of recording day numbers for given rat
    """
    return sorted(metadata[rat]['days'].keys())

def walk_days():
    """Generator for walking day nodes in the metadata tree
    """
    for rat in walk_rats():
        for day in get_day_list(rat):
            yield (rat, day)

def get_maze_list(rat, day):
    """Sorted list of session numbers for given rat and day
    """
    return sorted(metadata[rat]['days'][day]['sessions'].keys())
get_session_list = get_maze_list

def walk_sessions(*args, **kwds):
    return walk_mazes(*args, **kwds)

def walk_mazes():
    """Generator for walking maze nodes in the metadata tree
    
    Keywords are passed to get_maze_list().
    """
    for rat, day in walk_days():
        for maze in get_maze_list(rat, day):
            yield (rat, day, maze)

def get_tetrode_list(rat, day):
    """Sorted list of tetrode numbers for given rat and day
    """
    return sorted(metadata[rat]['days'][day]['tetrodes'].keys())

def walk_tetrodes():
    """Generator for walking tetrode nodes in the metadata tree
    """
    for rat, day in walk_days():
        for tt in get_tetrode_list(rat, day):
            yield (rat, day, tt)

def get_tetrode_comment(rat, day, tt):
    """Return the metadata string comment associated with the specified tetrode
    """
    try:
        ttstr = metadata[rat]['days'][day]['tetrodes'][tt]
    except KeyError, e:
        sys.stderr.write('tetrode_comment: bad tetrode specification: %s' %
            str((rat, day, tt)))
        raise e
    else:
        if ttstr is None:
            ttstr = ''
        return ttstr

def get_tetrode_comment(rat, day, tt):
    """Return the metadata string comment associated with the specified tetrode
    """
    try:
        ttstr = metadata[rat]['days'][day]['tetrodes'][tt]
    except KeyError, e:
        sys.stderr.write('tetrode_comment: bad tetrode specification: %s' %
            str((rat, day, tt)))
        raise e
    else:
        if ttstr is None:
            ttstr = ''
        return ttstr
        
def load_yaml_file(which, required=True):
    """Load the metadata tree
    """
    fn = get_data_file_path(which)
    if not fn:
        if required:
            raise IOError, 'could not find \'%s\' file at %s' % (which, fn)
        else:
            return None
    try:
        fd = file(fn, 'r')
    except IOError:
        raise IOError, 'could not open \'%s\' file at %s' % (which, fn)
    else:
        tree = yaml.load(fd)
        fd.close()
    return tree

def write_yaml_file(tree, which):
    """Write out a nested-dict tree to a YAML file specfied by *which*
    
    Arguments:
    tree -- nested dictionary containing hierarchical metadata
    which -- either 'experiment', 'location', 'timing', or 'master'
    """
    filename = get_data_file_path(which, search=False)
    if there(filename):
        backup = unique_path(filename, fmt='%s.%d')
        os.rename(filename, backup)
        sys.stdout.write('Backup: moved old %s file:\n\tFrom: %s\n\tTo: %s\n' % (
            which, filename, backup))
    try:
        fd = file(filename, 'w')
    except IOError:
        raise IOError, 'could not open \'%s\' file: %s' % (which, filename)
    else:
        sys.stdout.write('Writing %s file: %s\n' % (which, filename))
        if type(tree) is AutoVivification:
            tree_dict = tree.dict_copy()
        elif type(tree) is dict:
            tree_dict = tree
        yaml.dump(tree_dict, stream=fd, default_flow_style=False)
        fd.close()
