# encoding: utf-8
"""
plot_rate_map.py -- generate trajectory plots and rate maps to have a glance
at the data
"""

import numpy as np
import matplotlib.pylab as plt
from scipy.interpolate import interp1d
from scipy.ndimage.filters import gaussian_filter

from core.report import BaseReport
from tools.plot import quicktitle
from tools.images import masked_array_to_rgba
from tools.misc import memoize

from scanr.time import elapsed
from scanr.cluster import string_to_quality

from tdb.config import Config
from tdb.data import get_tdb_file, get_group
from tdb.meta import walk_mazes, stamp_to_time, get_maze_list
from tdb.spike import get_session_clusters, parse_cell_name
from tdb.td_maps import get_coordinate_range, plot_boundaries, bin_edges
from tdb.ratemap import get_ratemap, smooth_ratemap, get_session_occupancy
from tdb.tdb_misc import session_filtered_clusts, get_panel_session_list, get_panel_clust_list

#BIN_SIZE = 1.5
#MIN_VELOCITY = 10.0
#OCC_THRD = 0.05

BIN_SIZE = 2.0
OCC_THRD = 0.06
MIN_VELOCITY = 5.76 

def locate_spike(ts, x, y, t_spk, start=None, velocity=None, min_velocity=None):
    if start == None:
        start = ts[0]
    t_in_sec = stamp_to_time(ts,zero_stamp=start)
    interp_x = interp1d(t_in_sec,x,fill_value=0.0,bounds_error=False)
    interp_y = interp1d(t_in_sec,y,fill_value=0.0,bounds_error=False)
    x_spk = interp_x(stamp_to_time(t_spk,zero_stamp=start))
    y_spk = interp_y(stamp_to_time(t_spk,zero_stamp=start))

    if min_velocity is not None:
        if velocity is None:
            print 'Velocity missing, did not apply velocity filter'
            return x_spk, y_spk
        interp_v = interp1d(t_in_sec,velocity,fill_value=0.0,bounds_error=False)
        v_spk = interp_v(stamp_to_time(t_spk,zero_stamp=start))

        valid = v_spk > min_velocity
        x_spk = x_spk[valid]
        y_spk = y_spk[valid]

    return x_spk, y_spk
# end of FUNCTION locate_spike

def plot_traj_and_spike(ax, rds, x, y, x_spk, y_spk, **traits):
    xmin, xmax, ymin, ymax = get_coordinate_range( *rds, **traits )

    plt.sca(ax)
    plt.hold('on')
    plt.plot(x,y,'k-',lw=0.3)
    plt.plot(x_spk,y_spk,'r.',ms=1.5)
    plt.tick_params(axis='both',labelsize=7)
    ax.set( xlim=[xmin,xmax], ylim=[ymin,ymax] )

    return
# end of FUNCTION plot_traj_and_spike

def plot_ratemap(ax, session, spikes, rds, outlier=False, smooth=True, cbar=False, **traits ):
    H_occ, valid = get_session_occupancy(session, **traits)
    H_rate, mask = get_ratemap(session, spikes, rds, **traits)
    if smooth is not False:
        H_rate = smooth_ratemap( H_rate, mask, smooth=smooth )
    H_rate[mask] = 0

    if outlier:
        if np.sum(H_rate>0) > 0:
            half_rate = np.percentile(H_rate[H_rate>0],50)
            scale = half_rate / 0.5
            H_rate /= float(scale)
            H_rate[H_rate>1] = 1

    [y_edges, x_edges] = bin_edges(rds,**traits)
    img = masked_array_to_rgba(H_rate,mask=mask,mask_color='b',cmap='gray_r')
    im = ax.imshow(img,origin='lower',interpolation='none',aspect='auto',\
        extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )
    
    return H_rate, mask 
# end of FUNCTION plot_rate_map

def define_bin_edges(dmin, dmax):
    nbin = np.ceil( (dmax-dmin) / BIN_SIZE ) # actual bin size always <= BIN_SIZE
    edges = np.linspace(dmin,dmax,nbin+1)
    return edges
# end of FUNCTION define_bin_edges

def plot_heat_map(ax, rds, ts, x, y, fs, x_spk=None, y_spk=None, type='rate', smooth=None):
    xmin, xmax, ymin, ymax = get_coordinate_range( *rds )
    x_edges = define_bin_edges(xmin, xmax)
    y_edges = define_bin_edges(ymin, ymax)

    H_occ, ybins, xbins = np.histogram2d( y,x,bins=[y_edges,x_edges] )
    mask = H_occ == 0
    peak_rate = None

    if type == 'occ':
        if smooth is not None:
            H_out = gaussian_filter(H_occ,sigma=2,mode='nearest')
        else:
            H_out = H_occ
    elif type == 'rate':
        if smooth == 'adaptive_binning':
            Cx_list = (x_edges[:-1]+x_edges[1:]) / 2.0
            Cy_list = (y_edges[:-1]+y_edges[1:]) / 2.0
            alpha = 10**6
            len_traj, len_spk = len(x), len(x_spk)
            H_rate = []
            for Cy in Cy_list:
                H_rate.append([])
                for Cx in Cx_list:
                    r_traj = np.sort( np.sqrt((x-Cx)**2+(y-Cy)**2) )
                    r_spk = np.sort( np.sqrt((x_spk-Cx)**2+(y_spk-Cy)**2) ) 
                    ptr_traj, ptr_spk, Ntraj, Nspk = 0, 0, 0, 0
                    r, threshold = float(BIN_SIZE)/2, 1 
                    while (r<=10.0) and (Nspk<threshold):
                        while (ptr_traj<len_traj) and (r_traj[ptr_traj]<r):
                            ptr_traj += 1
                            Ntraj += 1
                        while (ptr_spk<len_spk) and (r_spk[ptr_spk]<r):
                            ptr_spk += 1
                            Nspk += 1
                        if Ntraj == 0:
                            break
                        threshold = alpha / ((Ntraj*r)**2)
                        r += BIN_SIZE
                    rate = 0 if Ntraj==0 else Nspk/Ntraj*fs
                    H_rate[-1].append( rate )
            H_rate = np.array( H_rate )
            print np.shape(H_rate)
            peak_rate = np.max( H_rate )
        else:
            H_spk, ybins, xbins = np.histogram2d(y_spk,x_spk,bins=[y_edges,x_edges])
            H_occ[mask] = 1
            H_rate = H_spk.astype('d') / H_occ * fs # in Hz
            H_rate[mask] = 0.0
            if smooth == 'gaussian':
                H_rate = gaussian_filter(H_rate,sigma=2,mode='nearest')
            peak_rate = np.max( H_rate )
        H_out = H_rate
        ax.text(0,-0.15,'%f'%peak_rate,fontsize=5,transform=ax.transAxes)
    img = masked_array_to_rgba( H_out,mask=mask,mask_color='b',cmap='gray_r' )
    ax.imshow( img,origin='lower',interpolation='none',aspect='auto', \
        extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )
    ax.set( xlim=[x_edges[0],x_edges[-1]], ylim=[y_edges[0],y_edges[-1]] )

    plot_boundaries(*rds,**{'c':'r'})
    plt.tick_params(axis='both',labelsize=7)

    return peak_rate
# end of FUNCTION plot_heat_map

class TrajReport(BaseReport):
    """
    For each cluster in each textured session, generate the trajectory and spike
    plot.  Combine all plots into one pdf report.
    """
    datadir = './TrajMap'
    label = 'trajectory map'

    ncols = 5
    nrows = 5
    figwidth = 10.0
    figheight = 11.0

    def collect_data( self, cluster_filter=True ):
        self.results = {} 

        data_list = []
        for rds in walk_mazes():
            dgroup = get_group(rat=rds[0],day=rds[1])
            tt_area = dgroup.tt_area

            session = get_group(rds=rds)
            tc_list = get_session_clusters(session)

            for tc in tc_list:
                cl_name = 't%dc%d' % tc
                if not cluster_filter:
                    data_list.append( (rds,cl_name) )
                    continue

                t_spk = session._v_children[cl_name]
                n = len(t_spk)

                # CA1 only
                if 'CA1' not in tt_area[ int(tc[0]) ]:
                    continue

                # remove low firing clusters
                if n < 50:
                    continue

                # remove putative interneurons
                start, end = session._v_attrs['start'], session._v_attrs['end']
                duration = elapsed(start, end)
                if float(n)/duration > 10.0: # in Hz
                    continue

                # remove Poor clusters
                quality = string_to_quality( t_spk._v_attrs['quality'] )
                if quality <= string_to_quality('Poor'):
                    continue

                data_list.append( (rds,cl_name) )

        self.results['data_list'] = data_list
        self.out('%d clusters collected!' % len(data_list))
        self.out('Done!')
    # end of METHOD filter_clust

    def process_data(self,bd=False):
        """
        For each cell-session generate the plots
        """
        data_list = self.results['data_list']

        def plot_data(rds, cl, ax):
            session = get_group(rds=rds)
            x, y, ts = \
                [ session._v_children[key] for key in ['x','y','t_pos'] ]
            t_spk = session._v_children[cl]

            x_spk, y_spk = locate_spike(ts,x,y,t_spk)
            plot_traj_and_spike(ax,rds,x,y,x_spk,y_spk)

            if bd:
                plot_boundaries(*rds,**{'c':'y','alpha':0.5,'lw':4})

            return t_spk._v_attrs['quality'] 
        # end of METHOD plot_data

        old_rds = None
        for item, ax in self.get_plot(data_list):
            rds, cl = item
            if old_rds != rds:
                self.out('Generating plots for rat %d day %d session %d...' % rds)
                old_rds = rds
                dgroup = get_group(rat=rds[0], day=rds[1])
                tt_area = dgroup.tt_area
            self.out('%s' % cl)

            tt = parse_cell_name(cl)[0]
            area = tt_area[int(tt)]

            quality = plot_data(rds,cl,ax)
            quicktitle(ax,'(%d, %d, %d) %s (%s) - %s'%(rds+(cl,quality,area)) \
                ,size='xx-small')
            #plt.subplots_adjust(hspace=1)
    # end of METHOD process_data

# end of CLASS TrajMapReport

class TrajNRateMapReport(BaseReport):
    """
    For each cluster in each session, generate the trajectory and spike plot, the
    occupancy map and the rate map, with or without velocity filtering.  Combine all
    plots into one pdf report.
    """
    datadir = './TrajNRateMap'
    label = 'trajectory and rate map'

    ncols = 5
    nrows = 5
    figwidth = 10.0
    figheight = 11.0

    def collect_data(self):
        """
        Collect session and cluster list
        """
        self.results = {}

        plt_types = ['Trajectory and Spike', \
            'Occupancy Map','Rate Map','Smoothed Rate Map', \
            'Speed Filtered']
        self.results['plt_types'] = plt_types

        data_list = []
        for rds in walk_mazes():
            session = get_group(rds=rds)
            tc_list = get_session_clusters(session)
            for tc in tc_list:
                cl_name = 't%dc%d' % tc
                for plt_type in plt_types:
                    data_list.append( (rds,cl_name,plt_type) )
        self.results['data_list'] = data_list
    # end of METHOD collect_data

    def process_data(self,smooth=None):
        """
        For each cell-session generate the plots
        """
        data_list = self.results['data_list']
        plt_types = self.results['plt_types']

        def plot_data(rds, cl, ax, plt_type):
            session = get_group(rds=rds)
            fs = session._v_attrs['tracking_sample_rate']
            x, y, ts = \
                [ session._v_children[key] for key in ['x','y','t_pos'] ]
            t_spk = session._v_children[cl]

            if plt_type == 'Trajectory and Spike':
                x_spk, y_spk = locate_spike(ts,x,y,t_spk)
                plot_traj_and_spike(ax,rds,x,y,x_spk,y_spk)
            elif plt_type == 'Occupancy Map':
                plot_heat_map(ax,rds,ts,x,y,fs,type='occ',smooth=smooth)
            elif plt_type == 'Rate Map':
                x_spk, y_spk = locate_spike(ts,x,y,t_spk)
                peak = plot_heat_map(ax,rds,ts,x,y,fs,x_spk=x_spk,y_spk=y_spk, \
                    type='rate',smooth=None)
            elif plt_type == 'Smoothed Rate Map':
                x_spk, y_spk = locate_spike(ts,x,y,t_spk)
                peak = plot_heat_map(ax,rds,ts,x,y,fs,x_spk=x_spk,y_spk=y_spk, \
                    type='rate',smooth='gaussian')
            elif plt_type == 'Speed Filtered':
                delta = np.sqrt(np.diff(x)**2 + np.diff(y)**2)
                dt = np.diff( stamp_to_time(ts,zero_stamp=ts[0]) )
                velocity = np.r_[0.0, delta/dt]
                
                x_spk, y_spk = locate_spike( ts,x,y,t_spk, \
                    velocity=velocity,min_velocity=MIN_VELOCITY)
                valid = velocity > MIN_VELOCITY
                peak = plot_heat_map( ax,rds,ts,x[valid],y[valid],fs, \
                    x_spk=x_spk,y_spk=y_spk,type='rate',smooth='gaussian')
        # end of METHOD plot_data

        old_rds = None

        for item, ax in self.get_plot(data_list):
            rds, cl, plt_type = item
            if old_rds != rds:
                self.out('Generating plots for rat %d day %d session %d...' % rds)
                old_rds = rds
            self.out('%s - %s' % (cl,plt_type))
            plot_data(rds,cl,ax,plt_type)
            if self.col == 0:
                quicktitle( ax,'(%d, %d, %d) %s' % \
                    (rds+(cl,)),size='xx-small' )
            elif self.firstrow:
                ax.set_title(plt_type,size='small')
            #plt.subplots_adjust(hspace=1)
    # end of METHOD process_data

# end of CLASS TrajNRateMapReport

class RateMapReport(BaseReport):
    """
    For each filtered session-cell plot its ratemap, and then merge the plots into
    one pdf report.
    """
    datadir = './RateMap'
    label = 'rate map'

    ncols = 5
    nrows = 5
    figwidth = 10.0
    figheight = 11.0

    def collect_data(self):
        """
        Collect session-cluster list
        """
        self.results = {}

        data_list = []
        for rds in walk_mazes():
            session = get_group(rds=rds)
            tc_list = get_session_clusters(session)
            for tc in tc_list:
                cl_name = 't%dc%d' % tc
                t_spk = session._v_children[cl_name]
                n = len(t_spk)

                # remove low firing clusters
                if n < 50:
                    continue

                # remove putative interneurons
                start, end = session._v_attrs['start'], session._v_attrs['end']
                duration = elapsed(start, end)
                if float(n)/duration > 10.0: # in Hz
                    continue

                # remove Poor clusters
                quality = string_to_quality( t_spk._v_attrs['quality'] )
                if quality <= string_to_quality('Poor'):
                    continue

                data_list.append( (rds,cl_name) )
        self.results['data_list'] = data_list
    # end of METHOD collect_data

    def process_data(self,bd=False,outlier=False,smooth=True,**traits):
        """
        For each cell-session generate the plots
        """
        data_list = self.results['data_list']

        def plot_data(rds, cl, ax):
            session = get_group(rds=rds)
            t_spk = session._v_children[cl]
            plot_ratemap( \
                ax,session,t_spk,rds,outlier=outlier,smooth=smooth,**traits)

            if bd:
                plot_boundaries(*rds,**{'c':'r'})
            plt.tick_params(axis='both',labelsize=7)
        # end of METHOD plot_data

        old_rds = None
        for item, ax in self.get_plot(data_list):
            rds, cl = item
            if old_rds != rds:
                self.out('Generating plots for rat %d day %d session %d...' % rds)
                old_rds = rds
                dgroup = get_group(rat=rds[0], day=rds[1])
                tt_area = dgroup.tt_area
            self.out('%s' % cl)

            tt = parse_cell_name(cl)[0]
            area = tt_area[int(tt)]

            plot_data(rds,cl,ax)
            quicktitle(ax,'(%d, %d, %d) %s - %s'%(rds+(cl,area)),size='xx-small')
            #plt.subplots_adjust(hspace=1)
    # end of METHOD process_data

# end of CLASS RateMapReport


class SingleLineClustReport(BaseReport):
    """
    For Single Line experiments, trace the same cluster across days/sessions and 
    generate its trajectory-spike plot or rate map. Combine all plots into one
    pdf report.
    """
    datadir = './SLClust'
    label = 'SL clust traj n rate'

    ncols = 4 
    nrows = 5
    figwidth = 8.0
    figheight = 11.0

    def collect_data( self, clust_filter='field' ):
        """
        Collect cluster list across sessions

        clust_filter: [ None, 'naive', 'field' ]
            - None
                collect the rat-clust list without any filtering
            - 'naive'
                collect the rat-clust list if the cell passes naive filter in at
                least one session , used only when the place field table is not
                constructed yet
            - 'field'
                collect the rat-clust list if the cell has place field in at
                least one session
        """
        self.results = {}

        if clust_filter == 'field':
            with_field_list = self._get_clust_with_field()

        data_list = {} 
        for rds in walk_mazes():
            session = get_group(rds=rds)
            # Collect only Single Line data
            if session._v_attrs['expt_type'] != 'SL':
                continue
            rat, day, maze = rds

            tc_list = get_session_clusters(session)
            for tc in tc_list:
                cl_name = 't%dc%d' % tc
                spk = session._v_children[cl_name]
                quality = string_to_quality( spk._v_attrs['quality'] )

                pass_filter = True # clust_filter is None, no filtering
                if clust_filter == 'field':
                    pass_filter = (rat,day,maze,cl_name) in with_field_list
                elif clust_filter == 'naive':
                    # label non-CA1 clusters 
                    dgroup = get_group(rat=rds[0],day=rds[1])
                    tt_area = dgroup.tt_area
                    if 'CA1' not in tt_area[ int(tc[0]) ]:
                        pass_filter = False

                    # label POOR clusters
                    if quality <= string_to_quality('Poor'):
                        pass_filter = False

                    # label low firing clusters
                    n = len(spk)
                    if n < 50:
                        pass_filter = False

                    # label putative interneurons
                    start, end = session._v_attrs['start'], session._v_attrs['end']
                    duration = elapsed(start, end)
                    if float(n)/duration > 10.0: # in Hz
                        pass_filter = False

                d_key = (rat,cl_name)
                rec = (day, maze, quality)
                if d_key not in data_list.keys():
                    data_list[d_key] = [ pass_filter, rec ]
                else:
                    data_list[d_key].append( rec )
                    # flag is set True if any session pass filter
                    data_list[d_key][0] = pass_filter or data_list[d_key][0]

        clust_list = []
        keys = sorted(data_list.keys())
        for key in keys:
            rat, tc = key
            if data_list[key][0]:
                for day, maze, quality in data_list[key][1:]:
                    clust_list.append( (rat,day,maze,tc,quality) )

        self.results['clust_list'] = clust_list

        self.out('%d rat-cluster collected' % len(keys))
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self,plt_type='spike',bd=True):
        """
        For each cell-session generate the plots
        """
        clust_list = self.results['clust_list']

        def plot_data(rds, tc, ax):
            session = get_group(rds=rds)
            fs = session._v_attrs['tracking_sample_rate']
            x, y, ts = \
                [ session._v_children[key] for key in ['x','y','t_pos'] ]
            t_spk = session._v_children[tc]

            if plt_type == 'spike':
                x_spk, y_spk = locate_spike(ts,x,y,t_spk)
                plot_traj_and_spike(ax,rds,x,y,x_spk,y_spk)
            elif plt_type == 'ratemap':
                x_spk, y_spk = locate_spike(ts,x,y,t_spk)
                plot_ratemap(ax,session,t_spk,rds)
                plt.tick_params(axis='both',labelsize=7)

            if bd:
                platform = session._v_attrs['platform']
                sess_type = session._v_attrs['type']
                pattern = platform + '_' + sess_type
                plot_boundaries(*rds,**{'c':'y','alpha':0.5,'lw':4})
                title = pattern[:]

            return pattern
        # end of METHOD plot_data

        cnt = 0
        new_tc = True
        for item, ax in self.get_plot(clust_list):
            rat, day, maze, tc, quality = item
            pattern = plot_data((rat,day,maze),tc,ax)
            if new_tc:
                self.out('Generating plots for rat %d %s...' % (rat,tc))
                plt.ylabel( '%d - %s' % (rat,tc), size='small' )
            quicktitle( ax, 'd%dm%d, %s, %s' % (day,maze,quality,pattern), \
                size='xx-small' )
            cnt += 1
            if cnt < len(clust_list):
                next_tc = ( clust_list[cnt][0], clust_list[cnt][3] )
                if next_tc != (rat,tc):
                    next_tc = (rat,tc)
                    new_tc = True
                    if not self.lastcol:
                        self.col = self.ncols-1
                else:
                    new_tc = False
            #plt.subplots_adjust(hspace=1)
    # end of METHOD process_data

    def _get_clust_with_field(self):
        tdb_file = get_tdb_file()
        field_table = tdb_file.root.spatial.fields
        query = '(area=="CA1")&(quality!="Poor")&(field_no==1)'

        with_field_list = []
        for field in field_table.where(query):
            tc = 't%dc%d'%(field['tt'],field['cl'])
            with_field_list.append( \
                (field['rat'],field['day'],field['session'],tc) )

        tdb_file.close()

        return with_field_list
    # end of METHOD 'collect_rat_clust_with_field'

# end of CLASS SingleLineClustReport



class SelClustReport(BaseReport):
    datadir = './SelClustMap'
    label = 'select cluster rate map'

    ncols = 4
    nrows = 5
    figwidth = 8.0
    figheight = 11.0

    def collect_data(self):
        self.results = {}

        plt_types = ['spike', 'spike_bd','rate','rate_bd']
        self.results['plt_types'] = plt_types

        cl_list = [ ((577,8,1),'t2c2'), ((577,8,2),'t1c9'), ((580,21,1),'t1c7'), \
                    ((576,4,3),'t3c1'), ((579,28,2),'t2c2'), ((576,9,1),'t6c5'), \
                    ((579,28,1),'t2c1'), ((740,5,1),'t14c8'), ((710,18,1),'t10c3'),\
                    ((740,6,2),'t13c10'), ((746,5,2),'t10c3'), ((740,5,2),'t13c9'),\
                    ((740,5,4),'t4c6'), ((710,17,4),'t8c11'), ((710,18,1),'t8c4'),\
                    ((740,5,2),'t5c4'), ((740,5,2),'t4c5') ]

        data_list = []
        for rds, tc in cl_list:
            for plt_type in plt_types:
                data_list.append( (rds, tc, plt_type) )
        self.results['data_list'] = data_list
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self):
        data_list = self.results['data_list']
        plt_types = self.results['plt_types']

        def plot_data(rds, cl, ax, plt_type):
            session = get_group(rds=rds)
            fs = session._v_attrs['tracking_sample_rate']
            x, y, ts = \
                [ session._v_children[key] for key in ['x','y','t_pos'] ]
            t_spk = session._v_children[cl]
            x_spk, y_spk = locate_spike(ts,x,y,t_spk)

            if 'spike' in plt_type:
                plot_traj_and_spike(ax,rds,x,y,x_spk,y_spk)
                if 'bd' in plt_type:
                    plot_boundaries(*rds,**{'c':'y','alpha':0.5,'lw':4})
            elif 'rate' in plt_type:
                plot_ratemap(ax,session,t_spk,rds)
                if 'bd' in plt_type:
                    plot_boundaries(*rds,**{'c':'r','alpha':0.5,'lw':1})
            plt.tick_params(axis='both',labelsize=6)
        # end of METHOD plot_data

        old_rds = None

        for item, ax in self.get_plot(data_list):
            rds, cl, plt_type = item
            if old_rds != rds:
                self.out('Generating plots for rat %d day %d session %d...' % rds)
                old_rds = rds
            self.out('%s - %s' % (cl,plt_type))
            plot_data(rds,cl,ax,plt_type)
            plt.axis([-70,70,-70,70])
            ax.set_axis_off()
            if self.col == 0:
                plt.title( 'rat %d, day %d, session %d, %s' \
                    % (rds+(cl,)), fontsize=7, va='top' )
            plt.subplots_adjust(wspace=0.05, hspace=0.05)
    # end of METHOD process_data

# end of CLASS SelClustReport

class SelClustXSessionReport(BaseReport):
    datadir = './SelClustSess'
    label = 'select cluster sessions'

    ncols = 4
    nrows = 5
    figwidth = 8.0
    figheight = 11.0

    def collect_data(self):
        self.results = {}

        cl_list = [ (740,5,'t13c9'), (740,5,'t13c4'), (710,18,'t10c2'), \
                    (740,5,'t14c3'), (746,5,'t15c21'), (710,17,'t8c1'), \
                    (710,17,'t15c2'), (740,6,'t5c3') ]

        data_list = []
        for rat, day, tc in cl_list:
            for maze in get_maze_list(rat, day):
                data_list.append( ((rat,day,maze),tc) )

        self.results['data_list'] = data_list
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self,plt_type='spike'):
        data_list = self.results['data_list']

        def plot_data(rds, cl, ax):
            session = get_group(rds=rds)
            fs = session._v_attrs['tracking_sample_rate']
            x, y, ts = \
                [ session._v_children[key] for key in ['x','y','t_pos'] ]
            t_spk = session._v_children[cl]
            x_spk, y_spk = locate_spike(ts,x,y,t_spk)

            if plt_type == 'spike':
                plot_traj_and_spike(ax,rds,x,y,x_spk,y_spk)
                plot_boundaries(*rds,**{'c':'y','alpha':0.5,'lw':4})
            elif plt_type == 'rate':
                plot_ratemap(ax,session,t_spk,rds)
                plot_boundaries(*rds,**{'c':'r','alpha':0.5,'lw':1})
            plt.tick_params(axis='both',labelsize=6)
        # end of METHOD plot_data

        old_rds = None

        for item, ax in self.get_plot(data_list):
            rds, cl = item
            if old_rds != rds:
                self.out('Generating plots for rat %d day %d session %d...' % rds)
                old_rds = rds
            self.out('%s' % cl)
            plot_data(rds,cl,ax)
            plt.axis([-70,70,-70,70])
            ax.set_axis_off()
            if (plt_type=='spike') and (self.col==0):
                plt.title( 'rat %d, day %d, session %d, %s' \
                    % (rds+(cl,)), fontsize=7, va='top' )
            plt.subplots_adjust(wspace=0.05, hspace=0.05)
    # end of METHOD process_data

# end of CLASS SelClustReport

def plot_select_cluster(board='simple'):
    plt.close('all')
    
    plt_types = ['spike','rate','rate_bd']
    if board == 'complex':
    #    cl_list = [ ((577,8,1),'t2c2'), ((577,8,2),'t1c9'), ((580,21,1),'t1c7'), \
    #                ((576,4,3),'t3c1'), ((579,28,2),'t2c2'), ((576,9,1),'t6c5'), \
    #                ((579,28,1),'t2c1') ]
        cl_list = [ ((740,3,3),'t4c11'), ((637,5,3),'t2c4'), ((576,6,2),'t6c5'), \
                    ((576,9,1),'t6c4'), ((579,28,1),'t5c3'), ((579,28,1),'t3c3'), \
                    ((576,4,3),'t3c1'), ((637,5,3),'t10c6'), ((576,9,1),'t6c5'), \
                    ((577,8,1),'t1c6'), ((740,4,3),'t2c2'), ((579,29,1),'t2c6') ]
    elif board == 'complex_full':
        cl_list = [ ((740,3,3),'t4c11'), ((637,5,3),'t2c4'), ((579,28,1),'t2c1'), \
                    ((576,6,2),'t6c5'), ((576,9,1),'t6c4'), ((577,7,1),'t1c1'), \
                    ((740,3,3),'t4c5'), ((576,8,1),'t3c2'), ((579,28,1),'t5c3'), \
                    ((579,28,1),'t3c3'), ((576,4,3),'t3c1'), ((637,5,3),'t10c6'), \
                    ((576,9,1),'t6c5'), ((577,8,1),'t1c6'), ((740,4,3),'t2c2'), \
                    ((579,29,1),'t2c6'), ((740,4,3),'t2c4'), ((576,5,3),'t3c10') ]
    elif board == 'simple':
        cl_list = [ ((740,5,1),'t14c8'), ((710,18,1),'t10c3'),\
                    ((740,6,2),'t13c10'), ((746,5,2),'t10c3'), ((740,5,2),'t13c9'),\
                    ((740,5,4),'t4c6'), ((710,17,4),'t8c11'), ((710,18,1),'t8c4'),\
                    ((740,5,2),'t5c4'), ((740,5,2),'t4c5') ]
    N = len(cl_list)

    data_list = []
    for rds, tc in cl_list:
        for plt_type in plt_types:
            data_list.append( (rds, tc, plt_type) )

    plt_cnt = 1
    for rds, cl, plt_type in data_list:
        session = get_group(rds=rds)
        x, y, ts = \
            [ session._v_children[key] for key in ['x','y','t_pos'] ]
        t_spk = session._v_children[cl]
        x_spk, y_spk = locate_spike(ts,x,y,t_spk)

        ax = plt.subplot(N,3,plt_cnt)
        if 'spike' in plt_type:
            plot_traj_and_spike(ax,rds,x,y,x_spk,y_spk)
        elif 'rate' in plt_type:
            #plot_ratemap(ax,session,t_spk,rds)
            plot_ratemap(ax,session,t_spk,rds,smooth='gaussian')
            if 'bd' in plt_type:
                plot_boundaries(*rds,**{'c':'r','alpha':0.5,'lw':1})

        ax.axis([-70,70,-70,70])
        ax.set_aspect('equal')
        ax.set_axis_off()
        if (plt_cnt%3) == 1:
            ttl = ax.set_title( 'cell %d: rat %d, day %d, session %d, %s' \
                % ((plt_cnt/3+1,)+rds+(cl,)),fontsize=8,fontname='Arial',loc='left')
            if board == 'complex':
                ttl.set_position([0,0.95])
            elif board == 'complex_full':
                ttl.set_position([0,0.95])
            elif board == 'simple':
                ttl.set_position([0,0.92])

        plt_cnt += 1

    #plt.subplots_adjust(left=0.01,right=0.99,bottom=0.01,top=0.95,wspace=0.02, hspace=0.05)
    plt.subplots_adjust(left=0.01,right=0.99,bottom=0.01,top=0.98,wspace=0.02, hspace=0.1)

    f = plt.gcf()
    if board == 'complex':
        #f.set_size_inches(3,8,forward=True)
        #fn = './complex.png'
        f.set_size_inches(2,9.5,forward=True)
        fn = './complex.png'
    elif board == 'complex_full':
        f.set_size_inches(2,15,forward=True)
        fn = './complex_full.png'
    elif board == 'simple':
        f.set_size_inches(2,8,forward=True)
        fn = './simple.png'
    f.savefig(fn,dpi=1000)

# end of CLASS SelClustReport

class QualifiedClustReport(BaseReport):
    datadir = './QualClust'
    label = 'qualified cluster'

    ncols = 3 
    nrows = 5

    def collect_data(self, repeat_filter=None):
        self.results = {}
        
        pc_list = self._get_pc_list()
        plt_types = ['spike','rate','rate_bd']

        data_list = []
        if repeat_filter is None:
            rds_list = get_panel_session_list('deStijl')
            clust_list = []
            for rds in rds_list:
                session = get_group(rds=rds)
                tc_list = get_session_clusters(session)
                for tt, cl in tc_list:
                    clust_list.append( rds+('t%dc%d'%(tt,cl),) )
        elif repeat_filter == 'highest_mean':
            full_list = get_panel_clust_list('deStijl',repeat_filter=repeat_filter)
            clust_list = list( set(zip(*zip(*full_list)[:-1])) ) 

        for clust in clust_list:
            if clust not in pc_list:
                continue
            for plt_type in plt_types:
                data_list.append( clust+(plt_type,) )

        self.out( \
            '%d place cell(s) collected from the complex board' % len(data_list) )

        self.results['data_list'] = data_list
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self):
        data_list = self.results['data_list']

        cnt = 1 
        for data, ax in self.get_plot(data_list):
            rat, day, maze, tc, plt_type = data
            rds = (rat, day, maze)
            session = get_group(rds=rds)
            x, y, ts = \
                [ session._v_children[key] for key in ['x','y','t_pos'] ]
            t_spk = session._v_children[tc]
            x_spk, y_spk = locate_spike(ts,x,y,t_spk)

            if 'spike' in plt_type:
                plot_traj_and_spike(ax,rds,x,y,x_spk,y_spk)
            elif 'rate' in plt_type:
                plot_ratemap(ax,session,t_spk,rds)
                if 'bd' in plt_type:
                    plot_boundaries(*rds,**{'c':'r','alpha':0.5,'lw':1})

            ax.axis([-70,70,-70,70])
            ax.set_aspect('equal')
            ax.set_axis_off()

            if plt_type == 'spike':
                ttl = ax.set_title( 'cell %d: rat %d, day %d, session %d, %s' \
                    % ((cnt/3+1,)+rds+(tc,)) )

            cnt += 1

    # end of METHOD process_data
    
    def _get_pc_list(self):
        tdb_file = get_tdb_file()
        field_table = tdb_file.root.spatial.fields
        query = '(area=="CA1")&(quality=="Fair")&(field_no==1)'

        pc_list = []
        for field in field_table.where(query):
            tc = 't%dc%d'%(field['tt'],field['cl'])
            pc_list.append( (field['rat'],field['day'],field['session'],tc) )

        tdb_file.close()

        return pc_list
    # end of METHOD _get_pc_list

# end of CLASS SelClustReport
