# encoding: utf-8
"""
population_field.py

population field analyses for Single Line platforms
"""

import os
import sys
import itertools

import tables as tb
import numpy as np
import matplotlib.pyplot as plt

from collections import Counter
from scipy.ndimage.filters import gaussian_filter
from scipy.signal import medfilt
from scipy.stats import ks_2samp, pearsonr, norm

from core.analysis import AbstractAnalysis
from tools.stats import pvalue

from tdb.data import get_tdb_file, get_group
from tdb.meta import walk_mazes, walk_tetrodes
from tdb.td_maps import bin_edges, get_coordinate_range, plot_boundaries, zone_filter, get_shuffled_zone_filter, plot_boundary_window, gaussian_template
from tdb.field import label_place_fields

from analysis_shan.misc import plot_resample_distribution

#=================================================================================

class PopFieldAna(AbstractAnalysis):
    datadir = './PopField'
    label = 'population field'

    def collect_data(self, ft='edge', **traits):
        self.results = {}

        H = {} # population map (summation of field binary maps)
        B = {} # stack of field binary maps 
        N = {} # number of fields

        # pattern: 'platform_session\ type', panel: other restrictions
        # single line
        platforms = ['leather', 'tape']
        stypes = ['STD', 'SHF']
        sl = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]
        # single line + deStijl
        tex, pla = 'TEX', 'square_PLA'
        panels = sl + ['deStijl', 'plain-all', 'plain-before', 'plain-after']
        patterns = sl + [tex, pla, pla, pla]

        for panel, pattern in zip(panels, patterns):
            H[panel], B[panel], N[panel] = self.collect_panel( \
                panel, pattern, ft=ft, **traits )
            
        self.results['H'] = H
        self.results['B'] = B
        self.results['N'] = N
        self.results['ft'] = ft
        self.results['panels'] = panels
        self.results['patterns'] = patterns

        self.out('Done!')
    # end of METHOD collect_data

    def plot_paper(self,smooth=True,norm=True,cmap='jet',control=None,**traits):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        self.figure = {}

        fg_cnt = 1
        for platform in ['leather', 'tape']:
            for stype in ['STD', 'SHF']:
                panel = platform + '_' + stype
                [y_edges, x_edges] = bin_edges(pattern=panel,**traits)

                ax = plt.subplot(1,4,fg_cnt)
                panel_h = panel if control is None else control
                H = self.results['H'][panel_h].copy().astype(float)
                if norm:
                    H /= float(self.results['N'][panel])
                if smooth:
                    H = gaussian_filter(H,2.5)
                ax.imshow( H,origin='lower',interpolation='none', \
                    aspect='equal',cmap=cmap,\
                        extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )

                bd_c = 'k'
                combine_kw = traits.copy()
                combine_kw.update( **{'c':bd_c,'alpha':0.5,'lw':2} )
                plot_boundaries( pattern=panel, **combine_kw )

                bd_c = 'k'
                combine_kw = traits.copy()
                combine_kw.update( **{'c':bd_c,'alpha':0.5,'lw':1} )
                plot_boundary_window( pattern=panel, edge_margin=10, **combine_kw )

                stype_str = 'Standard' if stype=='STD' else 'Shift'
                title_str = platform.capitalize() + '-' + stype_str
                ax.set_title(title_str,fontsize=12,fontname='Arial')
                plt.axis( [x_edges[0], x_edges[-1], y_edges[0], y_edges[-1]] )
                ax.set_axis_off()
                ax.set_aspect('equal')

                fg_cnt += 1

        plt.subplots_adjust(left=0.06,right=0.97,bottom=0.15,top=0.93,wspace=0.3)

        f = plt.gcf()
        f.set_size_inches(4.4,1,forward=True)
        f.savefig( os.path.join(self.datadir,'heatmap.png'),dpi=1000 )

        self.figure['paper'] = plt.gcf()
    # end of METHOD plot_paper

    def plot_paper_BPI(self,ctrl=None,test_type='boot_strap',xlim=[-1,1]):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        platforms = ['leather', 'tape']
        stypes = ['STD','SHF']
        panels = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]

        ttype = 'bootstrap' if test_type=='boot_strap' else 'permutation'
        # for correlation bootstrap (BCa)
        # tails = [0.4984, 0.2128, 0.2635, -0.2451]

        for i, panel in enumerate(panels):
            ax = plt.subplot(1,4,i+1)

            if ttype == 'bootstrap':
                bs_ptr = self.results['boot_strap'][panel]
                obs_ctrl = bs_ptr[ctrl][ 'obs_'+ctrl ]
                bs_ctrl = bs_ptr[ctrl][ 'bs_'+ctrl ]
                obs_exp = bs_ptr[panel]['obs']
                bs_exp = bs_ptr[panel]['bs']
                obs = obs_exp - obs_ctrl
                rs = bs_exp - bs_ctrl
            else:
                ptr = self.results['shuffle'][panel]
                obs = ptr[ 'obs_'+panel ]
                rs = ptr[ 'shuffle_'+panel ]
#            plot_resample_distribution( ax,obs,rs,exp=0,nbin=15, \
#                resample=ttype,xlim=xlim,one_side=True,multi_n=4,tail=tails[i] )
            plot_resample_distribution( ax,obs,rs,exp=0,nbin=15, \
                resample=ttype,xlim=xlim,one_side=True,multi_n=4 )

            bins = np.linspace(xlim[0],xlim[1],3)
            ax.set_xticks(bins)
            ax.set_xticklabels(bins,fontname='Arial',fontsize=8)

            ymin, ymax = ax.get_ylim()
            bins = np.round(np.linspace(ymin,ymax,3),0).astype(int)
            ax.set_yticks(bins)
            ax.set_yticklabels(bins,fontname='Arial',fontsize=8)

        plt.subplots_adjust(left=0.06,right=0.97,bottom=0.15,top=0.93,wspace=0.3)

        f = plt.gcf()
        f.set_size_inches(4.4,1,forward=True)
        fn = 'BPI_' + ttype + '.png'
        f.savefig( os.path.join(self.datadir,fn),dpi=1000 )
    # end of METHOD plot_paper_BPI

    def plot_pop_field_heatmap(self,smooth=True,norm=True,bd=True,cmap='jet',edge_margin=0,bs_bd=False,**traits):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        panels = self.results['panels']
        patterns = self.results['patterns']

        for i, panel in enumerate(panels):
            [y_edges, x_edges] = bin_edges(pattern=patterns[i],**traits)

            H = self.results['H'][panel].copy().astype(float)
            if norm:
                H /= float(self.results['N'][panel])
            if smooth:
                H = gaussian_filter(H,2.5)

            plt.subplot(2,4,i+1)
            plt.imshow( H,origin='lower',interpolation='none', \
                aspect='auto',cmap=cmap,\
                    extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )
            # cmap - jet/cool/gray_r
            if bd:
                bd_c = 'r' if cmap=='gray_r' else 'k'
                combine_kw = traits.copy()
                combine_kw.update( **{'c':bd_c,'alpha':0.5,'lw':5} )
                plot_boundaries( \
                    pattern=patterns[i], **combine_kw )
            if bs_bd:
                bd_c = 'r' if cmap=='gray_r' else 'k'
                combine_kw = traits.copy()
                combine_kw.update( **{'c':bd_c,'alpha':0.5,'lw':5} )
                plot_boundary_window( \
                    pattern=patterns[i], edge_margin=edge_margin, **combine_kw )
            plt.axis( [x_edges[0], x_edges[-1], y_edges[0], y_edges[-1]] )
            plt.title( panel )
            plt.tick_params(axis='both',labelsize=7)

        fname = 'c-' + cmap + '.png'
        if bd:
            fname = 'bd_' + fname
        if smooth:
            fname = 'smooth_' + fname
        if bs_bd:
            fname = 'window_' + fname

        fig = plt.gcf()
        fig.set_size_inches(20,10,forward=True)
        fig.savefig( os.path.join(self.datadir,fname) )
    # end of METHOD plot_pop_field_heatmap

    def plot_bootstrap_SL_panels(self,ctrl=None,**traits):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        self.out.timestamp = False
        self.log = False
        self.log = True

        # For SL sessions, panel == pattern
        panels = [ panel for panel in self.results['panels'] \
            if (not 'plain' in panel and not 'deStijl' in panel) ]

        if ctrl is None:
            for i, bs_panel in enumerate(panels):
                plt.subplot(2,2,i+1)
                self.bootstrap( bs_panel, ax=plt.gca(), **traits)
            plt.suptitle('SL Bootstrap', fontsize=9)
        else:
            ctrl_pairs = zip( [False,True,True], [False,False,True] )
            for i, bs_panel in enumerate(panels):
                for j, ctrl_pair in enumerate(ctrl_pairs):
                    plt.subplot(3,4,j*4+i+1)
                    ax = plt.gca()
                    ctrl_switch, ctrl_diff = ctrl_pair
                    self.bootstrap( bs_panel, ctrl_panel=ctrl, ctrl=ctrl_switch,\
                        ctrl_diff=ctrl_diff, ax=ax, **traits )
            plt.suptitle('SL Bootstrap (%s)'%ctrl, fontsize=24)

        fname = 'bootstrap'
        if ctrl is not None:
            fname += '_%s' % ctrl
        fname += '.png'

        f = plt.gcf()
        f.set_size_inches(20,15,forward=True)
        f.savefig( os.path.join(self.datadir,fname) )
    # end of METHOD plot_bootstrap_SL_panels

    def plot_bootstrap_corr_SL_panels(self,ctrl=None,**traits):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        self.out.timestamp = False
        self.log = False
        self.log = True

        # For SL sessions, panel == pattern
        panels = [ panel for panel in self.results['panels'] \
            if (not 'plain' in panel and not 'deStijl' in panel) ]

        if ctrl is None:
            for i, bs_panel in enumerate(panels):
                plt.subplot(2,2,i+1)
                self.bootstrap_corr( bs_panel, ax=plt.gca(), **traits)
            plt.suptitle('SL Bootstrap Corr', fontsize=9)
        else:
            ctrl_pairs = zip( [False,True,True], [False,False,True] )
            for i, bs_panel in enumerate(panels):
                for j, ctrl_pair in enumerate(ctrl_pairs):
                    plt.subplot(3,4,j*4+i+1)
                    ax = plt.gca()
                    ctrl_switch, ctrl_diff = ctrl_pair
                    self.bootstrap_corr( bs_panel, ctrl_panel=ctrl, \
                        ctrl=ctrl_switch, ctrl_diff=ctrl_diff, ax=ax, **traits )
            plt.suptitle('SL Bootstrap Corr (%s)'%ctrl, fontsize=24)

        fname = 'bootstrap_corr'
        if ctrl is not None:
            fname += '_%s' % ctrl
        fname += '.png'

        f = plt.gcf()
        f.set_size_inches(20,15,forward=True)
        f.savefig( os.path.join(self.datadir,fname) )
    # end of METHOD plot_bootstrap_corr_SL_panels

    def bootstrap(self, bs_panel, ctrl_panel=None, ctrl=False, ctrl_diff=False, nbin=30, plot_out=True, ax=None, sig='*', one_side=True, corr=True, **traits):

        self.out('\n')
        if not ctrl:
            self.out('Bootstrap data for %s sessions...' % bs_panel)
        elif ctrl_diff:
            self.out('Bootstrap difference between %s and %s...' \
                % (bs_panel,ctrl_panel) )
        else:
            self.out('Bootstrap control data (%s) for %s sessions...' \
                % (ctrl_panel,bs_panel) )
        self.out('\n')

        if ctrl_diff:
            title_str = 'PI Difference'
        else: 
            title_str = bs_panel if not ctrl else 'Control'  

        #obs_bs, bs_bs = self.get_bootstrap_data( bs_panel,bs_panel,**traits )
        obs_bs, bs_bs = self.get_bootstrap_auc( bs_panel,bs_panel,**traits )
        if ctrl_panel is not None:
           #obs_ctrl, bs_ctrl = self.get_bootstrap_data( \
           #     ctrl_panel,bs_panel,**traits )
           obs_ctrl, bs_ctrl = self.get_bootstrap_auc( \
                ctrl_panel,bs_panel,**traits )
        if ctrl_diff:
            obs = obs_bs - obs_ctrl
            bs = bs_bs - bs_ctrl
            x_min, x_max = min(min(obs,0),min(bs)), max(max(obs,0),max(bs))
        else:
            x_min = np.min( [0,obs_bs,obs_ctrl,bs_bs.min(),bs_ctrl.min()] )
            x_max = np.max( [0,obs_bs,obs_ctrl,bs_bs.max(),bs_ctrl.max()] )
            if ctrl:
                obs, bs = obs_ctrl, bs_ctrl
            else:
                obs, bs = obs_bs, bs_bs

        # output stats
        if corr:
            levels = np.array( [0.5/4.0, 1.0/4.0, 2.5/4.0, 5.0/4.0, 100-5.0/4.0, 100-2.5/4.0, 100-1/4.0, 100-0.5/4.0] )
        else:
            levels = np.array( [0.5, 1.0, 2.5, 5.0, 95.0, 97.5, 99.0, 99.5] )
        b5, b10, b25, b50, b950, b975, b990, b995 = \
            ( np.percentile(bs,level) for level in levels ) 
        self.out('Percentiles: B_5=%.2f, B_10=%.2f, B_25=%.2f, B_50=%.2f, B_950=%.2f, B_975=%.2f, B_990=%.2f, B_995=%.2f' % (b5,b10,b25,b50,b950,b975,b990,b995) )
        self.out('Observed boundary preference index = %.4f' % obs)
        self.out('%d out of %d (%.4f) bootstrap samples <= observed BPI' \
            % (np.sum(bs<=obs),len(bs),np.sum(bs<=obs)/float(len(bs))))
        self.out('\n')
        self.out('==== Centered Percentile Bootstrap ====')
        self.out('One-sided:')
        self.out('CI(0.05): (%.4f, ]' %(2*obs-b950))
        self.out('CI(0.01): (%.4f, ]' %(2*obs-b990))
        self.out('Two-sided:')
        self.out('CI(0.05): (%.4f, %.4f)' %(2*obs-b975,2*obs-b25))
        self.out('CI(0.01): (%.4f, %.4f)' %(2*obs-b995,2*obs-b5))
        self.out('\n')
        self.out('==== Percentile Bootstrap ====')
        self.out('One-sided:')
        self.out('CI(0.05): (%.4f, ]' %b50)
        self.out('CI(0.01): (%.4f, ]' %b10)
        self.out('Two-sided:')
        self.out('CI(0.05): (%.4f, %.4f)' %(b25,b975))
        self.out('CI(0.01): (%.4f, %.4f)' %(b5,b995))
        self.out('\n')

        # BCa
        if sig == '*':
            alpha = 0.05
        elif sig == '**':
            alpha = 0.01
        if not one_side:
            alpha /= 2.0
        if corr:
            alpha /= 4.0

        if ctrl_diff:
            obs_ctrl = self.jackknife_obs(ctrl_panel,bs_panel,**traits)
            obs_bs = self.jackknife_obs(bs_panel,bs_panel,**traits)
            obs_llo = []
            for i in obs_ctrl:
                for j in obs_bs:
                    obs_llo.append( j-i )
        else:
            if ctrl:
                obs_llo = self.jackknife_obs(ctrl_panel,bs_panel,**traits)
            else:
                obs_llo = self.jackknife_obs(bs_panel,bs_panel,**traits)
        diff_llo = np.mean(obs_llo) - obs_llo
        acc = np.sum(diff_llo**3) / (6.0*(np.sum(diff_llo**2)**1.5))  

        z_0 = norm.ppf( np.sum(bs<obs)/float(len(bs)) )
        alpha_1 = norm.cdf( z_0 + \
            (z_0+norm.ppf(alpha))/(1-acc*(z_0+norm.ppf(alpha))) )
        alpha_2 = norm.cdf( z_0 + \
            (z_0+norm.ppf(1-alpha))/(1-acc*(z_0+norm.ppf(1-alpha))) )
        lower = np.percentile(bs,alpha_1*100)
        upper = np.percentile(bs,alpha_2*100)

        self.out('==== BCa ====')
        if one_side:
            self.out('CI: (%.4f, ]' %lower)
        else:
            self.out('CI: (%.4f, %.4f)' %(lower,upper))

        # output bootstrap distribution histogram
        if plot_out:
            x_range = x_max - x_min
            bins = np.linspace(x_min-0.1*x_range, x_max+0.1*x_range, nbin+1)
            if ax is None:
                ax = plt.gca()
            ax.hist(bs,bins=bins,color='cyan')
            plt.axvline(x=obs,c='r',lw=2)
            plt.axvline(x=0,c='k',lw=2)
            #if sig == '**':
            #    if one_side:
            #        cut = (b950, )
            #    else:
            #        cut = (b975, b25)
            #elif sig == '***':
            #    if one_side:
            #        cut = (b990, )
            #    else:
            #        cut = (b995, b5)
            if one_side:
                cut = (lower, )
            else:
                cut = (lower, upper)
            for c in cut:
                #plt.axvline(x=2*obs-c,c='g',lw=2)
                plt.axvline(x=c,c='g',lw=2)
            plt.title(title_str,fontsize=16)
            plt.xlim( (x_min-0.1*x_range,x_max+0.1*x_range) )
            plt.tick_params(axis='both',labelsize=10)

    # end of METHOD bootstrap

    def bootstrap_corr(self, bs_panel, ctrl_panel=None, ctrl=False, ctrl_diff=False, nbin=30, plot_out=True, ax=None, sig='*', one_side=True, corr=True, **traits):

        self.out('\n')
        if not ctrl:
            self.out('Bootstrap data for %s sessions...' % bs_panel)
        elif ctrl_diff:
            self.out('Bootstrap difference between %s and %s...' \
                % (bs_panel,ctrl_panel) )
        else:
            self.out('Bootstrap control data (%s) for %s sessions...' \
                % (ctrl_panel,bs_panel) )
        self.out('\n')

        if ctrl_diff:
            title_str = 'PI Difference'
        else: 
            title_str = bs_panel if not ctrl else 'Control'  

        obs_bs, bs_bs = self.get_bootstrap_corr_data( bs_panel,bs_panel,**traits )
        if ctrl_panel is not None:
           obs_ctrl, bs_ctrl = self.get_bootstrap_corr_data( \
                ctrl_panel,bs_panel,**traits )
        if ctrl_diff:
            obs = obs_bs - obs_ctrl
            bs = bs_bs - bs_ctrl
            x_min, x_max = min(min(obs,0),min(bs)), max(max(obs,0),max(bs))
        else:
            x_min = np.min( [0,obs_bs,obs_ctrl,bs_bs.min(),bs_ctrl.min()] )
            x_max = np.max( [0,obs_bs,obs_ctrl,bs_bs.max(),bs_ctrl.max()] )
            if ctrl:
                obs, bs = obs_ctrl, bs_ctrl
            else:
                obs, bs = obs_bs, bs_bs

        # output stats
        if corr:
            levels = np.array( [0.5/4.0, 1.0/4.0, 2.5/4.0, 5.0/4.0, 100-5.0/4.0, 100-2.5/4.0, 100-1/4.0, 100-0.5/4.0] )
        else:
            levels = np.array( [0.5, 1.0, 2.5, 5.0, 95.0, 97.5, 99.0, 99.5] )
        b5, b10, b25, b50, b950, b975, b990, b995 = \
            ( np.percentile(bs,level) for level in levels ) 
        self.out('Percentiles: B_5=%.2f, B_10=%.2f, B_25=%.2f, B_50=%.2f, B_950=%.2f, B_975=%.2f, B_990=%.2f, B_995=%.2f' % (b5,b10,b25,b50,b950,b975,b990,b995) )
        self.out('Observed boundary preference index = %.4f' % obs)
        self.out('%d out of %d (%.4f) bootstrap samples <= observed BPI' \
            % (np.sum(bs<=obs),len(bs),np.sum(bs<=obs)/float(len(bs))))
        self.out('\n')
        self.out('==== Centered Percentile Bootstrap ====')
        self.out('One-sided:')
        self.out('CI(0.05): (%.4f, ]' %(2*obs-b950))
        self.out('CI(0.01): (%.4f, ]' %(2*obs-b990))
        self.out('Two-sided:')
        self.out('CI(0.05): (%.4f, %.4f)' %(2*obs-b975,2*obs-b25))
        self.out('CI(0.01): (%.4f, %.4f)' %(2*obs-b995,2*obs-b5))
        self.out('\n')
        self.out('==== Percentile Bootstrap ====')
        self.out('One-sided:')
        self.out('CI(0.05): (%.4f, ]' %b50)
        self.out('CI(0.01): (%.4f, ]' %b10)
        self.out('Two-sided:')
        self.out('CI(0.05): (%.4f, %.4f)' %(b25,b975))
        self.out('CI(0.01): (%.4f, %.4f)' %(b5,b995))
        self.out('\n')

        # BCa
        if sig == '**':
            alpha = 0.05
        elif sig == '***':
            alpha = 0.01
        if not one_side:
            alpha /= 2.0
        if corr:
            alpha /= 4.0

        if ctrl_diff:
            obs_ctrl = self.jackknife_obs_corr(ctrl_panel,bs_panel,**traits)
            obs_bs = self.jackknife_obs_corr(bs_panel,bs_panel,**traits)
            obs_llo = []
            for i in obs_ctrl:
                for j in obs_bs:
                    obs_llo.append( j-i )
        else:
            if ctrl:
                obs_llo = self.jackknife_obs_corr(ctrl_panel,bs_panel,**traits)
            else:
                obs_llo = self.jackknife_obs_corr(bs_panel,bs_panel,**traits)
        diff_llo = np.mean(obs_llo) - obs_llo
        acc = np.sum(diff_llo**3) / (6.0*(np.sum(diff_llo**2)**1.5))  

        z_0 = norm.ppf( np.sum(bs<obs)/float(len(bs)) )
        alpha_1 = norm.cdf( z_0 + \
            (z_0+norm.ppf(alpha))/(1-acc*(z_0+norm.ppf(alpha))) )
        alpha_2 = norm.cdf( z_0 + \
            (z_0+norm.ppf(1-alpha))/(1-acc*(z_0+norm.ppf(1-alpha))) )
        lower = np.percentile(bs,alpha_1*100)
        upper = np.percentile(bs,alpha_2*100)

        self.out('==== BCa ====')
        if one_side:
            self.out('CI: (%.4f, ]' %lower)
        else:
            self.out('CI: (%.4f, %.4f)' %(lower,upper))

        # output bootstrap distribution histogram
        if plot_out:
            x_range = x_max - x_min
            bins = np.linspace(x_min-0.1*x_range, x_max+0.1*x_range, nbin+1)
            if ax is None:
                ax = plt.gca()
            ax.hist(bs,bins=bins,color='cyan')
            plt.axvline(x=obs,c='r',lw=2)
            plt.axvline(x=0,c='k',lw=2)
            #if sig == '**':
            #    if one_side:
            #        cut = (b950, )
            #    else:
            #        cut = (b975, b25)
            #elif sig == '***':
            #    if one_side:
            #        cut = (b990, )
            #    else:
            #        cut = (b995, b5)
            if one_side:
                cut = (lower, )
            else:
                cut = (lower, upper)
            for c in cut:
                #plt.axvline(x=2*obs-c,c='g',lw=2)
                plt.axvline(x=c,c='g',lw=2)
            plt.title(title_str,fontsize=16)
            plt.xlim( (x_min-0.1*x_range,x_max+0.1*x_range) )
            plt.tick_params(axis='both',labelsize=10)

    # end of METHOD bootstrap

    def jackknife_obs(self, panel, bs_panel, smooth=True, N=None, overwrite=None, **traits):
        """
            bs_panel should be one of the SL panels
            panel is the data being bootstrapped
        """
        H_bool = self.results['B'][panel]
        length = H_bool.shape[2]

        llo = []
        index = np.arange(length)
        index = index[1:] - (index[:,None]>=index[1:])
        for idx in index:
            H = np.sum(H_bool[:,:,idx],2) / float(length-1)
            if smooth:
                H = gaussian_filter(H,2.5)
            llo.append( get_preference_index(H,bs_panel,**traits) )
        
        return np.array( llo )
    # end of METHOD jackknife_obs_corr 

    def jackknife_obs_corr(self, panel, bs_panel, smooth=True, N=None, overwrite=None, **traits):
        """
            bs_panel should be one of the SL panels
            panel is the data being bootstrapped
        """
        H_bool = self.results['B'][panel]
        length = H_bool.shape[2]

        llo = []
        index = np.arange(length)
        index = index[1:] - (index[:,None]>=index[1:])
        for idx in index:
            H = np.sum(H_bool[:,:,idx],2) / float(length-1)
            if smooth:
                H = gaussian_filter(H,2.5)
            llo.append( get_correlation(H,bs_panel,**traits) )
        
        return np.array( llo )
    # end of METHOD jackknife_obs_corr 

    def SL_shuffle_test(self, ctrl, N=1000, nbin=40, overwrite=False, smooth=True, **traits):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        self.out.timestamp = False
        self.log = False
        self.log = True

        update_pickle = False

        if 'shuffle' not in self.results.keys():
            self.results['shuffle'] = {} 
            update_pickle = True

        plt_cnt = 1
        for platform in ['leather', 'tape']:
            for exp_type in ['STD', 'SHF']:
                panel = platform + '_' + exp_type
                self.out('%s sessions:' % panel)

                if panel not in self.results['shuffle'].keys():
                    self.results['shuffle'][panel] = {} 
                    update_pickle = True
                ptr = self.results['shuffle'][panel]
                keys = ptr.keys()
                obs_str, shuffle_str = 'obs_'+panel, 'shuffle_'+panel

                if overwrite or (obs_str not in keys) or (shuffle_str not in keys):
                    H_obs = self.results['B'][panel].copy().astype(float)
                    H_ctrl = self.results['B'][ctrl].copy().astype(float)
                    H = np.dstack( (H_obs,H_ctrl) )
                N_obs = H_obs.shape[2]
                N_ctrl = H_ctrl.shape[2]
                index = np.array( range(N_obs+N_ctrl) )

                if not overwrite and (obs_str in keys):
                    obs = ptr[ obs_str ]
                else:
                    #H_in = np.sum(H_obs,2) / float(N_obs)
                    H_in = np.sum(H_obs,2)
                    if smooth:
                        H_in = gaussian_filter(H_in,2.5)
                    #pi_obs = get_preference_index(H_in,panel,**traits)
                    pi_obs = get_auc( H_in,panel,**traits )
                    #H_in = np.sum(H_ctrl,2) / float(N_ctrl)
                    H_in = np.sum(H_ctrl,2)
                    if smooth:
                        H_in = gaussian_filter(H_in,2.5)
                    #pi_ctrl = get_preference_index(H_in,panel,**traits)
                    pi_ctrl = get_auc( H_in,panel,**traits )
                    obs = pi_obs - pi_ctrl
                    ptr[obs_str] = obs
                    update_pickle = True

                if not overwrite and (shuffle_str in keys):
                    shuffle = ptr[ shuffle_str ]
                else:
                    shuffle = []
                    for i in range(N):
                        shuffle_index = np.random.permutation(index)
                        #H_in = np.sum(H[:,:,shuffle_index[:N_obs]],2)/float(N_obs)
                        H_in = np.sum(H[:,:,shuffle_index[:N_obs]],2)
                        if smooth:
                            H_in = gaussian_filter(H_in,2.5)
                        #pi_obs = get_preference_index(H_in,panel,**traits)
                        pi_obs = get_auc( H_in,panel,**traits )
                        #H_in = np.sum(H[:,:,shuffle_index[N_obs:]],2)/float(N_ctrl)
                        H_in = np.sum(H[:,:,shuffle_index[N_obs:]],2)
                        if smooth:
                            H_in = gaussian_filter(H_in,2.5)
                        #pi_ctrl = get_preference_index(H_in,panel,**traits)
                        pi_ctrl = get_auc( H_in,panel,**traits )
                        shuffle.append( pi_obs-pi_ctrl )
                        ptr[shuffle_str] = shuffle 
                        update_pickle = True

                self.out('Observed preference index difference = %.4f' % obs)
                self.out('%d out of %d (p=%.4f) shuffle samples >= observed PI' \
                    % (np.sum(shuffle>=obs),N,np.sum(shuffle>=obs)/float(N)))

                x_min = min( obs, min(shuffle) )
                x_max = max( obs, max(shuffle) )
                bins = np.linspace(x_min, x_max, nbin+1)
                bin_size = bins[1] - bins[0]

                plt.subplot(2,2,plt_cnt)
                plt.hist(shuffle,bins=bins,color='cyan')
                plt.axvline(x=obs,lw=2,color='r')
                plt.axvline(x=0,lw=2,color='k')
                plt.xlim( [x_min-bin_size, x_max+bin_size] )
                plt.tick_params(axis='both',labelsize=7)
                plt.title('%s - %s' % (platform,exp_type))
                plt.suptitle('Shuffle Test',fontsize=12)

                plt_cnt += 1

        # store shuffle results if updated
        if update_pickle:
            self.save_data()

        fig = plt.gcf()
        fig.set_size_inches(15,16,forward=True)
        fig.savefig( os.path.join(self.datadir,'shuffle.png') )
    # end of METHOD SL_shuffle_test

    def SL_shuffle_corr_test(self, ctrl, N=1000, nbin=40, overwrite=False, smooth=True, **traits):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        self.out.timestamp = False
        self.log = False
        self.log = True

        update_pickle = False

        if 'shuffle' not in self.results.keys():
            self.results['shuffle'] = {} 
            update_pickle = True

        plt_cnt = 1
        for platform in ['leather', 'tape']:
            for exp_type in ['STD', 'SHF']:
                panel = platform + '_' + exp_type
                self.out('%s sessions:' % panel)

                if panel not in self.results['shuffle'].keys():
                    self.results['shuffle'][panel] = {} 
                    updata_pickle = True
                ptr = self.results['shuffle'][panel]
                keys = ptr.keys()
                obs_str, shuffle_str = 'obs_'+panel, 'shuffle_'+panel

                if overwrite or (obs_str not in keys) or (shuffle_str not in keys):
                    H_obs = self.results['B'][panel].copy().astype(float)
                    H_ctrl = self.results['B'][ctrl].copy().astype(float)
                    H = np.dstack( (H_obs,H_ctrl) )
                N_obs = H_obs.shape[2]
                N_ctrl = H_ctrl.shape[2]
                index = np.array( range(N_obs+N_ctrl) )

                if not overwrite and (obs_str in keys):
                    obs = ptr[ obs_str ]
                else:
                    H_in = np.sum(H_obs,2) / float(N_obs)
                    if smooth:
                        H_in = gaussian_filter(H_in,2.5)
                    pi_obs = get_correlation(H_in,panel,**traits)
                    H_in = np.sum(H_ctrl,2) / float(N_ctrl)
                    if smooth:
                        H_in = gaussian_filter(H_in,2.5)
                    pi_ctrl = get_correlation(H_in,panel,**traits)
                    obs = pi_obs - pi_ctrl
                    ptr[obs_str] = obs
                    update_pickle = True

                if not overwrite and (shuffle_str in keys):
                    shuffle = ptr[ shuffle_str ]
                else:
                    shuffle = []
                    for i in range(N):
                        shuffle_index = np.random.permutation(index)
                        H_in = np.sum(H[:,:,shuffle_index[:N_obs]],2)/float(N_obs)
                        if smooth:
                            H_in = gaussian_filter(H_in,2.5)
                        pi_obs = get_correlation(H_in,panel,**traits)
                        H_in = np.sum(H[:,:,shuffle_index[N_obs:]],2)/float(N_ctrl)
                        if smooth:
                            H_in = gaussian_filter(H_in,2.5)
                        pi_ctrl = get_correlation(H_in,panel,**traits)
                        shuffle.append( pi_obs-pi_ctrl )
                        ptr[shuffle_str] = shuffle 
                        update_pickle = True

                self.out('Correlation coefficient difference = %.4f' % obs)
                self.out('%d out of %d (p=%.4f) shuffle samples >= observed PI' \
                    % (np.sum(shuffle>=obs),N,np.sum(shuffle>=obs)/float(N)))

                x_min = min( obs, min(shuffle) )
                x_max = max( obs, max(shuffle) )
                bins = np.linspace(x_min, x_max, nbin+1)
                bin_size = bins[1] - bins[0]

                plt.subplot(2,2,plt_cnt)
                plt.hist(shuffle,bins=bins,color='cyan')
                plt.axvline(x=obs,lw=2,color='r')
                plt.axvline(x=0,lw=2,color='k')
                plt.xlim( [x_min-bin_size, x_max+bin_size] )
                plt.tick_params(axis='both',labelsize=7)
                plt.title('%s - %s' % (platform,exp_type))
                plt.suptitle('Shuffle Corr Test',fontsize=12)

                plt_cnt += 1

        # store shuffle results if updated
        if update_pickle:
            self.save_data()

        fig = plt.gcf()
        fig.set_size_inches(15,16,forward=True)
        fig.savefig( os.path.join(self.datadir,'shuffle_corr.png') )
    # end of METHOD SL_shuffle_corr_test

    #-----------------------------------------------------------------------------

    def collect_panel( self, panel, pattern, ft='edge', min_quality='Marginal', repeat_filter=None, **traits):
        self.out('Collecting data for %s...' % panel)
        clust_list = collect_clust_list( \
            pattern,panel,min_quality=min_quality,repeat_filter=repeat_filter)
        if not len(clust_list):
            self.out('No data collected!')
            return None, 0
        
        self.out('Detecting place field contours...')
        N_field, H, H_bool = 0, None, None
        for i, clust in enumerate(clust_list):
            if np.mod(i,10) == 0:
                sys.stdout.write('.')
                sys.stdout.flush()
            rat, day, maze, tc, threshold = clust
            rds = (rat, day, maze)
            session = get_group(rds=rds)
            spikes = session._v_children[tc]
            #field_num, labels, contours, n_edges, tmp = label_place_fields( \
            #    rds,session,spikes,threshold=threshold,turn_off_rim=False,\
            #        smooth=True,smooth_kernel='gaussian',**traits )
            field_num, labels, contours, n_edges, tmp = label_place_fields( \
               rds,session,spikes,threshold=threshold,turn_off_rim=False,**traits )
            if ft == 'edge':
                unify_map = (contours>0).astype(int)
            elif ft == 'face':
                unify_map = (labels>0).astype(int)

            if H is None:
                H = unify_map.copy()
                H_bool = np.zeros( H.shape, dtype=bool ) # H[:,:,0] = False matrix
            else:
                H += unify_map
            for num in range(field_num):
                if ft == 'edge':
                    H_bool = np.dstack( (H_bool,contours==num+1) )
                elif ft == 'face':
                    H_bool = np.dstack( (H_bool,labels==num+1) )
            N_field += field_num

            if np.sum( abs(H-np.sum(H_bool,2)) ) > 0:
                print 'different!'
                print rds, tc
                print field_num, contours.max(), labels.max()
                print H.max()
                print np.sum(H_bool,2).max()
        H_bool = H_bool[:,:,1:] # remove the False matrix
        sys.stdout.write('\n')

        rats, days, mazes, tcs, tmp = zip(*clust_list)
        N_rat = len( list(set(rats)) )
        N_dataset = len( list(set(zip(rats,days))) )
        N_session = len( list(set(zip(rats,days,mazes))) )
        N_clust = len(clust_list)
        self.out('Data collected from %d rats, %d datasets, %d sessions, %d clusters and %d fields' % (N_rat,N_dataset,N_session,N_clust,N_field))

        return H, H_bool, N_field
    # end of METHOD collect_panel

    def get_bootstrap_data(self, panel, bs_panel, overwrite=False, N=1000, smooth=True, **traits):
        """
            bs_panel should be one of the SL panels
            panel is the data being bootstrapped
        """
        if 'boot_strap' not in self.results.keys():
            self.results['boot_strap'] = {} 
        if bs_panel not in self.results['boot_strap'].keys():
            self.results['boot_strap'][bs_panel] = {} 
        if panel not in self.results['boot_strap'][bs_panel].keys():
            self.results['boot_strap'][bs_panel][panel] = {}
        bs_ptr = self.results['boot_strap'][bs_panel][panel]
        #bs_ptr = self.results.get('boot_strap',{}).get(bs_panel,{})


        H_bool = self.results['B'][panel]
        length = H_bool.shape[2]
        update_pickle = False

        if 'plain' in panel: # control data 
            obs_str, bs_str, xmin, xmax, pattern, ctrl_pattern = \
                'obs_'+panel, 'bs_'+panel, 'min_'+panel, 'max_'+panel, \
                    bs_panel, 'square_PLA'
        else:
            obs_str, bs_str, xmin, xmax, pattern, ctrl_pattern = \
                'obs', 'bs', 'min', 'max', panel, None

        # calculate observed preference index if not stored 
        if not overwrite and obs_str in bs_ptr.keys():
            obs = bs_ptr[ obs_str ]
        else:
            H = np.sum(H_bool,2) / float(length)
            if smooth:
                H = gaussian_filter(H,2.5)
            obs = get_preference_index( H,pattern,**traits )
            bs_ptr[obs_str] = obs
            update_pickle = True

        # calculate bootstrapped preference index list if not stored
        #count = 1 
        if not overwrite and bs_str in bs_ptr.keys():
            bs = bs_ptr[ bs_str ]
        else:
            index = np.random.choice(range(length),size=(N,length),replace=True)
            bs = []
            #bd_freq, nbd_freq, colors = [], [], []
            for idx in index:
                H = np.sum(H_bool[:,:,idx],2) / float(length)
                if smooth:
                    H = gaussian_filter(H,2.5)
                bs.append( get_preference_index(H,pattern,**traits) )
                #if ('plain' in panel) and (bs_panel=='leather_STD'):
                #    freq = get_edge_freq(H,pattern,**traits)
                #    if bs[-1] > 0.15:
                #        colors.append('r')
                #    else:
                #        colors.append('b')
                #    bd_freq.append( freq[True] )
                #    nbd_freq.append( freq[False] )
                #if ('plain' in panel) and (bs_panel=='leather_STD') and (bs[-1]>0.15):
                #    plt.close('all')
                #    [y_edges, x_edges]=bin_edges(pattern=bs_panel,scale_plain=True)
                #    plt.imshow( H,origin='lower',interpolation='none', \
                #        aspect='equal',cmap='jet',\
                #           extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )
                #    plot_boundary_window( pattern=bs_panel, edge_margin=10, \
                #        **{'c':'k','alpha':0.5,'lw':1} )
                #    
                #    f = plt.gcf()
                #    f.savefig( './clips/%2d.png' % count )
                #    count += 1

            if ('plain' in panel) and (bs_panel=='leather_STD'):
                plt.close('all')
                plt.scatter(bd_freq, nbd_freq, c=colors)
                ax = plt.gca()
                ax.set_aspect('equal')
                plt.show()

            bs = np.array(bs)
            bs_ptr[bs_str] = bs
            bs_ptr['index'] = index
            update_pickle = True

        # store bootstrap results if updated
        if update_pickle:
            self.save_data()

        return obs, bs 
    # end of METHOD get_bootstrap_data
    
    def get_bootstrap_auc(self, panel, bs_panel, overwrite=False, N=1000, smooth=True, **traits):
        """
            bs_panel should be one of the SL panels
            panel is the data being bootstrapped
        """
        if 'boot_strap' not in self.results.keys():
            self.results['boot_strap'] = {} 
        if bs_panel not in self.results['boot_strap'].keys():
            self.results['boot_strap'][bs_panel] = {} 
        if panel not in self.results['boot_strap'][bs_panel].keys():
            self.results['boot_strap'][bs_panel][panel] = {}
        bs_ptr = self.results['boot_strap'][bs_panel][panel]
        #bs_ptr = self.results.get('boot_strap',{}).get(bs_panel,{})


        H_bool = self.results['B'][panel]
        length = H_bool.shape[2]
        update_pickle = False

        if 'plain' in panel: # control data 
            obs_str, bs_str, xmin, xmax, pattern, ctrl_pattern = \
                'obs_'+panel, 'bs_'+panel, 'min_'+panel, 'max_'+panel, \
                    bs_panel, 'square_PLA'
        else:
            obs_str, bs_str, xmin, xmax, pattern, ctrl_pattern = \
                'obs', 'bs', 'min', 'max', panel, None

        # calculate observed preference index if not stored 
        if not overwrite and obs_str in bs_ptr.keys():
            obs = bs_ptr[ obs_str ]
        else:
            H = np.sum(H_bool,2)# / float(length)
            if smooth:
                H = gaussian_filter(H,2.5)
            obs = get_auc( H,pattern,**traits )
            bs_ptr[obs_str] = obs
            update_pickle = True

        # calculate bootstrapped preference index list if not stored
        #count = 1 
        if not overwrite and bs_str in bs_ptr.keys():
            bs = bs_ptr[ bs_str ]
        else:
            index = np.random.choice(range(length),size=(N,length),replace=True)
            bs = []
            #bd_freq, nbd_freq, colors = [], [], []
            for idx in index:
                H = np.sum(H_bool[:,:,idx],2)# / float(length)
                if smooth:
                    H = gaussian_filter(H,2.5)
                bs.append( get_auc(H,pattern,**traits) )
            bs = np.array(bs)
            bs_ptr[bs_str] = bs
            bs_ptr['index'] = index
            update_pickle = True

        # store bootstrap results if updated
        if update_pickle:
            self.save_data()

        return obs, bs 
    # end of METHOD get_bootstrap_auc

    def get_bootstrap_corr_data(self, panel, bs_panel, overwrite=False, N=1000, smooth=True, **traits):
        """
            bs_panel should be one of the SL panels
            panel is the data being bootstrapped
        """
        if 'boot_strap' not in self.results.keys():
            self.results['boot_strap'] = {} 
        if bs_panel not in self.results['boot_strap'].keys():
            self.results['boot_strap'][bs_panel] = {} 
        if panel not in self.results['boot_strap'][bs_panel].keys():
            self.results['boot_strap'][bs_panel][panel] = {}
        bs_ptr = self.results['boot_strap'][bs_panel][panel]
        #bs_ptr = self.results.get('boot_strap',{}).get(bs_panel,{})

        H_bool = self.results['B'][panel]
        length = H_bool.shape[2]
        update_pickle = False

        if 'plain' in panel: # control data 
            obs_str, bs_str, xmin, xmax, pattern, ctrl_pattern = \
                'obs_'+panel, 'bs_'+panel, 'min_'+panel, 'max_'+panel, \
                    bs_panel, 'square_PLA'
        else:
            obs_str, bs_str, xmin, xmax, pattern, ctrl_pattern = \
                'obs', 'bs', 'min', 'max', panel, None

        # calculate observed preference index if not stored 
        if not overwrite and obs_str in bs_ptr.keys():
            obs = bs_ptr[ obs_str ]
        else:
            H = np.sum(H_bool,2) / float(length)
            if smooth:
                H = gaussian_filter(H,2.5)
            obs = get_correlation( H,pattern,**traits )
            bs_ptr[obs_str] = obs
            update_pickle = True

        # calculate bootstrapped preference index list if not stored
        if not overwrite and bs_str in bs_ptr.keys():
            bs = bs_ptr[ bs_str ]
        else:
            index = np.random.choice(range(length),size=(N,length),replace=True)
            bs = []
            for idx in index:
                H = np.sum(H_bool[:,:,idx],2) / float(length)
                if smooth:
                    H = gaussian_filter(H,2.5)
                bs.append( get_correlation(H,pattern,**traits) )
            bs = np.array(bs)
            bs_ptr[bs_str] = bs
            update_pickle = True

        # store bootstrap results if updated
        if update_pickle:
            self.save_data()

        return obs, bs 
    # end of METHOD get_bootstrap_data

# end of CLASS PopFieldEdgeAna

#=================================================================================

class CountourPIAna(AbstractAnalysis):
    datadir = './ContourPI'
    label = 'contour pi'

    def collect_data(self, **traits):
        self.results = {}

        B = {} # stack of field contour binary maps

        # pattern: tag stored (platform_type), panel: readable
        platforms = ['leather', 'tape']
        stypes = ['STD', 'SHF']
        sl = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]
        panels = sl + ['plain-all', 'plain-before', 'plain-after']
        patterns = sl + ['square_PLA']*3

        for panel, pattern in zip(panels, patterns):
            B[panel] = self.collect_panel( panel, pattern, **traits )
            
        self.results['B'] = B
        self.results['panels'] = panels
        self.results['patterns'] = patterns

        self.out('Done!')
    # end of METHOD collect_data

    def plot_bootstrap_SL_panels(self,ctrl=None,**traits):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        self.out.timestamp = False
        self.log = False
        self.log = True

        # For SL sessions, panel == pattern
        panels = [ panel for panel in self.results['panels'] \
            if (not 'plain' in panel and not 'deStijl' in panel) ]

        if ctrl is None:
            for i, bs_panel in enumerate(panels):
                plt.subplot(2,2,i+1)
                self.bootstrap( bs_panel, ax=plt.gca(), **traits)
            plt.suptitle('SL Bootstrap', fontsize=9)
        else:
            ctrl_pairs = zip( [False,True,True], [False,False,True] )
            for i, bs_panel in enumerate(panels):
                for j, ctrl_pair in enumerate(ctrl_pairs):
                    plt.subplot(3,4,j*4+i+1)
                    ax = plt.gca()
                    ctrl_switch, ctrl_diff = ctrl_pair
                    self.bootstrap( bs_panel, ctrl_panel=ctrl, ctrl=ctrl_switch,\
                        ctrl_diff=ctrl_diff, ax=ax, **traits )
            plt.suptitle('SL Bootstrap (%s)'%ctrl, fontsize=24)

        fname = 'bootstrap'
        if ctrl is not None:
            fname += '_%s' % ctrl
        fname += '.png'

        f = plt.gcf()
        f.set_size_inches(20,15,forward=True)
        f.savefig( os.path.join(self.datadir,fname) )
    # end of METHOD plot_bootstrap_SL_panels

    def plot_PI_distribution(self, ctrl=None, **traits):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        # For SL sessions, panel == pattern
        panels = [ panel for panel in self.results['panels'] \
            if (not 'plain' in panel and not 'deStijl' in panel) ]

        for i, panel in enumerate(panels):
            plt.subplot(2,2,i+1)
            data_panels = [panel, ] if ctrl is None else [panel, ctrl]
            colors = ['b', ] if ctrl is None else ['b', 'grey']
            ks_data = []
            for c, data_panel in zip(colors,data_panels):
                B = self.results['B'][data_panel]
                PIs = np.array( [get_preference_index(B[:,:,i],panel,**traits) \
                    for i in range(B.shape[2])] )
                self.out( '%d NaN found!' % np.sum(np.isnan(PIs)) )
                PIs = PIs[ ~np.isnan(PIs) ]
                plt.hist( PIs,bins=np.linspace(-1,1,20),alpha=0.5,color=c, \
                    normed=True,cumulative=True,histtype='step' )
                if ctrl is not None:
                    ks_data.append( PIs )
            if ctrl is not None:
                D, p = ks_2samp( ks_data[0], ks_data[1] )
                self.out( 'KS test: D=%.4f, p=%.4f' % (D, p) )
        plt.suptitle('PI Distributions CDF (%s)'%ctrl, fontsize=24)

        fname = 'PI_CDF'
        if ctrl is not None:
            fname += '_%s' % ctrl
        fname += '.png'

        f = plt.gcf()
        f.set_size_inches(20,15,forward=True)
        f.savefig( os.path.join(self.datadir,fname) )
    # end of METHOD plot_PI_distribution 

    #-----------------------------------------------------------------------------

    def collect_panel( self, panel, pattern, min_quality='Marginal', repeat_filter=None, **traits):
        self.out('Collecting data for %s...' % panel)
        clust_list = collect_clust_list( \
            pattern,panel,min_quality=min_quality,repeat_filter=repeat_filter)
        if not len(clust_list):
            self.out('No data collected!')
            return None, 0
        
        self.out('Detecting place fields and calculating preference indices...')
        N_field, B = 0, None
        for i, clust in enumerate(clust_list):
            if np.mod(i,10) == 0:
                sys.stdout.write('.')
                sys.stdout.flush()
            rat, day, maze, tc, threshold = clust
            rds = (rat, day, maze)
            session = get_group(rds=rds)
            spikes = session._v_children[tc]
            field_num, labels, contours, n_edges, tmp = label_place_fields( \
                rds,session,spikes,threshold=threshold,\
                    turn_off_rim=False,**traits )

            if B is None:
                B = np.zeros(labels.shape,dtype=bool) # H[:,:,0] = False matrix
            for num in range(field_num):
                B = np.dstack( (B,contours==num+1) )
            N_field += field_num

        B = B[:,:,1:] # remove the False matrix
        sys.stdout.write('\n')

        rats, days, mazes, tcs, tmp = zip(*clust_list)
        N_rat = len( list(set(rats)) )
        N_dataset = len( list(set(zip(rats,days))) )
        N_session = len( list(set(zip(rats,days,mazes))) )
        N_clust = len(clust_list)
        self.out('Data collected from %d rats, %d datasets, %d sessions, %d clusters and %d fields' % (N_rat,N_dataset,N_session,N_clust,N_field))

        return B
    # end of METHOD collect_panel

    def bootstrap(self, bs_panel, ctrl_panel=None, ctrl=False, ctrl_diff=False, nbin=30, plot_out=True, ax=None, sig='*', one_side=True, corr=True, **traits):

        self.out('\n')
        if not ctrl:
            self.out('Bootstrap data for %s sessions...' % bs_panel)
        elif ctrl_diff:
            self.out('Bootstrap difference between %s and %s...' \
                % (bs_panel,ctrl_panel) )
        else:
            self.out('Bootstrap control data (%s) for %s sessions...' \
                % (ctrl_panel,bs_panel) )
        self.out('\n')

        if ctrl_diff:
            title_str = 'PI Difference'
        else: 
            title_str = bs_panel if not ctrl else 'Control'  

        obs_bs, bs_bs = self.get_bootstrap_data( bs_panel,bs_panel,**traits )
        if ctrl_panel is not None:
           obs_ctrl, bs_ctrl = self.get_bootstrap_data( \
                ctrl_panel,bs_panel,**traits )
        if ctrl_diff:
            obs = obs_bs - obs_ctrl
            bs = bs_bs - bs_ctrl
            x_min, x_max = min(min(obs,0),min(bs)), max(max(obs,0),max(bs))
        else:
            x_min = np.min( [0,obs_bs,obs_ctrl,bs_bs.min(),bs_ctrl.min()] )
            x_max = np.max( [0,obs_bs,obs_ctrl,bs_bs.max(),bs_ctrl.max()] )
            if ctrl:
                obs, bs = obs_ctrl, bs_ctrl
            else:
                obs, bs = obs_bs, bs_bs

        # output stats
        if corr:
            levels = np.array( [0.5/4.0, 1.0/4.0, 2.5/4.0, 5.0/4.0, 100-5.0/4.0, 100-2.5/4.0, 100-1/4.0, 100-0.5/4.0] )
        else:
            levels = np.array( [0.5, 1.0, 2.5, 5.0, 95.0, 97.5, 99.0, 99.5] )
        b5, b10, b25, b50, b950, b975, b990, b995 = \
            ( np.percentile(bs,level) for level in levels ) 
        self.out('Percentiles: B_5=%.4f, B_10=%.4f, B_25=%.4f, B_50=%.4f, B_950=%.4f, B_975=%.4f, B_990=%.4f, B_995=%.4f' % (b5,b10,b25,b50,b950,b975,b990,b995) )
        self.out('Observed boundary preference index = %.4f' % obs)
        self.out('%d out of %d (%.4f) bootstrap samples <= 0' \
            % (np.sum(bs<=0),len(bs),np.sum(bs<=0)/float(len(bs))))
        self.out('\n')
        self.out('==== Centered Percentile Bootstrap ====')
        self.out('One-sided:')
        self.out('CI(0.05): (%.4f, ]' %(2*obs-b950))
        self.out('CI(0.01): (%.4f, ]' %(2*obs-b990))
        self.out('Two-sided:')
        self.out('CI(0.05): (%.4f, %.4f)' %(2*obs-b975,2*obs-b25))
        self.out('CI(0.01): (%.4f, %.4f)' %(2*obs-b995,2*obs-b5))
        self.out('\n')
        self.out('==== Percentile Bootstrap ====')
        self.out('One-sided:')
        self.out('CI(0.05): (%.4f, ]' %b50)
        self.out('CI(0.01): (%.4f, ]' %b10)
        self.out('Two-sided:')
        self.out('CI(0.05): (%.4f, %.4f)' %(b25,b975))
        self.out('CI(0.01): (%.4f, %.4f)' %(b5,b995))
        self.out('\n')

        # output bootstrap distribution histogram
        if plot_out:
            x_range = x_max - x_min
            bins = np.linspace(x_min-0.1*x_range, x_max+0.1*x_range, nbin+1)
            if ax is None:
                ax = plt.gca()
            ax.hist(bs,bins=bins,color='cyan')
            plt.axvline(x=obs,c='r',lw=2)
            plt.axvline(x=0,c='k',lw=2)
            if sig == '**':
                if one_side:
                    cut = (b950, )
                else:
                    cut = (b975, b25)
            elif sig == '***':
                if one_side:
                    cut = (b990, )
                else:
                    cut = (b995, b5)
            for c in cut:
                plt.axvline(x=2*obs-c,c='g',lw=2)
            plt.title(title_str,fontsize=16)
            plt.xlim( (x_min-0.1*x_range,x_max+0.1*x_range) )
            plt.tick_params(axis='both',labelsize=10)
    # end of METHOD bootstrap

    def get_bootstrap_data(self, panel, bs_panel, overwrite=False, N=1000, **traits):
        """
            bs_panel should be one of the SL panels
            panel is the data being bootstrapped
        """
        if 'boot_strap' not in self.results.keys():
            self.results['boot_strap'] = {} 
        if bs_panel not in self.results['boot_strap'].keys():
            self.results['boot_strap'][bs_panel] = {} 
        bs_ptr = self.results['boot_strap'][bs_panel]

        if 'plain' in panel: # control data 
            obs_str, bs_str, xmin, xmax, pattern = \
                'obs_'+panel, 'bs_'+panel, 'min_'+panel, 'max_'+panel, bs_panel
        else:
            obs_str, bs_str, xmin, xmax, pattern = \
                'obs', 'bs', 'min', 'max', panel

        B = self.results['B'][panel]
        update_pickle = False

        # calculate preference index for each field if not stored
        if overwrite or (obs_str not in bs_ptr) or (bs_str not in bs_ptr):
            PIs = np.array( [get_preference_index(B[:,:,i],pattern,**traits) \
                for i in range(B.shape[2])] )
            self.out( '%d NaN found!' % np.sum(np.isnan(PIs)) )
            PIs = PIs[ ~np.isnan(PIs) ]

        # calculate observed preference index mean if not stored 
        if not overwrite and obs_str in bs_ptr:
            obs = bs_ptr[ obs_str ]
        else:
            obs = np.mean( PIs )
            bs_ptr[obs_str] = obs
            update_pickle = True

        # calculate bootstrapped preference index mean if not stored
        if not overwrite and bs_str in bs_ptr:
            bs = bs_ptr[ bs_str ]
        else:
            length = len(PIs)
            index = np.random.choice(range(length),size=(N,length),replace=True)
            bs = np.array( [np.mean(PIs[idx]) for idx in index] )
            bs_ptr[bs_str] = bs
            update_pickle = True

        # store bootstrap results if updated
        if update_pickle:
            self.save_data()

        return obs, bs 
    # end of METHOD get_bootstrap_data

# end of CLASS ContourPIAna

#=================================================================================

def collect_clust_list(pattern, panel, min_quality='Marginal', repeat_filter=None):
    tdb_file = get_tdb_file()
    field_table = tdb_file.root.spatial.fields

    # Generate (rat,day,maze) list for specified session type
    # DS panels: ['deStijl', 'plain-all', 'plain-before', 'plain-after']
    if panel == 'plain-before':
        rds_list = [ (637,5,1), (637,5,2), (740,3,1), (740,3,2) ]
    elif panel == 'plain-after':
        rds_list = [ (539,10,2), (577,7,2), (577,8,3), (579,28,3), (579,29,2),\
                     (637,6,1), (637,6,2), (740,4,1), (740,4,2) ]
    else:
        platform, stype = (None,'TEX') if pattern=='TEX' else pattern.split('_')
        rds_list = []
        for rds in walk_mazes():
            session = get_group(rds=rds)
            if rds == (710,17,3): # platform placed in wrong orientation
                continue
            if session._v_attrs['platform'] != platform:
                continue
            if session._v_attrs['type'] != stype:
                continue
            rds_list.append( rds)

    # Select the day with most fields for each tetrode if repeat filter is on
    if repeat_filter is not None:
        # count for each (rat,day,tt), how many place fields were recorded
        tt_count_list = [] # (rat, day, tt, cell_counts)
        for rat, day in list( set(zip(*zip(*rds_list)[0:2])) ): # (rat,day) list
            query = '(rat==%d)&(day==%d)&(area=="CA1")&(quality!="Poor")'%(rat,day)
            if min_quality == 'Fair':
                query += '&(quality!="Marginal")'
            tt = [ int(field['tt']) for field in field_table.where(query) \
                if (rat,day,field['session']) in rds_list ]
            tt_counts = Counter(tt)
            for key in tt_counts.keys():
                tt_count_list.append( (rat,day,key,tt_counts[key]) )
        #print tt_count_list

        # growing the list based on field counts (high to low)
        rats, days, tts, counts = zip(*tt_count_list)
        insert_order = np.argsort(counts)[::-1] # from max counts to min
        tt_list = {} # tt_list[ (rat,day) ] = [tt1, tt2, ... ]
        rat_tt_list = [] # record whether (rat, tt) have encountered
        for i in insert_order:
            rat, day, tt = rats[i], days[i], tts[i]
            if (rat,tt) not in rat_tt_list: # current (rat,tt) has most fields
                key = (rat,day)
                if key not in tt_list.keys():
                    tt_list[key] = [tt, ]
                else:
                    tt_list[key].append( tt )
                rat_tt_list.append( (rat,tt) )
        #print tt_list

    clusts = []
    maze_list, highest_mean = {}, {}
    for rds in rds_list:
        rat, day, maze = rds
        if repeat_filter is not None and (rat,day) not in tt_list.keys():
            continue
        rds_query = '(rat==%d)&(day==%d)&(session==%d)' % rds
        query = rds_query + '&(area=="CA1")&(quality!="Poor")'
        if min_quality == 'Fair':
            query += '&(quality!="Marginal")'
        for field in field_table.where(query):
            tc = 't%dc%d' % (field['tt'], field['cl'])
            key = (rat, day, tc)
            threshold = field['threshold']
            if repeat_filter is None:
                clust = (int(rat), int(day), int(maze), tc, threshold)
                if clust not in clusts:
                    clusts.append( clust )
            else:
                if int(field['tt']) not in tt_list[(rat,day)]:
                    continue
                if key not in maze_list.keys():
                    maze_list[key] = (maze, threshold)
                    highest_mean[key] = field['mean_rate']
                else:
                    if repeat_filter == 'first':
                        if maze < maze_list[key][0]:
                            maze_list[key] = (maze, threshold)
                    elif repeat_filter == 'highest_mean':
                        if field['mean_rate'] > highest_mean[key]:
                            maze_list[key] = (maze, threshold)
                            highest_mean[key] = field['mean_rate']

    if repeat_filter is not None:
        for key in maze_list.keys():
            rat, day, tc = key
            maze, threshold = maze_list[(int(rat),int(day),tc)]
            clusts.append( (int(rat),int(day),int(maze),tc,threshold) )

    tdb_file.close()

    return clusts
# end of METHOD collect_clust_list

def get_rim_filter( map_shape, pattern, debug=False, edge_margin=0, **traits ):
    """
    return a boolean rim matrix which has out-of-arena bins filled with True
    """
    [y_edges, x_edges] = bin_edges(pattern=pattern,**traits)
    del_y, del_x = y_edges[1]-y_edges[0], x_edges[1]-x_edges[0]
    xmin, xmax, ymin, ymax = get_coordinate_range( \
        pattern=pattern, margin=0, **traits )
    rim_xmin = round( (xmin+edge_margin-x_edges[0])/float(del_x) )
    rim_xmax = round( (xmax-edge_margin-x_edges[0])/float(del_x) )
    rim_ymin = round( (ymin+edge_margin-y_edges[0])/float(del_y) )
    rim_ymax = round( (ymax-edge_margin-y_edges[0])/float(del_y) )

    H_rim = np.ones( map_shape ).astype(bool)
    H_rim[rim_ymin:rim_ymax+1,rim_xmin:rim_xmax+1] = False

    if debug:
        plt.imshow( H_rim,origin='lower',interpolation='none', \
            aspect='auto',cmap='gray_r',\
                extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )

    return H_rim
# end of FUNCTION get_rim_filter

def get_preference_index( H, pattern, debug=False, shuffle=False, scale_plain=False, absolute_diff=False, **traits ):
    """
    Calculate the (boundary) preference index (W_in-W_out)/(W_in+W_out)
    pattern: should be one of the SL pattern, the one decides the boundary window
    ctrl_pattern: the control data
    """
    H_rim = get_rim_filter( H.shape, pattern, scale_plain=scale_plain, **traits )

    if shuffle:
        pass
        # not updated
        #Z_i, Z_o = get_shuffled_zone_filter( \
        #    pattern=pattern, scale_plain=scale_plain )
        #F_i, F_o = (Z_i*~H_rim).astype(bool), (Z_o*~H_rim).astype(bool)
        #avg_i = H[F_i].sum() / float(F_i.sum())
        #avg_o = H[F_o].sum() / float(F_o.sum())
        #avg_diff = avg_i - avg_o
        #avg_sum = avg_i + avg_o
    else:
        cnt, area = [0, 0], [0, 0]
        for bd in [True, False]:
            Z = zone_filter( pattern=pattern, bd=bd, scale_plain=scale_plain )
            F = Z * ~H_rim
            if absolute_diff:
                cnt[bd] = (H*F).sum()
                area[bd] = F.sum()
            else:
                avg = (H*F).sum() / float(F.sum())
                avg_diff = avg if bd else avg_diff-avg
                avg_sum = avg if bd else avg_sum+avg
            if debug:
                plt.imshow( Z,origin='lower',interpolation='none', \
                    aspect='auto',cmap='gray_r' )
                plt.show()
                plt.close('all')
                plt.imshow( F,origin='lower',interpolation='none', \
                    aspect='auto',cmap='gray_r' )
                plt.show()
                plt.close('all')
                plt.imshow( H*F,origin='lower',interpolation='none', \
                    aspect='auto',cmap='gray_r' )
                plt.show()
                plt.close('all')
                print (H*F).sum(), F.sum(), avg

    if absolute_diff:
        r = float(area[False]) / area[True] # area scaling factor
        if r < 1:
            print 'r = %f!!' % r
        total_cnt = np.sum(cnt)
        pref_diff = np.nan if total_cnt==0 else \
            float(r*cnt[True]-cnt[False])/total_cnt 
        # pref_diff ranges from -1 (cnt[True]=0) to r (cnt[False]=0)
        pref_index = -1 + 2*(pref_diff-(-1))/float(r-(-1))
    else:
        pref_index = np.nan if avg_sum==0 else float(avg_diff)/avg_sum
        #pref_index = avg_diff

    return pref_index
# end of FUNCTION get_preference_index

def get_auc( H, pattern, debug=False, shuffle=False, scale_plain=False, absolute_diff=False, **traits ):
    H_rim = get_rim_filter( H.shape, pattern, scale_plain=scale_plain, **traits )

    cnt = [[], []]
    for bd in [True, False]:
        Z = zone_filter( pattern=pattern, bd=bd, scale_plain=scale_plain )
        F = (Z*~H_rim).astype(int)
        cnt[bd] = H[F.astype(bool)]

    Cmax = max( max(cnt[0]), max(cnt[1]) )
    bins = np.linspace( 0, Cmax+1, Cmax+2)
    diff = np.zeros( Cmax+1 )
    for bd in [True, False]:
        n,nbin,patches = plt.hist(cnt[bd],bins=bins,cumulative=True,visible=False)
        n = np.array(n).astype(float) / float(n[-1])
        diff = np.array(n) - diff
    auc = np.sum(diff) / float(Cmax+1)

    return auc
# end of FUNCTION get_auc

#def get_edge_freq( H, pattern, debug=False, scale_plain=False, **traits ):
def get_edge_freq( H, pattern, debug=False, shuffle=False, scale_plain=False, absolute_diff=False, **traits ):
    H_rim = get_rim_filter( H.shape, pattern, scale_plain=scale_plain, **traits )

    freq = [0, 0]
    for bd in [True, False]:
        Z = zone_filter( pattern=pattern, bd=bd, scale_plain=scale_plain )
        F = Z * ~H_rim
        freq[bd] = (H*F).sum() / float(F.sum())

    return freq 
# end of FUNCTION get_preference_index

def get_correlation( H, pattern, debug=False, scale_plain=False, sigma=5, cut_off=20, edge_margin=0, **traits ):

    [y_edges, x_edges] = bin_edges(pattern=pattern,**traits)
    del_y, del_x = y_edges[1]-y_edges[0], x_edges[1]-x_edges[0]
    xmin, xmax, ymin, ymax = get_coordinate_range( \
        pattern=pattern, margin=0, **traits )
    rim_xmin = round( (xmin+edge_margin-x_edges[0])/float(del_x) )
    rim_xmax = round( (xmax-edge_margin-x_edges[0])/float(del_x) )
    rim_ymin = round( (ymin+edge_margin-y_edges[0])/float(del_y) )
    rim_ymax = round( (ymax-edge_margin-y_edges[0])/float(del_y) )

    Z = gaussian_template( \
        pattern=pattern, scale_plain=scale_plain, sigma=sigma, cut_off=cut_off )
    corr, p = pearsonr( H[rim_ymin:rim_ymax+1,rim_xmin:rim_xmax+1].flatten(), \
                        Z[rim_ymin:rim_ymax+1,rim_xmin:rim_xmax+1].flatten() ) 

    if debug:
        plt.close('all')
        plt.imshow( Z[rim_ymin:rim_ymax+1,rim_xmin:rim_xmax+1], \
            origin='lower',interpolation='none',aspect='auto',cmap='gray_r' )
        plt.show()
        plt.imshow( H,origin='lower',interpolation='none',\
            aspect='auto',cmap='gray_r' )
        plt.show()
        plt.imshow( H[rim_ymin:rim_ymax+1,rim_xmin:rim_xmax+1], \
            origin='lower',interpolation='none',aspect='auto',cmap='gray_r' )
        plt.show()
        print corr, p

    return corr 
# end of FUNCTION get_correlation

#=================================================================================

# combine gradient map not done yet 

#class PopMapAna(AbstractAnalysis):
#    datadir = './PopField'
#    label = 'population field'
#
#    def collect_data(self, ft='edge', **traits):
#        self.results = {}
#
#        H = {} # population map (summation of cell-session maps)
#        B = {} # stack of cell-session maps 
#        N = {} # number of fields
#
#        # pattern: 'platform_session\ type', panel: other restrictions
#        # single line
#        platforms = ['leather', 'tape']
#        stypes = ['STD', 'SHF']
#        sl = [ '%s_%s'%(m,n) for m,n in itertools.product(platforms,stypes) ]
#        # single line + deStijl
#        tex, pla = 'TEX', 'square_PLA'
#        panels = sl + ['deStijl', 'plain-all', 'plain-before', 'plain-after']
#        patterns = sl + [tex, pla, pla, pla]
#
#        for panel, pattern in zip(panels, patterns):
#            H[panel], B[panel], N[panel] = self.collect_panel( \
#                panel, pattern, ft=ft, **traits )
#            
#        self.results['H'] = H
#        self.results['B'] = B
#        self.results['N'] = N
#        self.results['ft'] = ft
#        self.results['panels'] = panels
#        self.results['patterns'] = patterns
#
#        self.out('Done!')
#    # end of METHOD collect_data
#
#    def collect_clust_list(self, pattern, panel):
#        tdb_file = get_tdb_file()
#        field_table = tdb_file.root.spatial.fields
#
#        # DS panels: ['deStijl', 'plain-all', 'plain-before', 'plain-after']
#        if panel == 'plain-before':
#            rds_list = [ (637,5,1), (637,5,2), (740,3,1), (740,3,2) ]
#        elif panel == 'plain-after':
#            rds_list = [ (539,10,2), (577,7,2), (577,8,3), (579,28,3), (579,29,2),\
#                         (637,6,1), (637,6,2), (740,4,1), (740,4,2) ]
#        else:
#            platform, stype = (None,'TEX') if pattern=='TEX' else pattern.split('_')
#            rds_list = []
#            for rds in walk_mazes():
#                session = get_group(rds=rds)
#                if rds == (710,17,3): # platform placed in wrong orientation
#                    continue
#                if session._v_attrs['platform'] != platform:
#                    continue
#                if session._v_attrs['type'] != stype:
#                    continue
#                rds_list.append( rds)
#
#        clusts = []
#        for rds in rds_list:
#            rat, day, maze = rds
#            rds_query = '(rat==%d)&(day==%d)&(session==%d)' % rds
#            query = rds_query + '&(area=="CA1")&(quality!="Poor")'
#            for field in field_table.where(query):
#                tc = 't%dc%d' % (field['tt'], field['cl'])
#                clust = (rat, day, maze, tc, field['threshold'])
#                if clust not in clusts:
#                    clusts.append( clust )
#
#        tdb_file.close()
#        return clusts
#    # end of METHOD collect_pop_field
#
#
##=================================================================================
#
#    def collect_panel( self, panel, pattern, ft='edge', cell_norm=True, **traits):
#        self.out('Collecting data for %s...' % panel)
#        clust_list = self.collect_clust_list(pattern, panel)
#        if not len(clust_list):
#            self.out('No data collected!')
#            return None, 0
#
#        H, B = None, None
#        for i, clust in enumerate(clust_list):
#            if np.mod(i,10) == 0:
#                sys.stdout.write('.')
#                sys.stdout.flush()
#            rat, day, maze, tc, threshold = clust
#            rds = (rat, day, maze)
#            session = get_group(rds=rds)
#            spikes = session._v_children[tc]
#            G = get_gradient_map(rds,session,spikes,**traits)
#            if cell_norm:
#                G /= G.max()
#            if H is None:
#                H = G.copy()
#                B = np.zeros(H.shape)
#            else:
#                H += G
#                B = np.dstack( (B,G) )
#        B = B[:,:,1:] # remove the False matrix
#        print('\n')
#
#        rats, days, mazes, tcs, tmp = zip(*clust_list)
#        N_rat = len( list(set(rats)) )
#        N_dataset = len( list(set(zip(rats,days))) )
#        N_session = len( list(set(zip(rats,days,mazes))) )
#        N_clust = len(clust_list)
#        self.out( \
#            'Data collected from %d rats, %d datasets, %d sessions, %d clusters' \
#                % (N_rat,N_dataset,N_session,N_clust))
#
#        return H, B, None
#    # end of METHOD collect_panel
