# encoding: utf-8
"""
ratemap.py
"""

import numpy as np
import warnings
import pdb

from scipy import ndimage
from scipy.ndimage.filters import gaussian_filter1d, gaussian_filter, median_filter
from scipy.interpolate import interp1d, griddata
#from scipy.signal import medfilt
from scipy.linalg import toeplitz
from cv2 import bilateralFilter

from core.analysis import AbstractAnalysis
from tools.images import masked_array_to_rgba
from tools.plot import quicktitle
from tools.misc import memoize

from scanr.cluster import string_to_quality

from tdb.config import Config
from tdb.meta import walk_mazes, stamp_to_time
from tdb.data import get_tdb_file, new_table, get_group, get_node, get_unique_row
from tdb.spike import get_session_clusters, parse_cell_name
from tdb.td_maps import bin_edges


FieldCfg = Config['placefield']
BIN_SIZE = FieldCfg['bin_size'] # width of the spatial bin, 2.0 cm
OCC_THRD = FieldCfg['occ_thrd'] # occupancy threshold, 0.06 sec per bin
MIN_VELOCITY = 5.76 # test

#MIN_VELOCITY = 10.0
#BIN_SIZE = 2
#OCC_THRD = 0.06

# ================================================================================

def customize_hist(x, y, grids):
    x, y = np.array(x), np.array(y)

    x_coef, x_consts, y_coef, y_consts = grids
    nrow, ncol = len(y_consts)-1, len(x_consts)-1

    x_b = np.digitize( x_coef[0]*x + x_coef[1]*y, x_consts)
    x_b = x_b[(x_b>0)*(x_b<=ncol)] - 1
    y_b = np.digitize( y_coef[0]*x + y_coef[1]*y, y_consts)
    y_b = y_b[(y_b>0)*(y_b<=nrow)] - 1

    H = np.zeros( (nrow,ncol) ) 
    for i, j in zip(y_b,x_b):
        H[i][j] += 1

    return H
# end of FUNCTION custom_hist

def get_spike_heat_map(rds, session, spikes, grids=None, min_velocity=MIN_VELOCITY, scale_plain=False, **traits):
    """
    Calculate the spike counts in each bin for the specified session
    """
    spk_x, spk_y = get_spike_position( \
        session, spikes, min_velocity=min_velocity, scale_plain=scale_plain )
    if grids is not None:
        H_spk = customize_hist( spk_x, spk_y, grids )
    else:
        H_spk, ybins, xbins = np.histogram2d( \
            spk_y,spk_x,bins=bin_edges(rds,scale_plain=scale_plain,**traits) )

    return H_spk
# end of FUNCTION get_spike_heat_map

@memoize
def get_session_occupancy(session, grids=None, min_velocity=MIN_VELOCITY, scale_plain=False, **traits):
    """
    Calculate the occupancy map for the specified session
    """
    attrs = session._v_attrs
    rds = ( attrs['rat'], attrs['day'], attrs['session'] )

    x, y, ts = [ session._v_children[key] for key in ['x','y','t_pos'] ]
    if scale_plain and attrs['platform']=='square' and attrs['type']=='PLA':
        x = np.array(x) / 110.0 * 100
        y = np.array(y) / 110.0 * 100
    sec = stamp_to_time(ts, zero_stamp=ts[0]) 
    fs = session._v_attrs['tracking_sample_rate']

    velocity = compute_velocity(sec, x, y)
    valid = velocity >= min_velocity 
    if grids is not None:
        H_occ = customize_hist( x[valid], y[valid], grids )
    else:
        H_occ,ybins,xbins = np.histogram2d( y[valid], x[valid], \
            bins=bin_edges(rds,scale_plain=scale_plain,**traits) )
    H_occ = H_occ / float(fs) # in second

    return H_occ, valid
# end of FUNCTION get_session_occupancy

def get_ratemap(session, spikes, rds, occ_thrd=OCC_THRD, interp='nearest', H_spk=None, **traits):
    """
    Calculate the rate map for the specified session
    """
    H_occ, valid = get_session_occupancy(session, **traits)
    if H_spk is None:
        H_spk = get_spike_heat_map(rds, session, spikes, **traits)

    with warnings.catch_warnings(): # ignoring div-by-zero runtime warnings
        warnings.simplefilter("ignore")
        H_rate = 1.0 * H_spk / H_occ
    if np.sum( np.isinf(H_rate) ) > 0: # not zero / zero
        #if ('grids' not in traits.keys()) or (traits['grids'] is None):
            # there should not be any inf with 'nearest' interpolation
        #    print 'INF found in rate map!!'
        #    pdb.set_trace()
        H_rate[ np.isinf(H_rate) ] = 0.0
    H_rate[ np.isnan(H_rate) ] = 0.0

    mask = H_occ < occ_thrd 
    bin_y, bin_x = H_occ.shape
    gy, gx = np.mgrid[ 0:bin_y, 0:bin_x ]
    H_rate[mask] = griddata( zip(gy[~mask],gx[~mask]), H_rate[~mask].ravel(), \
        (gy[mask],gx[mask]), method=interp )

    return H_rate, mask
# end of FUNCTION get_ratemap

def smooth_ratemap( H_rate, mask, smooth=True ):
    if smooth is True or smooth=='kuwahara':
        H_rate = kuwahara_filter(H_rate,mask,k=11)
        H_rate = median_filter(H_rate,size=3,mode='constant',cval=0)
    elif smooth == 'gaussian':
        #H_rate = gaussian_filter(H_rate,sigma=2,mode='constant',cval=0)
        H_rate = gaussian_filter(H_rate,sigma=1.5,mode='constant',cval=0)

    return H_rate
# end of FUNCTION smooth_ratemap

# ================================================================================

def get_spike_position(session, spikes, min_velocity=MIN_VELOCITY, scale_plain=False):
    """
    Get rat position interpolation when spikes detected
    """
    ts = session._v_children['t_pos']
    spike_sec = stamp_to_time(spikes, zero_stamp=ts[0])
    if min_velocity is not None:
        spike_v = get_session_interpolant(session,'velocity')(spike_sec)
        spk_sec = spike_sec[ spike_v >= min_velocity ]
    else:
        spk_sec = spike_sec[:]
        print 'a'
    spk_x = get_session_interpolant(session,'x')(spk_sec)
    spk_y = get_session_interpolant(session,'y')(spk_sec)

    attrs = session._v_attrs
    if scale_plain and attrs['platform']=='square' and attrs['type']=='PLA':
        spk_x = np.array(spk_x) / 110.0 * 100
        spk_y = np.array(spk_y) / 110.0 * 100

    return spk_x, spk_y

def compute_heading(x, y):
    """
    Compute the instant heading degree based on the input coordinate sequences
    """
    diff_x, diff_y = np.diff(x), np.diff(y)

    deg_vec = []
    for dx, dy in zip(diff_x, diff_y):
        if dx == 0:
            if dy == 0:
                deg = np.nan
            elif dy > 0:
                deg = 90
            else: # dy < 0
                deg = 270
        elif dx < 0:
            deg = 180.0/np.pi*np.arctan(dy/dx) + 180
        else:
            deg = 180.0/np.pi*np.arctan(dy/dx)
        deg_vec.append( (deg+360.0)%360 )

    return np.r_[np.nan, np.array(deg_vec)]
# end of FUNCTION compute_heading

def compute_velocity(t, x, y):
    """
    Compute the instant velocity based on the input coordinate sequences
    
    t: time in second
    """
    delta = np.sqrt(np.diff(x)**2 + np.diff(y)**2)
    dt = np.diff(t) 
    velocity = np.r_[0.0, delta/dt]

    return velocity
# end of FUNCTION compute_velocity

@memoize
def get_session_interpolant(session, name):
    """
    Calculate the position and velocity interpolation for the specified session
    """
    x, y, ts = [ session._v_children[key] for key in ['x','y','t_pos'] ]
    sec = stamp_to_time(ts, zero_stamp=ts[0])

    if name == 'velocity':
        var = compute_velocity(sec, x, y)
    elif name == 'x':
        var = x
    elif name == 'y':
        var = y

    return interp1d(sec,var,fill_value=0.0,kind='nearest',bounds_error=False)
# end of FUNCTION get_session_interpolant

@memoize
def get_occ_rim(session,  margin=30, occ_thrd=OCC_THRD, **traits):
    """
    Label the non-visited outer zone when margin>0
    """
    if margin == 0:
        y_edges, x_edges = bin_edges(rds,margin=margin,**traits)
        h, w = len(y_edges)-1, len(x_edges)-1
        return np.zeros( (h,w) ).astype(bool)
    H_occ, valid = get_session_occupancy(session,margin=margin,**traits)
    mask = H_occ < occ_thrd 

    H_rim, N = ndimage.label(mask)
    H_rim[ H_rim != H_rim[0][0] ] = 0 # assume (0,0) is not visited
    H_rim[ H_rim == H_rim[0][0] ] = 1

    return H_rim.astype(bool)
# end of FUNCTION get_occ_rim

def kuwahara_filter(H, mask, k=5):
    """
    k should be an odd number
    """
    if np.max( H[~mask] ) == 0:
        return 0.0 * H

    H_mask = H.copy()
    H_mask[mask] = -1
    median_h = np.median( H_mask[H_mask>0] )
    if np.isnan(median_h):
        print 'NaN median!'

    quad_size = (k+1) / 2 # should be an integer
    delta = quad_size - 1
    offsets = range(quad_size)
    h, w = np.shape(H)

    Q_m = np.ones( (h+delta, w+delta) ) * np.nan # an nan array
    Q_v = Q_m.copy()
    # create a matrix with interleaving 0 and 1 to replace the 0-firing rate bins
    chess_grid = toeplitz( np.mod(range(h+delta),2), np.mod(range(w+delta),2) )
    
    for dx in offsets:
        for dy in offsets:
            S_m = np.lib.pad( H_mask,((dy,delta-dy),(dx,delta-dx)), \
                'constant', constant_values=((-1,-1),(-1,-1)) ).astype(float)
            S_m[S_m<0] = np.nan 
            Q_m = np.dstack( (Q_m,S_m) )

            S_v = S_m.copy()
            # artificially increase the variance for no-firing areas
            S_v[S_v==0] = median_h * chess_grid[S_v==0]
            Q_v = np.dstack( (Q_v,S_v))

    M = np.nanmean(Q_m,axis=2)
    M[ np.isnan(M) ] = 0
    V = np.nanvar(Q_v,axis=2)
    V[ np.isnan(V) ] = np.inf

    H_smooth = np.zeros( (h,w) )
    for y in range(h):
        for x in range(w):
            mean_array = np.array( \
                [ M[y][x], M[y][x+delta], M[y+delta][x], M[y+delta][x+delta] ] )
            var_array = np.array( \
                [ V[y][x], V[y][x+delta], V[y+delta][x], V[y+delta][x+delta] ] )

            if var_array.min() == np.inf:
                H_smooth[y][x] = 0
            else:
                # choose the mean closer to center if more than 1 min var found
                idx = np.where( var_array==var_array.min() )
                min_diff_idx = np.argmin( abs(H_mask[y][x]-mean_array[idx]) )
                H_smooth[y][x] = mean_array[idx][min_diff_idx]

    return H_smooth
# end of FUNCTION kuwahara_filter
