# encoding: utf-8
"""
shift_board.py
"""
import pdb
import numpy as np
import matplotlib.pyplot as plt

from scipy.stats import pearsonr, wilcoxon, ttest_rel
from itertools import product

from core.analysis import AbstractAnalysis
from core.report import BaseReport
from tools.plot import quicktitle

from scanr.cluster import string_to_quality

from tdb.data import get_tdb_file, get_group, get_unique_row
from tdb.td_maps import bin_edges, plot_boundaries
from tdb.ratemap import get_ratemap
from tdb.plot_rate_map import plot_ratemap

# ================================================================================

class ShiftCorrelationReport(BaseReport):

    datadir = './ShiftCorr'
    label = 'shift correlation report'

    ncols = 5 
    nrows = 7 
    figwidth = 8.0
    figheight = 11.0

    xnorm = False
    ynorm = False

    def collect_data(self, quality='Fair'):
        """
        Collect session and cluster list
        """
        self.results = {}

        tdb_file = get_tdb_file() 
        field_table = tdb_file.root.spatial.fields

        cluster_list = []
        if quality == 'Fair':
            query = '(area=="CA1")&(quality=="Fair")'
        elif quality == 'Marginal':
            query = '(area=="CA1")&((quality=="Fair")|(quality=="Marginal"))'
        for field in field_table.where(query):
            rat, day, maze = field['rat'], field['day'], field['session']
            if rat==710 and day==17:
                continue
            tc = 't%dc%d' % (field['tt'],field['cl'])
            session = get_group( rds=(rat,day,maze) )
            if session._v_attrs['expt_type'] != "SL":
                continue
            clust = (rat, day, tc, session._v_attrs['platform'] )
            if clust not in cluster_list:
                cluster_list.append( clust )

        tdb_file.close()

        data_list = [clust+(m,) for clust, m in product(cluster_list,[1,2,3,4,-1])]
        self.results['data_list'] = data_list
        self.results['quality'] = quality
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self):
        QUALITY = string_to_quality( self.results['quality'] )
        bin_traits = {'bin_size':2.0, 'margin':0} # bin size must be 2.0 cm

        def plot_rmap(ax, rat, day, maze, tc, session, spikes, H, M):
            rds = (rat,day,maze)
            H[maze-1], M[maze-1] = plot_ratemap(ax,session,spikes,rds,**bin_traits)
            M[maze-1] = ~M[maze-1]
            plot_boundaries(*rds,**{'c':'r','alpha':0.5})
            
            [y_edges, x_edges] = bin_edges(rds,**bin_traits)
            ax.set( xlim=[x_edges[0],x_edges[-1]], ylim=[y_edges[0],y_edges[-1]] )
        # end of FUNCTION plot_rmap

        def plot_r(ax, rat, day, tc, platform, H, M):
            x = np.linspace(-20,20,21) # x in cm, 21 points 
            if H[2] is None: # shift session
                return
            for i in [0, 1, 3]:
                if H[i] is None:
                    continue

                r, p = [], []
                for j in range(11): # STD shift -10~0 bin (-20~0 cm) 
                    if platform == 'leather':
                        mask = M[i][:,j:j+40] * M[2][:,:40]
                        std = H[i][:,j:j+40][mask]
                        shift = H[2][:,:40][mask]
                    elif platform == 'tape':
                        mask = M[i][10-j:50-j,:] * M[2][-40:,:]
                        std = H[i][10-j:50-j,:][mask]
                        shift = H[2][-40:,:][mask]
                    if (np.max(std)==0) or (np.max(shift)==0): # all-zero array
                        corr, p_value = 0, -1
                    else:
                        corr, p_value = pearsonr(std,shift)
                    if np.isnan(corr):
                        pdb.set_trace()
                    r.append(corr)
                    p.append(p_value)

                for j in range(1,11): # SHIFT shift 1~10 bin (2~20 cm)
                    if platform == 'leather':
                        mask = M[i][:,-40:] * M[2][:,j:j+40]
                        std = H[i][:,-40:][mask]
                        shift = H[2][:,j:j+40][mask]
                    elif platform == 'tape':
                        mask = M[i][:40,:] * M[2][10-j:50-j,:]
                        std = H[i][:40,:][mask]
                        shift = H[2][10-j:50-j,:][mask]
                    if (np.max(std)==0) or (np.max(shift)==0): # all-zero array
                        corr, p_value = 0, -1
                    else:
                        corr, p_value = pearsonr(std,shift)
                    if np.isnan(corr):
                        pdb.set_trace()
                    r.append(corr)
                    p.append(p_value)
                r, p = np.array(r), np.array(p)

                plt.plot(x,r+i*2,'b',lw=1,zorder=1)
                plt.scatter(x[p<0.01], r[p<0.01]+i*2, s=3, c='r', zorder=2)
                plt.scatter(x[p<0], r[p<0]+i*2, s=3, c='k', zorder=3)

            for i in [0, 1, 3]:
                plt.axhline(y=i*2+1,lw=0.5,c='grey',ls='--')
                plt.axhline(y=i*2-1,lw=0.5,c='grey',ls='--')
            plt.axvline(x=0,lw=0.5,c='grey')
            ax.set( xlim=[-20,20], ylim=[-1.5,7.5] )
        # end of FUNCTION plot_rmap

        for item, ax in self.get_plot(self.results['data_list']):
            rat, day, tc, platform, maze = item
            if maze < 0:
                for i in [0, 1, 3]:
                    plt.axhline(y=i*2+1,lw=0.5,c='grey',ls='--')
                    plt.axhline(y=i*2-1,lw=0.5,c='grey',ls='--')
                plt.axvline(x=0,lw=0.5,c='grey')
                ax.set( xlim=[-20,20], ylim=[-1.5,7.5] )
            elif maze == 1:
                self.out('Collecting data for rat %d, day %d, %s...' % (rat,day,tc))
                H, M = [None, None, None, None], [None, None, None, None]
            if self.col == 0:
                quicktitle( ax,'(%d, %d) %s' % (rat,day,tc),size='xx-small' )
            plt.tick_params(axis='both',labelsize=7)

            rds = (rat,day,maze) if maze>0 else (rat,day,3)
            session = get_group(rds=rds)
            if tc not in session._v_children.keys():
                continue
            spikes = session._v_children[tc]
            if string_to_quality(spikes._v_attrs['quality']) < QUALITY:
                continue

            if maze < 0: # shift-correlation
                self.out('Calculating cross-correlation coefficients...')
                plot_r(ax,rat,day,tc,platform,H,M)
            else:
                plot_rmap(ax,rat,day,maze,tc,session,spikes,H,M)

        self.out('Done')
    # end of METHOD process_data

# end of CLASS ShiftCorrelationReport

class CatRemapReport(BaseReport):

    datadir = './CatRemap'
    label = 'categorize remap report'

    ncols = 6 
    nrows = 8 
    figwidth = 8.0
    figheight = 11.0

    xnorm = False
    ynorm = False

    def collect_data(self, quality='Fair', case=1):
        """
        Collect session and cluster list
        """
        self.results = {}

        tdb_file = get_tdb_file() 
        field_table = tdb_file.root.spatial.fields

        query = '(area=="CA1")&(field_no==1)&(session<4)&(session>1)'
        if quality == 'Fair':
            query += '&(quality=="Fair")'
        elif quality == 'Marginal':
            query += '&((quality=="Fair")|(quality=="Marginal"))'

        # Collect list of clusters
        # data_list = [ (rat1,day1,tc1,platform1), (rat2,day2,tc2,platform2), ... ]
        if case == 1: # clusters with fields in m2 and m3
            clust_list = {} 
            for field in field_table.where(query):
                rat, day, maze = field['rat'], field['day'], field['session']
                if (rat==710) and (day==17):
                    continue
                session = get_group( rds=(rat,day,maze) )
                if session._v_attrs['expt_type'] != "SL":
                    continue
                clust = ( rat,day,'t%dc%d'%(field['tt'],field['cl']), \
                    session._v_attrs['platform'] )
                if clust not in clust_list.keys():
                    clust_list[clust] = 1
                else:
                    clust_list[clust] += 1
            data_list = [ clust+([True,True],) for clust in clust_list.keys() \
                if clust_list[clust]==2]
        elif case == 2: # clusters with at least one field in m2 or m3
            clust_list, has_field = [], []
            for field in field_table.where(query):
                rat, day, maze = field['rat'], field['day'], field['session']
                if (rat==710) and (day==17):
                    continue
                session = get_group( rds=(rat,day,maze) )
                if session._v_attrs['expt_type'] != "SL":
                    continue
                clust = ( rat,day,'t%dc%d'%(field['tt'],field['cl']), \
                    session._v_attrs['platform'] )
                if clust not in clust_list:
                    clust_list.append( clust )
                    has_field.append( list(np.array([2,3])==maze) )
                else:
                    has_field[ clust_list.index(clust) ][maze-2] = True
            data_list = [ i+(j,) for i,j in zip(clust_list,has_field)]
        self.out( '%d cluster collected' % len(data_list) )
        tdb_file.close()

        data_list = [clust+(m,) for clust, m in product(data_list,[2,3,-1])]

        self.results['data_list'] = data_list
        self.results['quality'] = quality
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self):
        QUALITY = string_to_quality( self.results['quality'] )
        bin_traits = {'bin_size':2.0, 'margin':0} # bin size must be 2.0 cm

        def plot_rmap(ax, rat, day, maze, tc, session, spikes, H, M):
            rds = (rat,day,maze)
            H[maze-2], M[maze-2] = plot_ratemap(ax,session,spikes,rds,**bin_traits)
            M[maze-2] = ~M[maze-2]
            plot_boundaries(*rds,**{'c':'r','alpha':0.5})
            
            [y_edges, x_edges] = bin_edges(rds,**bin_traits)
            ax.set( xlim=[x_edges[0],x_edges[-1]], ylim=[y_edges[0],y_edges[-1]] )
        # end of FUNCTION plot_rmap

        def plot_r(ax, rat, day, tc, platform, H, M):
            x = np.linspace(0,20,11) # x in cm, 11 points 

            if None in H: # do nothing if session 2 or 3 missing
                return 'none'

            r, p = [], []
            for j in range(11):
                if platform == 'leather':
                    mask = M[0][:,j:] * M[1][:,:50-j]
                    std = H[0][:,j:][mask]
                    shift = H[1][:,:50-j][mask]
                elif platform == 'tape':
                    mask = M[0][:50-j,:] * M[1][j:,:]
                    std = H[0][:50-j,:][mask]
                    shift = H[1][j:,:][mask]
                if (np.max(std)==0) or (np.max(shift)==0): # all-zero array
                    corr, p_value = 0, -1
                else:
                    corr, p_value = pearsonr(std,shift)
                if np.isnan(corr):
                    pdb.set_trace()
                r.append(corr)
                p.append(p_value)
            r = np.array( r )
            p = np.array( p )
            index = p < 0.01

            plt.plot(x,r,'b',lw=1,zorder=1)
            plt.scatter(x[index], r[index], s=3, c='r', zorder=2)
            plt.scatter(x[p<0], r[p<0], s=3, c='k', zorder=3)

            M_id = np.argmax(r)
            if not index[M_id]:
                return 'ambiguous'
            elif np.max(r) < 0.1:
                return 'ambiguous'
            elif (np.max(r)-np.min(r)) < 0.1:
                return 'ambiguous'
            elif (M_id>1) and (M_id<9):
                if index[M_id-1] and index[M_id+1]:
                    return 'intermediate' 
                else:
                    return 'ambiguous'
            elif M_id <= 1:
                if (r[M_id]-min(r[-1],r[-2])) < 0.1:
                    return 'ambiguous'
                elif (M_id==0) and not index[1]: 
                    return 'ambiguous'
                elif (M_id==1) and not (index[0] and index[2]): 
                    return 'ambiguous'
                else:
                    return 'same'
            elif M_id >= 9:
                if (r[M_id]-min(r[0],r[1])) < 0.1:
                    return 'ambiguous'
                elif (M_id==10) and not index[9]: 
                    return 'ambiguous'
                elif (M_id==9) and not (index[8] and index[10]): 
                    return 'ambiguous'
                else:
                    return 'shift'
        # end of FUNCTION plot_rmap

        type_cnt = { 'none'         : 0,
                     'ambiguous'    : 0,
                     'same'         : 0,
                     'intermediate' : 0,
                     'shift'        : 0,
                     'appear'       : 0,
                     'disappear'    : 0 }

        for item, ax in self.get_plot(self.results['data_list']):
            rat, day, tc, platform, has_field, maze = item
            has_field = np.array(has_field).astype(int)

            plt.tick_params(axis='both',labelsize=3)
            if maze == 2:
                self.out('Collecting data for rat %d, day %d, %s...' % (rat,day,tc))
                H, M = [None, None], [None, None]
                quicktitle( ax,'(%d, %d) %s' % (rat,day,tc),size='xx-small' )

            if maze < 0:
                self.out('Calculating cross-correlation coefficients...')
                field_type = plot_r(ax,rat,day,tc,platform,H,M)
                if field_type != 'none':
                    if np.diff(has_field) == 1: # [False, True]
                        field_type = 'appear'
                    elif np.diff(has_field) == -1: # [True, False] 
                        field_type = 'disappear'
                type_cnt[field_type] += 1
                plt.axhline(y=0,lw=0.5,c='grey')
                ax.set( xlim=[0,20], ylim=[-1,1] )
                quicktitle( ax, field_type, size='xx-small' )
            else:
                rds = (rat, day, maze)
                session = get_group(rds=rds)
                if tc not in session._v_children.keys():
                    continue
                spikes = session._v_children[tc]
                if string_to_quality(spikes._v_attrs['quality']) < QUALITY:
                    continue
                plot_rmap(ax,rat,day,maze,tc,session,spikes,H,M)

        sum_all, sum_valid, cnt_str = 0, 0, [] 
        for key in type_cnt.keys():
            sum_all += type_cnt[key]
            if key != 'none':
                sum_valid += type_cnt[key]
                cnt_str.append( '%d %s' % (type_cnt[key],key) )
        self.out( '%d cells collected (%d out of %d valid):' \
            % (sum_valid,sum_valid,sum_all) )
        self.out( '%s' % ','.join(cnt_str) )

        self.out('Done')
    # end of METHOD process_data

# end of CLASS CatRemapReport

class SimilarityAna(AbstractAnalysis):

    datadir = './Similarity'
    label = 'similarity analysis'

    def collect_data(self, quality='Fair', case=1):
        """
        Collect session and cluster list
        """
        self.results = {}

        tdb_file = get_tdb_file() 
        field_table = tdb_file.root.spatial.fields

        query = '(area=="CA1")&(field_no==1)&(session<4)'
        if quality == 'Fair':
            query += '&(quality=="Fair")'
        elif quality == 'Marginal':
            query += '&((quality=="Fair")|(quality=="Marginal"))'

        # Collect list of clusters
        # data_list = [ (rat1,day1,tc1,platform1), (rat2,day2,tc2,platform2), ... ]
        if case == 1: # clusters with fields in m1, m2, and m3
            clust_list = {} 
            for field in field_table.where(query):
                rat, day, maze = field['rat'], field['day'], field['session']
                if (rat==710) and (day==17):
                    continue
                session = get_group( rds=(rat,day,maze) )
                if session._v_attrs['expt_type'] != "SL":
                    continue
                clust = ( rat,day,'t%dc%d'%(field['tt'],field['cl']), \
                    session._v_attrs['platform'] )
                if clust not in clust_list.keys():
                    clust_list[clust] = 1
                else:
                    clust_list[clust] += 1
            data_list = [ clust for clust in clust_list.keys() \
                if clust_list[clust]==3]
        elif case == 2: # clusters with at least one field in m1, m2, or m3
            data_list = [] 
            for field in field_table.where(query):
                rat, day, maze = field['rat'], field['day'], field['session']
                if (rat==710) and (day==17):
                    continue
                session = get_group( rds=(rat,day,maze) )
                if session._v_attrs['expt_type'] != "SL":
                    continue
                clust = ( rat,day,'t%dc%d'%(field['tt'],field['cl']), \
                    session._v_attrs['platform'] )
                if clust not in data_list:
                    data_list.append( clust )
        self.out( '%d cluster collected' % len(data_list) )
        tdb_file.close()

        self.results['data_list'] = data_list
        self.results['quality'] = quality
        self.out('Done')
    # end of METHOD collect_data

    def process_data(self, nbin=15, platform='both'):
        plt.close('all')
        plt.rcdefaults()
        plt.hold(True)

        self.figure = {}

        data_list = self.results['data_list']
        if platform == 'both':
            data = [ clust[:-1] for clust in data_list ]
        else: # platform == 'leather' or 'tape'
            data = [ clust[:-1] for clust in data_list if clust[-1]==platform ]

        std, shift = self.get_correlation_list( data )
        self.out( '%d cluster used' % len(std) )
        T, p = wilcoxon( std, shift )
        self.out( 'Wilcoxon signed-rank test: T = %.4f, p = %.4f' % (T,p) )
        t, p = ttest_rel( std, shift )
        self.out( 'Paired t test: t = %.4f, p = %.4f' % (t,p) )

        title_str = ['Standard vs. Standard', 'Standard vs. Shift', 'Difference']
        for i, hist_data in enumerate( [std, shift, std-shift] ):
            plt.subplot(2,2,i+1)
            m, M = (-2,2) if i==2 else (-1,1)
            bins = np.linspace( m, M, nbin+1 )
            plt.hist( hist_data, bins=bins )
            if i == 2:
                plt.axvline(x=0,lw=1,c='r')
            plt.xlim( [m, M] )
            plt.tick_params(axis='both',labelsize=7)
            plt.title( title_str[i], size='medium' )

        plt.subplot(2,2,4)
        plt.scatter( std, shift )
        plt.plot( [-1,1], [-1,1], 'k', lw=0.5 ) # diagonal line
        plt.xlabel( 'Standard vs. Standard', size='small')
        plt.ylabel( 'Standard vs. Shift', size='small')
        plt.tick_params(axis='both',labelsize=7)
        #plt.axis( [-1, 1, -1, 1] )
        plt.axis( [-0.25, 0.75, -0.25, 0.75] )
        plt.title( 'Paired r Scatter Plot', size='medium' )

        self.figure['distr'] = plt.gcf()
        self.out('Done')
    # end of METHOD process_data
    
    def get_correlation_list(self, data_list):
        # data_list = [ (rat1,day1,tc1), (rat2,day2,tc2), ... ]
        bin_traits = {'bin_size':2.0, 'margin':0} # bin size must be 2.0 cm
        QUALITY = string_to_quality( self.results['quality'] )

        stdXstd, stdXshift = [], []
        for rat, day, tc in data_list:
            H, M = [None, None, None], [None, None, None] # rate maps and masks 
            exclude = False
            for maze in [1,2,3]:
                rds = (rat, day, maze)
                session = get_group( rds=rds )
                if tc not in session._v_children.keys():
                    exclude = True
                    break
                spikes = session._v_children[tc]
                if string_to_quality(spikes._v_attrs['quality']) < QUALITY:
                    exclude = True
                    break
                H[maze-1], M[maze-1] = get_ratemap(session,spikes,rds,**bin_traits)
            if exclude:
                continue

            for i, j in [ (0,1), (1,2) ]: # [ (std,std), (std,shift)]
                mask = M[i] + M[j] # or
                # check if all-zero array
                if (np.max(H[i][~mask])==0) or (np.max(H[j][~mask])==0):
                    corr, p_value = 0, -1
                else:
                    corr, p = pearsonr( H[i][~mask], H[j][~mask] )
                if i == 0:
                    stdXstd.append( corr )
                else:
                    stdXshift.append( corr )

        return np.array(stdXstd), np.array(stdXshift)
    # end of METHOD get_correlation_list

# end of CLASS SimilarityAna
