# encoding: utf-8
"""
spatial_info.py

Spatial information score calculation and place field detection
"""

# 3rd party lib
import tables as tb
import numpy as np
import warnings
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d

# database lib
from core.report import BaseReport
from tools.plot import quicktitle
from tools.stats import pvalue

from scanr.cluster import string_to_quality

from tdb.config import Config
from tdb.meta import walk_mazes, stamp_to_time
from tdb.data import get_tdb_file, new_table, get_group, get_node
from tdb.spike import get_session_clusters
from tdb.ratemap import get_session_interpolant, get_session_occupancy, get_spike_heat_map, smooth_ratemap, get_ratemap, compute_velocity
from tdb.td_maps import bin_edges
from tdb.plot_rate_map import plot_traj_and_spike


SkaggsColDescr = { 'rat'      : tb.UInt16Col(pos=1),
                   'day'      : tb.UInt8Col(pos=2),
                   'session'  : tb.UInt8Col(pos=3),
                   'tt'       : tb.UInt8Col(pos=4),
                   'cl'       : tb.UInt8Col(pos=5),
                   'area'     : tb.StringCol(itemsize=16,pos=6),
                   'quality'  : tb.StringCol(itemsize=16,pos=7),
                   'N'        : tb.UInt64Col(pos=8), # total spike number
                   'N_run'    : tb.UInt64Col(pos=9), # running spike number
                   'rate'     : tb.FloatCol(pos=10), # average firing rate
                   'Skaggs'   : tb.FloatCol(pos=11),
                   'p_value'  : tb.FloatCol(pos=12) }


# parameters selected by Shan on 2016/7/28
FieldCfg = Config['placefield']
BIN_SIZE = FieldCfg['bin_size'] # width of the spatial bin, 2.0 cm 

#MIN_VELOCITY = FieldCfg['min_velocity'] # velocity filter threshold, 10.0 cm/sec
# tuning min_velocity
MIN_VELOCITY = 5.76 
#MIN_VELOCITY = 0

OCC_THRD = FieldCfg['occ_thrd'] # occupancy threhold, 0.06 sec per bin
SKAGGS_THRD = FieldCfg['skaggs_thrd'] # minimum Skaggs information score, 0.6
SKAGGS_P = FieldCfg['skaggs_p'] # Skaggs score significance level, 0.01
MIN_SIZE = FieldCfg['min_size'] # minimum field size, 35 bin
MIN_RATE = FieldCfg['min_rate'] # minimum peak firing rate of field, 0.1 Hz
MIN_SPIKE = FieldCfg['min_spike'] # minimum spike count in field, 30 spikes
MIN_SESSION_SPIKE = FieldCfg['min_session_spike'] # minimum spike count per session, 50 spikes

# ================================================================================
# Create Skaggs information score table
# ================================================================================


class SkaggsInfoTable(BaseReport):
    """
    Calculate Skaggs information score for each session-cluster and create the
    Skaggs table

    Finalized in Nov, 2016, by Chia-Hsuan (Shan) Wang
    """

    datadir = './SkaggsInfo'
    label = 'skaggs info'

    ncols = 5
    nrows = 5
    figwidth = 10.0
    figheight = 11.0

    def collect_data(self, table='skaggs', debug=False, ratemap_out=True, margin=None, scale_plain=False, **traits):
        tdb_file = get_tdb_file() 
        tdb_root = get_node('/','')
        if 'spatial' not in tdb_root._v_children:
            tdb_file.create_group('/','spatial','Spatial Information')
        if 'skaggs' not in tdb_root.spatial._v_children:
        #    skaggs_table = new_table( \
        #        '/spatial',table,SkaggsColDescr,title='Skaggs Info')
            skaggs_table = tdb_file.createTable( \
                '/spatial',table,SkaggsColDescr,title='Skaggs Info' )
        else:
            skaggs_table = tdb_root.spatial.skaggs
        rec = skaggs_table.row

        clust_list = []
        for rds in walk_mazes():
            session = get_group(rds=rds)
            tc_list = get_session_clusters(session)
            for tt, cl in tc_list:
                clust_list.append( (rds,tt,cl) )
        self.out('%d cells collected' % len(clust_list))

        old_rds = None
        for clust, ax in self.get_plot(clust_list):
            rds, tt, cl = clust

            if old_rds != rds:
                rat, day, maze = rds
                dgroup = get_group(rat=rat, day=day)
                areas = dgroup.tt_area[:]

                [y_edges, x_edges] = bin_edges( \
                    rds, margin=margin, scale_plain=scale_plain )
                x_centers = (x_edges[:-1]+x_edges[1:]) / 2.0
                y_centers = (y_edges[:-1]+y_edges[1:]) / 2.0

                session = get_group(rds=rds)
                start, end = session._v_attrs['start'], session._v_attrs['end']
                ts = session._v_children['t_pos']
                sec = stamp_to_time(ts, zero_stamp=ts[0]) 

                old_rds = rds
            # end

            self.out( \
              'Calculating Skaggs score for rat %d, day %d, session %d, t%dc%d'\
                % (rat,day,maze,tt,cl) )
            spikes = session._v_children['t%dc%d' % (tt,cl)]
            quality = string_to_quality( spikes._v_attrs['quality'] )
            
            I, H_rate, N_run, r_avg = compute_skaggs( rds, session, spikes, sec, \
                margin=margin, scale_plain=scale_plain, **traits )
            if N_run > 0:
                self.out('Shuffling Data...')
                p, ncuts = pval( I, session, spikes, rds, sec, \
                    debug=debug, margin=margin, scale_plain=scale_plain, **traits )
                if ncuts < 100:
                    self.out('Too few cuts (ncuts = %d)' % ncuts)
            else:
                p = np.nan
            sig = '' if (I<SKAGGS_THRD)or(p>SKAGGS_P)or(N_run<MIN_SESSION_SPIKE) \
                else '*'
            self.out('I=%f, p=%f %s' % (I,p,sig))

            rec['rat'] = rat
            rec['day'] = day
            rec['session'] = maze
            rec['tt'] = tt
            rec['cl'] = cl
            rec['area'] = areas[tt]
            rec['quality'] = quality
            rec['N'] = len(spikes)
            rec['N_run'] = N_run
            rec['rate'] = r_avg
            rec['Skaggs'] = I 
            rec['p_value'] = p
            rec.append()

            skaggs_table.flush()

            if ratemap_out:
                ax.imshow( H_rate,origin='lower',interpolation='nearest', \
                    aspect='auto',cmap='gray_r', \
                    extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )
                ax.set(xlim=[x_edges[0],x_edges[-1]], ylim=[y_edges[0],y_edges[-1]])
                quicktitle( ax,'(%d,%d,%d) t%dc%d %s' % (rat,day,maze,tt,cl,sig),\
                    size='xx-small')
                plt.tick_params(axis='both',labelsize=7)

        # end of FOR LOOP rds

        tdb_file.close()
    # end of METHOD collect_data

# end of CLASS SkaggsInfoTable
    
# ================================================================================

def compute_skaggs(rds, session, spikes, sec, H_spk=None, adaptive_binning=False, **traits ):
    """
    Compute Skaggs information score and return the score and rate map
    Adaptive binning is used
    """

    # calculate raw rate map based on fixed size binning if H_spk not given
    if H_spk is None:
        H_spk = get_spike_heat_map(rds, session, spikes, **traits)

    # quickly check if there is any running spike on the platform. If not, 
    # return zero matrix without calculating ratemap to save time
    if H_spk.sum() == 0:
        return 0.0, 0.0*H_spk, 0, 0

    f, mask = get_ratemap(session, spikes, rds, H_spk=H_spk, **traits)
    if adaptive_binning:
        f = adaptive_binning()
    else:
        f = smooth_ratemap( f, mask )
        #f = smooth_ratemap( f, mask, smooth='gaussian' )
    f[mask] = 0

    H_spk[mask] = 0
    N = H_spk.sum() # on platform running spikes in valid spatial bins
    if N == 0:
        return 0.0, 0.0*H_spk, 0, 0

    occ, valid = get_session_occupancy(session,**traits)
    H_occ = occ.copy()
    H_occ[mask] = 0
    duration = float( H_occ.sum() )
    p = H_occ / duration    

    if f.sum() == 0:
        return 0.0, 0.0*H_spk, N, N/float(duration)

    # Calculate Skaggs information score based on adaptive binning rate map
    with warnings.catch_warnings(): # ignoring div-by-zero runtime warnings
        warnings.simplefilter("ignore")
        F = np.sum(f*H_occ) / duration
        I = p * f * np.log2(f/F)
        I[np.isnan(I)] = 0.0 # handle zero-spike bins
        I = I.sum() / F

    return I, f, N, N/float(duration)
# end of FUNCTION compute_skaggs


def adaptive_binning():
    pass
#    fs = session._v_attrs['tracking_sample_rate']
#    
#    def calc_adapt_rate(r):
#        idx = H_dist < r
#        N_occ = float( np.sum(H_nocc[idx]) )
#        N_spk = float( np.sum(H_spk[idx]) )
#        param = N_spk * N_occ * N_occ * r * r / alpha
#        rate = N_spk/N_occ*fs if param>1 else -1
#        return param, rate
#    
#    alpha = 5 * 1e5
#    H_nocc = H_occ * fs
#    H_mul = H_nocc * H_nocc * H_spk
#    xv, yv = np.meshgrid( np.arange(H_mul.shape[1]), np.arange(H_mul.shape[0]) )
#    
#    for y, x in zip( *np.where(H_mul<=alpha) ): # param<1 when r=1 for these bins
#        H_dist = np.sqrt( (xv-x)**2 + (yv-y)**2 )
#        mul = H_mul[ (y,x) ]
#        r_max = 15 if mul==0 else min(15, np.ceil(np.sqrt(alpha/mul)))
#        # binary search to speed up 
#        head, tail = 1, r_max
#        param, reg = calc_adapt_rate(tail)
#        if param < 1:
#            f[ (y,x) ] = 0.0
#            continue
#        while (tail-head) > 1:
#            cur = np.floor( (head+tail)/2.0 )
#            param, adapt_rate = calc_adapt_rate(cur)
#            if param > 1:
#                tail = cur
#                reg = adapt_rate
#            else:
#                head = cur
#        f[ (y,x) ] = reg
# end of FUNCTION adaptive binning

def pval(I_obs, session, spikes, rds, sec, debug=False, **traits): 
    """
    Shufflte the spike train (time shift) to get the p-value of the Skaggs score
    """
    if len(spikes) < MIN_SESSION_SPIKE:
        return 0.0, 100

    scale_xy = False
    if 'scale_plain' in traits.keys() and traits['scale_plain']:
        attrs = session._v_attrs
        if attrs['platform']=='square' and attrs['type']=='PLA':
            scale_xy = True

    x, y, ts = [session._v_children[key] for key in ['x','y','t_pos']]
    sec = stamp_to_time(ts, zero_stamp=ts[0])
    velocity = compute_velocity(sec, x, y)
    valid = velocity >= MIN_VELOCITY

    spk_ptr = 0
    while( spikes[spk_ptr]<ts[0] ):
        if spk_ptr == len(spikes)-1:
            return 0.0, 100
        spk_ptr += 1

    cat_x, cat_y, cat_spk, cat_ts, removed = [], [], [], [], 0
    for i, cur in enumerate(ts):
        while (spk_ptr<len(spikes)) and (spikes[spk_ptr]<cur):
            if valid[i]:
                cat_spk.append( spikes[spk_ptr]-removed )
            spk_ptr += 1

        if valid[i]:
            cat_x.append( x[i] )
            cat_y.append( y[i] )
            cat_ts.append( cur-removed )
        elif i > 0:
            removed += ts[i] - ts[i-1]
            
    start, end = cat_ts[0], cat_ts[-1]
    cat_sec = stamp_to_time(cat_ts, zero_stamp=start)
    # if need to reverse spike train:
    #spikes = (start+end-np.array(cat_spk))[::-1]
    spikes = np.array(cat_spk)
        
    time_margin = long(45 * 1e6) # the time shift is at least 30 second
    delta = end - start
    dt = max( 4*1e6, (end-start-2*time_margin)/100.1 )
    cuts = np.arange(time_margin, delta-time_margin, dt)

    if debug:
        [y_edges, x_edges] = bin_edges(rds,**traits)
        plt.close('all')
        ax = plt.gca()

    I = np.empty_like(cuts, dtype='d')
    for i, cut in enumerate(cuts):
        new_spikes = spikes - cut
        new_spikes[new_spikes<start] += delta
        new_spk_sec = stamp_to_time(new_spikes, zero_stamp=start)
        spk_x = interp1d(cat_sec,cat_x, \
            fill_value=0.0,kind='nearest',bounds_error=False)(new_spk_sec)
        spk_y = interp1d(cat_sec,cat_y, \
            fill_value=0.0,kind='nearest',bounds_error=False)(new_spk_sec)
        if scale_xy:
            spk_x = np.array(spk_x) / 110.0 * 100
            spk_y = np.array(spk_y) / 110.0 * 100
        H_spk, ybins, xbins = \
            np.histogram2d(spk_y,spk_x,bins=bin_edges(rds,**traits))

        #tmp = compute_skaggs(rds,session,new_spikes,sec,H_spk=H_spk,**traits)
        #I[i] = tmp[0]

        I[i], H_rate, N_run, r_avg = compute_skaggs( \
            rds, session, new_spikes, sec, H_spk=H_spk, **traits )

        if debug:
            ax = plt.gca()
            ax.imshow( H_spk,origin='lower',interpolation='nearest', \
                aspect='auto',cmap='gray_r', \
                extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )
            ax.set(xlim=[x_edges[0],x_edges[-1]], ylim=[y_edges[0],y_edges[-1]])
            plt.show()
        
    p = pvalue(I_obs, I)
    
    return p, len(cuts)
# end of FUNCTION pval
