# encoding: utf-8
"""
tdb.spike -- Utility functions for handling discrete spiking data
Modified from scanr.spike
"""

import sys, re
import numpy as np
from os import path

from tools.enum import Enum

from tdb.config import Config
from tdb.paths import get_path, get_data_file_path

ClusterQuality = Enum("None", "Poor", "Marginal", "Fair", "PrettyGood", "Good")

# ================================================================================
# Load spike trains and related information from files 
# ================================================================================

def find_spike_files(rat, day, session, tetrode):
    """Find and return a list of filenames and cluster id numbers (ints) for 
    the specified recording tetrode and session
    
    Returns (filenames, clusters) tuple of lists.
    """
    glob = get_data_file_path('cluster', rat=rat, day=day, session=session,
        tetrode=tetrode, do_glob=True)
    tmaze = re.compile('maze%d.(\d+)$'%session)
    filenames = []
    clusters = []
    for fn in glob:
        if fn.startswith('dummy'):
            continue
        match = tmaze.search(fn)
        if match:
            filenames.append(fn)
            clusters.append(int(match.groups()[0]))
    return filenames, clusters

def read_spike_file(fn):
    """Load spike train timestamp array from a cluster file
    
    Array of spike-time timestamps is returned. If an error occurs during
    load, then None is returned unless *error* keyword is set to True.
    """
    if not path.exists(fn):
        raise ValueError, 'invalid cluster file specified: %s.'%fn
    spikes = None
    try:
        spikes = np.loadtxt(fn, skiprows=13, delimiter=',', usecols=[-1])
    except IOError:
        sys.stderr.write('read_spike_file: error loading %s\n'%fn)
    else:
        spikes = spikes.astype('uint64')
        if spikes.size and not np.iterable(spikes): # single spike case
            spikes = spikes.reshape((1,))
    return spikes

def string_to_quality(rating):
    if type(rating) is type(ClusterQuality.None):
        return rating
    try:
        rating = str(rating)
    except TypeError:
        sys.stderr.write("string_to_quality: requires string argument\n")
        raise
    q = ClusterQuality.None
    if rating != "":
        if "pretty" in rating.lower():
            rating = 'Pretty Good'
        try:
            q = getattr(ClusterQuality, rating.title().replace(" ", ""))
        except AttributeError:
            sys.stderr.write('string_to_quality: bad rating: \"%s\"\n'%rating)
    return q

def load_cluster_quality(rat, day, tt, cl):
    """Retrieve the cluster isolation quality rating from a ClNotes file
    
    Returns (quality, comment) tuple of a ClusterQuality value and the string
    comment left by the experimenter.
    """
    fn = get_data_file_path('clnotes', rat=rat, day=day, tetrode=tt, 
        search=False)
    quality = ClusterQuality.None
    comment = "n/a"
    try:
        fd = file(fn, 'r')
    except IOError:
        sys.stderr.write('cluster_quality: error opening %s\n'%fn)
        return quality, comment
    try:    
        lines = fd.readlines()
    except IOError:
        sys.stderr.write('cluster_quality: error reading %s\n'%fn)
    else:
        found = False
        quality_pattern = re.compile('(\d+),"(.*)","(.*)"')
        for line in lines:
            search = re.match(quality_pattern, line)
            if not search:
                continue
            cl_str, quality_str, comment = search.groups()
            found_cl = int(cl_str)
            if found_cl == cl:
                quality = string_to_quality(quality_str)
                found = True
                break
        if not found:
            sys.stderr.write('cluster_quality: missing cluster %d in %s\n'%
                (cl,fn))
    finally:
        fd.close()
    return quality, comment        

def load_mean_spike_width(rat, day, session, tt, cl):
    """Retrieve the spike maxwidth data from a cluster file
    
    Returns mean spike width for the cluster in milliseconds.
    """
    fn = get_data_file_path('cluster', rat=rat, day=day, session=session, 
        tetrode=tt, cluster=cl)
    if not fn:
        sys.stderr.write('mean_spike_width: could not find cluster file\n')
        return 0.0
    fs = Config['sample_rate']['spikes']
    maxwidth = 0.0
    try:
        cldata = np.loadtxt(fn, comments='%', delimiter=',', skiprows=13, 
            usecols=[14]) # maxwidth is 15th column in cl-maze files
    except IOError:
        sys.stderr.write('mean_spike_width: error loading %s\n'%fn)
    else:
        nspikes = cldata.size
        if nspikes != 0:
            if nspikes == 1:
                nsamples = float(cldata)
            else:
                nsamples = cldata.mean()
            maxwidth = nsamples * (1000/fs)
    return maxwidth

# ================================================================================
# Misc. 
# ================================================================================

def parse_cell_name(tc):
    """Return (tetrode, cluster) tuple from string name of cell
    """
    search = re.match('t(\d+)c(\d+)$', tc)
    if not search:
        raise ValueError, 'bad cell name: %s'%str(tc)
    tokens = search.groups()
    return tuple(int(tok) for tok in tokens)

def get_session_clusters(maze, tt=None):
    """Get list of child spike-train arrays for parent maze group
    
    Keyword argumnets:
    tt -- restrict clusters to a particular tetrode or list of tetrodes
    
    Returns list of (t, c) tetrode-cluster tuples.
    """
    tc_list = []
    if tt is not None:
        if not np.iterable(tt):
            tt = [tt]
    for name in maze._v_children.keys():
        tc_match = re.match('t(\d+)c(\d+)$', name)
        if tc_match:
            tc = tc_match.groups()
            t, c = int(tc[0]), int(tc[1])
            if tt and t not in tt:
                continue
            tc_list.append((t, c))
    tc_list.sort()
    return tc_list
