# encoding: utf-8
"""
surface.py -- apparatus and textured surface configurations
"""

import numpy as np
import matplotlib.pyplot as plt

from tdb.config import Config
from tdb.data import get_group


class SurfacePattern(object):
    """
    Parent class of different surface patterns
    """
    pass

# end of CLASS SurfacePattern

class DeStijl(SurfacePattern):
    code = 'deStijl'        # pattern code
    patterned = True        # with patterns? 

    width = 100.0
    height = 100.0

    vertices = ( (50,-50), (50,-12.5), (50,50), (12.5,50), (-25,50), (-50,50),
                 (-50,12.5), (-50,-50), (-12.5,-50), (25,-50), (25,-12.5),
                 (12.5,12.5), (-25,12.5), (-12.5,-12.5) )
    edges = ( (2,7), (3,11), (4,12), (6,11), (9,10), (8,13), (1,13) )
    fringes = ( (1,3), (3,6), (6,8), (1,8) )
    segments = ( (2,11), (11,13), (7,13), (3,11), (4,12), (11,12), (6,12), (9,10), 
                 (8,13), (1,10), (10,13) )
    patches = ( (0,1,10,9), (9,10,13,8), (8,13,7), (1,2,11,13,10), (2,3,11),
                (11,3,4,12), (12,4,5,6), (7,13,11,12,6) )
    patch_v = ( (0,1,10,9), (9,10,13,8), (8,13,7), (1,2,13), (2,3,11), (11,3,4,12),
                (12,4,5,6), (7,11,6) )

    def __init__(self):
        self.Nv = len(self.vertices)
        self.Ne = len(self.edges)
        self.Ns = len(self.segments)
        self.Np = len(self.patches)

        # Whether two vertices are connected by edge segment, INT matrix
        # The corresponding segment number is stored if connected, otherwise -1
        connect = -1 * np.ones( (self.Nv,self.Nv) )
        connect[ zip(*self.segments) ] = range(self.Ns)
        connect[ zip(*self.segments)[::-1] ] = range(self.Ns)
        self.connect = connect

        # Whether two patches are adjacent to each other, BOOL matrix
        adjacent = np.zeros((self.Np,self.Np)).astype(bool)
        for i in range(self.Np):
            for j in range(i,self.Np):
                s1 = [ int(k) for k in self.connect[ self.patches[i], \
                    (self.patches[i][-1],)+self.patches[i][:-1] ] if k>=0 ]
                s2 = [ int(k) for k in self.connect[ self.patches[j], \
                    (self.patches[j][-1],)+self.patches[j][:-1] ] if k>=0 ]
                common = list( set(s1).intersection(s2) )
                if len(common):
                    adjacent[i][j] = True
                    adjacent[j][i] = True
        self.adjacent = adjacent
    # end of METHOD __init__

    def plot_fringe(self,**kwargs):
        for m, n in fringes:
            x, y = zip( *[self.vertices[m-1], self.vertices[n-1]] )
            plt.plot(x, y,**kwargs)
    # end of METHOD plot_fringe

# end of CLASS deStijl


def get_norm_vector(rds):
    """
    Return the norm vector (unit length) of the SL boundary 
    """
    session = get_group(rds=rds)
    platform = session._v_attrs['platform']

    if platform == 'tape':
        x1, y1, x2, y2 = -50, -25, 50, 0
    elif platform == 'leather':
        x1, y1, x2, y2 = 25, 50, 0, -50
    else:
        print 'Error!'

    dy, dx = y2-y1, x2-x1
    length = float( np.sqrt(dx**2+dy**2) )
    norm_vector = ( -dy/length, dx/length )

    return norm_vector
# end of FUNCTION get_norm_vector
