# encoding: utf-8
"""
td_map.py -- Collections of 2d map related functions
"""
import sys

# 3rd party lib
import numpy as np
import warnings
import matplotlib.pylab as plt

from random import uniform
from cv2 import bilateralFilter
from scipy.interpolate import interp1d, griddata
from scipy.signal import medfilt
from scipy.ndimage.filters import gaussian_filter
from shapely.geometry import box, LineString, Point

# database lib
from tools.misc import memoize
from tools.images import masked_array_to_rgba

from tdb.meta import stamp_to_time, get_session_metadata

BIN_SIZE = 2.0 
MARGIN = 30 # cm

# ================================================================================

@memoize
def bin_edges(rds=None, pattern=None, bin_size=BIN_SIZE, **traits):
    """
    Get x, y binning edges for specified session or surface pattern 
    """
    if rds is not None:
        xmin, xmax, ymin, ymax = get_coordinate_range( *rds, **traits )
    elif pattern is not None:
        xmin, xmax, ymin, ymax = get_coordinate_range( pattern=pattern, **traits )
    else:
        raise ValueError, 'Either rds or pattern code must be spcified!'

    nbin = np.ceil( (xmax-xmin) / bin_size ) # actual bin size always <= BIN_SIZE
    x_edges = np.linspace(xmin,xmax,nbin+1)

    nbin = np.ceil( (ymax-ymin) / bin_size ) # actual bin size always <= BIN_SIZE
    y_edges = np.linspace(ymin,ymax,nbin+1)

    return y_edges, x_edges
# end of FUNCTION bin_edges

# ================================================================================

def get_coordinate_range(rat=None, day=None, session=None, pattern=None, margin=None, shape=None, surface=None, scale_plain=False):
    """
    Returns xmin, xmax, ymin, ymax with margin
    """

    if margin is None:
        margin = MARGIN # default 30 cm

    if all( (rat,day,session) ):    # rds set by user
        metadata = get_session_metadata( rat, day, session )
        if metadata['platform'] is None :
            pattern = metadata['type']
        else:
            pattern = metadata['platform'] + '_' + metadata['type']

    if pattern is not None:
        #if pattern == 'deStijl':
        #    w, h = 100.0, 100.0
        #elif pattern == 'plain-black':
        #    w, h = 110.0, 110.0
        if pattern == 'TEX':
            w, h = 100.0, 100.0
        elif pattern == 'rectangle_PLA':
            w, h = 110.0, 80.0
        elif pattern == 'square_PLA':
            if scale_plain: # rescale the square plain board to fit textured data
                w, h = 100.0, 100.0
            else:
                w, h = 110.0, 110.0
        elif pattern == 'tape_STD':
            w, h = 100.0, 100.0
        elif pattern == 'tape_SHF':
            w, h = 100.0, 100.0
        elif pattern == 'leather_STD':
            w, h = 100.0, 100.0
        elif pattern == 'leather_SHF':
            w, h = 100.0, 100.0

    xmin, xmax, ymin, ymax = -w/2.0, w/2.0, -h/2.0, h/2.0

    return xmin-margin, xmax+margin, ymin-margin, ymax+margin

# ================================================================================

def plot_boundaries(rat=None, day=None, session=None, pattern=None, rotate=False, scale_plain=False, ax=None, **kw):
    kwargs = {'c':'k','lw':1}
    kwargs.update(kw)

    rds = (rat, day, session)
    if pattern is None: # pattern not specified, valid session number needed
        if None not in rds: # rds information is complete
            metadata = get_session_metadata( *rds )
            if metadata['platform'] is None:
                pattern = metadata['type']
            else:
                pattern = metadata['platform'] + '_' + metadata['type']
        else:
            sys.stdout.write(
                'boundary: warning: no pattern nor session specified\n')
            return
    else:
        if None not in rds:
            sys.stdout.write( 'boundary: warning: redundant information. the session information might be overwritten by the pattern parameter\n')
    
    if ax is None:
        ax = plt.gca()
    if rds == (710, 17, 3): # problematic shift session
        w, h = 100.0, 100.0
        ax.plot( (-50,50), (-20,5), **kwargs )
    elif pattern == 'TEX':
        w, h = 100.0, 100.0
        ax.plot( (-50,50), (-50,50), **kwargs )
        ax.plot( (-12.5,-12.5), (-50,-12.5), **kwargs )
        ax.plot( (12.5,12.5), (50,12.5), **kwargs )
        ax.plot( (25,25), (-50,-12.5), **kwargs )
        ax.plot( (-25,-25), (50,12.5), **kwargs )
        ax.plot( (-12.5,50), (-12.5,-12.5), **kwargs )
        ax.plot( (12.5,-50), (12.5,12.5), **kwargs )
    elif pattern == 'square_PLA':
        if scale_plain: # rescale the square plain board to fit textured data
            w, h = 100.0, 100.0
        else:
            w, h = 110.0, 110.0
    elif pattern == 'rectangle_PLA':
        w, h = 110.0, 80.0
    elif pattern == 'tape_STD':
        w, h = 100.0, 100.0
        ax.plot( (-50,50), (-25,0), **kwargs )
    elif pattern == 'tape_SHF':
        w, h = 100.0, 100.0
        ax.plot( (-50,50), (-5,20), **kwargs )
    elif pattern == 'leather_STD':
        w, h = 100.0, 100.0
        ax.plot( (25,0), (50,-50), **kwargs )
    elif pattern == 'leather_SHF':
        w, h = 100.0, 100.0
        ax.plot( (5,-20), (50,-50), **kwargs )
    else:
        sys.stdout.write( 'boundary: warning: unknown spec\n')

    hw, hh = w/2.0, h/2.0
    ax.plot( (-hw,hw), (hh,hh), **kwargs )
    ax.plot( (-hw,hw), (-hh,-hh), **kwargs )
    ax.plot( (-hw,-hw), (-hh,hh), **kwargs )
    ax.plot( (hw,hw), (-hh,hh), **kwargs )

    return

# ================================================================================

def get_boundary_line(rds=None, pattern=None):

    if pattern is None: # pattern not specified, valid session number needed
        if None not in rds: # rds information is complete
            metadata = get_session_metadata( *rds )
            pattern = metadata['platform'] + '_' + metadata['type'] 
        else:
            sys.stdout.write('Error: no pattern nor session specified\n')
            return

    if pattern == 'tape_STD':
        bd_line = LineString( [(-50,-25), (50,0)] )
        theta = np.arctan( float(0-(-25))/(50-(-50)) )
    elif pattern == 'leather_STD':
        bd_line = LineString( [(0,-50), (25,50)] )
        theta = np.arctan( float(50-(-50))/(25-0) )
    if pattern == 'tape_SHF':
        bd_line = LineString( [(-50,-5), (50,20)] )
        theta = np.arctan( float(20-(-5))/(50-(-50)) )
    elif pattern == 'leather_SHF':
        bd_line = LineString( [(-20,-50), (5,50)] )
        theta = np.arctan( float(50-(-50))/(5-(-20)) )

    return bd_line, theta*180.0/np.pi
# end of FUNCTION get_boundary_line

def get_boundary_line_params(rds=None, pattern=None):
    if pattern is None: # pattern not specified, valid session number needed
        if None not in rds: # rds information is complete
            metadata = get_session_metadata( *rds )
            pattern = metadata['platform'] + '_' + metadata['type'] 
        else:
            sys.stdout.write('Error: no pattern nor session specified\n')
            return

    if pattern == 'tape_STD': # x-4y-50=0
        a, b, c = 1, -4, -50
    elif pattern == 'leather_STD': # 4x-y-50=0
        a, b, c = 4, -1, -50
    if pattern == 'tape_SHF': # x-4y+30=0
        a, b, c = 1, -4, 30
    elif pattern == 'leather_SHF': # 4x-y+30=0
        a, b, c = 4, -1, 30

    return a, b, c
# end of FUNCTION get_boundary_line_params

def plot_boundary_window(rds=None, pattern=None, ctrl_pattern=None, hw=10, edge_margin=10, scale_plain=False, **kw):

    kwargs = {'c':'k','lw':1}
    kwargs.update(kw)

    if pattern is None: # pattern not specified, valid session number needed
        if None not in rds: # rds information is complete
            metadata = get_session_metadata( *rds )
            pattern = metadata['platform'] + '_' + metadata['type'] 
        else:
            sys.stdout.write('Error: no pattern nor session specified\n')
            return

    grid_pattern = pattern if ctrl_pattern is None else ctrl_pattern

    y_edges, x_edges = bin_edges( pattern=grid_pattern, scale_plain=scale_plain )
    xmin, xmax, ymin, ymax = get_coordinate_range( \
        pattern=grid_pattern, margin=0, scale_plain=scale_plain )
    xmin += edge_margin
    xmax -= edge_margin
    ymin += edge_margin
    ymax -= edge_margin
    delta = hw * 4.0 / np.sqrt(17)
    
    if pattern == 'tape_STD':
        f = lambda x : 0.25*x-12.5
        plt.plot( (xmin,xmax), (f(xmin)-delta,f(xmax)-delta), **kwargs )
        plt.plot( (xmin,xmax), (f(xmin)+delta,f(xmax)+delta), **kwargs )
    elif pattern == 'leather_STD':
        f = lambda y : 0.25*y+12.5
        plt.plot( (f(ymin)-delta,f(ymax)-delta), (ymin,ymax), **kwargs )
        plt.plot( (f(ymin)+delta,f(ymax)+delta), (ymin,ymax), **kwargs )
    if pattern == 'tape_SHF':
        f = lambda x : 0.25*x+7.5
        plt.plot( (xmin,xmax), (f(xmin)-delta,f(xmax)-delta), **kwargs )
        plt.plot( (xmin,xmax), (f(xmin)+delta,f(xmax)+delta), **kwargs )
    elif pattern == 'leather_SHF':
        f = lambda y : 0.25*y-7.5
        plt.plot( (f(ymin)-delta,f(ymax)-delta), (ymin,ymax), **kwargs )
        plt.plot( (f(ymin)+delta,f(ymax)+delta), (ymin,ymax), **kwargs )

    plt.plot( (xmin,xmax), (ymax,ymax), **kwargs )
    plt.plot( (xmin,xmax), (ymin,ymin), **kwargs )
    plt.plot( (xmin,xmin), (ymin,ymax), **kwargs )
    plt.plot( (xmax,xmax), (ymin,ymax), **kwargs )

    return

# end of FUNCTION plot_boundary_window

@memoize
def zone_filter(rds=None, pattern=None, bd=True, control=False, ctrl_pattern=None, hw=10, scale_plain=False ):

    if pattern is None: # pattern not specified, valid session number needed
        if None not in rds: # rds information is complete
            metadata = get_session_metadata( *rds )
            pattern = metadata['platform'] + '_' + metadata['type'] 
        else:
            sys.stdout.write('Error: no pattern nor session specified\n')
            return

    y_edges, x_edges = bin_edges( pattern=pattern, scale_plain=scale_plain )
    xmin, xmax, ymin, ymax = get_coordinate_range( \
        pattern=pattern, margin=0, scale_plain=scale_plain )
    platform_poly = box( xmin,ymin,xmax,ymax )
    
    if pattern == 'tape_STD':
        if control:
            bd_line = LineString( [(-100,-12.5), (100,37.5)] )
        else:
            bd_line = LineString( [(-100,-37.5), (100,12.5)] )
    elif pattern == 'leather_STD':
        if control:
            bd_line = LineString( [(12.5,100), (-37.5,-100)] )
        else:
            bd_line = LineString( [(37.5,100), (-12.5,-100)] )
    if pattern == 'tape_SHF':
        if control:
            bd_line = LineString( [(-100,-32.5), (100,17.5)] )
        else:
            bd_line = LineString( [(-100,-17.5), (100,32.5)] )
    elif pattern == 'leather_SHF':
        if control:
            bd_line = LineString( [(32.5,100), (-17.5,-100)] )
        else:
            bd_line = LineString( [(17.5,100), (-32.5,-100)] )

    y_centers = (y_edges[:-1]+y_edges[1:]) / 2.0
    x_centers = (x_edges[:-1]+x_edges[1:]) / 2.0

    w = x_centers[1] - x_centers[0] 
    alpha = float(5*w) / (2*np.sqrt(17))
    beta = float(3*w) / (2*np.sqrt(17))

    F = np.zeros( (len(y_centers),len(x_centers)) )
    for i, x in enumerate(x_centers):
        for j, y in enumerate(y_centers):
            pt = Point(x,y)
            D = pt.distance(bd_line)

            if not platform_poly.contains(pt):
                F[j,i] = 0
            else:
                if D <= hw-alpha:
                    F[j,i] = 1 
                elif (D>hw-alpha) and (D<hw-beta):
                    delta = hw - beta - D
                    F[j,i] = 1 - (w-delta*np.sqrt(17))**2 / (8.0*w*w)
                elif (D>=hw-beta) and (D<=hw+beta):
                    F[j,i] = 0.5 + (hw-D)*np.sqrt(17) / (4.0*w)
                elif (D>hw-beta) and (D<hw+alpha):
                    delta = D - hw - beta
                    F[j,i] = (w-delta*np.sqrt(17))**2 / (8.0*w*w)
                else:
                    F[j,i] = 0

                if not bd:
                    F[j,i] = 1 - F[j,i]

    return F
# end of FUNCTION zone_filter

def get_shuffled_zone_filter(rds=None,pattern=None,transp=False,orientation=True,shift=True,hw=10,return_params=False,debug=False,scale_plain=False):

    if pattern is None: # pattern not specified, valid session number needed
        if None not in rds: # rds information is complete
            metadata = get_session_metadata( *rds )
            pattern = metadata['platform'] + '_' + metadata['type'] 
        else:
            sys.stdout.write('Error: no pattern nor session specified\n')
            return

    y_edges, x_edges = bin_edges( pattern=pattern, scale_plain=scale_plain )
    xmin, xmax, ymin, ymax = get_coordinate_range( \
        pattern=pattern, margin=0, scale_plain=scale_plain )
    platform_poly = box( xmin,ymin,xmax,ymax )
    
    ew = np.ceil( np.sqrt((hw/4.0)**2+hw**2) )
    if pattern == 'tape_STD':
        x1, y1, x2, y2 = (-100, -37.5, 100, 12.5)
        shift_rg = (0, 0, -25+ew, 50-ew)
    elif pattern == 'leather_STD':
        x1, y1, x2, y2 = (37.5, 100, -12.5, -100)
        shift_rg = (-50+ew, 25-ew, 0, 0)
    elif pattern == 'tape_SHF':
        x1, y1, x2, y2 = (-100, -17.5, 100, 32.5)
        shift_rg = (0, 0, -45+ew, 30-ew)
    elif pattern == 'leather_SHF':
        x1, y1, x2, y2 = (17.5, 100, -32.5, -100)
        shift_rg = (-30+ew, 45-ew, 0, 0)

    if shift:
        x_shift = uniform(shift_rg[0], shift_rg[1])
        y_shift = uniform(shift_rg[2], shift_rg[3])
    bd_line = LineString( [(x1+x_shift,y1+y_shift), (x2+x_shift,y2+y_shift)] )

    y_centers = (y_edges[:-1]+y_edges[1:]) / 2.0
    x_centers = (x_edges[:-1]+x_edges[1:]) / 2.0

    w = x_centers[1] - x_centers[0] 
    alpha = float(5*w) / (2*np.sqrt(17))
    beta = float(3*w) / (2*np.sqrt(17))

    F_i = np.zeros( (len(y_centers),len(x_centers)) )
    F_o = np.zeros( (len(y_centers),len(x_centers)) )
    for i, x in enumerate(x_centers):
        for j, y in enumerate(y_centers):
            pt = Point(x,y)
            D = pt.distance(bd_line)

            if not platform_poly.contains(pt):
                F_i[j,i] = 0
            elif D <= hw-alpha:
                F_i[j,i] = 1
            elif (D>hw-alpha) and (D<hw-beta):
                delta = hw - beta - D
                F_i[j,i] = 1 - (w-delta*np.sqrt(17))**2 / (8.0*w*w)
            elif (D>=hw-beta) and (D<=hw+beta):
                F_i[j,i] = 0.5 + (hw-D)*np.sqrt(17) / (4.0*w)
            elif (D>hw-beta) and (D<hw+alpha):
                delta = D - hw - beta
                F_i[j,i] = (w-delta*np.sqrt(17))**2 / (8.0*w*w)
            elif D >= hw+alpha:
                F_i[j,i] = 0
            else:
                F_i[j,i] = 0

            if platform_poly.contains(pt):
                F_o[j,i] = 1 - F_i[j,i]
            else:
                F_o[j,i] = 0

    if transp:
        F_i = F_i.transpose()
        F_o = F_o.transpose()

    if orientation:
        mode = np.random.randint(3)
        if mode == 0:
            F_i = np.fliplr(F_i)
            F_o = np.fliplr(F_o)
        else:
            # mode == 1
            F_i = F_i.transpose()
            F_o = F_o.transpose()
            if mode == 2:
                F_i = np.flipud(F_i)
                F_o = np.flipud(F_o)

    if debug:
        print shift_rg
        print x_shift, y_shift
        print mode
        plt.close('all')
        plt.figure()
        plt.imshow( F_i,origin='lower',interpolation='none',cmap='gray_r',\
            extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )
        plt.figure()
        plt.imshow( F_o,origin='lower',interpolation='none',cmap='gray_r',\
            extent=[x_edges[0],x_edges[-1],y_edges[0],y_edges[-1]] )
        plt.show()

    if return_params:
        return x_shift, y_shift, mode
    else:
        return F_i, F_o
# end of FUNCTION get_shuffled_zone_filter

@memoize
def gaussian_template(rds=None, pattern=None, sigma=5, cut_off=20, scale_plain=False ):

    # sigma=5, cut_off=20: mean at d~=10 (hw~=10)

    if pattern is None: # pattern not specified, valid session number needed
        if None not in rds: # rds information is complete
            metadata = get_session_metadata( *rds )
            pattern = metadata['platform'] + '_' + metadata['type'] 
        else:
            sys.stdout.write('Error: no pattern nor session specified\n')
            return

    y_edges, x_edges = bin_edges( pattern=pattern, scale_plain=scale_plain )
    xmin, xmax, ymin, ymax = get_coordinate_range( \
        pattern=pattern, margin=0, scale_plain=scale_plain )
    platform_poly = box( xmin,ymin,xmax,ymax )
    
    if pattern == 'tape_STD':
        bd_line = LineString( [(-100,-37.5), (100,12.5)] )
    elif pattern == 'leather_STD':
        bd_line = LineString( [(37.5,100), (-12.5,-100)] )
    if pattern == 'tape_SHF':
        bd_line = LineString( [(-100,-17.5), (100,32.5)] )
    elif pattern == 'leather_SHF':
        bd_line = LineString( [(17.5,100), (-32.5,-100)] )

    y_centers = (y_edges[:-1]+y_edges[1:]) / 2.0
    x_centers = (x_edges[:-1]+x_edges[1:]) / 2.0
    var = sigma ** 2

    F = np.zeros( (len(y_centers),len(x_centers)) )
    for i, x in enumerate(x_centers):
        for j, y in enumerate(y_centers):
            pt = Point(x,y)
            D = pt.distance(bd_line)
            
            if not platform_poly.contains(pt):
                F[j,i] = 0
            else:
                if D > cut_off:
                    F[j,i] = 0
                else:
                    #F[j,i] = 1 
                    F[j,i] = 1.0/np.sqrt(2*np.pi*var) * np.exp(-(D**2)/(2*var))

    return F
# end of FUNCTION gaussian_template
