# encoding: utf-8
"""
tdb_misc.py

Collection of miscellaneous functions
"""

import numpy as np

from collections import Counter

from scanr.time import elapsed
from scanr.cluster import string_to_quality

from tdb.data import get_tdb_file, get_group
from tdb.meta import walk_mazes
from tdb.spike import get_session_clusters

#=================================================================================

def get_panel_session_list(panel):
    # SL panels: ['leather_STD', 'leather_SHF', 'tape_STD', 'tape_SHF']
    # DS panels: ['deStijl', 'plain-all', 'plain-before', 'plain-after']

    pla_before = [ (637,5,1), (637,5,2), (740,3,1), (740,3,2) ]
    pla_after = [ (539,10,2), (577,7,2), (577,8,3), (579,28,3), (579,29,2),\
                  (637,6,1), (637,6,2), (740,4,1), (740,4,2) ]

    if panel == 'plain-before':
        rds_list = pla_before[:]
    elif panel == 'plain-after':
        rds_list = pla_after[:]
    elif panel == 'plain-all':
        rds_list = pla_before + pla_after
    else:
        platform, stype = (None,'TEX') if panel=='deStijl' else panel.split('_')
        rds_list = []
        for rds in walk_mazes():
            session = get_group(rds=rds)
            if rds == (710,17,3): # platform placed in wrong orientation
                continue
            if session._v_attrs['platform'] != platform:
                continue
            if session._v_attrs['type'] != stype:
                continue
            rds_list.append( rds)

    return rds_list
# end of FUNCTION get_panel_session_list

def get_panel_clust_list(panel,**kwds):
    # field_list = ( rat, day, maze, tc, threshold, field_no )
    field_list = get_panel_field_list( panel, **kwds )
    # Transform field_list to clust_list ( rat, day, maze, tc, threshold )
    clust_list = list( set(zip(*zip(*field_list)[0:5])) )

    return clust_list
# end of FUNCTION get_reduced_clust_list

def get_panel_field_list(panel,min_quality='Fair',repeat_filter=None,debug=False):
    tdb_file = get_tdb_file()
    field_table = tdb_file.root.spatial.fields

    table_query = '(area=="CA1")&(quality!="Poor")'
    if min_quality == 'Fair':
        table_query += '&(quality!="Marginal")'

    # Generate (rat,day,maze) list for specified panel
    rds_list = get_panel_session_list( panel )
    field_list = []

    # Return all fields if no repeat filter needs to be applied 
    if repeat_filter is None:
        for rds in rds_list:
            rat, day, maze = rds
            rds_query = '&(rat==%d)&(day==%d)&(session==%d)' % rds
            query = table_query + rds_query
            for field in field_table.where(query):
                tc = 't%dc%d' % (field['tt'], field['cl'])
                field_list.append( (int(rat), int(day), int(maze), tc, \
                    field['threshold'], int(field['field_no'])) )
        tdb_file.close()
        return field_list

    # Otherwise, select the day with most fields for each tetrode
    # count for each (rat,day,tt), how many place fields were recorded
    tt_count_list = [] # (rat, day, tt, field_counts)
    for rat, day in list( set(zip(*zip(*rds_list)[0:2])) ): # (rat,day) list
        query = table_query + '&(rat==%d)&(day==%d)' % (rat,day)
        tt = [ int(field['tt']) for field in field_table.where(query) \
            if (rat,day,field['session']) in rds_list ]
        tt_counts = Counter(tt)
        for key in tt_counts.keys():
            tt_count_list.append( (rat,day,key,tt_counts[key]) )
    if debug:
        print tt_count_list

    # growing the list based on field counts (high to low)
    rats, days, tts, counts = zip(*tt_count_list)
    insert_order = np.argsort(counts)[::-1] # from max counts to min
    tt_list = {} # tt_list[ (rat,day) ] = [tt1, tt2, ... ]
    rat_tt_list = [] # record whether (rat, tt) have encountered
    for i in insert_order:
        rat, day, tt = rats[i], days[i], tts[i]
        if (rat,tt) not in rat_tt_list: # current (rat,tt) has most fields
            key = (rat,day)
            if key not in tt_list.keys():
                tt_list[key] = [tt, ]
            else:
                tt_list[key].append( tt )
            rat_tt_list.append( (rat,tt) )
    if debug:
        print tt_list

    # For each cluster select which session should be used based on repeat_filter
    maze_list, highest_mean = {}, {}
    for rds in rds_list:
        rat, day, maze = rds
        if (rat,day) not in tt_list.keys():
            continue
        rds_query = '&(rat==%d)&(day==%d)&(session==%d)' % rds
        query = table_query + rds_query
        for field in field_table.where(query):
            # Skip if no tetrode data selected for the specified data set 
            if int(field['tt']) not in tt_list[(rat,day)]:
                continue
            tc = 't%dc%d' % (field['tt'], field['cl'])
            threshold, fn = field['threshold'], int(field['field_no'])
            # For each cluster decide if it should be included
            key = (rat, day, tc)
            if key not in maze_list.keys(): # no record found for this cluster
                maze_list[key] = ( maze, threshold, [fn,] )
                highest_mean[key] = field['mean_rate']
            else:
                if maze == maze_list[key][0]: # same rds, append new field
                    maze_list[key][2].append( fn )
                else: # different maze encountered, check if need to replace
                    if (repeat_filter=='first') and (maze<maze_list[key][0]):
                        maze_list[key] = ( maze, threshold, [fn,] )
                    elif (repeat_filter=='highest_mean') and (field['mean_rate']>highest_mean[key]):
                        maze_list[key] = ( maze, threshold, [fn,] )
                        highest_mean[key] = field['mean_rate']
    tdb_file.close()

    # Generate field list
    for key in maze_list.keys():
        rat, day, tc = key
        maze, threshold, field_no = maze_list[ (int(rat),int(day),tc) ]
        for fn in field_no:
            field_list.append( (int(rat),int(day),int(maze),tc,threshold,fn) )

    return field_list
# end of FUNCTION get_panel_field_list 

#=================================================================================

def session_filtered_clusts( rds, w_rds=False ):
    dgroup = get_group(rat=rds[0],day=rds[1])
    tt_area = dgroup.tt_area

    session = get_group(rds=rds)
    tc_list = get_session_clusters(session)

    clusts = []
    for tc in tc_list:
        # only collect CA1 clusters 
        if 'CA1' not in tt_area[ int(tc[0]) ]:
            continue

        cl_name = 't%dc%d' % tc
        spk = session._v_children[cl_name]
        n = len(spk)

        # remove low firing clusters
        if n < 50:
            continue

        # remove putatitive interneurons
        start, end = session._v_attrs['start'], session._v_attrs['end']
        duration = elapsed(start, end)
        if float(n)/duration > 10.0: # in Hz
            continue

        # remove POOR clusters
        quality = string_to_quality( spk._v_attrs['quality'] )
        if quality <= string_to_quality('Poor'):
            continue
        
        if w_rds:
            clusts.append( (rds,cl_name) )
        else:
            clusts.append( cl_name )

    return clusts
# end of FUNCTION session_filtered_clusts

def highest_mean_filter(panel=None, min_quality='Fair', area='CA1', col=None):
    """
    Remove repeating clusters by selecting the cell-session with highest within
    field firing rate (only consider cells with field(s))

    If 'panel' is specified, return only clusters collected in such panel,
    otherwise return every cluster
    """
    tdb_file = get_tdb_file()
    field_table = tdb_file.root.spatial.fields

    if panel is None:
        rds_list = walk_mazes()
    else:
        rds_list = get_panel_session_list(panel)
    

    maze_list, highest_mean = {}, {}
    for rds in rds_list:
        rat, day, maze = rds
        query = '(area=="%s")&(rat==%d)&(day==%d)&(session==%d)' % ((area,)+rds)
        for field in field_table.where(query):
            quality = string_to_quality( field['quality'] )
            if quality < string_to_quality(min_quality):
                continue
            tc = 't%dc%d' % (field['tt'],field['cl'])
            rate = field['mean_rate']

            clust = (rat, day, tc)
            if (clust not in maze_list.keys()) or (rate>highest_mean[clust]):
                if col is None:
                    maze_list[clust] = maze
                else:
                    maze_list[clust] = (maze, field[col[0]])
                highest_mean[clust] = rate

    tdb_file.close()

    clust_list = []
    for key in maze_list.keys():
        rat, day, tc = key
        if col is None:
            clust_list.append( (rat, day, maze_list[key], tc) )
        else:
            maze, data = maze_list[key]
            clust_list.append( (rat, day, maze, tc, data) )

    dtype = [ ('rat',int), ('day',int), ('maze',int), ('tc','S10') ]
    if col is not None:
        dtype += [ (col[0],col[1]), ]
    clust_list = np.sort( \
        np.array(clust_list,dtype=dtype), order=['rat','day','tc'] )

    return clust_list
# end of FUNCTION highest_mean_filter
