# encoding: utf-8
"""
tdb.tracking -- Utility functions for handling position tracking data
Modified from scanr.tracking.py
"""

import sys
import numpy as np
import matplotlib.pylab as plt
from math import atan2
from scipy.interpolate import interp1d

from tools.misc import contiguous_groups, memoize
from tools.filters import quick_boxcar

from tdb.config import Config
from tdb.paths import get_data_file_path
from tdb.data import get_group, flush_file, close_file
from tdb.meta import get_start_end, get_maze_list, get_session_metadata, stamp_to_time

# Constants
DEBUG = Config['debug_mode']
INVALID = Config['track']['invalid_sample']
BOXCAR_KERNEL = Config['track']['boxcar_kernel']
STAMP_RATE = Config['sample_rate']['time']
WIDTH = 720
X_STRETCH = 640.0 / WIDTH 
PI = np.pi 

MIN_VELOCITY = 5.76

# ================================================================================
# Load and preprocess location data from pos.p file 
# ================================================================================

def load_tracking_data(rat=None, day=None, session=None, status={}):
    """Load tracking data for the specified experimental session
    
    Optional: pass in *status* namespace to extract fs, N_samples,
    hd_fix_success, hd_invalid, or hd_discontinuous.
    
    Returns (timestamp, x, y, direction) tuple of arrays.
    """
    if not (rat and day):
        raise ValueError, 'need (rat, day) to specify dataset'

    pos_file = get_data_file_path('tracking',rat=rat, day=day)
    ts, x, y, hd = read_position_file(pos_file)

    start, end = get_start_end(rat, day, session)
    ts, samples = time_slice_sample(ts,np.c_[x, y, hd],start=start,end=end)
    x, y, hd = samples.T
    fs = tracking_sample_rate(ts, verbose=False)

    remove_dupes(ts, x, y, hd)
    glitchy_samples = np.logical_or(x==0, y==0)
    fix_glitches(x, glitchy_samples, func='tracking_data')
    fix_glitches(y, glitchy_samples, func='tracking_data')
    smooth_tracking(x, y, fs=fs)

    x, y = calibrate_position( x, y, (rat,day,session) )

    status['fs'] = fs
    status['N_samples'] = ts.size

    return ts, x, y, hd

def time_slice_sample(t, samples=None, start=None, end=None):
    """Slice an interval out of time/timestamp and sample arrays
    
    Arguments:
    t -- time/timestamp array (length must match first dimension of samples)
    samples -- optional array (or column-wise matrix) of data samples
    start -- time/timestamp of start (default to first)
    end -- time/timestamp of end (default to last)
    
    Returns slice of time array, or (t, samples) tuple of sliced time and 
    sample arrays if additional sample arrays were passed in.
    """
    t = np.asarray(t)
    interval = time_slice(t, start, end)
    if samples is None:
        return t[interval]
    else:
        samples = np.asarray(samples)
        if t.shape[0] != samples.shape[0]:
            raise ValueError, 'shape mismatch for timestamp and sample arrays'
        return t[interval], samples[interval]
    
def time_slice(t, start=None, end=None):
    """Index slice object returned for given timestamp bounds 
    
    Arguments:
    t -- time/timestamp array
    start -- time/timestamp of start (default to first)
    end -- time/timestamp of end (default to last)
    
    Returns slice object which indexes the specified time/timestamp range if 
    it exists.
    """
    t = np.asarray(t).flatten()
    t0, t1 = 0, t.size
    if start is not None:
        if t[t0] <= start <= t[t1 - 1]:
            t0 = (t>=start).nonzero()[0].min()
        elif start > t[t1 - 1]:
            return slice(0, 0, None)
    if end is not None:
        if t[t0] <= end <= t[t1 - 1]:
            t1 = (t<=end).nonzero()[0].max() + 1
        elif end < t[t0]:
            return slice(0, 0, None)
    if DEBUG and t1 < t0:
        sys.stderr.write('time_slice: slice ends before start (%s < %s)\n'%
            (str(t1),str(t0)))
    return slice(t0, t1, None)

def find_sample_rate(ts):
    """Determine the sampling rate of the given timestamp array
    """
    dt = np.diff(ts)
    dt = dt[dt!=0]
    return float(STAMP_RATE/np.median(dt))
    
def remove_dupes(ts, x, y, hd):
    """Remove all tracking records with duplicate timestamps, treating the
    first occurrence of a given timestamp as the valid sample.
    
    Valid tracking arrays are returned.
    """
    valid = np.r_[True, (np.diff(ts) != 0)]
    if not valid.all():
        sys.stderr.write('remove_dupes: found %d dupes\n'%
            (np.sum(True - valid)))
        ts = ts[valid]
        x = x[valid]
        y = y[valid]    
        hd = hd[valid]
    return ts, x, y, hd
    
def smooth_tracking(x, y, fs=30.0):
    """In-place boxcar smoothing of tracking data to remove artifacting
    """
    box = int(round(BOXCAR_KERNEL * (fs / 30.0))) # sample frequency adjustment
    if box > 1:
        x[:] = quick_boxcar(x, M=box, centered=True)
        y[:] = quick_boxcar(y, M=box, centered=True)
    return

def tracking_sample_rate(ts, verbose=False):
    """Determine the sampling rate of tracking data based on the corresponding 
    timestamp array
    """
    r_bar = r0 = find_sample_rate(ts)
    if 29.5 < r0 < 30.5:
        r_bar = 30.0
    elif 59.0 < r0 < 61.0:
        r_bar = 60.0
        if DEBUG and verbose:
            sys.stdout.write('trajectory: detected 60Hz sampling\n')
    else:
        r_bar = r0
        if DEBUG and verbose:
            sys.stdout.write(
                'trajectory: warning: found %0.2fHz sampling\n'%r_bar)
    return r_bar
    
def fix_glitches(z, glitchy, wrapped=False, func='fix_glitches'):
    """Use linear interpolation to fix glitches in position sampling arrays
    
    Arguments:
    z -- array data to be fixed (x, y, or angle data)
    glitchy -- boolean array indicating which samples are glitches
    
    Array is fixed in place.
    """
    glitches = contiguous_groups(glitchy)
    for g in glitches:
        n = g[1] - g[0]
        if g[0] == 0:
            z0 = z[g[1]] - (n+1)*(z[g[1]+1]-z[g[1]])
            z1 = z[g[1]]
            if DEBUG:
                sys.stderr.write('%s: initial glitch: %d samples\n'%(func, n))
        else:
            z0 = z[g[0]-1]
            if g[1] == z.shape[0]:
                z1 = z0 + (n+1)*(z[g[0]-1]-z[g[0]-2])
                if DEBUG:
                    sys.stderr.write('%s: terminal glitch: %d samples\n'%(func, n))
            else:
                z1 = z[g[1]]
        z[slice(*g)] = np.linspace(z0, z1, n+2)[1:-1]
            
    if DEBUG:
        sys.stderr.write('%s: fixed %d tracking glitches\n'%
            (func, len(glitches)))

@memoize
def calibrate_factors(rds):
    """
    Calculate position calibration factors based on platform vertices
    """
    sgroup = get_group(rds=rds)
    st = sgroup._v_attrs
    expt_type, stype, platform, TL_corner = \
        st['expt_type'], st['type'], st['platform'], st['cam_TL']

    # set platform width and height
    if expt_type == 'DS': # deStijl
        if stype == 'TEX': # textured
            w, h = 100.0, 100.0
        elif stype == 'PLA': # plain
            if platform == 'rectangle':
                w, h = 110.0, 80.0
            elif platform == 'square':
                w, h = 110.0, 110.0
    elif expt_type == 'SL': # single line
        w, h = 100.0, 100.0
        
    # calculate the length of the platform diagonal line (in half) as scaling 
    # reference (in cm)
    hw, hh = w/2.0, h/2.0
    half_diags = np.sqrt(hw**2+hh**2) * np.ones(4)

    # calculate the angle of each vertex as rotation calibration reference
    # (0,0) set at the center of the platform
    angles = np.array( [ \
        atan2(j,i) for i,j in ((-hw,hh),(-hw,-hh),(hw,-hh),(hw,hh)) ] )

    x, y = read_vertices(*rds)
    x, y = x[:4], y[:4] # use only the four outer vertices

    # rescale vertices x since video got distorted in Cheetah 5 digital system
    x = x * X_STRETCH 

    # shift and flip, the new origin is set to be at the center of the platform
    #
    # The TL_corner is defined by the experimenter.  It should be placed at the
    # top-left corner when visualizing the data.  For 'deStijl' data, 
    # the TL_corner is the nearest corner to the sand paper, and the corresponding
    # corner for the plain platform.  For 'single line' data, check experiment
    # sheets for the orientations.  TL_corner is the top left corner under
    # Neuralynx color camera view
    #
    # Orientation in quotes follows Matlab/video convention, origin at top left
    # (origin at bottom left in python plots, so the y-axis needs to be flipped
    #  even if the TL_corner(cam_TL) is at top-left in video)
    CoM = np.mean([x,y], 1)
    if TL_corner == 'bottom right':
        x, y = -(x-CoM[0]), y-CoM[1] # x = -x
    elif TL_corner == 'top left':
        x, y = x-CoM[0], -(y-CoM[1]) # y = -y

    # rotate
    theta = np.mean( angles - np.array([atan2(j,i) for i, j in zip(x,y)]) )
    rotation_matrix = [ \
        [np.cos(theta),-np.sin(theta)], [np.sin(theta),np.cos(theta)] ]
    x, y = np.dot( rotation_matrix, [x,y] )

    # scale
    v_len  = np.array( [np.sqrt(i**2+j**2) for i, j in zip(x,y)] ) # in pixels
    scale_factor = np.mean( half_diags/v_len.astype('float') )

    return CoM, rotation_matrix, scale_factor, TL_corner

def calibrate_position(x, y, rds):
    CoM, rot, scale, TL = calibrate_factors(rds)

    # Nvt and pos.p files corrected in CalcPalm, March 2017
    # rescale vertices x since video got distorted in Cheetah 5 digital system
    #x = x * X_STRETCH 

    if TL == 'bottom right':
        x, y = -(x-CoM[0]), y-CoM[1] # x = -x
    elif TL == 'top left':
        x, y = x-CoM[0], -(y-CoM[1]) # y = -y
    x, y = np.dot( rot, [x,y] )
    x, y = x*scale, y*scale

    return x, y

def read_vertices(rat, day, session):
    """ 
    Read vertex coordinates from file. No adjustment.
    """
    v_file = get_data_file_path('vertices', rat=rat, day=day, session=session)
    vx, vy = np.loadtxt(v_file,comments='%',delimiter=',',unpack=True) 

    # Negative values are estimations, LED signal missing
    vx[vx<0] *= -1
    vy[vy<0] *= -1

    return vx, vy

def load_platform_vertices( rds ):
    """
    Return calibrated platform vertices coordinates
    """
    vx, vy = read_vertices( *rds )
    vx, vy = calibrate_position( vx, vy, rds )

    return vx, vy 

@memoize
def read_position_file(fn):
    """Load position tracking data from Pos.p (or Pos.p.ascii) file
    
    Returns (timestamp, x, y, direction) tuple of arrays.
    """
    if fn.endswith('.p'):
        import neuralynx
        ts, x, y, angle = neuralynx.read_position_file(fn)
    elif fn.endswith('.ascii'):
        ts, x, y, angle = np.loadtxt(fn, comments='%', delimiter=',', unpack=True)
    else:
        raise ValueError, 'bad position file name: %s'%fn
    ts = ts.astype(long)

    return ts, x, y, angle

# ================================================================================
# Misc 
# ================================================================================

def get_tracking_slice(start=None, end=None, rat=None, day=None, session=None):
    """Time-sliced tracking data
    
    Keyword arguments:
    start/end -- timestamp bounds of the slice
    rat/day/session -- specify the session
    
    Returns (t, x, y, hd) position tracking arrays.
    """
    maze = get_group(rat=rat, day=day, session=session)
    if not maze or (None in (rat, day, session)):
        raise ValueError, 'no session found for %s'%str((rat, day, session))
    t_slice = time_slice(maze.t_pos[:], start=start, end=end)
    return maze.t_pos[t_slice], maze.x[t_slice], maze.y[t_slice], \
        maze.hd[t_slice]

def get_tracking(rat, day, session=None):
    """Retrieve tracking data for specified session from H5 file
    
    If session is not specified, the data for the entire day is returned.
    
    Returns (ts, x, y, angle) tuple of arrays.
    """
    if session is not None:
        maze = get_group(rat=rat, day=day, session=session)
        if not maze:
            raise ValueError, 'no session found for %s'%str((rat, day, session))
        t_pos, x, y, hd = maze.t_pos[:], maze.x[:], maze.y[:], maze.hd[:]    
    else:
        t_pos, x, y, hd = np.array([], 'l'), np.array([], 'f'), np.array([], 'f'), \
            np.array([], 'f')
        for session in get_maze_list(rat, day):
            maze = get_group(rat=rat, day=day, session=session)
            if not maze:
                raise ValueError, 'no session found for %s'%str((rat, day, session))
            t_pos = np.r_[t_pos, maze.t_pos[:]]
            x = np.r_[x, maze.x[:]]
            y = np.r_[y, maze.y[:]]
            hd = np.r_[hd, maze.hd[:]]
    return t_pos, x, y, hd


def velocity_filter(ts, x, y, sig=None, min_velocity=MIN_VELOCITY):
    t_in_sec = stamp_to_time(ts,zero_stamp=ts[0])

    delta = np.sqrt(np.diff(x)**2 + np.diff(y)**2)
    dt = np.diff(t_in_sec) 
    velocity = np.r_[0.0, delta/dt]
    valid = velocity >= min_velocity

    if sig is None:
        return valid
    else:
        interp_velocity = interp1d( \
            t_in_sec, velocity, fill_value=0.0, bounds_error=False)
        sig_v = interp_velocity(stamp_to_time(sig,zero_stamp=ts[0]))
        pass_filter = sig_v >= min_velocity
        filtered = sig[pass_filter]

        return valid, filtered
